/* Generated by Cython 0.29.32 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "cloudtrace.read.packet",
        "sources": [
            "cloudtrace/read/packet.pyx"
        ]
    },
    "module_name": "cloudtrace.read.packet"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_32"
#define CYTHON_HEX_VERSION 0x001D20F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cloudtrace__read__packet
#define __PYX_HAVE_API__cloudtrace__read__packet
/* Early includes */
#include <stdint.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include "netinet/ip.h"
#include <string.h>
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "cloudtrace/read/packet.pyx",
  "stringsource",
  "cloudtrace/read/packet.pxd",
  "cloudtrace/read/combine.pxd",
  "cloudtrace/read/pcap.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_10cloudtrace_4read_7combine_Trace;
struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap;
struct __pyx_obj_10cloudtrace_4read_6packet_Packet;
struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe;
struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe;
struct __pyx_obj_10cloudtrace_4read_6packet_Reply;
struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply;
struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply;
struct __pyx_t_10cloudtrace_4read_5utils_ethhdr_s;
struct __pyx_t_10cloudtrace_4read_5utils_ip_s;
struct __pyx_t_10cloudtrace_4read_5utils_echo_s;
struct __pyx_t_10cloudtrace_4read_5utils_unused_s;
union __pyx_t_10cloudtrace_4read_5utils_icmpunion;
struct __pyx_t_10cloudtrace_4read_5utils_icmp_s;
struct __pyx_t_10cloudtrace_4read_5utils_udp_s;
struct __pyx_t_10cloudtrace_4read_5utils_udppseudo_s;
struct __pyx_t_10cloudtrace_4read_5utils_ip6_hdrctl;
union __pyx_t_10cloudtrace_4read_5utils_ip6_un;
struct __pyx_t_10cloudtrace_4read_5utils_icmpexthdr_s;
struct __pyx_t_10cloudtrace_4read_5utils_icmpext_s;
struct __pyx_t_10cloudtrace_4read_5utils_ip6_s;

/* "cloudtrace/read/utils.pxd":24
 * cdef uint16_t checksum(void *data, int length) except -1;
 * 
 * cdef struct ethhdr_s:             # <<<<<<<<<<<<<<
 *     uint8_t h_dest[6]
 *     uint8_t h_source[6]
 */
struct __pyx_t_10cloudtrace_4read_5utils_ethhdr_s {
  uint8_t h_dest[6];
  uint8_t h_source[6];
  uint16_t h_proto;
};

/* "cloudtrace/read/utils.pxd":28
 *     uint8_t h_source[6]
 *     uint16_t h_proto
 * ctypedef ethhdr_s ethhdr             # <<<<<<<<<<<<<<
 * 
 * cdef struct ip_s:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_ethhdr_s __pyx_t_10cloudtrace_4read_5utils_ethhdr;

/* "cloudtrace/read/utils.pxd":30
 * ctypedef ethhdr_s ethhdr
 * 
 * cdef struct ip_s:             # <<<<<<<<<<<<<<
 *     uint8_t ip_hl
 *     uint8_t ip_tos
 */
struct __pyx_t_10cloudtrace_4read_5utils_ip_s {
  uint8_t ip_hl;
  uint8_t ip_tos;
  uint16_t ip_len;
  uint16_t id;
  uint16_t ip_off;
  uint8_t ip_ttl;
  uint8_t ip_p;
  uint16_t ip_sum;
  uint8_t ip_src[4];
  uint8_t ip_dst[4];
};

/* "cloudtrace/read/utils.pxd":41
 *     uint8_t ip_src[4]
 *     uint8_t ip_dst[4]
 * ctypedef ip_s ip             # <<<<<<<<<<<<<<
 * 
 * cdef void iptohost(ip *iphdr);
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_ip_s __pyx_t_10cloudtrace_4read_5utils_ip;

/* "cloudtrace/read/utils.pxd":45
 * cdef void iptohost(ip *iphdr);
 * 
 * cdef struct echo_s:             # <<<<<<<<<<<<<<
 *     uint16_t id
 *     uint16_t seq
 */
struct __pyx_t_10cloudtrace_4read_5utils_echo_s {
  uint16_t id;
  uint16_t seq;
};

/* "cloudtrace/read/utils.pxd":48
 *     uint16_t id
 *     uint16_t seq
 * ctypedef echo_s echo             # <<<<<<<<<<<<<<
 * 
 * cdef struct unused_s:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_echo_s __pyx_t_10cloudtrace_4read_5utils_echo;

/* "cloudtrace/read/utils.pxd":50
 * ctypedef echo_s echo
 * 
 * cdef struct unused_s:             # <<<<<<<<<<<<<<
 *     uint16_t __unused
 *     uint16_t mtu
 */
struct __pyx_t_10cloudtrace_4read_5utils_unused_s {
  uint16_t __pyx___unused;
  uint16_t mtu;
};

/* "cloudtrace/read/utils.pxd":53
 *     uint16_t __unused
 *     uint16_t mtu
 * ctypedef unused_s unused             # <<<<<<<<<<<<<<
 * 
 * cdef union icmpunion:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_unused_s __pyx_t_10cloudtrace_4read_5utils_unused;

/* "cloudtrace/read/utils.pxd":55
 * ctypedef unused_s unused
 * 
 * cdef union icmpunion:             # <<<<<<<<<<<<<<
 *     echo_s echo
 *     unsigned int gateway
 */
union __pyx_t_10cloudtrace_4read_5utils_icmpunion {
  struct __pyx_t_10cloudtrace_4read_5utils_echo_s echo;
  unsigned int gateway;
  struct __pyx_t_10cloudtrace_4read_5utils_unused_s frag;
};

/* "cloudtrace/read/utils.pxd":60
 *     unused_s frag
 * 
 * cdef struct icmp_s:             # <<<<<<<<<<<<<<
 *     uint8_t type
 *     uint8_t code
 */
struct __pyx_t_10cloudtrace_4read_5utils_icmp_s {
  uint8_t type;
  uint8_t code;
  uint16_t checksum;
  union __pyx_t_10cloudtrace_4read_5utils_icmpunion un;
};

/* "cloudtrace/read/utils.pxd":65
 *     uint16_t checksum
 *     icmpunion un
 * ctypedef icmp_s icmp             # <<<<<<<<<<<<<<
 * 
 * cdef void icmptohost(icmp *hdr);
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_icmp_s __pyx_t_10cloudtrace_4read_5utils_icmp;

/* "cloudtrace/read/utils.pxd":69
 * cdef void icmptohost(icmp *hdr);
 * 
 * cdef struct udp_s:             # <<<<<<<<<<<<<<
 *     uint16_t source
 *     uint16_t dest
 */
struct __pyx_t_10cloudtrace_4read_5utils_udp_s {
  uint16_t source;
  uint16_t dest;
  short len;
  uint16_t check;
};

/* "cloudtrace/read/utils.pxd":74
 *     short len
 *     uint16_t check
 * ctypedef udp_s udp             # <<<<<<<<<<<<<<
 * 
 * cdef void udptohost(udp *hdr);
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_udp_s __pyx_t_10cloudtrace_4read_5utils_udp;

/* "cloudtrace/read/utils.pxd":78
 * cdef void udptohost(udp *hdr);
 * 
 * cdef struct udppseudo_s:             # <<<<<<<<<<<<<<
 *     uint8_t src[4]
 *     uint8_t dst[4]
 */
struct __pyx_t_10cloudtrace_4read_5utils_udppseudo_s {
  uint8_t src[4];
  uint8_t dst[4];
  uint8_t zeros;
  uint8_t proto;
  uint16_t length;
  uint16_t sport;
  uint16_t dport;
  uint16_t length2;
  uint16_t chksum;
  uint8_t data[0x5C0];
};

/* "cloudtrace/read/utils.pxd":89
 *     uint16_t chksum
 *     uint8_t data[1472]
 * ctypedef udppseudo_s udppseudo             # <<<<<<<<<<<<<<
 * 
 * cdef struct ip6_hdrctl:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_udppseudo_s __pyx_t_10cloudtrace_4read_5utils_udppseudo;

/* "cloudtrace/read/utils.pxd":91
 * ctypedef udppseudo_s udppseudo
 * 
 * cdef struct ip6_hdrctl:             # <<<<<<<<<<<<<<
 *     uint32_t ip6_un1_flow
 *     uint16_t ip6_un1_plen
 */
struct __pyx_t_10cloudtrace_4read_5utils_ip6_hdrctl {
  uint32_t ip6_un1_flow;
  uint16_t ip6_un1_plen;
  uint8_t ip6_un1_nxt;
  uint8_t ip6_un1_hlim;
};

/* "cloudtrace/read/utils.pxd":97
 *     uint8_t ip6_un1_hlim
 * 
 * cdef union ip6_un:             # <<<<<<<<<<<<<<
 *     ip6_hdrctl ip6_un1
 *     uint8_t ip6_un2_vfc
 */
union __pyx_t_10cloudtrace_4read_5utils_ip6_un {
  struct __pyx_t_10cloudtrace_4read_5utils_ip6_hdrctl ip6_un1;
  uint8_t ip6_un2_vfc;
};

/* "cloudtrace/read/utils.pxd":101
 *     uint8_t ip6_un2_vfc
 * 
 * cdef struct icmpexthdr_s:             # <<<<<<<<<<<<<<
 *     uint8_t version
 *     uint8_t reserved
 */
struct __pyx_t_10cloudtrace_4read_5utils_icmpexthdr_s {
  uint8_t version;
  uint8_t reserved;
  uint16_t checksum;
};

/* "cloudtrace/read/utils.pxd":105
 *     uint8_t reserved
 *     uint16_t checksum
 * ctypedef icmpexthdr_s icmpexthdr             # <<<<<<<<<<<<<<
 * 
 * cdef struct icmpext_s:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_icmpexthdr_s __pyx_t_10cloudtrace_4read_5utils_icmpexthdr;

/* "cloudtrace/read/utils.pxd":107
 * ctypedef icmpexthdr_s icmpexthdr
 * 
 * cdef struct icmpext_s:             # <<<<<<<<<<<<<<
 *     uint16_t length
 *     uint8_t classnum
 */
struct __pyx_t_10cloudtrace_4read_5utils_icmpext_s {
  uint16_t length;
  uint8_t classnum;
  uint8_t ctype;
};

/* "cloudtrace/read/utils.pxd":111
 *     uint8_t classnum
 *     uint8_t ctype
 * ctypedef icmpext_s icmpext             # <<<<<<<<<<<<<<
 * 
 * cdef struct ip6_s:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_icmpext_s __pyx_t_10cloudtrace_4read_5utils_icmpext;

/* "cloudtrace/read/utils.pxd":113
 * ctypedef icmpext_s icmpext
 * 
 * cdef struct ip6_s:             # <<<<<<<<<<<<<<
 *     ip6_un ip6_ctlun
 *     uint8_t ip6_src[16]
 */
struct __pyx_t_10cloudtrace_4read_5utils_ip6_s {
  union __pyx_t_10cloudtrace_4read_5utils_ip6_un ip6_ctlun;
  uint8_t ip6_src[16];
  uint8_t ip6_dst[16];
};
struct __pyx_t_10cloudtrace_4read_7combine_hop;

/* "cloudtrace/read/combine.pxd":4
 * from posix.time cimport timeval
 * 
 * cdef struct hop:             # <<<<<<<<<<<<<<
 *     char hop_addr[16]
 *     uint8_t hop_flags
 */
struct __pyx_t_10cloudtrace_4read_7combine_hop {
  char hop_addr[16];
  uint8_t hop_flags;
  uint8_t hop_probe_ttl;
  uint16_t hop_probe_id;
  uint16_t hop_probe_size;
  uint8_t hop_reply_ttl;
  uint8_t hop_reply_tos;
  uint16_t hop_reply_size;
  uint16_t hop_reply_ipid;
  uint8_t hop_icmp_type;
  uint8_t hop_icmp_code;
  uint8_t hop_icmp_q_ttl;
  uint8_t hop_icmp_q_tos;
  uint16_t hop_icmp_q_ipl;
  uint16_t hop_icmp_nhmtu;
  struct timeval hop_tx;
  struct timeval hop_rtt;
};

/* "cloudtrace/read/combine.pxd":22
 *     timeval hop_tx
 *     timeval hop_rtt
 * ctypedef hop hop_t             # <<<<<<<<<<<<<<
 * 
 * cdef class Trace:
 */
typedef struct __pyx_t_10cloudtrace_4read_7combine_hop __pyx_t_10cloudtrace_4read_7combine_hop_t;
struct __pyx_t_10cloudtrace_4read_4pcap_global_hdr_s;
struct __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_s;

/* "cloudtrace/read/pcap.pxd":4
 * 
 * # /[inserted by cython to avoid comment start]* Global header (http://v2.nat32.com/pcap.htm) *[inserted by cython to avoid comment closer]/
 * cdef struct global_hdr_s:             # <<<<<<<<<<<<<<
 *     uint32_t magic_number   #/[inserted by cython to avoid comment start]* magic number *[inserted by cython to avoid comment closer]/
 *     uint16_t version_major  #/[inserted by cython to avoid comment start]* major version number *[inserted by cython to avoid comment closer]/
 */
struct __pyx_t_10cloudtrace_4read_4pcap_global_hdr_s {
  uint32_t magic_number;
  uint16_t version_major;
  uint16_t version_minor;
  int32_t thiszone;
  uint32_t sigfigs;
  uint32_t snaplen;
  uint32_t network;
};

/* "cloudtrace/read/pcap.pxd":12
 *     uint32_t snaplen        #/[inserted by cython to avoid comment start]* max length of captured packets, in octets *[inserted by cython to avoid comment closer]/
 *     uint32_t network        #/[inserted by cython to avoid comment start]* data link type *[inserted by cython to avoid comment closer]/
 * ctypedef global_hdr_s global_hdr_t             # <<<<<<<<<<<<<<
 * 
 * # /[inserted by cython to avoid comment start]* Packet header (http://v2.nat32.com/pcap.htm) *[inserted by cython to avoid comment closer]/
 */
typedef struct __pyx_t_10cloudtrace_4read_4pcap_global_hdr_s __pyx_t_10cloudtrace_4read_4pcap_global_hdr_t;

/* "cloudtrace/read/pcap.pxd":15
 * 
 * # /[inserted by cython to avoid comment start]* Packet header (http://v2.nat32.com/pcap.htm) *[inserted by cython to avoid comment closer]/
 * cdef struct packet_hdr_s:             # <<<<<<<<<<<<<<
 *     uint32_t ts_sec         #/[inserted by cython to avoid comment start]* timestamp seconds *[inserted by cython to avoid comment closer]/
 *     uint32_t ts_usec        #/[inserted by cython to avoid comment start]* timestamp microseconds *[inserted by cython to avoid comment closer]/
 */
struct __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_s {
  uint32_t ts_sec;
  uint32_t ts_usec;
  uint32_t incl_len;
  uint32_t orig_len;
};

/* "cloudtrace/read/pcap.pxd":20
 *     uint32_t incl_len       #/[inserted by cython to avoid comment start]* number of octets of packet saved in file *[inserted by cython to avoid comment closer]/
 *     uint32_t orig_len       #/[inserted by cython to avoid comment start]* actual length of packet *[inserted by cython to avoid comment closer]/
 * ctypedef packet_hdr_s packet_hdr_t             # <<<<<<<<<<<<<<
 * 
 * cdef class Pcap:
 */
typedef struct __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_s __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_t;

/* "cloudtrace/read/combine.pxd":24
 * ctypedef hop hop_t
 * 
 * cdef class Trace:             # <<<<<<<<<<<<<<
 *     cdef:
 *         char src[16]
 */
struct __pyx_obj_10cloudtrace_4read_7combine_Trace {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_4read_7combine_Trace *__pyx_vtab;
  char src[16];
  char dst[16];
  struct timeval start;
  __pyx_t_10cloudtrace_4read_7combine_hop_t **hops;
  uint16_t hop_count;
  uint16_t probec;
  uint8_t stop_reason;
  uint8_t stop_data;
  uint8_t type;
  uint8_t flags;
  uint8_t attempts;
  uint8_t hoplimit;
  uint8_t gaplimit;
  uint8_t gapaction;
  uint8_t firsthop;
  uint8_t tos;
  uint8_t wait;
  uint8_t wait_probe;
  uint8_t loops;
  uint8_t loopaction;
  uint8_t confidence;
  uint16_t probe_size;
  uint16_t sport;
  uint16_t dport;
  uint16_t offset;
};


/* "cloudtrace/read/pcap.pxd":22
 * ctypedef packet_hdr_s packet_hdr_t
 * 
 * cdef class Pcap:             # <<<<<<<<<<<<<<
 *     cdef:
 *         global_hdr_t *global_hdr
 */
struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap *__pyx_vtab;
  __pyx_t_10cloudtrace_4read_4pcap_global_hdr_t *global_hdr;
  __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_t *packet_hdr;
  char *frame;
  PyObject *infile;
  PyObject *proc;
};


/* "cloudtrace/read/packet.pxd":10
 *     public uint8_t ICMP_PROBE, UDP_PROBE, ECHO_REPLY, ICMP_REPLY
 * 
 * cdef class Packet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public uint8_t ptype
 */
struct __pyx_obj_10cloudtrace_4read_6packet_Packet {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *__pyx_vtab;
  uint8_t ptype;
  int isprobe;
  __pyx_t_10cloudtrace_4read_5utils_ip *iphdr;
  PyObject *dst;
  struct timeval *time;
  PyObject *pid_b;
  PyObject *dst_b;
};


/* "cloudtrace/read/packet.pxd":26
 *     cpdef bytes get_dst(self);
 * 
 * cdef class ICMPProbe(Packet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         icmp *icmphdr
 */
struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet __pyx_base;
  __pyx_t_10cloudtrace_4read_5utils_icmp *icmphdr;
};


/* "cloudtrace/read/packet.pxd":34
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr);
 * 
 * cdef class UDPProbe(Packet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         udp *udphdr
 */
struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet __pyx_base;
  __pyx_t_10cloudtrace_4read_5utils_udp *udphdr;
};


/* "cloudtrace/read/packet.pxd":40
 *     cdef void set_udp(self, ip *iphdr, udp *hdr);
 * 
 * cdef class Reply(Packet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         icmp *icmphdr
 */
struct __pyx_obj_10cloudtrace_4read_6packet_Reply {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet __pyx_base;
  __pyx_t_10cloudtrace_4read_5utils_icmp *icmphdr;
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *orig_probe;
};


/* "cloudtrace/read/packet.pxd":56
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr);
 * 
 * cdef class EchoReply(Reply):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply {
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply __pyx_base;
};


/* "cloudtrace/read/packet.pxd":59
 *     pass
 * 
 * cdef class ICMPReply(Reply):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Packet probe
 */
struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply {
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply __pyx_base;
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *probe;
};



/* "cloudtrace/read/combine.pxd":24
 * ctypedef hop hop_t
 * 
 * cdef class Trace:             # <<<<<<<<<<<<<<
 *     cdef:
 *         char src[16]
 */

struct __pyx_vtabstruct_10cloudtrace_4read_7combine_Trace {
  __pyx_t_10cloudtrace_4read_7combine_hop_t *(*new_hop)(struct __pyx_obj_10cloudtrace_4read_7combine_Trace *, uint8_t);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_7combine_Trace *__pyx_vtabptr_10cloudtrace_4read_7combine_Trace;


/* "cloudtrace/read/pcap.pxd":22
 * ctypedef packet_hdr_s packet_hdr_t
 * 
 * cdef class Pcap:             # <<<<<<<<<<<<<<
 *     cdef:
 *         global_hdr_t *global_hdr
 */

struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap {
  void (*close)(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch);
  int (*next_packet)(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch);
  void (*read_all)(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap *__pyx_vtabptr_10cloudtrace_4read_4pcap_Pcap;


/* "cloudtrace/read/packet.pyx":10
 *     public int ICMP_REPLY = 4
 * 
 * cdef class Packet:             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.iphdr = <ip *>calloc(1, sizeof(ip))
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet {
  uint8_t *(*get_dst_c)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *);
  uint8_t (*probe_ttl)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
  uint16_t (*get_pid)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
  uint16_t (*ipid)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
  void (*set_ip)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, __pyx_t_10cloudtrace_4read_5utils_ip *);
  PyObject *(*get_dst)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *__pyx_vtabptr_10cloudtrace_4read_6packet_Packet;


/* "cloudtrace/read/packet.pyx":45
 *         return <bytes> dst
 * 
 * cdef class ICMPProbe(Packet):             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.ptype = ICMP_PROBE
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPProbe {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet __pyx_base;
  uint8_t (*get_type)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *, int __pyx_skip_dispatch);
  uint8_t (*get_code)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *, int __pyx_skip_dispatch);
  void (*set_icmp)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_icmp *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPProbe *__pyx_vtabptr_10cloudtrace_4read_6packet_ICMPProbe;


/* "cloudtrace/read/packet.pyx":75
 *         self.dst = <bytes>dst
 * 
 * cdef class UDPProbe(Packet):             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.ptype = UDP_PROBE
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_UDPProbe {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet __pyx_base;
  void (*set_udp)(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_udp *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_UDPProbe *__pyx_vtabptr_10cloudtrace_4read_6packet_UDPProbe;


/* "cloudtrace/read/packet.pyx":99
 *         self.dst = <bytes>dst
 * 
 * cdef class Reply(Packet):             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.isprobe = False
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet __pyx_base;
  void (*_create_hop)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_7combine_hop_t *);
  PyObject *(*addr)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  void (*create_hop)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_7combine_hop_t *);
  uint16_t (*probe_ipid)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint16_t (*probe_size)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint8_t (*reply_ttl)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint8_t (*get_type)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint8_t (*get_code)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint8_t (*get_icmp_q_ttl)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  void (*set_icmp)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_icmp *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply *__pyx_vtabptr_10cloudtrace_4read_6packet_Reply;


/* "cloudtrace/read/packet.pyx":158
 *         self._create_hop(th)
 * 
 * cdef class EchoReply(Reply):             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.ptype = ECHO_REPLY
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_EchoReply {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply __pyx_base;
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_EchoReply *__pyx_vtabptr_10cloudtrace_4read_6packet_EchoReply;


/* "cloudtrace/read/packet.pyx":184
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 * 
 * cdef class ICMPReply(Reply):             # <<<<<<<<<<<<<<
 *     def __cinit__(self):
 *         self.ptype = ICMP_REPLY
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPReply {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply __pyx_base;
  void (*create_pseudo)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *, __pyx_t_10cloudtrace_4read_5utils_udppseudo *, unsigned char *);
  void (*set_probe)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPReply *__pyx_vtabptr_10cloudtrace_4read_6packet_ICMPReply;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    return __Pyx_decode_c_bytes(
        PyBytes_AS_STRING(string), PyBytes_GET_SIZE(string),
        start, stop, encoding, errors, decode_func);
}

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

struct __pyx_t_10cloudtrace_4read_5utils_ip_s;
static PyObject* __pyx_convert__to_py_struct____pyx_t_10cloudtrace_4read_5utils_ip_s(struct __pyx_t_10cloudtrace_4read_5utils_ip_s s);
/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static uint8_t *__pyx_f_10cloudtrace_4read_6packet_6Packet_get_dst_c(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_6Packet_set_ip(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_6Packet_probe_ttl(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_6Packet_get_pid(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_6Packet_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10cloudtrace_4read_6packet_6Packet_get_dst(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_type(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_code(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_set_icmp(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr, __pyx_t_10cloudtrace_4read_5utils_icmp *__pyx_v_hdr); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_8UDPProbe_probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_8UDPProbe_get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_8UDPProbe_set_udp(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr, __pyx_t_10cloudtrace_4read_5utils_udp *__pyx_v_hdr); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_5Reply__create_hop(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, CYTHON_UNUSED __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_th); /* proto*/
static PyObject *__pyx_f_10cloudtrace_4read_6packet_5Reply_addr(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_5Reply_probe_ipid(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_5Reply_probe_size(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_5Reply_reply_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_5Reply_get_type(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_5Reply_get_code(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_5Reply_get_icmp_q_ttl(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_5Reply_set_icmp(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr, __pyx_t_10cloudtrace_4read_5utils_icmp *__pyx_v_hdr); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_5Reply_create_hop(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_th); /* proto*/
static uint8_t *__pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_dst_c(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_icmp_q_ttl(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_9EchoReply_set_icmp(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr, __pyx_t_10cloudtrace_4read_5utils_icmp *__pyx_v_hdr); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_9ICMPReply__create_hop(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_th); /* proto*/
static uint8_t *__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_dst_c(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_create_pseudo(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_udppseudo *__pyx_v_p, unsigned char *__pyx_v_src); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_icmp_q_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_set_probe(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_probe, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'posix.types' */

/* Module declarations from 'posix.signal' */

/* Module declarations from 'posix.time' */

/* Module declarations from 'cloudtrace.read.utils' */
static uint16_t (*__pyx_f_10cloudtrace_4read_5utils_checksum)(void *, int); /*proto*/
static void (*__pyx_f_10cloudtrace_4read_5utils_iptohost)(__pyx_t_10cloudtrace_4read_5utils_ip *); /*proto*/
static void (*__pyx_f_10cloudtrace_4read_5utils_icmptohost)(__pyx_t_10cloudtrace_4read_5utils_icmp *); /*proto*/
static void (*__pyx_f_10cloudtrace_4read_5utils_udptohost)(__pyx_t_10cloudtrace_4read_5utils_udp *); /*proto*/

/* Module declarations from 'cloudtrace.read.combine' */
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_7combine_Trace = 0;

/* Module declarations from 'cloudtrace.read.pcap' */
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_4pcap_Pcap = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cloudtrace.read.packet' */
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_Packet = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_ICMPProbe = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_UDPProbe = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_Reply = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_EchoReply = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_ICMPReply = 0;
__PYX_EXTERN_C DL_EXPORT(uint8_t) ICMP_PROBE;
__PYX_EXTERN_C DL_EXPORT(uint8_t) UDP_PROBE;
__PYX_EXTERN_C DL_EXPORT(uint8_t) ECHO_REPLY;
__PYX_EXTERN_C DL_EXPORT(uint8_t) ICMP_REPLY;
static struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_f_10cloudtrace_4read_6packet_disassemble_packet(char *, int, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_f_10cloudtrace_4read_6packet_disassemble(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "cloudtrace.read.packet"
extern int __pyx_module_is_main_cloudtrace__read__packet;
int __pyx_module_is_main_cloudtrace__read__packet = 0;

/* Implementation of 'cloudtrace.read.packet' */
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_TypeError;
static const char __pyx_k_id[] = "id";
static const char __pyx_k_addr[] = "addr";
static const char __pyx_k_ip_p[] = "ip_p";
static const char __pyx_k_ipid[] = "ipid";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_Reply[] = "Reply";
static const char __pyx_k_ip_hl[] = "ip_hl";
static const char __pyx_k_Packet[] = "Packet";
static const char __pyx_k_ip_dst[] = "ip_dst";
static const char __pyx_k_ip_len[] = "ip_len";
static const char __pyx_k_ip_off[] = "ip_off";
static const char __pyx_k_ip_src[] = "ip_src";
static const char __pyx_k_ip_sum[] = "ip_sum";
static const char __pyx_k_ip_tos[] = "ip_tos";
static const char __pyx_k_ip_ttl[] = "ip_ttl";
static const char __pyx_k_packet[] = "packet";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_get_dst[] = "get_dst";
static const char __pyx_k_get_pid[] = "get_pid";
static const char __pyx_k_UDPProbe[] = "UDPProbe";
static const char __pyx_k_get_code[] = "get_code";
static const char __pyx_k_get_type[] = "get_type";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_incl_len[] = "incl_len";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_EchoReply[] = "EchoReply";
static const char __pyx_k_ICMPProbe[] = "ICMPProbe";
static const char __pyx_k_ICMPReply[] = "ICMPReply";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_UDP_PROBE[] = "UDP_PROBE";
static const char __pyx_k_probe_ttl[] = "probe_ttl";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_reply_ttl[] = "reply_ttl";
static const char __pyx_k_set_probe[] = "set_probe";
static const char __pyx_k_ECHO_REPLY[] = "ECHO_REPLY";
static const char __pyx_k_ICMP_PROBE[] = "ICMP_PROBE";
static const char __pyx_k_ICMP_REPLY[] = "ICMP_REPLY";
static const char __pyx_k_probe_ipid[] = "probe_ipid";
static const char __pyx_k_probe_size[] = "probe_size";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_get_icmp_q_ttl[] = "get_icmp_q_ttl";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_Probe_is_not_set_yet[] = "Probe is not set yet.";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_n_s_ECHO_REPLY;
static PyObject *__pyx_n_s_EchoReply;
static PyObject *__pyx_n_s_ICMPProbe;
static PyObject *__pyx_n_s_ICMPReply;
static PyObject *__pyx_n_s_ICMP_PROBE;
static PyObject *__pyx_n_s_ICMP_REPLY;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_Packet;
static PyObject *__pyx_kp_u_Probe_is_not_set_yet;
static PyObject *__pyx_n_s_Reply;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UDPProbe;
static PyObject *__pyx_n_s_UDP_PROBE;
static PyObject *__pyx_n_s_addr;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_get_code;
static PyObject *__pyx_n_s_get_dst;
static PyObject *__pyx_n_s_get_icmp_q_ttl;
static PyObject *__pyx_n_s_get_pid;
static PyObject *__pyx_n_s_get_type;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_incl_len;
static PyObject *__pyx_n_s_ip_dst;
static PyObject *__pyx_n_s_ip_hl;
static PyObject *__pyx_n_s_ip_len;
static PyObject *__pyx_n_s_ip_off;
static PyObject *__pyx_n_s_ip_p;
static PyObject *__pyx_n_s_ip_src;
static PyObject *__pyx_n_s_ip_sum;
static PyObject *__pyx_n_s_ip_tos;
static PyObject *__pyx_n_s_ip_ttl;
static PyObject *__pyx_n_s_ipid;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_packet;
static PyObject *__pyx_n_s_probe_ipid;
static PyObject *__pyx_n_s_probe_size;
static PyObject *__pyx_n_s_probe_ttl;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_reply_ttl;
static PyObject *__pyx_n_s_set_probe;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_test;
static int __pyx_pf_10cloudtrace_4read_6packet_6Packet___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static void __pyx_pf_10cloudtrace_4read_6packet_6Packet_2__dealloc__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_4get_hdr(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_6probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_8get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_10ipid(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_12get_dst(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_5ptype___get__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_6packet_6Packet_5ptype_2__set__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_7isprobe___get__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_6packet_6Packet_7isprobe_2__set__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_3dst___get__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_6packet_6Packet_3dst_2__set__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10cloudtrace_4read_6packet_6Packet_3dst_4__del__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self); /* proto */
static void __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_2__dealloc__(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_4probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_6get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_8get_type(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_10get_code(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self); /* proto */
static void __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_2__dealloc__(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_4probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_6get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10cloudtrace_4read_6packet_5Reply___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static void __pyx_pf_10cloudtrace_4read_6packet_5Reply_2__dealloc__(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_4addr(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_6probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_8probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_10reply_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_12get_type(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_14get_code(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_16get_icmp_q_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10cloudtrace_4read_6packet_9EchoReply___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_2probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_4get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_6probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_8probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_10get_icmp_q_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self); /* proto */
static void __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_4probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_6probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_8get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_10probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_12get_icmp_q_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_14set_probe(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_probe); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_disassemble_packet(CYTHON_UNUSED PyObject *__pyx_self, char *__pyx_v_packet, int __pyx_v_incl_len); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_6packet_2disassemble(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *__pyx_v_pcap); /* proto */
static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_Packet(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_ICMPProbe(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_UDPProbe(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_Reply(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_EchoReply(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_ICMPReply(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
/* Late includes */
uint8_t ICMP_PROBE;
uint8_t UDP_PROBE;
uint8_t ECHO_REPLY;
uint8_t ICMP_REPLY;

/* "cloudtrace/read/packet.pyx":11
 * 
 * cdef class Packet:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.iphdr = <ip *>calloc(1, sizeof(ip))
 *         self.time = <timeval *>calloc(1, sizeof(timeval))
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet___cinit__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_6Packet___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/read/packet.pyx":12
 * cdef class Packet:
 *     def __cinit__(self):
 *         self.iphdr = <ip *>calloc(1, sizeof(ip))             # <<<<<<<<<<<<<<
 *         self.time = <timeval *>calloc(1, sizeof(timeval))
 * 
 */
  __pyx_v_self->iphdr = ((__pyx_t_10cloudtrace_4read_5utils_ip *)calloc(1, (sizeof(__pyx_t_10cloudtrace_4read_5utils_ip))));

  /* "cloudtrace/read/packet.pyx":13
 *     def __cinit__(self):
 *         self.iphdr = <ip *>calloc(1, sizeof(ip))
 *         self.time = <timeval *>calloc(1, sizeof(timeval))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->time = ((struct timeval *)calloc(1, (sizeof(struct timeval))));

  /* "cloudtrace/read/packet.pyx":11
 * 
 * cdef class Packet:
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.iphdr = <ip *>calloc(1, sizeof(ip))
 *         self.time = <timeval *>calloc(1, sizeof(timeval))
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":15
 *         self.time = <timeval *>calloc(1, sizeof(timeval))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.iphdr is not NULL:
 *             free(self.iphdr)
 */

/* Python wrapper */
static void __pyx_pw_10cloudtrace_4read_6packet_6Packet_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10cloudtrace_4read_6packet_6Packet_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10cloudtrace_4read_6packet_6Packet_2__dealloc__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10cloudtrace_4read_6packet_6Packet_2__dealloc__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "cloudtrace/read/packet.pyx":16
 * 
 *     def __dealloc__(self):
 *         if self.iphdr is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.iphdr)
 *         if self.time is not NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->iphdr != NULL) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/packet.pyx":17
 *     def __dealloc__(self):
 *         if self.iphdr is not NULL:
 *             free(self.iphdr)             # <<<<<<<<<<<<<<
 *         if self.time is not NULL:
 *             free(self.time)
 */
    free(__pyx_v_self->iphdr);

    /* "cloudtrace/read/packet.pyx":16
 * 
 *     def __dealloc__(self):
 *         if self.iphdr is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.iphdr)
 *         if self.time is not NULL:
 */
  }

  /* "cloudtrace/read/packet.pyx":18
 *         if self.iphdr is not NULL:
 *             free(self.iphdr)
 *         if self.time is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.time)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->time != NULL) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/packet.pyx":19
 *             free(self.iphdr)
 *         if self.time is not NULL:
 *             free(self.time)             # <<<<<<<<<<<<<<
 * 
 *     cdef uint8_t *get_dst_c(self):
 */
    free(__pyx_v_self->time);

    /* "cloudtrace/read/packet.pyx":18
 *         if self.iphdr is not NULL:
 *             free(self.iphdr)
 *         if self.time is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.time)
 * 
 */
  }

  /* "cloudtrace/read/packet.pyx":15
 *         self.time = <timeval *>calloc(1, sizeof(timeval))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.iphdr is not NULL:
 *             free(self.iphdr)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":21
 *             free(self.time)
 * 
 *     cdef uint8_t *get_dst_c(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_dst
 * 
 */

static uint8_t *__pyx_f_10cloudtrace_4read_6packet_6Packet_get_dst_c(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  uint8_t *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_dst_c", 0);

  /* "cloudtrace/read/packet.pyx":22
 * 
 *     cdef uint8_t *get_dst_c(self):
 *         return self.iphdr.ip_dst             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_ip(self, ip *iphdr):
 */
  __pyx_r = __pyx_v_self->iphdr->ip_dst;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":21
 *             free(self.time)
 * 
 *     cdef uint8_t *get_dst_c(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_dst
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":24
 *         return self.iphdr.ip_dst
 * 
 *     cdef void set_ip(self, ip *iphdr):             # <<<<<<<<<<<<<<
 *         memcpy(self.iphdr, iphdr, sizeof(ip))
 * 
 */

static void __pyx_f_10cloudtrace_4read_6packet_6Packet_set_ip(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_ip", 0);

  /* "cloudtrace/read/packet.pyx":25
 * 
 *     cdef void set_ip(self, ip *iphdr):
 *         memcpy(self.iphdr, iphdr, sizeof(ip))             # <<<<<<<<<<<<<<
 * 
 *     def get_hdr(self):
 */
  (void)(memcpy(__pyx_v_self->iphdr, __pyx_v_iphdr, (sizeof(__pyx_t_10cloudtrace_4read_5utils_ip))));

  /* "cloudtrace/read/packet.pyx":24
 *         return self.iphdr.ip_dst
 * 
 *     cdef void set_ip(self, ip *iphdr):             # <<<<<<<<<<<<<<
 *         memcpy(self.iphdr, iphdr, sizeof(ip))
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":27
 *         memcpy(self.iphdr, iphdr, sizeof(ip))
 * 
 *     def get_hdr(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr[0]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_5get_hdr(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_6Packet_4get_hdr[] = "Packet.get_hdr(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_5get_hdr(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_hdr (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_4get_hdr(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_4get_hdr(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_hdr", 0);

  /* "cloudtrace/read/packet.pyx":28
 * 
 *     def get_hdr(self):
 *         return self.iphdr[0]             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t probe_ttl(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_convert__to_py_struct____pyx_t_10cloudtrace_4read_5utils_ip_s((__pyx_v_self->iphdr[0])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":27
 *         memcpy(self.iphdr, iphdr, sizeof(ip))
 * 
 *     def get_hdr(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr[0]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.get_hdr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":30
 *         return self.iphdr[0]
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_7probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_6Packet_probe_ttl(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_6Packet_7probe_ttl)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 30, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 30, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":31
 * 
 *     cpdef uint8_t probe_ttl(self):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t get_pid(self):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 31, __pyx_L1_error)

  /* "cloudtrace/read/packet.pyx":30
 *         return self.iphdr[0]
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.Packet.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_7probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_6Packet_6probe_ttl[] = "Packet.probe_ttl(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_7probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_ttl (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_6probe_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_6probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_6Packet_probe_ttl(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":33
 *         raise NotImplementedError()
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_9get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_6Packet_get_pid(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_pid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_6Packet_9get_pid)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 33, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":34
 * 
 *     cpdef uint16_t get_pid(self):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t ipid(self):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 34, __pyx_L1_error)

  /* "cloudtrace/read/packet.pyx":33
 *         raise NotImplementedError()
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.Packet.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_9get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_6Packet_8get_pid[] = "Packet.get_pid(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_9get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_pid (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_8get_pid(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_8get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_6Packet_get_pid(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":36
 *         raise NotImplementedError()
 * 
 *     cpdef uint16_t ipid(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.id
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_11ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_6Packet_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ipid", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ipid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_6Packet_11ipid)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 36, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":37
 * 
 *     cpdef uint16_t ipid(self):
 *         return self.iphdr.id             # <<<<<<<<<<<<<<
 * 
 *     cpdef bytes get_dst(self):
 */
  __pyx_r = __pyx_v_self->iphdr->id;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":36
 *         raise NotImplementedError()
 * 
 *     cpdef uint16_t ipid(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.id
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.Packet.ipid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_11ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_6Packet_10ipid[] = "Packet.ipid(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_11ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ipid (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_10ipid(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_10ipid(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ipid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_6Packet_ipid(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.ipid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":39
 *         return self.iphdr.id
 * 
 *     cpdef bytes get_dst(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char dst[16]
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_13get_dst(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_10cloudtrace_4read_6packet_6Packet_get_dst(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, int __pyx_skip_dispatch) {
  char __pyx_v_dst[16];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_dst", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_dst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_6Packet_13get_dst)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 39, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":42
 *         cdef:
 *             char dst[16]
 *         inet_ntop(AF_INET, self.get_dst_c(), dst, INET_ADDRSTRLEN)             # <<<<<<<<<<<<<<
 *         return <bytes> dst
 * 
 */
  (void)(inet_ntop(AF_INET, ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *)__pyx_v_self->__pyx_vtab)->get_dst_c(__pyx_v_self), __pyx_v_dst, INET_ADDRSTRLEN));

  /* "cloudtrace/read/packet.pyx":43
 *             char dst[16]
 *         inet_ntop(AF_INET, self.get_dst_c(), dst, INET_ADDRSTRLEN)
 *         return <bytes> dst             # <<<<<<<<<<<<<<
 * 
 * cdef class ICMPProbe(Packet):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_FromString(__pyx_v_dst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":39
 *         return self.iphdr.id
 * 
 *     cpdef bytes get_dst(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char dst[16]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.get_dst", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_13get_dst(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_6Packet_12get_dst[] = "Packet.get_dst(self) -> bytes";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_13get_dst(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_dst (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_12get_dst(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_12get_dst(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_dst", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10cloudtrace_4read_6packet_6Packet_get_dst(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.get_dst", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pxd":12
 * cdef class Packet:
 *     cdef:
 *         public uint8_t ptype             # <<<<<<<<<<<<<<
 *         public bint isprobe
 *         ip *iphdr
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_5ptype_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_5ptype_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_5ptype___get__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_5ptype___get__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_v_self->ptype); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.ptype.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_5ptype_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_5ptype_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_5ptype_2__set__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_6Packet_5ptype_2__set__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  uint8_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_uint8_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __pyx_v_self->ptype = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.ptype.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pxd":13
 *     cdef:
 *         public uint8_t ptype
 *         public bint isprobe             # <<<<<<<<<<<<<<
 *         ip *iphdr
 *         public bytes dst
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_7isprobe_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_7isprobe_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_7isprobe___get__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_7isprobe___get__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->isprobe); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.isprobe.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_7isprobe_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_7isprobe_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_7isprobe_2__set__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_6Packet_7isprobe_2__set__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_v_self->isprobe = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.isprobe.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pxd":15
 *         public bint isprobe
 *         ip *iphdr
 *         public bytes dst             # <<<<<<<<<<<<<<
 *         timeval *time
 *         bytes pid_b, dst_b
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_3dst_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_3dst_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_3dst___get__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_3dst___get__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->dst);
  __pyx_r = __pyx_v_self->dst;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_3dst_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_3dst_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_3dst_2__set__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_6Packet_3dst_2__set__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyBytes_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 15, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->dst);
  __Pyx_DECREF(__pyx_v_self->dst);
  __pyx_v_self->dst = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.dst.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_3dst_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_6Packet_3dst_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_3dst_4__del__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_6Packet_3dst_4__del__(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->dst);
  __Pyx_DECREF(__pyx_v_self->dst);
  __pyx_v_self->dst = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_6Packet_14__reduce_cython__[] = "Packet.__reduce_cython__(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_14__reduce_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_6Packet_16__setstate_cython__[] = "Packet.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_6Packet_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_6Packet_16__setstate_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_6Packet_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Packet.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":46
 * 
 * cdef class ICMPProbe(Packet):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.ptype = ICMP_PROBE
 *         self.isprobe = True
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe___cinit__(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/read/packet.pyx":47
 * cdef class ICMPProbe(Packet):
 *     def __cinit__(self):
 *         self.ptype = ICMP_PROBE             # <<<<<<<<<<<<<<
 *         self.isprobe = True
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))
 */
  __pyx_v_self->__pyx_base.ptype = ICMP_PROBE;

  /* "cloudtrace/read/packet.pyx":48
 *     def __cinit__(self):
 *         self.ptype = ICMP_PROBE
 *         self.isprobe = True             # <<<<<<<<<<<<<<
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))
 * 
 */
  __pyx_v_self->__pyx_base.isprobe = 1;

  /* "cloudtrace/read/packet.pyx":49
 *         self.ptype = ICMP_PROBE
 *         self.isprobe = True
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->icmphdr = ((__pyx_t_10cloudtrace_4read_5utils_icmp *)calloc(1, (sizeof(__pyx_t_10cloudtrace_4read_5utils_icmp))));

  /* "cloudtrace/read/packet.pyx":46
 * 
 * cdef class ICMPProbe(Packet):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.ptype = ICMP_PROBE
 *         self.isprobe = True
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":51
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.icmphdr is not NULL:
 *             free(self.icmphdr)
 */

/* Python wrapper */
static void __pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_2__dealloc__(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_2__dealloc__(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "cloudtrace/read/packet.pyx":52
 * 
 *     def __dealloc__(self):
 *         if self.icmphdr is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.icmphdr)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->icmphdr != NULL) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/packet.pyx":53
 *     def __dealloc__(self):
 *         if self.icmphdr is not NULL:
 *             free(self.icmphdr)             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t probe_ttl(self):
 */
    free(__pyx_v_self->icmphdr);

    /* "cloudtrace/read/packet.pyx":52
 * 
 *     def __dealloc__(self):
 *         if self.icmphdr is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.icmphdr)
 * 
 */
  }

  /* "cloudtrace/read/packet.pyx":51
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.icmphdr is not NULL:
 *             free(self.icmphdr)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":55
 *             free(self.icmphdr)
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.un.echo.seq
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_5probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_5probe_ttl)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 55, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":56
 * 
 *     cpdef uint8_t probe_ttl(self):
 *         return self.icmphdr.un.echo.seq             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t get_pid(self):
 */
  __pyx_r = __pyx_v_self->icmphdr->un.echo.seq;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":55
 *             free(self.icmphdr)
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.un.echo.seq
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPProbe.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_5probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_4probe_ttl[] = "ICMPProbe.probe_ttl(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_5probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_ttl (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_4probe_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_4probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_probe_ttl(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPProbe.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":58
 *         return self.icmphdr.un.echo.seq
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.un.echo.id
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_7get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_pid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_7get_pid)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":59
 * 
 *     cpdef uint16_t get_pid(self):
 *         return self.icmphdr.un.echo.id             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t get_type(self):
 */
  __pyx_r = __pyx_v_self->icmphdr->un.echo.id;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":58
 *         return self.icmphdr.un.echo.seq
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.un.echo.id
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPProbe.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_7get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_6get_pid[] = "ICMPProbe.get_pid(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_7get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_pid (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_6get_pid(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_6get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_pid(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPProbe.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":61
 *         return self.icmphdr.un.echo.id
 * 
 *     cpdef uint8_t get_type(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.type
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_9get_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_type(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_type", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_9get_type)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":62
 * 
 *     cpdef uint8_t get_type(self):
 *         return self.icmphdr.type             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t get_code(self):
 */
  __pyx_r = __pyx_v_self->icmphdr->type;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":61
 *         return self.icmphdr.un.echo.id
 * 
 *     cpdef uint8_t get_type(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.type
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPProbe.get_type", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_9get_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_8get_type[] = "ICMPProbe.get_type(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_9get_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_type (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_8get_type(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_8get_type(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_type", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_type(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPProbe.get_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":64
 *         return self.icmphdr.type
 * 
 *     cpdef uint8_t get_code(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.code
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_11get_code(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_code(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_code", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_11get_code)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":65
 * 
 *     cpdef uint8_t get_code(self):
 *         return self.icmphdr.code             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):
 */
  __pyx_r = __pyx_v_self->icmphdr->code;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":64
 *         return self.icmphdr.type
 * 
 *     cpdef uint8_t get_code(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.code
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPProbe.get_code", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_11get_code(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_10get_code[] = "ICMPProbe.get_code(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_11get_code(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_code (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_10get_code(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_10get_code(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_code", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_code(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPProbe.get_code", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":67
 *         return self.icmphdr.code
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char dst[16]
 */

static void __pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_set_icmp(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr, __pyx_t_10cloudtrace_4read_5utils_icmp *__pyx_v_hdr) {
  char __pyx_v_dst[16];
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_icmp", 0);

  /* "cloudtrace/read/packet.pyx":70
 *         cdef:
 *             char dst[16]
 *         self.set_ip(iphdr)             # <<<<<<<<<<<<<<
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 *         inet_ntop(AF_INET, self.iphdr.ip_dst, dst, INET_ADDRSTRLEN)
 */
  ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.set_ip(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self), __pyx_v_iphdr);

  /* "cloudtrace/read/packet.pyx":71
 *             char dst[16]
 *         self.set_ip(iphdr)
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))             # <<<<<<<<<<<<<<
 *         inet_ntop(AF_INET, self.iphdr.ip_dst, dst, INET_ADDRSTRLEN)
 *         self.dst = <bytes>dst
 */
  (void)(memcpy(__pyx_v_self->icmphdr, __pyx_v_hdr, (sizeof(__pyx_t_10cloudtrace_4read_5utils_icmp))));

  /* "cloudtrace/read/packet.pyx":72
 *         self.set_ip(iphdr)
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 *         inet_ntop(AF_INET, self.iphdr.ip_dst, dst, INET_ADDRSTRLEN)             # <<<<<<<<<<<<<<
 *         self.dst = <bytes>dst
 * 
 */
  (void)(inet_ntop(AF_INET, __pyx_v_self->__pyx_base.iphdr->ip_dst, __pyx_v_dst, INET_ADDRSTRLEN));

  /* "cloudtrace/read/packet.pyx":73
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 *         inet_ntop(AF_INET, self.iphdr.ip_dst, dst, INET_ADDRSTRLEN)
 *         self.dst = <bytes>dst             # <<<<<<<<<<<<<<
 * 
 * cdef class UDPProbe(Packet):
 */
  __pyx_t_1 = __Pyx_PyObject_FromString(__pyx_v_dst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.dst);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.dst);
  __pyx_v_self->__pyx_base.dst = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "cloudtrace/read/packet.pyx":67
 *         return self.icmphdr.code
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char dst[16]
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPProbe.set_icmp", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_12__reduce_cython__[] = "ICMPProbe.__reduce_cython__(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_12__reduce_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPProbe.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_14__setstate_cython__[] = "ICMPProbe.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_14__setstate_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPProbe_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPProbe.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":76
 * 
 * cdef class UDPProbe(Packet):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.ptype = UDP_PROBE
 *         self.isprobe = True
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe___cinit__(((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/read/packet.pyx":77
 * cdef class UDPProbe(Packet):
 *     def __cinit__(self):
 *         self.ptype = UDP_PROBE             # <<<<<<<<<<<<<<
 *         self.isprobe = True
 *         self.udphdr = <udp *>calloc(1, sizeof(udp))
 */
  __pyx_v_self->__pyx_base.ptype = UDP_PROBE;

  /* "cloudtrace/read/packet.pyx":78
 *     def __cinit__(self):
 *         self.ptype = UDP_PROBE
 *         self.isprobe = True             # <<<<<<<<<<<<<<
 *         self.udphdr = <udp *>calloc(1, sizeof(udp))
 * 
 */
  __pyx_v_self->__pyx_base.isprobe = 1;

  /* "cloudtrace/read/packet.pyx":79
 *         self.ptype = UDP_PROBE
 *         self.isprobe = True
 *         self.udphdr = <udp *>calloc(1, sizeof(udp))             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->udphdr = ((__pyx_t_10cloudtrace_4read_5utils_udp *)calloc(1, (sizeof(__pyx_t_10cloudtrace_4read_5utils_udp))));

  /* "cloudtrace/read/packet.pyx":76
 * 
 * cdef class UDPProbe(Packet):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.ptype = UDP_PROBE
 *         self.isprobe = True
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":81
 *         self.udphdr = <udp *>calloc(1, sizeof(udp))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.udphdr is not NULL:
 *             free(self.udphdr)
 */

/* Python wrapper */
static void __pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_2__dealloc__(((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_2__dealloc__(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "cloudtrace/read/packet.pyx":82
 * 
 *     def __dealloc__(self):
 *         if self.udphdr is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.udphdr)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->udphdr != NULL) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/packet.pyx":83
 *     def __dealloc__(self):
 *         if self.udphdr is not NULL:
 *             free(self.udphdr)             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t probe_ttl(self):
 */
    free(__pyx_v_self->udphdr);

    /* "cloudtrace/read/packet.pyx":82
 * 
 *     def __dealloc__(self):
 *         if self.udphdr is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.udphdr)
 * 
 */
  }

  /* "cloudtrace/read/packet.pyx":81
 *         self.udphdr = <udp *>calloc(1, sizeof(udp))
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.udphdr is not NULL:
 *             free(self.udphdr)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":85
 *             free(self.udphdr)
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.udphdr.check
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_5probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_8UDPProbe_probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_5probe_ttl)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 85, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":86
 * 
 *     cpdef uint8_t probe_ttl(self):
 *         return self.udphdr.check             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t get_pid(self):
 */
  __pyx_r = __pyx_v_self->udphdr->check;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":85
 *             free(self.udphdr)
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.udphdr.check
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.UDPProbe.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_5probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_8UDPProbe_4probe_ttl[] = "UDPProbe.probe_ttl(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_5probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_ttl (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_4probe_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_4probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_8UDPProbe_probe_ttl(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.UDPProbe.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":88
 *         return self.udphdr.check
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         return self.udphdr.source
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_7get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_8UDPProbe_get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_pid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_7get_pid)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 88, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 88, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":89
 * 
 *     cpdef uint16_t get_pid(self):
 *         return self.udphdr.source             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_udp(self, ip *iphdr, udp *hdr):
 */
  __pyx_r = __pyx_v_self->udphdr->source;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":88
 *         return self.udphdr.check
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         return self.udphdr.source
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.UDPProbe.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_7get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_8UDPProbe_6get_pid[] = "UDPProbe.get_pid(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_7get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_pid (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_6get_pid(((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_6get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_8UDPProbe_get_pid(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.UDPProbe.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":91
 *         return self.udphdr.source
 * 
 *     cdef void set_udp(self, ip *iphdr, udp *hdr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char dst[16]
 */

static void __pyx_f_10cloudtrace_4read_6packet_8UDPProbe_set_udp(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr, __pyx_t_10cloudtrace_4read_5utils_udp *__pyx_v_hdr) {
  char __pyx_v_dst[16];
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_udp", 0);

  /* "cloudtrace/read/packet.pyx":94
 *         cdef:
 *             char dst[16]
 *         self.set_ip(iphdr)             # <<<<<<<<<<<<<<
 *         memcpy(self.udphdr, hdr, sizeof(udp))
 *         inet_ntop(AF_INET, self.iphdr.ip_dst, dst, INET_ADDRSTRLEN)
 */
  ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.set_ip(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self), __pyx_v_iphdr);

  /* "cloudtrace/read/packet.pyx":95
 *             char dst[16]
 *         self.set_ip(iphdr)
 *         memcpy(self.udphdr, hdr, sizeof(udp))             # <<<<<<<<<<<<<<
 *         inet_ntop(AF_INET, self.iphdr.ip_dst, dst, INET_ADDRSTRLEN)
 *         self.dst = <bytes>dst
 */
  (void)(memcpy(__pyx_v_self->udphdr, __pyx_v_hdr, (sizeof(__pyx_t_10cloudtrace_4read_5utils_udp))));

  /* "cloudtrace/read/packet.pyx":96
 *         self.set_ip(iphdr)
 *         memcpy(self.udphdr, hdr, sizeof(udp))
 *         inet_ntop(AF_INET, self.iphdr.ip_dst, dst, INET_ADDRSTRLEN)             # <<<<<<<<<<<<<<
 *         self.dst = <bytes>dst
 * 
 */
  (void)(inet_ntop(AF_INET, __pyx_v_self->__pyx_base.iphdr->ip_dst, __pyx_v_dst, INET_ADDRSTRLEN));

  /* "cloudtrace/read/packet.pyx":97
 *         memcpy(self.udphdr, hdr, sizeof(udp))
 *         inet_ntop(AF_INET, self.iphdr.ip_dst, dst, INET_ADDRSTRLEN)
 *         self.dst = <bytes>dst             # <<<<<<<<<<<<<<
 * 
 * cdef class Reply(Packet):
 */
  __pyx_t_1 = __Pyx_PyObject_FromString(__pyx_v_dst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_t_1;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.dst);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.dst);
  __pyx_v_self->__pyx_base.dst = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "cloudtrace/read/packet.pyx":91
 *         return self.udphdr.source
 * 
 *     cdef void set_udp(self, ip *iphdr, udp *hdr):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char dst[16]
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.UDPProbe.set_udp", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_8UDPProbe_8__reduce_cython__[] = "UDPProbe.__reduce_cython__(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_8__reduce_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.UDPProbe.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_8UDPProbe_10__setstate_cython__[] = "UDPProbe.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_10__setstate_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_8UDPProbe_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.UDPProbe.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":100
 * 
 * cdef class Reply(Packet):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.isprobe = False
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_5Reply_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_5Reply_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply___cinit__(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_5Reply___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/read/packet.pyx":101
 * cdef class Reply(Packet):
 *     def __cinit__(self):
 *         self.isprobe = False             # <<<<<<<<<<<<<<
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))
 *         self.orig_probe = None
 */
  __pyx_v_self->__pyx_base.isprobe = 0;

  /* "cloudtrace/read/packet.pyx":102
 *     def __cinit__(self):
 *         self.isprobe = False
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))             # <<<<<<<<<<<<<<
 *         self.orig_probe = None
 * 
 */
  __pyx_v_self->icmphdr = ((__pyx_t_10cloudtrace_4read_5utils_icmp *)calloc(1, (sizeof(__pyx_t_10cloudtrace_4read_5utils_icmp))));

  /* "cloudtrace/read/packet.pyx":103
 *         self.isprobe = False
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))
 *         self.orig_probe = None             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->orig_probe);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->orig_probe));
  __pyx_v_self->orig_probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None);

  /* "cloudtrace/read/packet.pyx":100
 * 
 * cdef class Reply(Packet):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.isprobe = False
 *         self.icmphdr = <icmp *>calloc(1, sizeof(icmp))
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":105
 *         self.orig_probe = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.icmphdr is not NULL:
 *             free(self.icmphdr)
 */

/* Python wrapper */
static void __pyx_pw_10cloudtrace_4read_6packet_5Reply_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10cloudtrace_4read_6packet_5Reply_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10cloudtrace_4read_6packet_5Reply_2__dealloc__(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10cloudtrace_4read_6packet_5Reply_2__dealloc__(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "cloudtrace/read/packet.pyx":106
 * 
 *     def __dealloc__(self):
 *         if self.icmphdr is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.icmphdr)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->icmphdr != NULL) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/packet.pyx":107
 *     def __dealloc__(self):
 *         if self.icmphdr is not NULL:
 *             free(self.icmphdr)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _create_hop(self, hop_t *th):
 */
    free(__pyx_v_self->icmphdr);

    /* "cloudtrace/read/packet.pyx":106
 * 
 *     def __dealloc__(self):
 *         if self.icmphdr is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.icmphdr)
 * 
 */
  }

  /* "cloudtrace/read/packet.pyx":105
 *         self.orig_probe = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.icmphdr is not NULL:
 *             free(self.icmphdr)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":109
 *             free(self.icmphdr)
 * 
 *     cdef void _create_hop(self, hop_t *th):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static void __pyx_f_10cloudtrace_4read_6packet_5Reply__create_hop(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, CYTHON_UNUSED __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_th) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_hop", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":112
 *         pass
 * 
 *     cpdef str addr(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char dst[16]
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_5addr(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_10cloudtrace_4read_6packet_5Reply_addr(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch) {
  char __pyx_v_dst[16];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("addr", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_addr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_5Reply_5addr)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 112, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 112, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":115
 *         cdef:
 *             char dst[16]
 *         inet_ntop(AF_INET, self.iphdr.ip_src, dst, 16)             # <<<<<<<<<<<<<<
 *         return (<bytes>dst).decode()
 * 
 */
  (void)(inet_ntop(AF_INET, __pyx_v_self->__pyx_base.iphdr->ip_src, __pyx_v_dst, 16));

  /* "cloudtrace/read/packet.pyx":116
 *             char dst[16]
 *         inet_ntop(AF_INET, self.iphdr.ip_src, dst, 16)
 *         return (<bytes>dst).decode()             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t probe_ipid(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_FromString(__pyx_v_dst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 116, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_decode_bytes(((PyObject*)__pyx_t_1), 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":112
 *         pass
 * 
 *     cpdef str addr(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             char dst[16]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.addr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_5addr(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_5Reply_4addr[] = "Reply.addr(self) -> unicode";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_5addr(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("addr (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply_4addr(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_4addr(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("addr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10cloudtrace_4read_6packet_5Reply_addr(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.addr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":118
 *         return (<bytes>dst).decode()
 * 
 *     cpdef uint16_t probe_ipid(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_7probe_ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_5Reply_probe_ipid(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ipid", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_ipid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_5Reply_7probe_ipid)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 118, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":119
 * 
 *     cpdef uint16_t probe_ipid(self):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t probe_size(self):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 119, __pyx_L1_error)

  /* "cloudtrace/read/packet.pyx":118
 *         return (<bytes>dst).decode()
 * 
 *     cpdef uint16_t probe_ipid(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.Reply.probe_ipid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_7probe_ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_5Reply_6probe_ipid[] = "Reply.probe_ipid(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_7probe_ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_ipid (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply_6probe_ipid(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_6probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ipid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_5Reply_probe_ipid(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.probe_ipid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":121
 *         raise NotImplementedError()
 * 
 *     cpdef uint16_t probe_size(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_9probe_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_5Reply_probe_size(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_size", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_5Reply_9probe_size)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 121, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":122
 * 
 *     cpdef uint16_t probe_size(self):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t reply_ttl(self):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 122, __pyx_L1_error)

  /* "cloudtrace/read/packet.pyx":121
 *         raise NotImplementedError()
 * 
 *     cpdef uint16_t probe_size(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.Reply.probe_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_9probe_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_5Reply_8probe_size[] = "Reply.probe_size(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_9probe_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_size (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply_8probe_size(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_8probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_size", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_5Reply_probe_size(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.probe_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":124
 *         raise NotImplementedError()
 * 
 *     cpdef uint8_t reply_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_ttl
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_11reply_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_5Reply_reply_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reply_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reply_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_5Reply_11reply_ttl)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 124, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":125
 * 
 *     cpdef uint8_t reply_ttl(self):
 *         return self.iphdr.ip_ttl             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t get_type(self):
 */
  __pyx_r = __pyx_v_self->__pyx_base.iphdr->ip_ttl;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":124
 *         raise NotImplementedError()
 * 
 *     cpdef uint8_t reply_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_ttl
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.Reply.reply_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_11reply_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_5Reply_10reply_ttl[] = "Reply.reply_ttl(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_11reply_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reply_ttl (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply_10reply_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_10reply_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reply_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_5Reply_reply_ttl(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.reply_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":127
 *         return self.iphdr.ip_ttl
 * 
 *     cpdef uint8_t get_type(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.type
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_13get_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_5Reply_get_type(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_type", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_5Reply_13get_type)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 127, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":128
 * 
 *     cpdef uint8_t get_type(self):
 *         return self.icmphdr.type             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t get_code(self):
 */
  __pyx_r = __pyx_v_self->icmphdr->type;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":127
 *         return self.iphdr.ip_ttl
 * 
 *     cpdef uint8_t get_type(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.type
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.Reply.get_type", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_13get_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_5Reply_12get_type[] = "Reply.get_type(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_13get_type(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_type (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply_12get_type(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_12get_type(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_type", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_5Reply_get_type(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.get_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":130
 *         return self.icmphdr.type
 * 
 *     cpdef uint8_t get_code(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.code
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_15get_code(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_5Reply_get_code(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_code", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_5Reply_15get_code)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 130, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":131
 * 
 *     cpdef uint8_t get_code(self):
 *         return self.icmphdr.code             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):
 */
  __pyx_r = __pyx_v_self->icmphdr->code;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":130
 *         return self.icmphdr.type
 * 
 *     cpdef uint8_t get_code(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.code
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.Reply.get_code", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_15get_code(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_5Reply_14get_code[] = "Reply.get_code(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_15get_code(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_code (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply_14get_code(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_14get_code(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_code", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_5Reply_get_code(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.get_code", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":133
 *         return self.icmphdr.code
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_17get_icmp_q_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_5Reply_get_icmp_q_ttl(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_icmp_q_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_icmp_q_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_5Reply_17get_icmp_q_ttl)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 133, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":134
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 134, __pyx_L1_error)

  /* "cloudtrace/read/packet.pyx":133
 *         return self.icmphdr.code
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.Reply.get_icmp_q_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_17get_icmp_q_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_5Reply_16get_icmp_q_ttl[] = "Reply.get_icmp_q_ttl(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_17get_icmp_q_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_icmp_q_ttl (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply_16get_icmp_q_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_16get_icmp_q_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_icmp_q_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_5Reply_get_icmp_q_ttl(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.get_icmp_q_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":136
 *         raise NotImplementedError()
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):             # <<<<<<<<<<<<<<
 *         self.set_ip(iphdr)
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 */

static void __pyx_f_10cloudtrace_4read_6packet_5Reply_set_icmp(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr, __pyx_t_10cloudtrace_4read_5utils_icmp *__pyx_v_hdr) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_icmp", 0);

  /* "cloudtrace/read/packet.pyx":137
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):
 *         self.set_ip(iphdr)             # <<<<<<<<<<<<<<
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 * 
 */
  ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.set_ip(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self), __pyx_v_iphdr);

  /* "cloudtrace/read/packet.pyx":138
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):
 *         self.set_ip(iphdr)
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))             # <<<<<<<<<<<<<<
 * 
 *     cdef void create_hop(self, hop_t *th):
 */
  (void)(memcpy(__pyx_v_self->icmphdr, __pyx_v_hdr, (sizeof(__pyx_t_10cloudtrace_4read_5utils_icmp))));

  /* "cloudtrace/read/packet.pyx":136
 *         raise NotImplementedError()
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):             # <<<<<<<<<<<<<<
 *         self.set_ip(iphdr)
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":140
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 * 
 *     cdef void create_hop(self, hop_t *th):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Packet probe = self.orig_probe
 */

static void __pyx_f_10cloudtrace_4read_6packet_5Reply_create_hop(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_th) {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_probe = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint16_t __pyx_t_2;
  uint8_t __pyx_t_3;
  __Pyx_RefNannySetupContext("create_hop", 0);

  /* "cloudtrace/read/packet.pyx":142
 *     cdef void create_hop(self, hop_t *th):
 *         cdef:
 *             Packet probe = self.orig_probe             # <<<<<<<<<<<<<<
 *         memcpy(th.hop_addr, self.iphdr.ip_src, sizeof(th.hop_addr))
 *         th.hop_flags = 0
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->orig_probe);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/read/packet.pyx":143
 *         cdef:
 *             Packet probe = self.orig_probe
 *         memcpy(th.hop_addr, self.iphdr.ip_src, sizeof(th.hop_addr))             # <<<<<<<<<<<<<<
 *         th.hop_flags = 0
 *         th.hop_probe_id = probe.iphdr.id
 */
  (void)(memcpy(__pyx_v_th->hop_addr, __pyx_v_self->__pyx_base.iphdr->ip_src, (sizeof(__pyx_v_th->hop_addr))));

  /* "cloudtrace/read/packet.pyx":144
 *             Packet probe = self.orig_probe
 *         memcpy(th.hop_addr, self.iphdr.ip_src, sizeof(th.hop_addr))
 *         th.hop_flags = 0             # <<<<<<<<<<<<<<
 *         th.hop_probe_id = probe.iphdr.id
 *         th.hop_probe_ttl = probe.iphdr.ip_ttl
 */
  __pyx_v_th->hop_flags = 0;

  /* "cloudtrace/read/packet.pyx":145
 *         memcpy(th.hop_addr, self.iphdr.ip_src, sizeof(th.hop_addr))
 *         th.hop_flags = 0
 *         th.hop_probe_id = probe.iphdr.id             # <<<<<<<<<<<<<<
 *         th.hop_probe_ttl = probe.iphdr.ip_ttl
 *         th.hop_probe_size = probe.iphdr.ip_len
 */
  __pyx_t_2 = __pyx_v_probe->iphdr->id;
  __pyx_v_th->hop_probe_id = __pyx_t_2;

  /* "cloudtrace/read/packet.pyx":146
 *         th.hop_flags = 0
 *         th.hop_probe_id = probe.iphdr.id
 *         th.hop_probe_ttl = probe.iphdr.ip_ttl             # <<<<<<<<<<<<<<
 *         th.hop_probe_size = probe.iphdr.ip_len
 *         th.hop_reply_ttl = self.iphdr.ip_ttl
 */
  __pyx_t_3 = __pyx_v_probe->iphdr->ip_ttl;
  __pyx_v_th->hop_probe_ttl = __pyx_t_3;

  /* "cloudtrace/read/packet.pyx":147
 *         th.hop_probe_id = probe.iphdr.id
 *         th.hop_probe_ttl = probe.iphdr.ip_ttl
 *         th.hop_probe_size = probe.iphdr.ip_len             # <<<<<<<<<<<<<<
 *         th.hop_reply_ttl = self.iphdr.ip_ttl
 *         th.hop_reply_tos = self.iphdr.ip_tos
 */
  __pyx_t_2 = __pyx_v_probe->iphdr->ip_len;
  __pyx_v_th->hop_probe_size = __pyx_t_2;

  /* "cloudtrace/read/packet.pyx":148
 *         th.hop_probe_ttl = probe.iphdr.ip_ttl
 *         th.hop_probe_size = probe.iphdr.ip_len
 *         th.hop_reply_ttl = self.iphdr.ip_ttl             # <<<<<<<<<<<<<<
 *         th.hop_reply_tos = self.iphdr.ip_tos
 *         th.hop_reply_size = self.iphdr.ip_len
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.iphdr->ip_ttl;
  __pyx_v_th->hop_reply_ttl = __pyx_t_3;

  /* "cloudtrace/read/packet.pyx":149
 *         th.hop_probe_size = probe.iphdr.ip_len
 *         th.hop_reply_ttl = self.iphdr.ip_ttl
 *         th.hop_reply_tos = self.iphdr.ip_tos             # <<<<<<<<<<<<<<
 *         th.hop_reply_size = self.iphdr.ip_len
 *         th.hop_reply_ipid = self.iphdr.id
 */
  __pyx_t_3 = __pyx_v_self->__pyx_base.iphdr->ip_tos;
  __pyx_v_th->hop_reply_tos = __pyx_t_3;

  /* "cloudtrace/read/packet.pyx":150
 *         th.hop_reply_ttl = self.iphdr.ip_ttl
 *         th.hop_reply_tos = self.iphdr.ip_tos
 *         th.hop_reply_size = self.iphdr.ip_len             # <<<<<<<<<<<<<<
 *         th.hop_reply_ipid = self.iphdr.id
 *         th.hop_icmp_type = self.icmphdr.type
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.iphdr->ip_len;
  __pyx_v_th->hop_reply_size = __pyx_t_2;

  /* "cloudtrace/read/packet.pyx":151
 *         th.hop_reply_tos = self.iphdr.ip_tos
 *         th.hop_reply_size = self.iphdr.ip_len
 *         th.hop_reply_ipid = self.iphdr.id             # <<<<<<<<<<<<<<
 *         th.hop_icmp_type = self.icmphdr.type
 *         th.hop_icmp_code = self.icmphdr.code
 */
  __pyx_t_2 = __pyx_v_self->__pyx_base.iphdr->id;
  __pyx_v_th->hop_reply_ipid = __pyx_t_2;

  /* "cloudtrace/read/packet.pyx":152
 *         th.hop_reply_size = self.iphdr.ip_len
 *         th.hop_reply_ipid = self.iphdr.id
 *         th.hop_icmp_type = self.icmphdr.type             # <<<<<<<<<<<<<<
 *         th.hop_icmp_code = self.icmphdr.code
 *         memcpy(&th.hop_tx, probe.time, sizeof(timeval))
 */
  __pyx_t_3 = __pyx_v_self->icmphdr->type;
  __pyx_v_th->hop_icmp_type = __pyx_t_3;

  /* "cloudtrace/read/packet.pyx":153
 *         th.hop_reply_ipid = self.iphdr.id
 *         th.hop_icmp_type = self.icmphdr.type
 *         th.hop_icmp_code = self.icmphdr.code             # <<<<<<<<<<<<<<
 *         memcpy(&th.hop_tx, probe.time, sizeof(timeval))
 *         timersub(self.time, probe.time, &th.hop_rtt)
 */
  __pyx_t_3 = __pyx_v_self->icmphdr->code;
  __pyx_v_th->hop_icmp_code = __pyx_t_3;

  /* "cloudtrace/read/packet.pyx":154
 *         th.hop_icmp_type = self.icmphdr.type
 *         th.hop_icmp_code = self.icmphdr.code
 *         memcpy(&th.hop_tx, probe.time, sizeof(timeval))             # <<<<<<<<<<<<<<
 *         timersub(self.time, probe.time, &th.hop_rtt)
 *         self._create_hop(th)
 */
  (void)(memcpy((&__pyx_v_th->hop_tx), __pyx_v_probe->time, (sizeof(struct timeval))));

  /* "cloudtrace/read/packet.pyx":155
 *         th.hop_icmp_code = self.icmphdr.code
 *         memcpy(&th.hop_tx, probe.time, sizeof(timeval))
 *         timersub(self.time, probe.time, &th.hop_rtt)             # <<<<<<<<<<<<<<
 *         self._create_hop(th)
 * 
 */
  timersub(__pyx_v_self->__pyx_base.time, __pyx_v_probe->time, (&__pyx_v_th->hop_rtt));

  /* "cloudtrace/read/packet.pyx":156
 *         memcpy(&th.hop_tx, probe.time, sizeof(timeval))
 *         timersub(self.time, probe.time, &th.hop_rtt)
 *         self._create_hop(th)             # <<<<<<<<<<<<<<
 * 
 * cdef class EchoReply(Reply):
 */
  ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply *)__pyx_v_self->__pyx_base.__pyx_vtab)->_create_hop(__pyx_v_self, __pyx_v_th);

  /* "cloudtrace/read/packet.pyx":140
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 * 
 *     cdef void create_hop(self, hop_t *th):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Packet probe = self.orig_probe
 */

  /* function exit code */
  __Pyx_XDECREF((PyObject *)__pyx_v_probe);
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_5Reply_18__reduce_cython__[] = "Reply.__reduce_cython__(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply_18__reduce_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_5Reply_20__setstate_cython__[] = "Reply.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_5Reply_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_5Reply_20__setstate_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_5Reply_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.Reply.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":159
 * 
 * cdef class EchoReply(Reply):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.ptype = ECHO_REPLY
 * 
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_9EchoReply_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_9EchoReply_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9EchoReply___cinit__(((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_9EchoReply___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/read/packet.pyx":160
 * cdef class EchoReply(Reply):
 *     def __cinit__(self):
 *         self.ptype = ECHO_REPLY             # <<<<<<<<<<<<<<
 * 
 *     cdef uint8_t *get_dst_c(self):
 */
  __pyx_v_self->__pyx_base.__pyx_base.ptype = ECHO_REPLY;

  /* "cloudtrace/read/packet.pyx":159
 * 
 * cdef class EchoReply(Reply):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.ptype = ECHO_REPLY
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":162
 *         self.ptype = ECHO_REPLY
 * 
 *     cdef uint8_t *get_dst_c(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_src
 * 
 */

static uint8_t *__pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_dst_c(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self) {
  uint8_t *__pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_dst_c", 0);

  /* "cloudtrace/read/packet.pyx":163
 * 
 *     cdef uint8_t *get_dst_c(self):
 *         return self.iphdr.ip_src             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t probe_ttl(self):
 */
  __pyx_r = __pyx_v_self->__pyx_base.__pyx_base.iphdr->ip_src;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":162
 *         self.ptype = ECHO_REPLY
 * 
 *     cdef uint8_t *get_dst_c(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_src
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":165
 *         return self.iphdr.ip_src
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.un.echo.seq
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_3probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_3probe_ttl)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 165, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":166
 * 
 *     cpdef uint8_t probe_ttl(self):
 *         return self.icmphdr.un.echo.seq             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t get_pid(self):
 */
  __pyx_r = __pyx_v_self->__pyx_base.icmphdr->un.echo.seq;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":165
 *         return self.iphdr.ip_src
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.un.echo.seq
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.EchoReply.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_3probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_2probe_ttl[] = "EchoReply.probe_ttl(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_3probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_ttl (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9EchoReply_2probe_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_2probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_ttl(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.EchoReply.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":168
 *         return self.icmphdr.un.echo.seq
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.un.echo.id
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_5get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_pid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_5get_pid)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 168, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":169
 * 
 *     cpdef uint16_t get_pid(self):
 *         return self.icmphdr.un.echo.id             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t probe_ipid(self):
 */
  __pyx_r = __pyx_v_self->__pyx_base.icmphdr->un.echo.id;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":168
 *         return self.icmphdr.un.echo.seq
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         return self.icmphdr.un.echo.id
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.EchoReply.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_5get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_4get_pid[] = "EchoReply.get_pid(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_5get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_pid (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9EchoReply_4get_pid(((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_4get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_pid(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.EchoReply.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":171
 *         return self.icmphdr.un.echo.id
 * 
 *     cpdef uint16_t probe_ipid(self):             # <<<<<<<<<<<<<<
 *         return self.ipid()
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_7probe_ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ipid", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_ipid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_7probe_ipid)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 171, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 171, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":172
 * 
 *     cpdef uint16_t probe_ipid(self):
 *         return self.ipid()             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t probe_size(self):
 */
  __pyx_r = ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.ipid(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self), 0);
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":171
 *         return self.icmphdr.un.echo.id
 * 
 *     cpdef uint16_t probe_ipid(self):             # <<<<<<<<<<<<<<
 *         return self.ipid()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.EchoReply.probe_ipid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_7probe_ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_6probe_ipid[] = "EchoReply.probe_ipid(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_7probe_ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_ipid (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9EchoReply_6probe_ipid(((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_6probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ipid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_ipid(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.EchoReply.probe_ipid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":174
 *         return self.ipid()
 * 
 *     cpdef uint16_t probe_size(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_len
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_9probe_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_size", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_9probe_size)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 174, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":175
 * 
 *     cpdef uint16_t probe_size(self):
 *         return self.iphdr.ip_len             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):
 */
  __pyx_r = __pyx_v_self->__pyx_base.__pyx_base.iphdr->ip_len;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":174
 *         return self.ipid()
 * 
 *     cpdef uint16_t probe_size(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_len
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.EchoReply.probe_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_9probe_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_8probe_size[] = "EchoReply.probe_size(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_9probe_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_size (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9EchoReply_8probe_size(((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_8probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_size", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_size(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.EchoReply.probe_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":177
 *         return self.iphdr.ip_len
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_11get_icmp_q_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_icmp_q_ttl(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_icmp_q_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_icmp_q_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_11get_icmp_q_ttl)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 177, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":178
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):
 *         return 1             # <<<<<<<<<<<<<<
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":177
 *         return self.iphdr.ip_len
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):             # <<<<<<<<<<<<<<
 *         return 1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.EchoReply.get_icmp_q_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_11get_icmp_q_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_10get_icmp_q_ttl[] = "EchoReply.get_icmp_q_ttl(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_11get_icmp_q_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_icmp_q_ttl (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9EchoReply_10get_icmp_q_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_10get_icmp_q_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_icmp_q_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_icmp_q_ttl(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.EchoReply.get_icmp_q_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":180
 *         return 1
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):             # <<<<<<<<<<<<<<
 *         self.set_ip(iphdr)
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 */

static void __pyx_f_10cloudtrace_4read_6packet_9EchoReply_set_icmp(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr, __pyx_t_10cloudtrace_4read_5utils_icmp *__pyx_v_hdr) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_icmp", 0);

  /* "cloudtrace/read/packet.pyx":181
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):
 *         self.set_ip(iphdr)             # <<<<<<<<<<<<<<
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 * 
 */
  ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.set_ip(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_self), __pyx_v_iphdr);

  /* "cloudtrace/read/packet.pyx":182
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):
 *         self.set_ip(iphdr)
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))             # <<<<<<<<<<<<<<
 * 
 * cdef class ICMPReply(Reply):
 */
  (void)(memcpy(__pyx_v_self->__pyx_base.icmphdr, __pyx_v_hdr, (sizeof(__pyx_t_10cloudtrace_4read_5utils_icmp))));

  /* "cloudtrace/read/packet.pyx":180
 *         return 1
 * 
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr):             # <<<<<<<<<<<<<<
 *         self.set_ip(iphdr)
 *         memcpy(self.icmphdr, hdr, sizeof(icmp))
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_12__reduce_cython__[] = "EchoReply.__reduce_cython__(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9EchoReply_12__reduce_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.EchoReply.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_14__setstate_cython__[] = "EchoReply.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9EchoReply_14__setstate_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9EchoReply_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.EchoReply.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":185
 * 
 * cdef class ICMPReply(Reply):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.ptype = ICMP_REPLY
 *         self.probe = None
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply___cinit__(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply___cinit__(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/read/packet.pyx":186
 * cdef class ICMPReply(Reply):
 *     def __cinit__(self):
 *         self.ptype = ICMP_REPLY             # <<<<<<<<<<<<<<
 *         self.probe = None
 *         # self.ie = NULL
 */
  __pyx_v_self->__pyx_base.__pyx_base.ptype = ICMP_REPLY;

  /* "cloudtrace/read/packet.pyx":187
 *     def __cinit__(self):
 *         self.ptype = ICMP_REPLY
 *         self.probe = None             # <<<<<<<<<<<<<<
 *         # self.ie = NULL
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->probe);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->probe));
  __pyx_v_self->probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None);

  /* "cloudtrace/read/packet.pyx":185
 * 
 * cdef class ICMPReply(Reply):
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.ptype = ICMP_REPLY
 *         self.probe = None
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":190
 *         # self.ie = NULL
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         pass
 *         # if self.ie is not NULL:
 */

/* Python wrapper */
static void __pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_2__dealloc__(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":195
 *         #     scamper_icmpext_free(self.ie)
 * 
 *     cdef void _create_hop(self, hop_t *th):             # <<<<<<<<<<<<<<
 *         th.hop_icmp_q_ttl = self.probe.iphdr.ip_ttl
 *         th.hop_icmp_q_tos = self.probe.iphdr.ip_tos
 */

static void __pyx_f_10cloudtrace_4read_6packet_9ICMPReply__create_hop(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_th) {
  __Pyx_RefNannyDeclarations
  uint8_t __pyx_t_1;
  uint16_t __pyx_t_2;
  __Pyx_RefNannySetupContext("_create_hop", 0);

  /* "cloudtrace/read/packet.pyx":196
 * 
 *     cdef void _create_hop(self, hop_t *th):
 *         th.hop_icmp_q_ttl = self.probe.iphdr.ip_ttl             # <<<<<<<<<<<<<<
 *         th.hop_icmp_q_tos = self.probe.iphdr.ip_tos
 *         th.hop_icmp_q_ipl = self.probe.iphdr.ip_len
 */
  __pyx_t_1 = __pyx_v_self->probe->iphdr->ip_ttl;
  __pyx_v_th->hop_icmp_q_ttl = __pyx_t_1;

  /* "cloudtrace/read/packet.pyx":197
 *     cdef void _create_hop(self, hop_t *th):
 *         th.hop_icmp_q_ttl = self.probe.iphdr.ip_ttl
 *         th.hop_icmp_q_tos = self.probe.iphdr.ip_tos             # <<<<<<<<<<<<<<
 *         th.hop_icmp_q_ipl = self.probe.iphdr.ip_len
 *         # if self.ie is not NULL:
 */
  __pyx_t_1 = __pyx_v_self->probe->iphdr->ip_tos;
  __pyx_v_th->hop_icmp_q_tos = __pyx_t_1;

  /* "cloudtrace/read/packet.pyx":198
 *         th.hop_icmp_q_ttl = self.probe.iphdr.ip_ttl
 *         th.hop_icmp_q_tos = self.probe.iphdr.ip_tos
 *         th.hop_icmp_q_ipl = self.probe.iphdr.ip_len             # <<<<<<<<<<<<<<
 *         # if self.ie is not NULL:
 *         #     th.hop_icmpext = scamper_icmpext_alloc(self.ie.ie_cn, self.ie.ie_ct, self.ie.ie_dl, self.ie.ie_data)
 */
  __pyx_t_2 = __pyx_v_self->probe->iphdr->ip_len;
  __pyx_v_th->hop_icmp_q_ipl = __pyx_t_2;

  /* "cloudtrace/read/packet.pyx":195
 *         #     scamper_icmpext_free(self.ie)
 * 
 *     cdef void _create_hop(self, hop_t *th):             # <<<<<<<<<<<<<<
 *         th.hop_icmp_q_ttl = self.probe.iphdr.ip_ttl
 *         th.hop_icmp_q_tos = self.probe.iphdr.ip_tos
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":202
 *         #     th.hop_icmpext = scamper_icmpext_alloc(self.ie.ie_cn, self.ie.ie_ct, self.ie.ie_dl, self.ie.ie_data)
 * 
 *     cdef uint8_t *get_dst_c(self):             # <<<<<<<<<<<<<<
 *         if self.probe is None:
 *             raise Exception('Probe is not set yet.')
 */

static uint8_t *__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_dst_c(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self) {
  uint8_t *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_dst_c", 0);

  /* "cloudtrace/read/packet.pyx":203
 * 
 *     cdef uint8_t *get_dst_c(self):
 *         if self.probe is None:             # <<<<<<<<<<<<<<
 *             raise Exception('Probe is not set yet.')
 *         return self.probe.iphdr.ip_dst
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->probe) == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "cloudtrace/read/packet.pyx":204
 *     cdef uint8_t *get_dst_c(self):
 *         if self.probe is None:
 *             raise Exception('Probe is not set yet.')             # <<<<<<<<<<<<<<
 *         return self.probe.iphdr.ip_dst
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 204, __pyx_L1_error)

    /* "cloudtrace/read/packet.pyx":203
 * 
 *     cdef uint8_t *get_dst_c(self):
 *         if self.probe is None:             # <<<<<<<<<<<<<<
 *             raise Exception('Probe is not set yet.')
 *         return self.probe.iphdr.ip_dst
 */
  }

  /* "cloudtrace/read/packet.pyx":205
 *         if self.probe is None:
 *             raise Exception('Probe is not set yet.')
 *         return self.probe.iphdr.ip_dst             # <<<<<<<<<<<<<<
 * 
 *     cdef void create_pseudo(self, udppseudo *p, unsigned char *src):
 */
  __pyx_r = __pyx_v_self->probe->iphdr->ip_dst;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":202
 *         #     th.hop_icmpext = scamper_icmpext_alloc(self.ie.ie_cn, self.ie.ie_ct, self.ie.ie_dl, self.ie.ie_data)
 * 
 *     cdef uint8_t *get_dst_c(self):             # <<<<<<<<<<<<<<
 *         if self.probe is None:
 *             raise Exception('Probe is not set yet.')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPReply.get_dst_c", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":207
 *         return self.probe.iphdr.ip_dst
 * 
 *     cdef void create_pseudo(self, udppseudo *p, unsigned char *src):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ip *iphdr = (<UDPProbe>self.probe).iphdr
 */

static void __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_create_pseudo(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, __pyx_t_10cloudtrace_4read_5utils_udppseudo *__pyx_v_p, unsigned char *__pyx_v_src) {
  __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr;
  __pyx_t_10cloudtrace_4read_5utils_udp *__pyx_v_udphdr;
  __Pyx_RefNannyDeclarations
  __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_t_1;
  __pyx_t_10cloudtrace_4read_5utils_udp *__pyx_t_2;
  __Pyx_RefNannySetupContext("create_pseudo", 0);

  /* "cloudtrace/read/packet.pyx":209
 *     cdef void create_pseudo(self, udppseudo *p, unsigned char *src):
 *         cdef:
 *             ip *iphdr = (<UDPProbe>self.probe).iphdr             # <<<<<<<<<<<<<<
 *             udp *udphdr = (<UDPProbe>self.probe).udphdr
 *         memcpy(p.src, src, 4)
 */
  __pyx_t_1 = ((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_self->probe)->__pyx_base.iphdr;
  __pyx_v_iphdr = __pyx_t_1;

  /* "cloudtrace/read/packet.pyx":210
 *         cdef:
 *             ip *iphdr = (<UDPProbe>self.probe).iphdr
 *             udp *udphdr = (<UDPProbe>self.probe).udphdr             # <<<<<<<<<<<<<<
 *         memcpy(p.src, src, 4)
 *         memcpy(p.dst, iphdr.ip_dst, 4)
 */
  __pyx_t_2 = ((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_self->probe)->udphdr;
  __pyx_v_udphdr = __pyx_t_2;

  /* "cloudtrace/read/packet.pyx":211
 *             ip *iphdr = (<UDPProbe>self.probe).iphdr
 *             udp *udphdr = (<UDPProbe>self.probe).udphdr
 *         memcpy(p.src, src, 4)             # <<<<<<<<<<<<<<
 *         memcpy(p.dst, iphdr.ip_dst, 4)
 *         p.zeros = 0
 */
  (void)(memcpy(__pyx_v_p->src, __pyx_v_src, 4));

  /* "cloudtrace/read/packet.pyx":212
 *             udp *udphdr = (<UDPProbe>self.probe).udphdr
 *         memcpy(p.src, src, 4)
 *         memcpy(p.dst, iphdr.ip_dst, 4)             # <<<<<<<<<<<<<<
 *         p.zeros = 0
 *         p.proto = IPPROTO_UDP
 */
  (void)(memcpy(__pyx_v_p->dst, __pyx_v_iphdr->ip_dst, 4));

  /* "cloudtrace/read/packet.pyx":213
 *         memcpy(p.src, src, 4)
 *         memcpy(p.dst, iphdr.ip_dst, 4)
 *         p.zeros = 0             # <<<<<<<<<<<<<<
 *         p.proto = IPPROTO_UDP
 *         p.length = htons(udphdr.len)
 */
  __pyx_v_p->zeros = 0;

  /* "cloudtrace/read/packet.pyx":214
 *         memcpy(p.dst, iphdr.ip_dst, 4)
 *         p.zeros = 0
 *         p.proto = IPPROTO_UDP             # <<<<<<<<<<<<<<
 *         p.length = htons(udphdr.len)
 *         p.sport = htons(udphdr.source)
 */
  __pyx_v_p->proto = IPPROTO_UDP;

  /* "cloudtrace/read/packet.pyx":215
 *         p.zeros = 0
 *         p.proto = IPPROTO_UDP
 *         p.length = htons(udphdr.len)             # <<<<<<<<<<<<<<
 *         p.sport = htons(udphdr.source)
 *         p.dport = htons(udphdr.dest)
 */
  __pyx_v_p->length = htons(__pyx_v_udphdr->len);

  /* "cloudtrace/read/packet.pyx":216
 *         p.proto = IPPROTO_UDP
 *         p.length = htons(udphdr.len)
 *         p.sport = htons(udphdr.source)             # <<<<<<<<<<<<<<
 *         p.dport = htons(udphdr.dest)
 *         p.length2 = htons(udphdr.len)
 */
  __pyx_v_p->sport = htons(__pyx_v_udphdr->source);

  /* "cloudtrace/read/packet.pyx":217
 *         p.length = htons(udphdr.len)
 *         p.sport = htons(udphdr.source)
 *         p.dport = htons(udphdr.dest)             # <<<<<<<<<<<<<<
 *         p.length2 = htons(udphdr.len)
 *         p.chksum = 0
 */
  __pyx_v_p->dport = htons(__pyx_v_udphdr->dest);

  /* "cloudtrace/read/packet.pyx":218
 *         p.sport = htons(udphdr.source)
 *         p.dport = htons(udphdr.dest)
 *         p.length2 = htons(udphdr.len)             # <<<<<<<<<<<<<<
 *         p.chksum = 0
 * 
 */
  __pyx_v_p->length2 = htons(__pyx_v_udphdr->len);

  /* "cloudtrace/read/packet.pyx":219
 *         p.dport = htons(udphdr.dest)
 *         p.length2 = htons(udphdr.len)
 *         p.chksum = 0             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t probe_ttl(self):
 */
  __pyx_v_p->chksum = 0;

  /* "cloudtrace/read/packet.pyx":207
 *         return self.probe.iphdr.ip_dst
 * 
 *     cdef void create_pseudo(self, udppseudo *p, unsigned char *src):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ip *iphdr = (<UDPProbe>self.probe).iphdr
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/packet.pyx":221
 *         p.chksum = 0
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             udppseudo *np
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_5probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch) {
  __pyx_t_10cloudtrace_4read_5utils_udppseudo *__pyx_v_np;
  __pyx_t_10cloudtrace_4read_5utils_udppseudo *__pyx_v_op;
  struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *__pyx_v_probe = 0;
  unsigned short __pyx_v_oc;
  unsigned short __pyx_v_ttl;
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_t_6;
  uint16_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_5probe_ttl)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 221, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":227
 *             UDPProbe probe
 *             unsigned short oc, ttl
 *         if self.probe.ptype == ICMP_PROBE:             # <<<<<<<<<<<<<<
 *             return self.probe.probe_ttl()
 *         probe = <UDPProbe>self.probe
 */
  __pyx_t_6 = ((__pyx_v_self->probe->ptype == ICMP_PROBE) != 0);
  if (__pyx_t_6) {

    /* "cloudtrace/read/packet.pyx":228
 *             unsigned short oc, ttl
 *         if self.probe.ptype == ICMP_PROBE:
 *             return self.probe.probe_ttl()             # <<<<<<<<<<<<<<
 *         probe = <UDPProbe>self.probe
 *         np = <udppseudo *>calloc(1, sizeof(udppseudo))
 */
    __pyx_r = ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *)__pyx_v_self->probe->__pyx_vtab)->probe_ttl(__pyx_v_self->probe, 0);
    goto __pyx_L0;

    /* "cloudtrace/read/packet.pyx":227
 *             UDPProbe probe
 *             unsigned short oc, ttl
 *         if self.probe.ptype == ICMP_PROBE:             # <<<<<<<<<<<<<<
 *             return self.probe.probe_ttl()
 *         probe = <UDPProbe>self.probe
 */
  }

  /* "cloudtrace/read/packet.pyx":229
 *         if self.probe.ptype == ICMP_PROBE:
 *             return self.probe.probe_ttl()
 *         probe = <UDPProbe>self.probe             # <<<<<<<<<<<<<<
 *         np = <udppseudo *>calloc(1, sizeof(udppseudo))
 *         op = <udppseudo *>calloc(1, sizeof(udppseudo))
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->probe);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/read/packet.pyx":230
 *             return self.probe.probe_ttl()
 *         probe = <UDPProbe>self.probe
 *         np = <udppseudo *>calloc(1, sizeof(udppseudo))             # <<<<<<<<<<<<<<
 *         op = <udppseudo *>calloc(1, sizeof(udppseudo))
 *         self.create_pseudo(np, probe.iphdr.ip_src)
 */
  __pyx_v_np = ((__pyx_t_10cloudtrace_4read_5utils_udppseudo *)calloc(1, (sizeof(__pyx_t_10cloudtrace_4read_5utils_udppseudo))));

  /* "cloudtrace/read/packet.pyx":231
 *         probe = <UDPProbe>self.probe
 *         np = <udppseudo *>calloc(1, sizeof(udppseudo))
 *         op = <udppseudo *>calloc(1, sizeof(udppseudo))             # <<<<<<<<<<<<<<
 *         self.create_pseudo(np, probe.iphdr.ip_src)
 *         self.create_pseudo(op, self.iphdr.ip_dst)
 */
  __pyx_v_op = ((__pyx_t_10cloudtrace_4read_5utils_udppseudo *)calloc(1, (sizeof(__pyx_t_10cloudtrace_4read_5utils_udppseudo))));

  /* "cloudtrace/read/packet.pyx":232
 *         np = <udppseudo *>calloc(1, sizeof(udppseudo))
 *         op = <udppseudo *>calloc(1, sizeof(udppseudo))
 *         self.create_pseudo(np, probe.iphdr.ip_src)             # <<<<<<<<<<<<<<
 *         self.create_pseudo(op, self.iphdr.ip_dst)
 *         oc = htons(checksum(<unsigned char *>op, 20))
 */
  ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->create_pseudo(__pyx_v_self, __pyx_v_np, __pyx_v_probe->__pyx_base.iphdr->ip_src);

  /* "cloudtrace/read/packet.pyx":233
 *         op = <udppseudo *>calloc(1, sizeof(udppseudo))
 *         self.create_pseudo(np, probe.iphdr.ip_src)
 *         self.create_pseudo(op, self.iphdr.ip_dst)             # <<<<<<<<<<<<<<
 *         oc = htons(checksum(<unsigned char *>op, 20))
 *         memcpy(np.data, &oc, 2)
 */
  ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_vtab)->create_pseudo(__pyx_v_self, __pyx_v_op, __pyx_v_self->__pyx_base.__pyx_base.iphdr->ip_dst);

  /* "cloudtrace/read/packet.pyx":234
 *         self.create_pseudo(np, probe.iphdr.ip_src)
 *         self.create_pseudo(op, self.iphdr.ip_dst)
 *         oc = htons(checksum(<unsigned char *>op, 20))             # <<<<<<<<<<<<<<
 *         memcpy(np.data, &oc, 2)
 *         ttl = probe.udphdr.check - checksum(<unsigned char *>np, 22)
 */
  __pyx_t_7 = __pyx_f_10cloudtrace_4read_5utils_checksum(((unsigned char *)__pyx_v_op), 20); if (unlikely(__pyx_t_7 == ((uint16_t)-1))) __PYX_ERR(0, 234, __pyx_L1_error)
  __pyx_v_oc = htons(__pyx_t_7);

  /* "cloudtrace/read/packet.pyx":235
 *         self.create_pseudo(op, self.iphdr.ip_dst)
 *         oc = htons(checksum(<unsigned char *>op, 20))
 *         memcpy(np.data, &oc, 2)             # <<<<<<<<<<<<<<
 *         ttl = probe.udphdr.check - checksum(<unsigned char *>np, 22)
 *         if np is not NULL:
 */
  (void)(memcpy(__pyx_v_np->data, (&__pyx_v_oc), 2));

  /* "cloudtrace/read/packet.pyx":236
 *         oc = htons(checksum(<unsigned char *>op, 20))
 *         memcpy(np.data, &oc, 2)
 *         ttl = probe.udphdr.check - checksum(<unsigned char *>np, 22)             # <<<<<<<<<<<<<<
 *         if np is not NULL:
 *             free(np)
 */
  __pyx_t_7 = __pyx_f_10cloudtrace_4read_5utils_checksum(((unsigned char *)__pyx_v_np), 22); if (unlikely(__pyx_t_7 == ((uint16_t)-1))) __PYX_ERR(0, 236, __pyx_L1_error)
  __pyx_v_ttl = (__pyx_v_probe->udphdr->check - __pyx_t_7);

  /* "cloudtrace/read/packet.pyx":237
 *         memcpy(np.data, &oc, 2)
 *         ttl = probe.udphdr.check - checksum(<unsigned char *>np, 22)
 *         if np is not NULL:             # <<<<<<<<<<<<<<
 *             free(np)
 *         if op is not NULL:
 */
  __pyx_t_6 = ((__pyx_v_np != NULL) != 0);
  if (__pyx_t_6) {

    /* "cloudtrace/read/packet.pyx":238
 *         ttl = probe.udphdr.check - checksum(<unsigned char *>np, 22)
 *         if np is not NULL:
 *             free(np)             # <<<<<<<<<<<<<<
 *         if op is not NULL:
 *             free(op)
 */
    free(__pyx_v_np);

    /* "cloudtrace/read/packet.pyx":237
 *         memcpy(np.data, &oc, 2)
 *         ttl = probe.udphdr.check - checksum(<unsigned char *>np, 22)
 *         if np is not NULL:             # <<<<<<<<<<<<<<
 *             free(np)
 *         if op is not NULL:
 */
  }

  /* "cloudtrace/read/packet.pyx":239
 *         if np is not NULL:
 *             free(np)
 *         if op is not NULL:             # <<<<<<<<<<<<<<
 *             free(op)
 *         return ttl
 */
  __pyx_t_6 = ((__pyx_v_op != NULL) != 0);
  if (__pyx_t_6) {

    /* "cloudtrace/read/packet.pyx":240
 *             free(np)
 *         if op is not NULL:
 *             free(op)             # <<<<<<<<<<<<<<
 *         return ttl
 * 
 */
    free(__pyx_v_op);

    /* "cloudtrace/read/packet.pyx":239
 *         if np is not NULL:
 *             free(np)
 *         if op is not NULL:             # <<<<<<<<<<<<<<
 *             free(op)
 *         return ttl
 */
  }

  /* "cloudtrace/read/packet.pyx":241
 *         if op is not NULL:
 *             free(op)
 *         return ttl             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t probe_size(self):
 */
  __pyx_r = __pyx_v_ttl;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":221
 *         p.chksum = 0
 * 
 *     cpdef uint8_t probe_ttl(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             udppseudo *np
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPReply.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_probe);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_5probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_4probe_ttl[] = "ICMPReply.probe_ttl(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_5probe_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_ttl (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_4probe_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_4probe_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_ttl(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPReply.probe_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":243
 *         return ttl
 * 
 *     cpdef uint16_t probe_size(self):             # <<<<<<<<<<<<<<
 *         return self.probe.iphdr.ip_len
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_7probe_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_size", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_7probe_size)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 243, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 243, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":244
 * 
 *     cpdef uint16_t probe_size(self):
 *         return self.probe.iphdr.ip_len             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t get_pid(self):
 */
  __pyx_r = __pyx_v_self->probe->iphdr->ip_len;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":243
 *         return ttl
 * 
 *     cpdef uint16_t probe_size(self):             # <<<<<<<<<<<<<<
 *         return self.probe.iphdr.ip_len
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPReply.probe_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_7probe_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_6probe_size[] = "ICMPReply.probe_size(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_7probe_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_size (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_6probe_size(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_6probe_size(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_size", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_size(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPReply.probe_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":246
 *         return self.probe.iphdr.ip_len
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         return self.probe.get_pid()
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_9get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_pid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_9get_pid)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 246, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":247
 * 
 *     cpdef uint16_t get_pid(self):
 *         return self.probe.get_pid()             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint16_t probe_ipid(self):
 */
  __pyx_r = ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *)__pyx_v_self->probe->__pyx_vtab)->get_pid(__pyx_v_self->probe, 0);
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":246
 *         return self.probe.iphdr.ip_len
 * 
 *     cpdef uint16_t get_pid(self):             # <<<<<<<<<<<<<<
 *         return self.probe.get_pid()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPReply.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_9get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_8get_pid[] = "ICMPReply.get_pid(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_9get_pid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_pid (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_8get_pid(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_8get_pid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_pid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_pid(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPReply.get_pid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":249
 *         return self.probe.get_pid()
 * 
 *     cpdef uint16_t probe_ipid(self):             # <<<<<<<<<<<<<<
 *         return self.probe.ipid()
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_11probe_ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint16_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint16_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ipid", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probe_ipid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_11probe_ipid)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 249, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint16_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 249, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":250
 * 
 *     cpdef uint16_t probe_ipid(self):
 *         return self.probe.ipid()             # <<<<<<<<<<<<<<
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):
 */
  __pyx_r = ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *)__pyx_v_self->probe->__pyx_vtab)->ipid(__pyx_v_self->probe, 0);
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":249
 *         return self.probe.get_pid()
 * 
 *     cpdef uint16_t probe_ipid(self):             # <<<<<<<<<<<<<<
 *         return self.probe.ipid()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPReply.probe_ipid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_11probe_ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_10probe_ipid[] = "ICMPReply.probe_ipid(self) -> uint16_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_11probe_ipid(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("probe_ipid (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_10probe_ipid(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_10probe_ipid(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("probe_ipid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_ipid(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPReply.probe_ipid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":252
 *         return self.probe.ipid()
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_ttl
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_13get_icmp_q_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static uint8_t __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_icmp_q_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_icmp_q_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_icmp_q_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_13get_icmp_q_ttl)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_2); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 252, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":253
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):
 *         return self.iphdr.ip_ttl             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_probe(self, Packet probe):
 */
  __pyx_r = __pyx_v_self->__pyx_base.__pyx_base.iphdr->ip_ttl;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":252
 *         return self.probe.ipid()
 * 
 *     cpdef uint8_t get_icmp_q_ttl(self):             # <<<<<<<<<<<<<<
 *         return self.iphdr.ip_ttl
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPReply.get_icmp_q_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_13get_icmp_q_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_12get_icmp_q_ttl[] = "ICMPReply.get_icmp_q_ttl(self) -> uint8_t";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_13get_icmp_q_ttl(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_icmp_q_ttl (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_12get_icmp_q_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_12get_icmp_q_ttl(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_icmp_q_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_icmp_q_ttl(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPReply.get_icmp_q_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":255
 *         return self.iphdr.ip_ttl
 * 
 *     cpdef void set_probe(self, Packet probe):             # <<<<<<<<<<<<<<
 *         self.probe = probe
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_15set_probe(PyObject *__pyx_v_self, PyObject *__pyx_v_probe); /*proto*/
static void __pyx_f_10cloudtrace_4read_6packet_9ICMPReply_set_probe(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_probe, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_probe", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_probe); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_15set_probe)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_probe)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_probe));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/packet.pyx":256
 * 
 *     cpdef void set_probe(self, Packet probe):
 *         self.probe = probe             # <<<<<<<<<<<<<<
 * 
 * cpdef Packet disassemble_packet(char *packet, int incl_len):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_probe));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_probe));
  __Pyx_GOTREF(__pyx_v_self->probe);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->probe));
  __pyx_v_self->probe = __pyx_v_probe;

  /* "cloudtrace/read/packet.pyx":255
 *         return self.iphdr.ip_ttl
 * 
 *     cpdef void set_probe(self, Packet probe):             # <<<<<<<<<<<<<<
 *         self.probe = probe
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.read.packet.ICMPReply.set_probe", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_15set_probe(PyObject *__pyx_v_self, PyObject *__pyx_v_probe); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_14set_probe[] = "ICMPReply.set_probe(self, Packet probe) -> void";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_15set_probe(PyObject *__pyx_v_self, PyObject *__pyx_v_probe) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_probe (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_probe), __pyx_ptype_10cloudtrace_4read_6packet_Packet, 1, "probe", 0))) __PYX_ERR(0, 255, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_14set_probe(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self), ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_probe));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_14set_probe(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_probe) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_probe", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_set_probe(__pyx_v_self, __pyx_v_probe, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPReply.set_probe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_16__reduce_cython__[] = "ICMPReply.__reduce_cython__(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_16__reduce_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPReply.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_18__setstate_cython__[] = "ICMPReply.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_18__setstate_cython__(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_9ICMPReply_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.ICMPReply.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":258
 *         self.probe = probe
 * 
 * cpdef Packet disassemble_packet(char *packet, int incl_len):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char *packet_orig = packet
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_1disassemble_packet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_f_10cloudtrace_4read_6packet_disassemble_packet(char *__pyx_v_packet, int __pyx_v_incl_len, CYTHON_UNUSED int __pyx_skip_dispatch) {
  CYTHON_UNUSED char *__pyx_v_packet_orig;
  __pyx_t_10cloudtrace_4read_5utils_ip *__pyx_v_iphdr;
  unsigned int __pyx_v_iphdr_len;
  __pyx_t_10cloudtrace_4read_5utils_icmp *__pyx_v_icmphdr;
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_probe = 0;
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_reply = 0;
  __pyx_t_10cloudtrace_4read_5utils_udp *__pyx_v_udphdr;
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("disassemble_packet", 0);

  /* "cloudtrace/read/packet.pyx":260
 * cpdef Packet disassemble_packet(char *packet, int incl_len):
 *     cdef:
 *         char *packet_orig = packet             # <<<<<<<<<<<<<<
 *         ip *iphdr
 *         unsigned int iphdr_len
 */
  __pyx_v_packet_orig = __pyx_v_packet;

  /* "cloudtrace/read/packet.pyx":268
 *         icmpexthdr *icmpexthdr_s
 *         icmpext *icmpext_s
 *     if incl_len < <int>sizeof(ip):             # <<<<<<<<<<<<<<
 *         return None
 *     # memcpy(iphdr, packet, sizeof(ip))
 */
  __pyx_t_1 = ((__pyx_v_incl_len < ((int)(sizeof(__pyx_t_10cloudtrace_4read_5utils_ip)))) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/packet.pyx":269
 *         icmpext *icmpext_s
 *     if incl_len < <int>sizeof(ip):
 *         return None             # <<<<<<<<<<<<<<
 *     # memcpy(iphdr, packet, sizeof(ip))
 *     iphdr = <ip *>packet
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "cloudtrace/read/packet.pyx":268
 *         icmpexthdr *icmpexthdr_s
 *         icmpext *icmpext_s
 *     if incl_len < <int>sizeof(ip):             # <<<<<<<<<<<<<<
 *         return None
 *     # memcpy(iphdr, packet, sizeof(ip))
 */
  }

  /* "cloudtrace/read/packet.pyx":271
 *         return None
 *     # memcpy(iphdr, packet, sizeof(ip))
 *     iphdr = <ip *>packet             # <<<<<<<<<<<<<<
 *     iptohost(iphdr)
 *     iphdr_len = (iphdr.ip_hl & 0x0f) * 4
 */
  __pyx_v_iphdr = ((__pyx_t_10cloudtrace_4read_5utils_ip *)__pyx_v_packet);

  /* "cloudtrace/read/packet.pyx":272
 *     # memcpy(iphdr, packet, sizeof(ip))
 *     iphdr = <ip *>packet
 *     iptohost(iphdr)             # <<<<<<<<<<<<<<
 *     iphdr_len = (iphdr.ip_hl & 0x0f) * 4
 *     packet += iphdr_len
 */
  __pyx_f_10cloudtrace_4read_5utils_iptohost(__pyx_v_iphdr);

  /* "cloudtrace/read/packet.pyx":273
 *     iphdr = <ip *>packet
 *     iptohost(iphdr)
 *     iphdr_len = (iphdr.ip_hl & 0x0f) * 4             # <<<<<<<<<<<<<<
 *     packet += iphdr_len
 *     incl_len -= iphdr_len
 */
  __pyx_v_iphdr_len = ((__pyx_v_iphdr->ip_hl & 0x0f) * 4);

  /* "cloudtrace/read/packet.pyx":274
 *     iptohost(iphdr)
 *     iphdr_len = (iphdr.ip_hl & 0x0f) * 4
 *     packet += iphdr_len             # <<<<<<<<<<<<<<
 *     incl_len -= iphdr_len
 *     if iphdr.ip_p == IPPROTO_ICMP:
 */
  __pyx_v_packet = (__pyx_v_packet + __pyx_v_iphdr_len);

  /* "cloudtrace/read/packet.pyx":275
 *     iphdr_len = (iphdr.ip_hl & 0x0f) * 4
 *     packet += iphdr_len
 *     incl_len -= iphdr_len             # <<<<<<<<<<<<<<
 *     if iphdr.ip_p == IPPROTO_ICMP:
 *         if incl_len < <int>sizeof(icmp):
 */
  __pyx_v_incl_len = (__pyx_v_incl_len - __pyx_v_iphdr_len);

  /* "cloudtrace/read/packet.pyx":276
 *     packet += iphdr_len
 *     incl_len -= iphdr_len
 *     if iphdr.ip_p == IPPROTO_ICMP:             # <<<<<<<<<<<<<<
 *         if incl_len < <int>sizeof(icmp):
 *             return None
 */
  __pyx_t_1 = ((__pyx_v_iphdr->ip_p == IPPROTO_ICMP) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/packet.pyx":277
 *     incl_len -= iphdr_len
 *     if iphdr.ip_p == IPPROTO_ICMP:
 *         if incl_len < <int>sizeof(icmp):             # <<<<<<<<<<<<<<
 *             return None
 *         icmphdr = <icmp *>packet
 */
    __pyx_t_1 = ((__pyx_v_incl_len < ((int)(sizeof(__pyx_t_10cloudtrace_4read_5utils_icmp)))) != 0);
    if (__pyx_t_1) {

      /* "cloudtrace/read/packet.pyx":278
 *     if iphdr.ip_p == IPPROTO_ICMP:
 *         if incl_len < <int>sizeof(icmp):
 *             return None             # <<<<<<<<<<<<<<
 *         icmphdr = <icmp *>packet
 *         icmptohost(icmphdr)
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "cloudtrace/read/packet.pyx":277
 *     incl_len -= iphdr_len
 *     if iphdr.ip_p == IPPROTO_ICMP:
 *         if incl_len < <int>sizeof(icmp):             # <<<<<<<<<<<<<<
 *             return None
 *         icmphdr = <icmp *>packet
 */
    }

    /* "cloudtrace/read/packet.pyx":279
 *         if incl_len < <int>sizeof(icmp):
 *             return None
 *         icmphdr = <icmp *>packet             # <<<<<<<<<<<<<<
 *         icmptohost(icmphdr)
 *         if icmphdr.type == 8:
 */
    __pyx_v_icmphdr = ((__pyx_t_10cloudtrace_4read_5utils_icmp *)__pyx_v_packet);

    /* "cloudtrace/read/packet.pyx":280
 *             return None
 *         icmphdr = <icmp *>packet
 *         icmptohost(icmphdr)             # <<<<<<<<<<<<<<
 *         if icmphdr.type == 8:
 *             probe = ICMPProbe.__new__(ICMPProbe)
 */
    __pyx_f_10cloudtrace_4read_5utils_icmptohost(__pyx_v_icmphdr);

    /* "cloudtrace/read/packet.pyx":281
 *         icmphdr = <icmp *>packet
 *         icmptohost(icmphdr)
 *         if icmphdr.type == 8:             # <<<<<<<<<<<<<<
 *             probe = ICMPProbe.__new__(ICMPProbe)
 *             (<ICMPProbe>probe).set_icmp(iphdr, icmphdr)
 */
    switch (__pyx_v_icmphdr->type) {
      case 8:

      /* "cloudtrace/read/packet.pyx":282
 *         icmptohost(icmphdr)
 *         if icmphdr.type == 8:
 *             probe = ICMPProbe.__new__(ICMPProbe)             # <<<<<<<<<<<<<<
 *             (<ICMPProbe>probe).set_icmp(iphdr, icmphdr)
 *             return probe
 */
      __pyx_t_2 = ((PyObject *)__pyx_tp_new_10cloudtrace_4read_6packet_ICMPProbe(((PyTypeObject *)__pyx_ptype_10cloudtrace_4read_6packet_ICMPProbe), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 282, __pyx_L1_error)
      __Pyx_GOTREF(((PyObject *)__pyx_t_2));
      if (!(likely(__Pyx_TypeTest(((PyObject *)__pyx_t_2), __pyx_ptype_10cloudtrace_4read_6packet_Packet)))) __PYX_ERR(0, 282, __pyx_L1_error)
      __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "cloudtrace/read/packet.pyx":283
 *         if icmphdr.type == 8:
 *             probe = ICMPProbe.__new__(ICMPProbe)
 *             (<ICMPProbe>probe).set_icmp(iphdr, icmphdr)             # <<<<<<<<<<<<<<
 *             return probe
 *         elif icmphdr.type == 0:
 */
      ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPProbe *)((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_probe)->__pyx_base.__pyx_vtab)->set_icmp(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)__pyx_v_probe), __pyx_v_iphdr, __pyx_v_icmphdr);

      /* "cloudtrace/read/packet.pyx":284
 *             probe = ICMPProbe.__new__(ICMPProbe)
 *             (<ICMPProbe>probe).set_icmp(iphdr, icmphdr)
 *             return probe             # <<<<<<<<<<<<<<
 *         elif icmphdr.type == 0:
 *             reply = EchoReply.__new__(EchoReply)
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(((PyObject *)__pyx_v_probe));
      __pyx_r = __pyx_v_probe;
      goto __pyx_L0;

      /* "cloudtrace/read/packet.pyx":281
 *         icmphdr = <icmp *>packet
 *         icmptohost(icmphdr)
 *         if icmphdr.type == 8:             # <<<<<<<<<<<<<<
 *             probe = ICMPProbe.__new__(ICMPProbe)
 *             (<ICMPProbe>probe).set_icmp(iphdr, icmphdr)
 */
      break;
      case 0:

      /* "cloudtrace/read/packet.pyx":286
 *             return probe
 *         elif icmphdr.type == 0:
 *             reply = EchoReply.__new__(EchoReply)             # <<<<<<<<<<<<<<
 *             (<EchoReply>reply).set_icmp(iphdr, icmphdr)
 *             return reply
 */
      __pyx_t_2 = ((PyObject *)__pyx_tp_new_10cloudtrace_4read_6packet_EchoReply(((PyTypeObject *)__pyx_ptype_10cloudtrace_4read_6packet_EchoReply), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(((PyObject *)__pyx_t_2));
      if (!(likely(__Pyx_TypeTest(((PyObject *)__pyx_t_2), __pyx_ptype_10cloudtrace_4read_6packet_Reply)))) __PYX_ERR(0, 286, __pyx_L1_error)
      __pyx_v_reply = ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "cloudtrace/read/packet.pyx":287
 *         elif icmphdr.type == 0:
 *             reply = EchoReply.__new__(EchoReply)
 *             (<EchoReply>reply).set_icmp(iphdr, icmphdr)             # <<<<<<<<<<<<<<
 *             return reply
 *         elif icmphdr.type == 3 or icmphdr.type == 11:
 */
      ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_EchoReply *)((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_reply)->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.set_icmp(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)__pyx_v_reply)), __pyx_v_iphdr, __pyx_v_icmphdr);

      /* "cloudtrace/read/packet.pyx":288
 *             reply = EchoReply.__new__(EchoReply)
 *             (<EchoReply>reply).set_icmp(iphdr, icmphdr)
 *             return reply             # <<<<<<<<<<<<<<
 *         elif icmphdr.type == 3 or icmphdr.type == 11:
 *             packet += sizeof(icmp)
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(((PyObject *)__pyx_v_reply));
      __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_reply);
      goto __pyx_L0;

      /* "cloudtrace/read/packet.pyx":285
 *             (<ICMPProbe>probe).set_icmp(iphdr, icmphdr)
 *             return probe
 *         elif icmphdr.type == 0:             # <<<<<<<<<<<<<<
 *             reply = EchoReply.__new__(EchoReply)
 *             (<EchoReply>reply).set_icmp(iphdr, icmphdr)
 */
      break;
      case 3:

      /* "cloudtrace/read/packet.pyx":289
 *             (<EchoReply>reply).set_icmp(iphdr, icmphdr)
 *             return reply
 *         elif icmphdr.type == 3 or icmphdr.type == 11:             # <<<<<<<<<<<<<<
 *             packet += sizeof(icmp)
 *             incl_len -= sizeof(icmp)
 */
      case 11:

      /* "cloudtrace/read/packet.pyx":290
 *             return reply
 *         elif icmphdr.type == 3 or icmphdr.type == 11:
 *             packet += sizeof(icmp)             # <<<<<<<<<<<<<<
 *             incl_len -= sizeof(icmp)
 *             probe = disassemble_packet(packet, incl_len)
 */
      __pyx_v_packet = (__pyx_v_packet + (sizeof(__pyx_t_10cloudtrace_4read_5utils_icmp)));

      /* "cloudtrace/read/packet.pyx":291
 *         elif icmphdr.type == 3 or icmphdr.type == 11:
 *             packet += sizeof(icmp)
 *             incl_len -= sizeof(icmp)             # <<<<<<<<<<<<<<
 *             probe = disassemble_packet(packet, incl_len)
 *             if probe is None:
 */
      __pyx_v_incl_len = (__pyx_v_incl_len - (sizeof(__pyx_t_10cloudtrace_4read_5utils_icmp)));

      /* "cloudtrace/read/packet.pyx":292
 *             packet += sizeof(icmp)
 *             incl_len -= sizeof(icmp)
 *             probe = disassemble_packet(packet, incl_len)             # <<<<<<<<<<<<<<
 *             if probe is None:
 *                 return None
 */
      __pyx_t_2 = ((PyObject *)__pyx_f_10cloudtrace_4read_6packet_disassemble_packet(__pyx_v_packet, __pyx_v_incl_len, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "cloudtrace/read/packet.pyx":293
 *             incl_len -= sizeof(icmp)
 *             probe = disassemble_packet(packet, incl_len)
 *             if probe is None:             # <<<<<<<<<<<<<<
 *                 return None
 *             reply = ICMPReply.__new__(ICMPReply)
 */
      __pyx_t_1 = (((PyObject *)__pyx_v_probe) == Py_None);
      __pyx_t_3 = (__pyx_t_1 != 0);
      if (__pyx_t_3) {

        /* "cloudtrace/read/packet.pyx":294
 *             probe = disassemble_packet(packet, incl_len)
 *             if probe is None:
 *                 return None             # <<<<<<<<<<<<<<
 *             reply = ICMPReply.__new__(ICMPReply)
 *             (<ICMPReply>reply).set_icmp(iphdr, icmphdr)
 */
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); __Pyx_INCREF(Py_None);
        goto __pyx_L0;

        /* "cloudtrace/read/packet.pyx":293
 *             incl_len -= sizeof(icmp)
 *             probe = disassemble_packet(packet, incl_len)
 *             if probe is None:             # <<<<<<<<<<<<<<
 *                 return None
 *             reply = ICMPReply.__new__(ICMPReply)
 */
      }

      /* "cloudtrace/read/packet.pyx":295
 *             if probe is None:
 *                 return None
 *             reply = ICMPReply.__new__(ICMPReply)             # <<<<<<<<<<<<<<
 *             (<ICMPReply>reply).set_icmp(iphdr, icmphdr)
 *             (<ICMPReply>reply).probe = probe
 */
      __pyx_t_2 = ((PyObject *)__pyx_tp_new_10cloudtrace_4read_6packet_ICMPReply(((PyTypeObject *)__pyx_ptype_10cloudtrace_4read_6packet_ICMPReply), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
      __Pyx_GOTREF(((PyObject *)__pyx_t_2));
      if (!(likely(__Pyx_TypeTest(((PyObject *)__pyx_t_2), __pyx_ptype_10cloudtrace_4read_6packet_Reply)))) __PYX_ERR(0, 295, __pyx_L1_error)
      __pyx_v_reply = ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "cloudtrace/read/packet.pyx":296
 *                 return None
 *             reply = ICMPReply.__new__(ICMPReply)
 *             (<ICMPReply>reply).set_icmp(iphdr, icmphdr)             # <<<<<<<<<<<<<<
 *             (<ICMPReply>reply).probe = probe
 *             # if iphdr.ip_len >= 136 + sizeof(icmpexthdr) + sizeof(icmpext):
 */
      ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPReply *)((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_reply)->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.set_icmp(((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_reply)), __pyx_v_iphdr, __pyx_v_icmphdr);

      /* "cloudtrace/read/packet.pyx":297
 *             reply = ICMPReply.__new__(ICMPReply)
 *             (<ICMPReply>reply).set_icmp(iphdr, icmphdr)
 *             (<ICMPReply>reply).probe = probe             # <<<<<<<<<<<<<<
 *             # if iphdr.ip_len >= 136 + sizeof(icmpexthdr) + sizeof(icmpext):
 *             #     icmpext_s = <icmpext *>(packet_orig + 136 + sizeof(icmpexthdr))
 */
      __Pyx_INCREF(((PyObject *)__pyx_v_probe));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_probe));
      __Pyx_GOTREF(((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_reply)->probe);
      __Pyx_DECREF(((PyObject *)((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_reply)->probe));
      ((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)__pyx_v_reply)->probe = __pyx_v_probe;

      /* "cloudtrace/read/packet.pyx":303
 *             #     if icmpext_s.classnum == 1 and icmpext_s.ctype == 1:
 *             #         (<ICMPReply>reply).ie = scamper_icmpext_alloc(icmpext_s.classnum, icmpext_s.ctype, ntohs(icmpext_s.length), packet_orig + 136 + sizeof(icmpexthdr) + sizeof(icmpext))
 *             return reply             # <<<<<<<<<<<<<<
 *     elif iphdr.ip_p == IPPROTO_UDP:
 *         if incl_len < <int>sizeof(udp):
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(((PyObject *)__pyx_v_reply));
      __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_reply);
      goto __pyx_L0;

      /* "cloudtrace/read/packet.pyx":289
 *             (<EchoReply>reply).set_icmp(iphdr, icmphdr)
 *             return reply
 *         elif icmphdr.type == 3 or icmphdr.type == 11:             # <<<<<<<<<<<<<<
 *             packet += sizeof(icmp)
 *             incl_len -= sizeof(icmp)
 */
      break;
      default: break;
    }

    /* "cloudtrace/read/packet.pyx":276
 *     packet += iphdr_len
 *     incl_len -= iphdr_len
 *     if iphdr.ip_p == IPPROTO_ICMP:             # <<<<<<<<<<<<<<
 *         if incl_len < <int>sizeof(icmp):
 *             return None
 */
    goto __pyx_L4;
  }

  /* "cloudtrace/read/packet.pyx":304
 *             #         (<ICMPReply>reply).ie = scamper_icmpext_alloc(icmpext_s.classnum, icmpext_s.ctype, ntohs(icmpext_s.length), packet_orig + 136 + sizeof(icmpexthdr) + sizeof(icmpext))
 *             return reply
 *     elif iphdr.ip_p == IPPROTO_UDP:             # <<<<<<<<<<<<<<
 *         if incl_len < <int>sizeof(udp):
 *             return None
 */
  __pyx_t_3 = ((__pyx_v_iphdr->ip_p == IPPROTO_UDP) != 0);
  if (__pyx_t_3) {

    /* "cloudtrace/read/packet.pyx":305
 *             return reply
 *     elif iphdr.ip_p == IPPROTO_UDP:
 *         if incl_len < <int>sizeof(udp):             # <<<<<<<<<<<<<<
 *             return None
 *         udphdr = <udp *>packet
 */
    __pyx_t_3 = ((__pyx_v_incl_len < ((int)(sizeof(__pyx_t_10cloudtrace_4read_5utils_udp)))) != 0);
    if (__pyx_t_3) {

      /* "cloudtrace/read/packet.pyx":306
 *     elif iphdr.ip_p == IPPROTO_UDP:
 *         if incl_len < <int>sizeof(udp):
 *             return None             # <<<<<<<<<<<<<<
 *         udphdr = <udp *>packet
 *         udptohost(udphdr)
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "cloudtrace/read/packet.pyx":305
 *             return reply
 *     elif iphdr.ip_p == IPPROTO_UDP:
 *         if incl_len < <int>sizeof(udp):             # <<<<<<<<<<<<<<
 *             return None
 *         udphdr = <udp *>packet
 */
    }

    /* "cloudtrace/read/packet.pyx":307
 *         if incl_len < <int>sizeof(udp):
 *             return None
 *         udphdr = <udp *>packet             # <<<<<<<<<<<<<<
 *         udptohost(udphdr)
 *         probe = UDPProbe.__new__(UDPProbe)
 */
    __pyx_v_udphdr = ((__pyx_t_10cloudtrace_4read_5utils_udp *)__pyx_v_packet);

    /* "cloudtrace/read/packet.pyx":308
 *             return None
 *         udphdr = <udp *>packet
 *         udptohost(udphdr)             # <<<<<<<<<<<<<<
 *         probe = UDPProbe.__new__(UDPProbe)
 *         (<UDPProbe>probe).set_udp(iphdr, udphdr)
 */
    __pyx_f_10cloudtrace_4read_5utils_udptohost(__pyx_v_udphdr);

    /* "cloudtrace/read/packet.pyx":309
 *         udphdr = <udp *>packet
 *         udptohost(udphdr)
 *         probe = UDPProbe.__new__(UDPProbe)             # <<<<<<<<<<<<<<
 *         (<UDPProbe>probe).set_udp(iphdr, udphdr)
 *         return probe
 */
    __pyx_t_2 = ((PyObject *)__pyx_tp_new_10cloudtrace_4read_6packet_UDPProbe(((PyTypeObject *)__pyx_ptype_10cloudtrace_4read_6packet_UDPProbe), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GOTREF(((PyObject *)__pyx_t_2));
    if (!(likely(__Pyx_TypeTest(((PyObject *)__pyx_t_2), __pyx_ptype_10cloudtrace_4read_6packet_Packet)))) __PYX_ERR(0, 309, __pyx_L1_error)
    __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "cloudtrace/read/packet.pyx":310
 *         udptohost(udphdr)
 *         probe = UDPProbe.__new__(UDPProbe)
 *         (<UDPProbe>probe).set_udp(iphdr, udphdr)             # <<<<<<<<<<<<<<
 *         return probe
 *     return None
 */
    ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_UDPProbe *)((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_probe)->__pyx_base.__pyx_vtab)->set_udp(((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_probe), __pyx_v_iphdr, __pyx_v_udphdr);

    /* "cloudtrace/read/packet.pyx":311
 *         probe = UDPProbe.__new__(UDPProbe)
 *         (<UDPProbe>probe).set_udp(iphdr, udphdr)
 *         return probe             # <<<<<<<<<<<<<<
 *     return None
 * 
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_probe));
    __pyx_r = __pyx_v_probe;
    goto __pyx_L0;

    /* "cloudtrace/read/packet.pyx":304
 *             #         (<ICMPReply>reply).ie = scamper_icmpext_alloc(icmpext_s.classnum, icmpext_s.ctype, ntohs(icmpext_s.length), packet_orig + 136 + sizeof(icmpexthdr) + sizeof(icmpext))
 *             return reply
 *     elif iphdr.ip_p == IPPROTO_UDP:             # <<<<<<<<<<<<<<
 *         if incl_len < <int>sizeof(udp):
 *             return None
 */
  }
  __pyx_L4:;

  /* "cloudtrace/read/packet.pyx":312
 *         (<UDPProbe>probe).set_udp(iphdr, udphdr)
 *         return probe
 *     return None             # <<<<<<<<<<<<<<
 * 
 * cpdef Packet disassemble(Pcap pcap):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":258
 *         self.probe = probe
 * 
 * cpdef Packet disassemble_packet(char *packet, int incl_len):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char *packet_orig = packet
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cloudtrace.read.packet.disassemble_packet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_probe);
  __Pyx_XDECREF((PyObject *)__pyx_v_reply);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_1disassemble_packet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_disassemble_packet[] = "disassemble_packet(char *packet, int incl_len) -> Packet";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_1disassemble_packet(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  char *__pyx_v_packet;
  int __pyx_v_incl_len;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("disassemble_packet (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_packet,&__pyx_n_s_incl_len,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_packet)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_incl_len)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("disassemble_packet", 1, 2, 2, 1); __PYX_ERR(0, 258, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "disassemble_packet") < 0)) __PYX_ERR(0, 258, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_packet = __Pyx_PyObject_AsWritableString(values[0]); if (unlikely((!__pyx_v_packet) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)
    __pyx_v_incl_len = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_incl_len == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("disassemble_packet", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 258, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.read.packet.disassemble_packet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_disassemble_packet(__pyx_self, __pyx_v_packet, __pyx_v_incl_len);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_disassemble_packet(CYTHON_UNUSED PyObject *__pyx_self, char *__pyx_v_packet, int __pyx_v_incl_len) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("disassemble_packet", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10cloudtrace_4read_6packet_disassemble_packet(__pyx_v_packet, __pyx_v_incl_len, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.disassemble_packet", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/packet.pyx":314
 *     return None
 * 
 * cpdef Packet disassemble(Pcap pcap):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t incl_len = pcap.packet_hdr.incl_len
 */

static PyObject *__pyx_pw_10cloudtrace_4read_6packet_3disassemble(PyObject *__pyx_self, PyObject *__pyx_v_pcap); /*proto*/
static struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_f_10cloudtrace_4read_6packet_disassemble(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *__pyx_v_pcap, CYTHON_UNUSED int __pyx_skip_dispatch) {
  uint32_t __pyx_v_incl_len;
  char *__pyx_v_packet;
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_pkt = 0;
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  uint32_t __pyx_t_1;
  char *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("disassemble", 0);

  /* "cloudtrace/read/packet.pyx":316
 * cpdef Packet disassemble(Pcap pcap):
 *     cdef:
 *         uint32_t incl_len = pcap.packet_hdr.incl_len             # <<<<<<<<<<<<<<
 *         char *packet = pcap.frame
 *         Packet pkt = None
 */
  __pyx_t_1 = __pyx_v_pcap->packet_hdr->incl_len;
  __pyx_v_incl_len = __pyx_t_1;

  /* "cloudtrace/read/packet.pyx":317
 *     cdef:
 *         uint32_t incl_len = pcap.packet_hdr.incl_len
 *         char *packet = pcap.frame             # <<<<<<<<<<<<<<
 *         Packet pkt = None
 *         timeval time
 */
  __pyx_t_2 = __pyx_v_pcap->frame;
  __pyx_v_packet = __pyx_t_2;

  /* "cloudtrace/read/packet.pyx":318
 *         uint32_t incl_len = pcap.packet_hdr.incl_len
 *         char *packet = pcap.frame
 *         Packet pkt = None             # <<<<<<<<<<<<<<
 *         timeval time
 *     if incl_len < sizeof(ethhdr):
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_pkt = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None);

  /* "cloudtrace/read/packet.pyx":320
 *         Packet pkt = None
 *         timeval time
 *     if incl_len < sizeof(ethhdr):             # <<<<<<<<<<<<<<
 *         return None
 *     packet += sizeof(ethhdr)
 */
  __pyx_t_3 = ((__pyx_v_incl_len < (sizeof(__pyx_t_10cloudtrace_4read_5utils_ethhdr))) != 0);
  if (__pyx_t_3) {

    /* "cloudtrace/read/packet.pyx":321
 *         timeval time
 *     if incl_len < sizeof(ethhdr):
 *         return None             # <<<<<<<<<<<<<<
 *     packet += sizeof(ethhdr)
 *     incl_len -= sizeof(ethhdr)
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "cloudtrace/read/packet.pyx":320
 *         Packet pkt = None
 *         timeval time
 *     if incl_len < sizeof(ethhdr):             # <<<<<<<<<<<<<<
 *         return None
 *     packet += sizeof(ethhdr)
 */
  }

  /* "cloudtrace/read/packet.pyx":322
 *     if incl_len < sizeof(ethhdr):
 *         return None
 *     packet += sizeof(ethhdr)             # <<<<<<<<<<<<<<
 *     incl_len -= sizeof(ethhdr)
 *     pkt = disassemble_packet(packet, incl_len)
 */
  __pyx_v_packet = (__pyx_v_packet + (sizeof(__pyx_t_10cloudtrace_4read_5utils_ethhdr)));

  /* "cloudtrace/read/packet.pyx":323
 *         return None
 *     packet += sizeof(ethhdr)
 *     incl_len -= sizeof(ethhdr)             # <<<<<<<<<<<<<<
 *     pkt = disassemble_packet(packet, incl_len)
 *     if pkt is None:
 */
  __pyx_v_incl_len = (__pyx_v_incl_len - (sizeof(__pyx_t_10cloudtrace_4read_5utils_ethhdr)));

  /* "cloudtrace/read/packet.pyx":324
 *     packet += sizeof(ethhdr)
 *     incl_len -= sizeof(ethhdr)
 *     pkt = disassemble_packet(packet, incl_len)             # <<<<<<<<<<<<<<
 *     if pkt is None:
 *         return None
 */
  __pyx_t_4 = ((PyObject *)__pyx_f_10cloudtrace_4read_6packet_disassemble_packet(__pyx_v_packet, __pyx_v_incl_len, 0)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF_SET(__pyx_v_pkt, ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_t_4));
  __pyx_t_4 = 0;

  /* "cloudtrace/read/packet.pyx":325
 *     incl_len -= sizeof(ethhdr)
 *     pkt = disassemble_packet(packet, incl_len)
 *     if pkt is None:             # <<<<<<<<<<<<<<
 *         return None
 *     # time.tv_sec = pcap.ts_sec
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_pkt) == Py_None);
  __pyx_t_5 = (__pyx_t_3 != 0);
  if (__pyx_t_5) {

    /* "cloudtrace/read/packet.pyx":326
 *     pkt = disassemble_packet(packet, incl_len)
 *     if pkt is None:
 *         return None             # <<<<<<<<<<<<<<
 *     # time.tv_sec = pcap.ts_sec
 *     # time.tv_usec = pcap.ts_usec
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "cloudtrace/read/packet.pyx":325
 *     incl_len -= sizeof(ethhdr)
 *     pkt = disassemble_packet(packet, incl_len)
 *     if pkt is None:             # <<<<<<<<<<<<<<
 *         return None
 *     # time.tv_sec = pcap.ts_sec
 */
  }

  /* "cloudtrace/read/packet.pyx":330
 *     # time.tv_usec = pcap.ts_usec
 *     # memcpy(&pkt.time, &time, sizeof(timeval))
 *     pkt.time.tv_sec = pcap.packet_hdr.ts_sec             # <<<<<<<<<<<<<<
 *     pkt.time.tv_usec = pcap.packet_hdr.ts_usec
 *     return pkt
 */
  __pyx_t_1 = __pyx_v_pcap->packet_hdr->ts_sec;
  __pyx_v_pkt->time->tv_sec = __pyx_t_1;

  /* "cloudtrace/read/packet.pyx":331
 *     # memcpy(&pkt.time, &time, sizeof(timeval))
 *     pkt.time.tv_sec = pcap.packet_hdr.ts_sec
 *     pkt.time.tv_usec = pcap.packet_hdr.ts_usec             # <<<<<<<<<<<<<<
 *     return pkt
 */
  __pyx_t_1 = __pyx_v_pcap->packet_hdr->ts_usec;
  __pyx_v_pkt->time->tv_usec = __pyx_t_1;

  /* "cloudtrace/read/packet.pyx":332
 *     pkt.time.tv_sec = pcap.packet_hdr.ts_sec
 *     pkt.time.tv_usec = pcap.packet_hdr.ts_usec
 *     return pkt             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_pkt));
  __pyx_r = __pyx_v_pkt;
  goto __pyx_L0;

  /* "cloudtrace/read/packet.pyx":314
 *     return None
 * 
 * cpdef Packet disassemble(Pcap pcap):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t incl_len = pcap.packet_hdr.incl_len
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cloudtrace.read.packet.disassemble", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_pkt);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_3disassemble(PyObject *__pyx_self, PyObject *__pyx_v_pcap); /*proto*/
static char __pyx_doc_10cloudtrace_4read_6packet_2disassemble[] = "disassemble(Pcap pcap) -> Packet";
static PyObject *__pyx_pw_10cloudtrace_4read_6packet_3disassemble(PyObject *__pyx_self, PyObject *__pyx_v_pcap) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("disassemble (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_pcap), __pyx_ptype_10cloudtrace_4read_4pcap_Pcap, 1, "pcap", 0))) __PYX_ERR(0, 314, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_4read_6packet_2disassemble(__pyx_self, ((struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *)__pyx_v_pcap));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_6packet_2disassemble(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *__pyx_v_pcap) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("disassemble", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10cloudtrace_4read_6packet_disassemble(__pyx_v_pcap, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.packet.disassemble", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet __pyx_vtable_10cloudtrace_4read_6packet_Packet;

static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_Packet(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)o);
  p->__pyx_vtab = __pyx_vtabptr_10cloudtrace_4read_6packet_Packet;
  p->dst = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->pid_b = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->dst_b = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_10cloudtrace_4read_6packet_6Packet_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10cloudtrace_4read_6packet_Packet(PyObject *o) {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *p = (struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10cloudtrace_4read_6packet_6Packet_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->dst);
  Py_CLEAR(p->pid_b);
  Py_CLEAR(p->dst_b);
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_10cloudtrace_4read_6packet_6Packet_ptype(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_6packet_6Packet_5ptype_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_6packet_6Packet_ptype(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_6packet_6Packet_5ptype_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_6packet_6Packet_isprobe(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_6packet_6Packet_7isprobe_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_6packet_6Packet_isprobe(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_6packet_6Packet_7isprobe_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_6packet_6Packet_dst(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_6packet_6Packet_3dst_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_6packet_6Packet_dst(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_6packet_6Packet_3dst_3__set__(o, v);
  }
  else {
    return __pyx_pw_10cloudtrace_4read_6packet_6Packet_3dst_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_10cloudtrace_4read_6packet_Packet[] = {
  {"get_hdr", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_6Packet_5get_hdr, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_6Packet_4get_hdr},
  {"probe_ttl", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_6Packet_7probe_ttl, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_6Packet_6probe_ttl},
  {"get_pid", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_6Packet_9get_pid, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_6Packet_8get_pid},
  {"ipid", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_6Packet_11ipid, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_6Packet_10ipid},
  {"get_dst", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_6Packet_13get_dst, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_6Packet_12get_dst},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_6Packet_15__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_6Packet_14__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_6Packet_17__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_4read_6packet_6Packet_16__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10cloudtrace_4read_6packet_Packet[] = {
  {(char *)"ptype", __pyx_getprop_10cloudtrace_4read_6packet_6Packet_ptype, __pyx_setprop_10cloudtrace_4read_6packet_6Packet_ptype, (char *)"ptype: 'uint8_t'", 0},
  {(char *)"isprobe", __pyx_getprop_10cloudtrace_4read_6packet_6Packet_isprobe, __pyx_setprop_10cloudtrace_4read_6packet_6Packet_isprobe, (char *)"isprobe: 'bool'", 0},
  {(char *)"dst", __pyx_getprop_10cloudtrace_4read_6packet_6Packet_dst, __pyx_setprop_10cloudtrace_4read_6packet_6Packet_dst, (char *)"dst: bytes", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_4read_6packet_Packet = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.read.packet.Packet", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_Packet), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_Packet, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_4read_6packet_Packet, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10cloudtrace_4read_6packet_Packet, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_4read_6packet_Packet, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPProbe __pyx_vtable_10cloudtrace_4read_6packet_ICMPProbe;

static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_ICMPProbe(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *p;
  PyObject *o = __pyx_tp_new_10cloudtrace_4read_6packet_Packet(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet*)__pyx_vtabptr_10cloudtrace_4read_6packet_ICMPProbe;
  if (unlikely(__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10cloudtrace_4read_6packet_ICMPProbe(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_Packet(o);
}

static PyMethodDef __pyx_methods_10cloudtrace_4read_6packet_ICMPProbe[] = {
  {"probe_ttl", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_5probe_ttl, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_4probe_ttl},
  {"get_pid", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_7get_pid, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_6get_pid},
  {"get_type", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_9get_type, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_8get_type},
  {"get_code", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_11get_code, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_10get_code},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_13__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_12__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPProbe_15__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_4read_6packet_9ICMPProbe_14__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_4read_6packet_ICMPProbe = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.read.packet.ICMPProbe", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_ICMPProbe, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_4read_6packet_ICMPProbe, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_4read_6packet_ICMPProbe, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_UDPProbe __pyx_vtable_10cloudtrace_4read_6packet_UDPProbe;

static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_UDPProbe(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *p;
  PyObject *o = __pyx_tp_new_10cloudtrace_4read_6packet_Packet(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet*)__pyx_vtabptr_10cloudtrace_4read_6packet_UDPProbe;
  if (unlikely(__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10cloudtrace_4read_6packet_UDPProbe(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_Packet(o);
}

static PyMethodDef __pyx_methods_10cloudtrace_4read_6packet_UDPProbe[] = {
  {"probe_ttl", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_5probe_ttl, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_8UDPProbe_4probe_ttl},
  {"get_pid", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_7get_pid, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_8UDPProbe_6get_pid},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_9__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_8UDPProbe_8__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_8UDPProbe_11__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_4read_6packet_8UDPProbe_10__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_4read_6packet_UDPProbe = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.read.packet.UDPProbe", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_UDPProbe, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_4read_6packet_UDPProbe, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_4read_6packet_UDPProbe, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply __pyx_vtable_10cloudtrace_4read_6packet_Reply;

static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_Reply(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply *p;
  PyObject *o = __pyx_tp_new_10cloudtrace_4read_6packet_Packet(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet*)__pyx_vtabptr_10cloudtrace_4read_6packet_Reply;
  p->orig_probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_10cloudtrace_4read_6packet_5Reply_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10cloudtrace_4read_6packet_Reply(PyObject *o) {
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply *p = (struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10cloudtrace_4read_6packet_5Reply_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->orig_probe);
  #if CYTHON_USE_TYPE_SLOTS
  if (PyType_IS_GC(Py_TYPE(o)->tp_base))
  #endif
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_Packet(o);
}

static int __pyx_tp_traverse_10cloudtrace_4read_6packet_Reply(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply *p = (struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)o;
  e = ((likely(__pyx_ptype_10cloudtrace_4read_6packet_Packet)) ? ((__pyx_ptype_10cloudtrace_4read_6packet_Packet->tp_traverse) ? __pyx_ptype_10cloudtrace_4read_6packet_Packet->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_10cloudtrace_4read_6packet_Reply)); if (e) return e;
  if (p->orig_probe) {
    e = (*v)(((PyObject *)p->orig_probe), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10cloudtrace_4read_6packet_Reply(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply *p = (struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)o;
  if (likely(__pyx_ptype_10cloudtrace_4read_6packet_Packet)) { if (__pyx_ptype_10cloudtrace_4read_6packet_Packet->tp_clear) __pyx_ptype_10cloudtrace_4read_6packet_Packet->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_10cloudtrace_4read_6packet_Reply);
  tmp = ((PyObject*)p->orig_probe);
  p->orig_probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_10cloudtrace_4read_6packet_Reply[] = {
  {"addr", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_5Reply_5addr, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_5Reply_4addr},
  {"probe_ipid", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_5Reply_7probe_ipid, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_5Reply_6probe_ipid},
  {"probe_size", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_5Reply_9probe_size, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_5Reply_8probe_size},
  {"reply_ttl", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_5Reply_11reply_ttl, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_5Reply_10reply_ttl},
  {"get_type", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_5Reply_13get_type, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_5Reply_12get_type},
  {"get_code", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_5Reply_15get_code, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_5Reply_14get_code},
  {"get_icmp_q_ttl", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_5Reply_17get_icmp_q_ttl, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_5Reply_16get_icmp_q_ttl},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_5Reply_19__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_5Reply_18__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_5Reply_21__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_4read_6packet_5Reply_20__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_4read_6packet_Reply = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.read.packet.Reply", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_Reply), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_Reply, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10cloudtrace_4read_6packet_Reply, /*tp_traverse*/
  __pyx_tp_clear_10cloudtrace_4read_6packet_Reply, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_4read_6packet_Reply, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_4read_6packet_Reply, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_EchoReply __pyx_vtable_10cloudtrace_4read_6packet_EchoReply;

static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_EchoReply(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *p;
  PyObject *o = __pyx_tp_new_10cloudtrace_4read_6packet_Reply(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet*)__pyx_vtabptr_10cloudtrace_4read_6packet_EchoReply;
  if (unlikely(__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_10cloudtrace_4read_6packet_EchoReply[] = {
  {"probe_ttl", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_3probe_ttl, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_2probe_ttl},
  {"get_pid", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_5get_pid, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_4get_pid},
  {"probe_ipid", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_7probe_ipid, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_6probe_ipid},
  {"probe_size", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_9probe_size, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_8probe_size},
  {"get_icmp_q_ttl", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_11get_icmp_q_ttl, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_10get_icmp_q_ttl},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_13__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_12__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9EchoReply_15__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_4read_6packet_9EchoReply_14__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_4read_6packet_EchoReply = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.read.packet.EchoReply", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_Reply, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10cloudtrace_4read_6packet_Reply, /*tp_traverse*/
  __pyx_tp_clear_10cloudtrace_4read_6packet_Reply, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_4read_6packet_EchoReply, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_4read_6packet_EchoReply, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPReply __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply;

static PyObject *__pyx_tp_new_10cloudtrace_4read_6packet_ICMPReply(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *p;
  PyObject *o = __pyx_tp_new_10cloudtrace_4read_6packet_Reply(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet*)__pyx_vtabptr_10cloudtrace_4read_6packet_ICMPReply;
  p->probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10cloudtrace_4read_6packet_ICMPReply(PyObject *o) {
  struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *p = (struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->probe);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_Reply(o);
}

static int __pyx_tp_traverse_10cloudtrace_4read_6packet_ICMPReply(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *p = (struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)o;
  e = __pyx_tp_traverse_10cloudtrace_4read_6packet_Reply(o, v, a); if (e) return e;
  if (p->probe) {
    e = (*v)(((PyObject *)p->probe), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10cloudtrace_4read_6packet_ICMPReply(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *p = (struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *)o;
  __pyx_tp_clear_10cloudtrace_4read_6packet_Reply(o);
  tmp = ((PyObject*)p->probe);
  p->probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_10cloudtrace_4read_6packet_ICMPReply[] = {
  {"probe_ttl", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_5probe_ttl, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_4probe_ttl},
  {"probe_size", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_7probe_size, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_6probe_size},
  {"get_pid", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_9get_pid, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_8get_pid},
  {"probe_ipid", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_11probe_ipid, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_10probe_ipid},
  {"get_icmp_q_ttl", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_13get_icmp_q_ttl, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_12get_icmp_q_ttl},
  {"set_probe", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_15set_probe, METH_O, __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_14set_probe},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_17__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_16__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_9ICMPReply_19__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_4read_6packet_9ICMPReply_18__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_4read_6packet_ICMPReply = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.read.packet.ICMPReply", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_4read_6packet_ICMPReply, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10cloudtrace_4read_6packet_ICMPReply, /*tp_traverse*/
  __pyx_tp_clear_10cloudtrace_4read_6packet_ICMPReply, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_4read_6packet_ICMPReply, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_4read_6packet_ICMPReply, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"disassemble_packet", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10cloudtrace_4read_6packet_1disassemble_packet, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10cloudtrace_4read_6packet_disassemble_packet},
  {"disassemble", (PyCFunction)__pyx_pw_10cloudtrace_4read_6packet_3disassemble, METH_O, __pyx_doc_10cloudtrace_4read_6packet_2disassemble},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_packet(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_packet},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "packet",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ECHO_REPLY, __pyx_k_ECHO_REPLY, sizeof(__pyx_k_ECHO_REPLY), 0, 0, 1, 1},
  {&__pyx_n_s_EchoReply, __pyx_k_EchoReply, sizeof(__pyx_k_EchoReply), 0, 0, 1, 1},
  {&__pyx_n_s_ICMPProbe, __pyx_k_ICMPProbe, sizeof(__pyx_k_ICMPProbe), 0, 0, 1, 1},
  {&__pyx_n_s_ICMPReply, __pyx_k_ICMPReply, sizeof(__pyx_k_ICMPReply), 0, 0, 1, 1},
  {&__pyx_n_s_ICMP_PROBE, __pyx_k_ICMP_PROBE, sizeof(__pyx_k_ICMP_PROBE), 0, 0, 1, 1},
  {&__pyx_n_s_ICMP_REPLY, __pyx_k_ICMP_REPLY, sizeof(__pyx_k_ICMP_REPLY), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_Packet, __pyx_k_Packet, sizeof(__pyx_k_Packet), 0, 0, 1, 1},
  {&__pyx_kp_u_Probe_is_not_set_yet, __pyx_k_Probe_is_not_set_yet, sizeof(__pyx_k_Probe_is_not_set_yet), 0, 1, 0, 0},
  {&__pyx_n_s_Reply, __pyx_k_Reply, sizeof(__pyx_k_Reply), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe, __pyx_k_UDPProbe, sizeof(__pyx_k_UDPProbe), 0, 0, 1, 1},
  {&__pyx_n_s_UDP_PROBE, __pyx_k_UDP_PROBE, sizeof(__pyx_k_UDP_PROBE), 0, 0, 1, 1},
  {&__pyx_n_s_addr, __pyx_k_addr, sizeof(__pyx_k_addr), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_get_code, __pyx_k_get_code, sizeof(__pyx_k_get_code), 0, 0, 1, 1},
  {&__pyx_n_s_get_dst, __pyx_k_get_dst, sizeof(__pyx_k_get_dst), 0, 0, 1, 1},
  {&__pyx_n_s_get_icmp_q_ttl, __pyx_k_get_icmp_q_ttl, sizeof(__pyx_k_get_icmp_q_ttl), 0, 0, 1, 1},
  {&__pyx_n_s_get_pid, __pyx_k_get_pid, sizeof(__pyx_k_get_pid), 0, 0, 1, 1},
  {&__pyx_n_s_get_type, __pyx_k_get_type, sizeof(__pyx_k_get_type), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_incl_len, __pyx_k_incl_len, sizeof(__pyx_k_incl_len), 0, 0, 1, 1},
  {&__pyx_n_s_ip_dst, __pyx_k_ip_dst, sizeof(__pyx_k_ip_dst), 0, 0, 1, 1},
  {&__pyx_n_s_ip_hl, __pyx_k_ip_hl, sizeof(__pyx_k_ip_hl), 0, 0, 1, 1},
  {&__pyx_n_s_ip_len, __pyx_k_ip_len, sizeof(__pyx_k_ip_len), 0, 0, 1, 1},
  {&__pyx_n_s_ip_off, __pyx_k_ip_off, sizeof(__pyx_k_ip_off), 0, 0, 1, 1},
  {&__pyx_n_s_ip_p, __pyx_k_ip_p, sizeof(__pyx_k_ip_p), 0, 0, 1, 1},
  {&__pyx_n_s_ip_src, __pyx_k_ip_src, sizeof(__pyx_k_ip_src), 0, 0, 1, 1},
  {&__pyx_n_s_ip_sum, __pyx_k_ip_sum, sizeof(__pyx_k_ip_sum), 0, 0, 1, 1},
  {&__pyx_n_s_ip_tos, __pyx_k_ip_tos, sizeof(__pyx_k_ip_tos), 0, 0, 1, 1},
  {&__pyx_n_s_ip_ttl, __pyx_k_ip_ttl, sizeof(__pyx_k_ip_ttl), 0, 0, 1, 1},
  {&__pyx_n_s_ipid, __pyx_k_ipid, sizeof(__pyx_k_ipid), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_packet, __pyx_k_packet, sizeof(__pyx_k_packet), 0, 0, 1, 1},
  {&__pyx_n_s_probe_ipid, __pyx_k_probe_ipid, sizeof(__pyx_k_probe_ipid), 0, 0, 1, 1},
  {&__pyx_n_s_probe_size, __pyx_k_probe_size, sizeof(__pyx_k_probe_size), 0, 0, 1, 1},
  {&__pyx_n_s_probe_ttl, __pyx_k_probe_ttl, sizeof(__pyx_k_probe_ttl), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_reply_ttl, __pyx_k_reply_ttl, sizeof(__pyx_k_reply_ttl), 0, 0, 1, 1},
  {&__pyx_n_s_set_probe, __pyx_k_set_probe, sizeof(__pyx_k_set_probe), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 31, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "cloudtrace/read/packet.pyx":204
 *     cdef uint8_t *get_dst_c(self):
 *         if self.probe is None:
 *             raise Exception('Probe is not set yet.')             # <<<<<<<<<<<<<<
 *         return self.probe.iphdr.ip_dst
 * 
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_Probe_is_not_set_yet); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_ICMP_PROBE, (void *)&ICMP_PROBE, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_UDP_PROBE, (void *)&UDP_PROBE, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_ECHO_REPLY, (void *)&ECHO_REPLY, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_ICMP_REPLY, (void *)&ICMP_REPLY, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("disassemble_packet", (void (*)(void))__pyx_f_10cloudtrace_4read_6packet_disassemble_packet, "struct __pyx_obj_10cloudtrace_4read_6packet_Packet *(char *, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("disassemble", (void (*)(void))__pyx_f_10cloudtrace_4read_6packet_disassemble, "struct __pyx_obj_10cloudtrace_4read_6packet_Packet *(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_10cloudtrace_4read_6packet_Packet = &__pyx_vtable_10cloudtrace_4read_6packet_Packet;
  __pyx_vtable_10cloudtrace_4read_6packet_Packet.get_dst_c = (uint8_t *(*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *))__pyx_f_10cloudtrace_4read_6packet_6Packet_get_dst_c;
  __pyx_vtable_10cloudtrace_4read_6packet_Packet.probe_ttl = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_6Packet_probe_ttl;
  __pyx_vtable_10cloudtrace_4read_6packet_Packet.get_pid = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_6Packet_get_pid;
  __pyx_vtable_10cloudtrace_4read_6packet_Packet.ipid = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_6Packet_ipid;
  __pyx_vtable_10cloudtrace_4read_6packet_Packet.set_ip = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, __pyx_t_10cloudtrace_4read_5utils_ip *))__pyx_f_10cloudtrace_4read_6packet_6Packet_set_ip;
  __pyx_vtable_10cloudtrace_4read_6packet_Packet.get_dst = (PyObject *(*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_6Packet_get_dst;
  if (PyType_Ready(&__pyx_type_10cloudtrace_4read_6packet_Packet) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_4read_6packet_Packet.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_4read_6packet_Packet.tp_dictoffset && __pyx_type_10cloudtrace_4read_6packet_Packet.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_4read_6packet_Packet.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_4read_6packet_Packet.tp_dict, __pyx_vtabptr_10cloudtrace_4read_6packet_Packet) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Packet, (PyObject *)&__pyx_type_10cloudtrace_4read_6packet_Packet) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_4read_6packet_Packet) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_Packet = &__pyx_type_10cloudtrace_4read_6packet_Packet;
  __pyx_vtabptr_10cloudtrace_4read_6packet_ICMPProbe = &__pyx_vtable_10cloudtrace_4read_6packet_ICMPProbe;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPProbe.__pyx_base = *__pyx_vtabptr_10cloudtrace_4read_6packet_Packet;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPProbe.__pyx_base.probe_ttl = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_probe_ttl;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPProbe.__pyx_base.get_pid = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_pid;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPProbe.get_type = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_type;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPProbe.get_code = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_get_code;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPProbe.set_icmp = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_icmp *))__pyx_f_10cloudtrace_4read_6packet_9ICMPProbe_set_icmp;
  __pyx_type_10cloudtrace_4read_6packet_ICMPProbe.tp_base = __pyx_ptype_10cloudtrace_4read_6packet_Packet;
  if (PyType_Ready(&__pyx_type_10cloudtrace_4read_6packet_ICMPProbe) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_4read_6packet_ICMPProbe.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_4read_6packet_ICMPProbe.tp_dictoffset && __pyx_type_10cloudtrace_4read_6packet_ICMPProbe.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_4read_6packet_ICMPProbe.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_4read_6packet_ICMPProbe.tp_dict, __pyx_vtabptr_10cloudtrace_4read_6packet_ICMPProbe) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ICMPProbe, (PyObject *)&__pyx_type_10cloudtrace_4read_6packet_ICMPProbe) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_4read_6packet_ICMPProbe) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_ICMPProbe = &__pyx_type_10cloudtrace_4read_6packet_ICMPProbe;
  __pyx_vtabptr_10cloudtrace_4read_6packet_UDPProbe = &__pyx_vtable_10cloudtrace_4read_6packet_UDPProbe;
  __pyx_vtable_10cloudtrace_4read_6packet_UDPProbe.__pyx_base = *__pyx_vtabptr_10cloudtrace_4read_6packet_Packet;
  __pyx_vtable_10cloudtrace_4read_6packet_UDPProbe.__pyx_base.probe_ttl = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_8UDPProbe_probe_ttl;
  __pyx_vtable_10cloudtrace_4read_6packet_UDPProbe.__pyx_base.get_pid = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_8UDPProbe_get_pid;
  __pyx_vtable_10cloudtrace_4read_6packet_UDPProbe.set_udp = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_udp *))__pyx_f_10cloudtrace_4read_6packet_8UDPProbe_set_udp;
  __pyx_type_10cloudtrace_4read_6packet_UDPProbe.tp_base = __pyx_ptype_10cloudtrace_4read_6packet_Packet;
  if (PyType_Ready(&__pyx_type_10cloudtrace_4read_6packet_UDPProbe) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_4read_6packet_UDPProbe.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_4read_6packet_UDPProbe.tp_dictoffset && __pyx_type_10cloudtrace_4read_6packet_UDPProbe.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_4read_6packet_UDPProbe.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_4read_6packet_UDPProbe.tp_dict, __pyx_vtabptr_10cloudtrace_4read_6packet_UDPProbe) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UDPProbe, (PyObject *)&__pyx_type_10cloudtrace_4read_6packet_UDPProbe) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_4read_6packet_UDPProbe) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_UDPProbe = &__pyx_type_10cloudtrace_4read_6packet_UDPProbe;
  __pyx_vtabptr_10cloudtrace_4read_6packet_Reply = &__pyx_vtable_10cloudtrace_4read_6packet_Reply;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.__pyx_base = *__pyx_vtabptr_10cloudtrace_4read_6packet_Packet;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply._create_hop = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_7combine_hop_t *))__pyx_f_10cloudtrace_4read_6packet_5Reply__create_hop;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.addr = (PyObject *(*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_5Reply_addr;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.create_hop = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_7combine_hop_t *))__pyx_f_10cloudtrace_4read_6packet_5Reply_create_hop;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.probe_ipid = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_5Reply_probe_ipid;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.probe_size = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_5Reply_probe_size;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.reply_ttl = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_5Reply_reply_ttl;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.get_type = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_5Reply_get_type;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.get_code = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_5Reply_get_code;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.get_icmp_q_ttl = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_5Reply_get_icmp_q_ttl;
  __pyx_vtable_10cloudtrace_4read_6packet_Reply.set_icmp = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_icmp *))__pyx_f_10cloudtrace_4read_6packet_5Reply_set_icmp;
  __pyx_type_10cloudtrace_4read_6packet_Reply.tp_base = __pyx_ptype_10cloudtrace_4read_6packet_Packet;
  if (PyType_Ready(&__pyx_type_10cloudtrace_4read_6packet_Reply) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_4read_6packet_Reply.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_4read_6packet_Reply.tp_dictoffset && __pyx_type_10cloudtrace_4read_6packet_Reply.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_4read_6packet_Reply.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_4read_6packet_Reply.tp_dict, __pyx_vtabptr_10cloudtrace_4read_6packet_Reply) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Reply, (PyObject *)&__pyx_type_10cloudtrace_4read_6packet_Reply) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_4read_6packet_Reply) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_Reply = &__pyx_type_10cloudtrace_4read_6packet_Reply;
  __pyx_vtabptr_10cloudtrace_4read_6packet_EchoReply = &__pyx_vtable_10cloudtrace_4read_6packet_EchoReply;
  __pyx_vtable_10cloudtrace_4read_6packet_EchoReply.__pyx_base = *__pyx_vtabptr_10cloudtrace_4read_6packet_Reply;
  __pyx_vtable_10cloudtrace_4read_6packet_EchoReply.__pyx_base.__pyx_base.get_dst_c = (uint8_t *(*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *))__pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_dst_c;
  __pyx_vtable_10cloudtrace_4read_6packet_EchoReply.__pyx_base.__pyx_base.probe_ttl = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_ttl;
  __pyx_vtable_10cloudtrace_4read_6packet_EchoReply.__pyx_base.__pyx_base.get_pid = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_pid;
  __pyx_vtable_10cloudtrace_4read_6packet_EchoReply.__pyx_base.probe_ipid = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_ipid;
  __pyx_vtable_10cloudtrace_4read_6packet_EchoReply.__pyx_base.probe_size = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9EchoReply_probe_size;
  __pyx_vtable_10cloudtrace_4read_6packet_EchoReply.__pyx_base.get_icmp_q_ttl = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9EchoReply_get_icmp_q_ttl;
  __pyx_vtable_10cloudtrace_4read_6packet_EchoReply.__pyx_base.set_icmp = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_icmp *))__pyx_f_10cloudtrace_4read_6packet_9EchoReply_set_icmp;
  __pyx_type_10cloudtrace_4read_6packet_EchoReply.tp_base = __pyx_ptype_10cloudtrace_4read_6packet_Reply;
  if (PyType_Ready(&__pyx_type_10cloudtrace_4read_6packet_EchoReply) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_4read_6packet_EchoReply.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_4read_6packet_EchoReply.tp_dictoffset && __pyx_type_10cloudtrace_4read_6packet_EchoReply.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_4read_6packet_EchoReply.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_4read_6packet_EchoReply.tp_dict, __pyx_vtabptr_10cloudtrace_4read_6packet_EchoReply) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_EchoReply, (PyObject *)&__pyx_type_10cloudtrace_4read_6packet_EchoReply) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_4read_6packet_EchoReply) < 0) __PYX_ERR(0, 158, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_EchoReply = &__pyx_type_10cloudtrace_4read_6packet_EchoReply;
  __pyx_vtabptr_10cloudtrace_4read_6packet_ICMPReply = &__pyx_vtable_10cloudtrace_4read_6packet_ICMPReply;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.__pyx_base = *__pyx_vtabptr_10cloudtrace_4read_6packet_Reply;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.__pyx_base.__pyx_base.get_dst_c = (uint8_t *(*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *))__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_dst_c;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.__pyx_base.__pyx_base.probe_ttl = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_ttl;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.__pyx_base.__pyx_base.get_pid = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_pid;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.__pyx_base._create_hop = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_7combine_hop_t *))__pyx_f_10cloudtrace_4read_6packet_9ICMPReply__create_hop;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.__pyx_base.probe_ipid = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_ipid;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.__pyx_base.probe_size = (uint16_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_probe_size;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.__pyx_base.get_icmp_q_ttl = (uint8_t (*)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_get_icmp_q_ttl;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.create_pseudo = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *, __pyx_t_10cloudtrace_4read_5utils_udppseudo *, unsigned char *))__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_create_pseudo;
  __pyx_vtable_10cloudtrace_4read_6packet_ICMPReply.set_probe = (void (*)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_6packet_9ICMPReply_set_probe;
  __pyx_type_10cloudtrace_4read_6packet_ICMPReply.tp_base = __pyx_ptype_10cloudtrace_4read_6packet_Reply;
  if (PyType_Ready(&__pyx_type_10cloudtrace_4read_6packet_ICMPReply) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_4read_6packet_ICMPReply.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_4read_6packet_ICMPReply.tp_dictoffset && __pyx_type_10cloudtrace_4read_6packet_ICMPReply.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_4read_6packet_ICMPReply.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_4read_6packet_ICMPReply.tp_dict, __pyx_vtabptr_10cloudtrace_4read_6packet_ICMPReply) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ICMPReply, (PyObject *)&__pyx_type_10cloudtrace_4read_6packet_ICMPReply) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_4read_6packet_ICMPReply) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_ICMPReply = &__pyx_type_10cloudtrace_4read_6packet_ICMPReply;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.read.combine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10cloudtrace_4read_7combine_Trace = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.combine", "Trace", sizeof(struct __pyx_obj_10cloudtrace_4read_7combine_Trace), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_7combine_Trace) __PYX_ERR(3, 24, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_7combine_Trace = (struct __pyx_vtabstruct_10cloudtrace_4read_7combine_Trace*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_7combine_Trace->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_7combine_Trace)) __PYX_ERR(3, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.read.pcap"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10cloudtrace_4read_4pcap_Pcap = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.pcap", "Pcap", sizeof(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_4pcap_Pcap) __PYX_ERR(4, 22, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_4pcap_Pcap = (struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_4pcap_Pcap->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_4pcap_Pcap)) __PYX_ERR(4, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.read.utils"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "checksum", (void (**)(void))&__pyx_f_10cloudtrace_4read_5utils_checksum, "uint16_t (void *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "iptohost", (void (**)(void))&__pyx_f_10cloudtrace_4read_5utils_iptohost, "void (__pyx_t_10cloudtrace_4read_5utils_ip *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "icmptohost", (void (**)(void))&__pyx_f_10cloudtrace_4read_5utils_icmptohost, "void (__pyx_t_10cloudtrace_4read_5utils_icmp *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "udptohost", (void (**)(void))&__pyx_f_10cloudtrace_4read_5utils_udptohost, "void (__pyx_t_10cloudtrace_4read_5utils_udp *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initpacket(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initpacket(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_packet(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_packet(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_packet(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'packet' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_packet(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("packet", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_cloudtrace__read__packet) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cloudtrace.read.packet")) {
      if (unlikely(PyDict_SetItemString(modules, "cloudtrace.read.packet", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely(__Pyx_modinit_variable_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "cloudtrace/read/packet.pyx":5
 * 
 * cdef:
 *     public int ICMP_PROBE = 1             # <<<<<<<<<<<<<<
 *     public int UDP_PROBE = 2
 *     public int ECHO_REPLY = 3
 */
  ICMP_PROBE = 1;

  /* "cloudtrace/read/packet.pyx":6
 * cdef:
 *     public int ICMP_PROBE = 1
 *     public int UDP_PROBE = 2             # <<<<<<<<<<<<<<
 *     public int ECHO_REPLY = 3
 *     public int ICMP_REPLY = 4
 */
  UDP_PROBE = 2;

  /* "cloudtrace/read/packet.pyx":7
 *     public int ICMP_PROBE = 1
 *     public int UDP_PROBE = 2
 *     public int ECHO_REPLY = 3             # <<<<<<<<<<<<<<
 *     public int ICMP_REPLY = 4
 * 
 */
  ECHO_REPLY = 3;

  /* "cloudtrace/read/packet.pyx":8
 *     public int UDP_PROBE = 2
 *     public int ECHO_REPLY = 3
 *     public int ICMP_REPLY = 4             # <<<<<<<<<<<<<<
 * 
 * cdef class Packet:
 */
  ICMP_REPLY = 4;

  /* "cloudtrace/read/packet.pyx":1
 * from libc.string cimport memcpy, strcpy, memset, memcmp             # <<<<<<<<<<<<<<
 * from libc.stdlib cimport free, calloc
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init cloudtrace.read.packet", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cloudtrace.read.packet");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* decode_c_bytes */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

static PyObject* __pyx_convert__to_py_struct____pyx_t_10cloudtrace_4read_5utils_ip_s(struct __pyx_t_10cloudtrace_4read_5utils_ip_s s) {
  PyObject* res;
  PyObject* member;
  res = __Pyx_PyDict_NewPresized(10); if (unlikely(!res)) return NULL;
  member = __Pyx_PyInt_From_uint8_t(s.ip_hl); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_ip_hl, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_uint8_t(s.ip_tos); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_ip_tos, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_uint16_t(s.ip_len); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_ip_len, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_uint16_t(s.id); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_id, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_uint16_t(s.ip_off); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_ip_off, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_uint8_t(s.ip_ttl); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_ip_ttl, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_uint8_t(s.ip_p); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_ip_p, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyInt_From_uint16_t(s.ip_sum); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_ip_sum, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyObject_FromCString(s.ip_src); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_ip_src, member) < 0)) goto bad;
  Py_DECREF(member);
  member = __Pyx_PyObject_FromCString(s.ip_dst); if (unlikely(!member)) goto bad;
  if (unlikely(PyDict_SetItem(res, __pyx_n_s_ip_dst, member) < 0)) goto bad;
  Py_DECREF(member);
  return res;
  bad:
  Py_XDECREF(member);
  Py_DECREF(res);
  return NULL;
}
/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint16_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint16_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint16_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint16_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint8_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint8_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint8_t) -1;
        }
    } else {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint16_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint16_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint16_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint16_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint16_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 2 * PyLong_SHIFT) {
                            return (uint16_t) (((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 3 * PyLong_SHIFT) {
                            return (uint16_t) (((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 4 * PyLong_SHIFT) {
                            return (uint16_t) (((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint16_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint16_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint16_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint16_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint16_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint16_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint16_t) ((((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint16_t) ((((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint16_t) ((((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint16_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint16_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint16_t) -1;
        }
    } else {
        uint16_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint16_t) -1;
        val = __Pyx_PyInt_As_uint16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint16_t");
    return (uint16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint16_t");
    return (uint16_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
