/* Generated by Cython 0.29.32 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "cloudtrace.read.convert",
        "sources": [
            "cloudtrace/read/convert.pyx"
        ]
    },
    "module_name": "cloudtrace.read.convert"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_32"
#define CYTHON_HEX_VERSION 0x001D20F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cloudtrace__read__convert
#define __PYX_HAVE_API__cloudtrace__read__convert
/* Early includes */
#include <stdint.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include "netinet/ip.h"
#include <string.h>
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "cloudtrace/read/convert.pyx",
  "stringsource",
  "cloudtrace/read/convert.pxd",
  "cloudtrace/read/pcap.pxd",
  "cloudtrace/read/combine.pxd",
  "cloudtrace/read/packet.pxd",
  "cloudtrace/read/linkedlist.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap;
struct __pyx_obj_10cloudtrace_4read_7combine_Trace;
struct __pyx_obj_10cloudtrace_4read_6packet_Packet;
struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe;
struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe;
struct __pyx_obj_10cloudtrace_4read_6packet_Reply;
struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply;
struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply;
struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList;
struct __pyx_obj_10cloudtrace_4read_10linkedlist_Node;
struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList2;
struct __pyx_obj_10cloudtrace_4read_7convert_Convert;
struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace;
struct __pyx_t_10cloudtrace_4read_5utils_ethhdr_s;
struct __pyx_t_10cloudtrace_4read_5utils_ip_s;
struct __pyx_t_10cloudtrace_4read_5utils_echo_s;
struct __pyx_t_10cloudtrace_4read_5utils_unused_s;
union __pyx_t_10cloudtrace_4read_5utils_icmpunion;
struct __pyx_t_10cloudtrace_4read_5utils_icmp_s;
struct __pyx_t_10cloudtrace_4read_5utils_udp_s;
struct __pyx_t_10cloudtrace_4read_5utils_udppseudo_s;
struct __pyx_t_10cloudtrace_4read_5utils_ip6_hdrctl;
union __pyx_t_10cloudtrace_4read_5utils_ip6_un;
struct __pyx_t_10cloudtrace_4read_5utils_icmpexthdr_s;
struct __pyx_t_10cloudtrace_4read_5utils_icmpext_s;
struct __pyx_t_10cloudtrace_4read_5utils_ip6_s;

/* "cloudtrace/read/utils.pxd":24
 * cdef uint16_t checksum(void *data, int length) except -1;
 * 
 * cdef struct ethhdr_s:             # <<<<<<<<<<<<<<
 *     uint8_t h_dest[6]
 *     uint8_t h_source[6]
 */
struct __pyx_t_10cloudtrace_4read_5utils_ethhdr_s {
  uint8_t h_dest[6];
  uint8_t h_source[6];
  uint16_t h_proto;
};

/* "cloudtrace/read/utils.pxd":28
 *     uint8_t h_source[6]
 *     uint16_t h_proto
 * ctypedef ethhdr_s ethhdr             # <<<<<<<<<<<<<<
 * 
 * cdef struct ip_s:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_ethhdr_s __pyx_t_10cloudtrace_4read_5utils_ethhdr;

/* "cloudtrace/read/utils.pxd":30
 * ctypedef ethhdr_s ethhdr
 * 
 * cdef struct ip_s:             # <<<<<<<<<<<<<<
 *     uint8_t ip_hl
 *     uint8_t ip_tos
 */
struct __pyx_t_10cloudtrace_4read_5utils_ip_s {
  uint8_t ip_hl;
  uint8_t ip_tos;
  uint16_t ip_len;
  uint16_t id;
  uint16_t ip_off;
  uint8_t ip_ttl;
  uint8_t ip_p;
  uint16_t ip_sum;
  uint8_t ip_src[4];
  uint8_t ip_dst[4];
};

/* "cloudtrace/read/utils.pxd":41
 *     uint8_t ip_src[4]
 *     uint8_t ip_dst[4]
 * ctypedef ip_s ip             # <<<<<<<<<<<<<<
 * 
 * cdef void iptohost(ip *iphdr);
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_ip_s __pyx_t_10cloudtrace_4read_5utils_ip;

/* "cloudtrace/read/utils.pxd":45
 * cdef void iptohost(ip *iphdr);
 * 
 * cdef struct echo_s:             # <<<<<<<<<<<<<<
 *     uint16_t id
 *     uint16_t seq
 */
struct __pyx_t_10cloudtrace_4read_5utils_echo_s {
  uint16_t id;
  uint16_t seq;
};

/* "cloudtrace/read/utils.pxd":48
 *     uint16_t id
 *     uint16_t seq
 * ctypedef echo_s echo             # <<<<<<<<<<<<<<
 * 
 * cdef struct unused_s:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_echo_s __pyx_t_10cloudtrace_4read_5utils_echo;

/* "cloudtrace/read/utils.pxd":50
 * ctypedef echo_s echo
 * 
 * cdef struct unused_s:             # <<<<<<<<<<<<<<
 *     uint16_t __unused
 *     uint16_t mtu
 */
struct __pyx_t_10cloudtrace_4read_5utils_unused_s {
  uint16_t __pyx___unused;
  uint16_t mtu;
};

/* "cloudtrace/read/utils.pxd":53
 *     uint16_t __unused
 *     uint16_t mtu
 * ctypedef unused_s unused             # <<<<<<<<<<<<<<
 * 
 * cdef union icmpunion:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_unused_s __pyx_t_10cloudtrace_4read_5utils_unused;

/* "cloudtrace/read/utils.pxd":55
 * ctypedef unused_s unused
 * 
 * cdef union icmpunion:             # <<<<<<<<<<<<<<
 *     echo_s echo
 *     unsigned int gateway
 */
union __pyx_t_10cloudtrace_4read_5utils_icmpunion {
  struct __pyx_t_10cloudtrace_4read_5utils_echo_s echo;
  unsigned int gateway;
  struct __pyx_t_10cloudtrace_4read_5utils_unused_s frag;
};

/* "cloudtrace/read/utils.pxd":60
 *     unused_s frag
 * 
 * cdef struct icmp_s:             # <<<<<<<<<<<<<<
 *     uint8_t type
 *     uint8_t code
 */
struct __pyx_t_10cloudtrace_4read_5utils_icmp_s {
  uint8_t type;
  uint8_t code;
  uint16_t checksum;
  union __pyx_t_10cloudtrace_4read_5utils_icmpunion un;
};

/* "cloudtrace/read/utils.pxd":65
 *     uint16_t checksum
 *     icmpunion un
 * ctypedef icmp_s icmp             # <<<<<<<<<<<<<<
 * 
 * cdef void icmptohost(icmp *hdr);
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_icmp_s __pyx_t_10cloudtrace_4read_5utils_icmp;

/* "cloudtrace/read/utils.pxd":69
 * cdef void icmptohost(icmp *hdr);
 * 
 * cdef struct udp_s:             # <<<<<<<<<<<<<<
 *     uint16_t source
 *     uint16_t dest
 */
struct __pyx_t_10cloudtrace_4read_5utils_udp_s {
  uint16_t source;
  uint16_t dest;
  short len;
  uint16_t check;
};

/* "cloudtrace/read/utils.pxd":74
 *     short len
 *     uint16_t check
 * ctypedef udp_s udp             # <<<<<<<<<<<<<<
 * 
 * cdef void udptohost(udp *hdr);
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_udp_s __pyx_t_10cloudtrace_4read_5utils_udp;

/* "cloudtrace/read/utils.pxd":78
 * cdef void udptohost(udp *hdr);
 * 
 * cdef struct udppseudo_s:             # <<<<<<<<<<<<<<
 *     uint8_t src[4]
 *     uint8_t dst[4]
 */
struct __pyx_t_10cloudtrace_4read_5utils_udppseudo_s {
  uint8_t src[4];
  uint8_t dst[4];
  uint8_t zeros;
  uint8_t proto;
  uint16_t length;
  uint16_t sport;
  uint16_t dport;
  uint16_t length2;
  uint16_t chksum;
  uint8_t data[0x5C0];
};

/* "cloudtrace/read/utils.pxd":89
 *     uint16_t chksum
 *     uint8_t data[1472]
 * ctypedef udppseudo_s udppseudo             # <<<<<<<<<<<<<<
 * 
 * cdef struct ip6_hdrctl:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_udppseudo_s __pyx_t_10cloudtrace_4read_5utils_udppseudo;

/* "cloudtrace/read/utils.pxd":91
 * ctypedef udppseudo_s udppseudo
 * 
 * cdef struct ip6_hdrctl:             # <<<<<<<<<<<<<<
 *     uint32_t ip6_un1_flow
 *     uint16_t ip6_un1_plen
 */
struct __pyx_t_10cloudtrace_4read_5utils_ip6_hdrctl {
  uint32_t ip6_un1_flow;
  uint16_t ip6_un1_plen;
  uint8_t ip6_un1_nxt;
  uint8_t ip6_un1_hlim;
};

/* "cloudtrace/read/utils.pxd":97
 *     uint8_t ip6_un1_hlim
 * 
 * cdef union ip6_un:             # <<<<<<<<<<<<<<
 *     ip6_hdrctl ip6_un1
 *     uint8_t ip6_un2_vfc
 */
union __pyx_t_10cloudtrace_4read_5utils_ip6_un {
  struct __pyx_t_10cloudtrace_4read_5utils_ip6_hdrctl ip6_un1;
  uint8_t ip6_un2_vfc;
};

/* "cloudtrace/read/utils.pxd":101
 *     uint8_t ip6_un2_vfc
 * 
 * cdef struct icmpexthdr_s:             # <<<<<<<<<<<<<<
 *     uint8_t version
 *     uint8_t reserved
 */
struct __pyx_t_10cloudtrace_4read_5utils_icmpexthdr_s {
  uint8_t version;
  uint8_t reserved;
  uint16_t checksum;
};

/* "cloudtrace/read/utils.pxd":105
 *     uint8_t reserved
 *     uint16_t checksum
 * ctypedef icmpexthdr_s icmpexthdr             # <<<<<<<<<<<<<<
 * 
 * cdef struct icmpext_s:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_icmpexthdr_s __pyx_t_10cloudtrace_4read_5utils_icmpexthdr;

/* "cloudtrace/read/utils.pxd":107
 * ctypedef icmpexthdr_s icmpexthdr
 * 
 * cdef struct icmpext_s:             # <<<<<<<<<<<<<<
 *     uint16_t length
 *     uint8_t classnum
 */
struct __pyx_t_10cloudtrace_4read_5utils_icmpext_s {
  uint16_t length;
  uint8_t classnum;
  uint8_t ctype;
};

/* "cloudtrace/read/utils.pxd":111
 *     uint8_t classnum
 *     uint8_t ctype
 * ctypedef icmpext_s icmpext             # <<<<<<<<<<<<<<
 * 
 * cdef struct ip6_s:
 */
typedef struct __pyx_t_10cloudtrace_4read_5utils_icmpext_s __pyx_t_10cloudtrace_4read_5utils_icmpext;

/* "cloudtrace/read/utils.pxd":113
 * ctypedef icmpext_s icmpext
 * 
 * cdef struct ip6_s:             # <<<<<<<<<<<<<<
 *     ip6_un ip6_ctlun
 *     uint8_t ip6_src[16]
 */
struct __pyx_t_10cloudtrace_4read_5utils_ip6_s {
  union __pyx_t_10cloudtrace_4read_5utils_ip6_un ip6_ctlun;
  uint8_t ip6_src[16];
  uint8_t ip6_dst[16];
};
struct __pyx_t_10cloudtrace_4read_4pcap_global_hdr_s;
struct __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_s;

/* "cloudtrace/read/pcap.pxd":4
 * 
 * # /[inserted by cython to avoid comment start]* Global header (http://v2.nat32.com/pcap.htm) *[inserted by cython to avoid comment closer]/
 * cdef struct global_hdr_s:             # <<<<<<<<<<<<<<
 *     uint32_t magic_number   #/[inserted by cython to avoid comment start]* magic number *[inserted by cython to avoid comment closer]/
 *     uint16_t version_major  #/[inserted by cython to avoid comment start]* major version number *[inserted by cython to avoid comment closer]/
 */
struct __pyx_t_10cloudtrace_4read_4pcap_global_hdr_s {
  uint32_t magic_number;
  uint16_t version_major;
  uint16_t version_minor;
  int32_t thiszone;
  uint32_t sigfigs;
  uint32_t snaplen;
  uint32_t network;
};

/* "cloudtrace/read/pcap.pxd":12
 *     uint32_t snaplen        #/[inserted by cython to avoid comment start]* max length of captured packets, in octets *[inserted by cython to avoid comment closer]/
 *     uint32_t network        #/[inserted by cython to avoid comment start]* data link type *[inserted by cython to avoid comment closer]/
 * ctypedef global_hdr_s global_hdr_t             # <<<<<<<<<<<<<<
 * 
 * # /[inserted by cython to avoid comment start]* Packet header (http://v2.nat32.com/pcap.htm) *[inserted by cython to avoid comment closer]/
 */
typedef struct __pyx_t_10cloudtrace_4read_4pcap_global_hdr_s __pyx_t_10cloudtrace_4read_4pcap_global_hdr_t;

/* "cloudtrace/read/pcap.pxd":15
 * 
 * # /[inserted by cython to avoid comment start]* Packet header (http://v2.nat32.com/pcap.htm) *[inserted by cython to avoid comment closer]/
 * cdef struct packet_hdr_s:             # <<<<<<<<<<<<<<
 *     uint32_t ts_sec         #/[inserted by cython to avoid comment start]* timestamp seconds *[inserted by cython to avoid comment closer]/
 *     uint32_t ts_usec        #/[inserted by cython to avoid comment start]* timestamp microseconds *[inserted by cython to avoid comment closer]/
 */
struct __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_s {
  uint32_t ts_sec;
  uint32_t ts_usec;
  uint32_t incl_len;
  uint32_t orig_len;
};

/* "cloudtrace/read/pcap.pxd":20
 *     uint32_t incl_len       #/[inserted by cython to avoid comment start]* number of octets of packet saved in file *[inserted by cython to avoid comment closer]/
 *     uint32_t orig_len       #/[inserted by cython to avoid comment start]* actual length of packet *[inserted by cython to avoid comment closer]/
 * ctypedef packet_hdr_s packet_hdr_t             # <<<<<<<<<<<<<<
 * 
 * cdef class Pcap:
 */
typedef struct __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_s __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_t;
struct __pyx_t_10cloudtrace_4read_7combine_hop;

/* "cloudtrace/read/combine.pxd":4
 * from posix.time cimport timeval
 * 
 * cdef struct hop:             # <<<<<<<<<<<<<<
 *     char hop_addr[16]
 *     uint8_t hop_flags
 */
struct __pyx_t_10cloudtrace_4read_7combine_hop {
  char hop_addr[16];
  uint8_t hop_flags;
  uint8_t hop_probe_ttl;
  uint16_t hop_probe_id;
  uint16_t hop_probe_size;
  uint8_t hop_reply_ttl;
  uint8_t hop_reply_tos;
  uint16_t hop_reply_size;
  uint16_t hop_reply_ipid;
  uint8_t hop_icmp_type;
  uint8_t hop_icmp_code;
  uint8_t hop_icmp_q_ttl;
  uint8_t hop_icmp_q_tos;
  uint16_t hop_icmp_q_ipl;
  uint16_t hop_icmp_nhmtu;
  struct timeval hop_tx;
  struct timeval hop_rtt;
};

/* "cloudtrace/read/combine.pxd":22
 *     timeval hop_tx
 *     timeval hop_rtt
 * ctypedef hop hop_t             # <<<<<<<<<<<<<<
 * 
 * cdef class Trace:
 */
typedef struct __pyx_t_10cloudtrace_4read_7combine_hop __pyx_t_10cloudtrace_4read_7combine_hop_t;
struct __pyx_t_10cloudtrace_4read_10linkedlist_node_s;

/* "cloudtrace/read/linkedlist.pxd":5
 * from cloudtrace.read.utils cimport *
 * 
 * cdef struct node_s:             # <<<<<<<<<<<<<<
 *     void *dst
 *     uint8_t dstlen
 */
struct __pyx_t_10cloudtrace_4read_10linkedlist_node_s {
  void *dst;
  uint8_t dstlen;
  struct timeval time;
  struct __pyx_t_10cloudtrace_4read_10linkedlist_node_s *next;
};

/* "cloudtrace/read/linkedlist.pxd":10
 *     timeval time
 *     node_s *next
 * ctypedef node_s node             # <<<<<<<<<<<<<<
 * 
 * cdef class LinkedList:
 */
typedef struct __pyx_t_10cloudtrace_4read_10linkedlist_node_s __pyx_t_10cloudtrace_4read_10linkedlist_node;
struct __pyx_t_10cloudtrace_4read_7convert_probe_id_s;

/* "cloudtrace/read/convert.pxd":12
 *     public uint8_t TRACE_STOP_GAPLIMIT, TRACE_STOP_COMPLETED, TRACE_STOP_UNREACH, TRACE_STOP_LOOP, TRACE_STOP_HOPLIMIT
 * 
 * cdef struct probe_id_s:             # <<<<<<<<<<<<<<
 *     char dst[4]
 *     unsigned char ttl
 */
struct __pyx_t_10cloudtrace_4read_7convert_probe_id_s {
  char dst[4];
  unsigned char ttl;
  unsigned short pid;
};

/* "cloudtrace/read/convert.pxd":16
 *     unsigned char ttl
 *     unsigned short pid
 * ctypedef probe_id_s probe_id             # <<<<<<<<<<<<<<
 * 
 * # cdef unsigned short checksum(unsigned char *data, int length) except -1;
 */
typedef struct __pyx_t_10cloudtrace_4read_7convert_probe_id_s __pyx_t_10cloudtrace_4read_7convert_probe_id;

/* "cloudtrace/read/pcap.pxd":22
 * ctypedef packet_hdr_s packet_hdr_t
 * 
 * cdef class Pcap:             # <<<<<<<<<<<<<<
 *     cdef:
 *         global_hdr_t *global_hdr
 */
struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap *__pyx_vtab;
  __pyx_t_10cloudtrace_4read_4pcap_global_hdr_t *global_hdr;
  __pyx_t_10cloudtrace_4read_4pcap_packet_hdr_t *packet_hdr;
  char *frame;
  PyObject *infile;
  PyObject *proc;
};


/* "cloudtrace/read/combine.pxd":24
 * ctypedef hop hop_t
 * 
 * cdef class Trace:             # <<<<<<<<<<<<<<
 *     cdef:
 *         char src[16]
 */
struct __pyx_obj_10cloudtrace_4read_7combine_Trace {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_4read_7combine_Trace *__pyx_vtab;
  char src[16];
  char dst[16];
  struct timeval start;
  __pyx_t_10cloudtrace_4read_7combine_hop_t **hops;
  uint16_t hop_count;
  uint16_t probec;
  uint8_t stop_reason;
  uint8_t stop_data;
  uint8_t type;
  uint8_t flags;
  uint8_t attempts;
  uint8_t hoplimit;
  uint8_t gaplimit;
  uint8_t gapaction;
  uint8_t firsthop;
  uint8_t tos;
  uint8_t wait;
  uint8_t wait_probe;
  uint8_t loops;
  uint8_t loopaction;
  uint8_t confidence;
  uint16_t probe_size;
  uint16_t sport;
  uint16_t dport;
  uint16_t offset;
};


/* "cloudtrace/read/packet.pxd":10
 *     public uint8_t ICMP_PROBE, UDP_PROBE, ECHO_REPLY, ICMP_REPLY
 * 
 * cdef class Packet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public uint8_t ptype
 */
struct __pyx_obj_10cloudtrace_4read_6packet_Packet {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *__pyx_vtab;
  uint8_t ptype;
  int isprobe;
  __pyx_t_10cloudtrace_4read_5utils_ip *iphdr;
  PyObject *dst;
  struct timeval *time;
  PyObject *pid_b;
  PyObject *dst_b;
};


/* "cloudtrace/read/packet.pxd":26
 *     cpdef bytes get_dst(self);
 * 
 * cdef class ICMPProbe(Packet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         icmp *icmphdr
 */
struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet __pyx_base;
  __pyx_t_10cloudtrace_4read_5utils_icmp *icmphdr;
};


/* "cloudtrace/read/packet.pxd":34
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr);
 * 
 * cdef class UDPProbe(Packet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         udp *udphdr
 */
struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet __pyx_base;
  __pyx_t_10cloudtrace_4read_5utils_udp *udphdr;
};


/* "cloudtrace/read/packet.pxd":40
 *     cdef void set_udp(self, ip *iphdr, udp *hdr);
 * 
 * cdef class Reply(Packet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         icmp *icmphdr
 */
struct __pyx_obj_10cloudtrace_4read_6packet_Reply {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet __pyx_base;
  __pyx_t_10cloudtrace_4read_5utils_icmp *icmphdr;
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *orig_probe;
};


/* "cloudtrace/read/packet.pxd":56
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr);
 * 
 * cdef class EchoReply(Reply):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply {
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply __pyx_base;
};


/* "cloudtrace/read/packet.pxd":59
 *     pass
 * 
 * cdef class ICMPReply(Reply):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Packet probe
 */
struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply {
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply __pyx_base;
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *probe;
};


/* "cloudtrace/read/linkedlist.pxd":12
 * ctypedef node_s node
 * 
 * cdef class LinkedList:             # <<<<<<<<<<<<<<
 *     cdef:
 *         node *start
 */
struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList *__pyx_vtab;
  __pyx_t_10cloudtrace_4read_10linkedlist_node *start;
  __pyx_t_10cloudtrace_4read_10linkedlist_node *end;
};


/* "cloudtrace/read/linkedlist.pxd":22
 * cdef void free_node(node *n);
 * 
 * cdef class Node:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes dst
 */
struct __pyx_obj_10cloudtrace_4read_10linkedlist_Node {
  PyObject_HEAD
  PyObject *dst;
  struct timeval time;
  struct __pyx_obj_10cloudtrace_4read_10linkedlist_Node *next;
};


/* "cloudtrace/read/linkedlist.pxd":28
 *         Node next
 * 
 * cdef class LinkedList2:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Node start
 */
struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList2 {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList2 *__pyx_vtab;
  struct __pyx_obj_10cloudtrace_4read_10linkedlist_Node *start;
  struct __pyx_obj_10cloudtrace_4read_10linkedlist_Node *end;
};


/* "cloudtrace/read/convert.pxd":20
 * # cdef unsigned short checksum(unsigned char *data, int length) except -1;
 * 
 * cdef class Convert:             # <<<<<<<<<<<<<<
 *     cdef:
 *         str infile, outfile, hostname
 */
struct __pyx_obj_10cloudtrace_4read_7convert_Convert {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert *__pyx_vtab;
  PyObject *infile;
  PyObject *outfile;
  PyObject *hostname;
  long total;
  long nprobes;
  long nreplies;
  long skipped;
  long pidprobe;
  long pidreply;
  long ttlhigh;
  long wrote;
  struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList2 *times;
  PyObject *probes;
  PyObject *replies;
  PyObject *writefile;
  struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *pcap;
  int more;
};


/* "cloudtrace/read/convert.pxd":41
 * cdef str addr_tostr(char *addr);
 * 
 * cdef class ConvertTrace(Convert):             # <<<<<<<<<<<<<<
 *     cdef Trace initialize_trace(self, Packet probe);
 *     cpdef void convert_trace(self, bint fix, uint8_t verbose) except *;
 */
struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace {
  struct __pyx_obj_10cloudtrace_4read_7convert_Convert __pyx_base;
};



/* "cloudtrace/read/pcap.pxd":22
 * ctypedef packet_hdr_s packet_hdr_t
 * 
 * cdef class Pcap:             # <<<<<<<<<<<<<<
 *     cdef:
 *         global_hdr_t *global_hdr
 */

struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap {
  void (*close)(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch);
  int (*next_packet)(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch);
  void (*read_all)(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap *__pyx_vtabptr_10cloudtrace_4read_4pcap_Pcap;


/* "cloudtrace/read/combine.pxd":24
 * ctypedef hop hop_t
 * 
 * cdef class Trace:             # <<<<<<<<<<<<<<
 *     cdef:
 *         char src[16]
 */

struct __pyx_vtabstruct_10cloudtrace_4read_7combine_Trace {
  __pyx_t_10cloudtrace_4read_7combine_hop_t *(*new_hop)(struct __pyx_obj_10cloudtrace_4read_7combine_Trace *, uint8_t);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_7combine_Trace *__pyx_vtabptr_10cloudtrace_4read_7combine_Trace;


/* "cloudtrace/read/packet.pxd":10
 *     public uint8_t ICMP_PROBE, UDP_PROBE, ECHO_REPLY, ICMP_REPLY
 * 
 * cdef class Packet:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public uint8_t ptype
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet {
  uint8_t *(*get_dst_c)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *);
  uint8_t (*probe_ttl)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
  uint16_t (*get_pid)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
  uint16_t (*ipid)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
  void (*set_ip)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, __pyx_t_10cloudtrace_4read_5utils_ip *);
  PyObject *(*get_dst)(struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *__pyx_vtabptr_10cloudtrace_4read_6packet_Packet;


/* "cloudtrace/read/packet.pxd":26
 *     cpdef bytes get_dst(self);
 * 
 * cdef class ICMPProbe(Packet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         icmp *icmphdr
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPProbe {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet __pyx_base;
  uint8_t (*get_type)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *, int __pyx_skip_dispatch);
  uint8_t (*get_code)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *, int __pyx_skip_dispatch);
  void (*set_icmp)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_icmp *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPProbe *__pyx_vtabptr_10cloudtrace_4read_6packet_ICMPProbe;


/* "cloudtrace/read/packet.pxd":34
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr);
 * 
 * cdef class UDPProbe(Packet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         udp *udphdr
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_UDPProbe {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet __pyx_base;
  void (*set_udp)(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_udp *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_UDPProbe *__pyx_vtabptr_10cloudtrace_4read_6packet_UDPProbe;


/* "cloudtrace/read/packet.pxd":40
 *     cdef void set_udp(self, ip *iphdr, udp *hdr);
 * 
 * cdef class Reply(Packet):             # <<<<<<<<<<<<<<
 *     cdef:
 *         icmp *icmphdr
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet __pyx_base;
  void (*_create_hop)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_7combine_hop_t *);
  PyObject *(*addr)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  void (*create_hop)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_7combine_hop_t *);
  uint16_t (*probe_ipid)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint16_t (*probe_size)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint8_t (*reply_ttl)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint8_t (*get_type)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint8_t (*get_code)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  uint8_t (*get_icmp_q_ttl)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, int __pyx_skip_dispatch);
  void (*set_icmp)(struct __pyx_obj_10cloudtrace_4read_6packet_Reply *, __pyx_t_10cloudtrace_4read_5utils_ip *, __pyx_t_10cloudtrace_4read_5utils_icmp *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply *__pyx_vtabptr_10cloudtrace_4read_6packet_Reply;


/* "cloudtrace/read/packet.pxd":56
 *     cdef void set_icmp(self, ip *iphdr, icmp *hdr);
 * 
 * cdef class EchoReply(Reply):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_EchoReply {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply __pyx_base;
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_EchoReply *__pyx_vtabptr_10cloudtrace_4read_6packet_EchoReply;


/* "cloudtrace/read/packet.pxd":59
 *     pass
 * 
 * cdef class ICMPReply(Reply):             # <<<<<<<<<<<<<<
 *     cdef:
 *         Packet probe
 */

struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPReply {
  struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply __pyx_base;
  void (*create_pseudo)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *, __pyx_t_10cloudtrace_4read_5utils_udppseudo *, unsigned char *);
  void (*set_probe)(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply *, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPReply *__pyx_vtabptr_10cloudtrace_4read_6packet_ICMPReply;


/* "cloudtrace/read/linkedlist.pxd":12
 * ctypedef node_s node
 * 
 * cdef class LinkedList:             # <<<<<<<<<<<<<<
 *     cdef:
 *         node *start
 */

struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList {
  void (*add)(struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList *, uint8_t, void *, struct timeval *);
  __pyx_t_10cloudtrace_4read_10linkedlist_node *(*pop)(struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList *__pyx_vtabptr_10cloudtrace_4read_10linkedlist_LinkedList;


/* "cloudtrace/read/linkedlist.pxd":28
 *         Node next
 * 
 * cdef class LinkedList2:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Node start
 */

struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList2 {
  void (*add)(struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList2 *, uint8_t, char *, struct timeval *);
  struct __pyx_obj_10cloudtrace_4read_10linkedlist_Node *(*pop)(struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList2 *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList2 *__pyx_vtabptr_10cloudtrace_4read_10linkedlist_LinkedList2;


/* "cloudtrace/read/convert.pyx":20
 *     public uint8_t TRACE_STOP_HOPLIMIT = 5
 * 
 * cdef class Convert:             # <<<<<<<<<<<<<<
 *     def __cinit__(self, str infile, str outfile, str hostname=None):
 *         self.total = 0
 */

struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert {
  void (*write_dst)(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *, PyObject *);
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply *(*next)(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *, struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *);
  void (*write_old)(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *, struct timeval *, int);
  void (*write_all)(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert *__pyx_vtabptr_10cloudtrace_4read_7convert_Convert;


/* "cloudtrace/read/convert.pyx":161
 *     return (<bytes>dst).decode()
 * 
 * cdef class ConvertTrace(Convert):             # <<<<<<<<<<<<<<
 *     cdef dict hopdict(self, hop_t *hop):
 *         rtt = round(hop.hop_rtt.tv_sec * 1000 + (hop.hop_rtt.tv_usec / 1000), 3)
 */

struct __pyx_vtabstruct_10cloudtrace_4read_7convert_ConvertTrace {
  struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert __pyx_base;
  struct __pyx_obj_10cloudtrace_4read_7combine_Trace *(*initialize_trace)(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *);
  void (*convert_trace)(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *, int, uint8_t, int __pyx_skip_dispatch);
  PyObject *(*hopdict)(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *, __pyx_t_10cloudtrace_4read_7combine_hop_t *);
  PyObject *(*tracedict)(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *, struct __pyx_obj_10cloudtrace_4read_7combine_Trace *);
};
static struct __pyx_vtabstruct_10cloudtrace_4read_7convert_ConvertTrace *__pyx_vtabptr_10cloudtrace_4read_7convert_ConvertTrace;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* tp_new.proto */
#define __Pyx_tp_new(type_obj, args) __Pyx_tp_new_kwargs(type_obj, args, NULL)
static CYTHON_INLINE PyObject* __Pyx_tp_new_kwargs(PyObject* type_obj, PyObject* args, PyObject* kwargs) {
    return (PyObject*) (((PyTypeObject*)type_obj)->tp_new((PyTypeObject*)type_obj, args, kwargs));
}

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    return __Pyx_decode_c_bytes(
        PyBytes_AS_STRING(string), PyBytes_GET_SIZE(string),
        start, stop, encoding, errors, decode_func);
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* ModInt[long].proto */
static CYTHON_INLINE long __Pyx_mod_long(long, long);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_time_t(time_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_suseconds_t(suseconds_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static void __pyx_f_10cloudtrace_4read_7convert_7Convert_write_dst(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_dst_b); /* proto*/
static void __pyx_f_10cloudtrace_4read_7convert_7Convert_write_old(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, struct timeval *__pyx_v_currtime, int __pyx_v_delta); /* proto*/
static void __pyx_f_10cloudtrace_4read_7convert_7Convert_write_all(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto*/
static struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_f_10cloudtrace_4read_7convert_7Convert_next(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *__pyx_v_pcap); /* proto*/
static PyObject *__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_hopdict(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_hop); /* proto*/
static PyObject *__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_tracedict(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_7combine_Trace *__pyx_v_trace); /* proto*/
static void __pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_write_dst(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, PyObject *__pyx_v_dst_b); /* proto*/
static struct __pyx_obj_10cloudtrace_4read_7combine_Trace *__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_initialize_trace(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_probe); /* proto*/
static void __pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_convert_trace(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, int __pyx_v_fix, uint8_t __pyx_v_verbose, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'posix.types' */

/* Module declarations from 'posix.signal' */

/* Module declarations from 'posix.time' */

/* Module declarations from 'cloudtrace.read.utils' */

/* Module declarations from 'cloudtrace.read.pcap' */
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_4pcap_Pcap = 0;

/* Module declarations from 'cloudtrace.read.combine' */
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_7combine_Trace = 0;

/* Module declarations from 'cloudtrace.read.packet' */
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_Packet = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_ICMPProbe = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_UDPProbe = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_Reply = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_EchoReply = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_6packet_ICMPReply = 0;
static uint8_t *__pyx_vp_10cloudtrace_4read_6packet_ICMP_PROBE = 0;
#define __pyx_v_10cloudtrace_4read_6packet_ICMP_PROBE (*__pyx_vp_10cloudtrace_4read_6packet_ICMP_PROBE)
static uint8_t *__pyx_vp_10cloudtrace_4read_6packet_UDP_PROBE = 0;
#define __pyx_v_10cloudtrace_4read_6packet_UDP_PROBE (*__pyx_vp_10cloudtrace_4read_6packet_UDP_PROBE)
static uint8_t *__pyx_vp_10cloudtrace_4read_6packet_ECHO_REPLY = 0;
#define __pyx_v_10cloudtrace_4read_6packet_ECHO_REPLY (*__pyx_vp_10cloudtrace_4read_6packet_ECHO_REPLY)
static uint8_t *__pyx_vp_10cloudtrace_4read_6packet_ICMP_REPLY = 0;
#define __pyx_v_10cloudtrace_4read_6packet_ICMP_REPLY (*__pyx_vp_10cloudtrace_4read_6packet_ICMP_REPLY)
static struct __pyx_obj_10cloudtrace_4read_6packet_Packet *(*__pyx_f_10cloudtrace_4read_6packet_disassemble)(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'cloudtrace.read.linkedlist' */
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_10linkedlist_Node = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList2 = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cloudtrace.read.convert' */
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_7convert_Convert = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_4read_7convert_ConvertTrace = 0;
__PYX_EXTERN_C DL_EXPORT(uint8_t) TRACE_TYPE_ICMP_ECHO_PARIS;
__PYX_EXTERN_C DL_EXPORT(uint8_t) TRACE_TYPE_UDP_PARIS;
__PYX_EXTERN_C DL_EXPORT(uint8_t) TRACE_STOP_GAPLIMIT;
__PYX_EXTERN_C DL_EXPORT(uint8_t) TRACE_STOP_COMPLETED;
__PYX_EXTERN_C DL_EXPORT(uint8_t) TRACE_STOP_UNREACH;
__PYX_EXTERN_C DL_EXPORT(uint8_t) TRACE_STOP_LOOP;
__PYX_EXTERN_C DL_EXPORT(uint8_t) TRACE_STOP_HOPLIMIT;
static int __pyx_f_10cloudtrace_4read_7convert_loop_check(struct __pyx_obj_10cloudtrace_4read_7combine_Trace *); /*proto*/
static PyObject *__pyx_f_10cloudtrace_4read_7convert_addr_tostr(char *); /*proto*/
#define __Pyx_MODULE_NAME "cloudtrace.read.convert"
extern int __pyx_module_is_main_cloudtrace__read__convert;
int __pyx_module_is_main_cloudtrace__read__convert = 0;

/* Implementation of 'cloudtrace.read.convert' */
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_round;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_print;
static const char __pyx_k_[] = "\000\000\000\000";
static const char __pyx_k__4[] = "{}\n";
static const char __pyx_k_tx[] = "tx";
static const char __pyx_k_wt[] = "wt";
static const char __pyx_k_0_1[] = "0.1";
static const char __pyx_k_dst[] = "dst";
static const char __pyx_k_fix[] = "fix";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_rtt[] = "rtt";
static const char __pyx_k_sec[] = "sec";
static const char __pyx_k_src[] = "src";
static const char __pyx_k_tos[] = "tos";
static const char __pyx_k_LOOP[] = "LOOP";
static const char __pyx_k_addr[] = "addr";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_fake[] = "fake";
static const char __pyx_k_hops[] = "hops";
static const char __pyx_k_json[] = "json";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_usec[] = "usec";
static const char __pyx_k_wait[] = "wait";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_dumps[] = "dumps";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_fopen[] = "fopen";
static const char __pyx_k_ftime[] = "ftime";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_round[] = "round";
static const char __pyx_k_sleep[] = "sleep";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_trace[] = "trace";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_infile[] = "infile";
static const char __pyx_k_method[] = "method";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_userid[] = "userid";
static const char __pyx_k_Convert[] = "Convert";
static const char __pyx_k_UNREACH[] = "UNREACH";
static const char __pyx_k_outfile[] = "outfile";
static const char __pyx_k_verbose[] = "verbose";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_GAPLIMIT[] = "GAPLIMIT";
static const char __pyx_k_HOPLIMIT[] = "HOPLIMIT";
static const char __pyx_k_attempts[] = "attempts";
static const char __pyx_k_firsthop[] = "firsthop";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_hoplimit[] = "hoplimit";
static const char __pyx_k_hostname[] = "hostname";
static const char __pyx_k_icmp_sum[] = "icmp_sum";
static const char __pyx_k_probe_id[] = "probe_id";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_readline[] = "readline";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_COMPLETED[] = "COMPLETED";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_hop_count[] = "hop_count";
static const char __pyx_k_icmp_code[] = "icmp_code";
static const char __pyx_k_icmp_type[] = "icmp_type";
static const char __pyx_k_probe_ttl[] = "probe_ttl";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_reply_tos[] = "reply_tos";
static const char __pyx_k_reply_ttl[] = "reply_ttl";
static const char __pyx_k_stop_data[] = "stop_data";
static const char __pyx_k_udp_paris[] = "udp-paris";
static const char __pyx_k_icmp_q_ipl[] = "icmp_q_ipl";
static const char __pyx_k_icmp_q_tos[] = "icmp_q_tos";
static const char __pyx_k_icmp_q_ttl[] = "icmp_q_ttl";
static const char __pyx_k_probe_size[] = "probe_size";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_reply_ipid[] = "reply_ipid";
static const char __pyx_k_reply_size[] = "reply_size";
static const char __pyx_k_wait_probe[] = "wait_probe";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_defaultdict[] = "defaultdict";
static const char __pyx_k_probe_count[] = "probe_count";
static const char __pyx_k_stop_reason[] = "stop_reason";
static const char __pyx_k_ConvertTrace[] = "ConvertTrace";
static const char __pyx_k_convert_trace[] = "convert_trace";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_Loop_d_Wrote_d[] = "Loop {:,d} Wrote {:,d}";
static const char __pyx_k_TRACE_STOP_LOOP[] = "TRACE_STOP_LOOP";
static const char __pyx_k_icmp_echo_paris[] = "icmp-echo-paris";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_TRACE_STOP_UNREACH[] = "TRACE_STOP_UNREACH";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_TRACE_STOP_GAPLIMIT[] = "TRACE_STOP_GAPLIMIT";
static const char __pyx_k_TRACE_STOP_HOPLIMIT[] = "TRACE_STOP_HOPLIMIT";
static const char __pyx_k_TRACE_STOP_COMPLETED[] = "TRACE_STOP_COMPLETED";
static const char __pyx_k_TRACE_TYPE_UDP_PARIS[] = "TRACE_TYPE_UDP_PARIS";
static const char __pyx_k_cloudtrace_read_pcap[] = "cloudtrace.read.pcap";
static const char __pyx_k_TRACE_TYPE_ICMP_ECHO_PARIS[] = "TRACE_TYPE_ICMP_ECHO_PARIS";
static const char __pyx_k_Total_d_Probes_d_Replies_d_PIDP[] = "Total {:,d} Probes {:,d} Replies {:,d} PIDP {:,d} PIDR {:,d} Skipped {:,d} Wrote {:,d}";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_kp_b_;
static PyObject *__pyx_kp_u_0_1;
static PyObject *__pyx_n_u_COMPLETED;
static PyObject *__pyx_n_s_Convert;
static PyObject *__pyx_n_s_ConvertTrace;
static PyObject *__pyx_n_u_GAPLIMIT;
static PyObject *__pyx_n_u_HOPLIMIT;
static PyObject *__pyx_n_u_LOOP;
static PyObject *__pyx_kp_u_Loop_d_Wrote_d;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_TRACE_STOP_COMPLETED;
static PyObject *__pyx_n_s_TRACE_STOP_GAPLIMIT;
static PyObject *__pyx_n_s_TRACE_STOP_HOPLIMIT;
static PyObject *__pyx_n_s_TRACE_STOP_LOOP;
static PyObject *__pyx_n_s_TRACE_STOP_UNREACH;
static PyObject *__pyx_n_s_TRACE_TYPE_ICMP_ECHO_PARIS;
static PyObject *__pyx_n_s_TRACE_TYPE_UDP_PARIS;
static PyObject *__pyx_kp_u_Total_d_Probes_d_Replies_d_PIDP;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_u_UNREACH;
static PyObject *__pyx_kp_u__4;
static PyObject *__pyx_n_u_addr;
static PyObject *__pyx_n_u_attempts;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cloudtrace_read_pcap;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_convert_trace;
static PyObject *__pyx_n_s_defaultdict;
static PyObject *__pyx_n_u_dst;
static PyObject *__pyx_n_s_dumps;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_u_fake;
static PyObject *__pyx_n_u_firsthop;
static PyObject *__pyx_n_s_fix;
static PyObject *__pyx_n_s_fopen;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_u_ftime;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_u_hop_count;
static PyObject *__pyx_n_u_hoplimit;
static PyObject *__pyx_n_u_hops;
static PyObject *__pyx_n_s_hostname;
static PyObject *__pyx_n_u_icmp_code;
static PyObject *__pyx_kp_u_icmp_echo_paris;
static PyObject *__pyx_n_u_icmp_q_ipl;
static PyObject *__pyx_n_u_icmp_q_tos;
static PyObject *__pyx_n_u_icmp_q_ttl;
static PyObject *__pyx_n_u_icmp_sum;
static PyObject *__pyx_n_u_icmp_type;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_infile;
static PyObject *__pyx_n_s_json;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_u_method;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_outfile;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_u_probe_count;
static PyObject *__pyx_n_u_probe_id;
static PyObject *__pyx_n_u_probe_size;
static PyObject *__pyx_n_u_probe_ttl;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_readline;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_u_reply_ipid;
static PyObject *__pyx_n_u_reply_size;
static PyObject *__pyx_n_u_reply_tos;
static PyObject *__pyx_n_u_reply_ttl;
static PyObject *__pyx_n_s_round;
static PyObject *__pyx_n_u_rtt;
static PyObject *__pyx_n_u_sec;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_sleep;
static PyObject *__pyx_n_u_src;
static PyObject *__pyx_n_u_start;
static PyObject *__pyx_n_u_stop_data;
static PyObject *__pyx_n_u_stop_reason;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_u_tos;
static PyObject *__pyx_n_u_trace;
static PyObject *__pyx_n_u_tx;
static PyObject *__pyx_n_u_type;
static PyObject *__pyx_kp_u_udp_paris;
static PyObject *__pyx_n_u_usec;
static PyObject *__pyx_n_u_userid;
static PyObject *__pyx_n_s_verbose;
static PyObject *__pyx_n_u_version;
static PyObject *__pyx_n_u_wait;
static PyObject *__pyx_n_u_wait_probe;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_u_wt;
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert___cinit__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_infile, PyObject *__pyx_v_outfile, PyObject *__pyx_v_hostname); /* proto */
static void __pyx_pf_10cloudtrace_4read_7convert_7Convert_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_4__repr__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_6close(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_5total___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_5total_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_7nprobes___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_7nprobes_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_8nreplies___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_8nreplies_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_7skipped___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_7skipped_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidprobe___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidprobe_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidreply___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidreply_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_7ttlhigh___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_7ttlhigh_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_5wrote___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_5wrote_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_9writefile___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_9writefile_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_9writefile_4__del__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_12ConvertTrace_convert_trace(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, int __pyx_v_fix, uint8_t __pyx_v_verbose); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_12ConvertTrace_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_4read_7convert_12ConvertTrace_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_10cloudtrace_4read_7convert_Convert(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10cloudtrace_4read_7convert_ConvertTrace(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_47652;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
/* Late includes */
uint8_t TRACE_TYPE_ICMP_ECHO_PARIS;
uint8_t TRACE_TYPE_UDP_PARIS;
uint8_t TRACE_STOP_GAPLIMIT;
uint8_t TRACE_STOP_COMPLETED;
uint8_t TRACE_STOP_UNREACH;
uint8_t TRACE_STOP_LOOP;
uint8_t TRACE_STOP_HOPLIMIT;

/* "cloudtrace/read/convert.pyx":21
 * 
 * cdef class Convert:
 *     def __cinit__(self, str infile, str outfile, str hostname=None):             # <<<<<<<<<<<<<<
 *         self.total = 0
 *         self.nprobes = 0
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_infile = 0;
  PyObject *__pyx_v_outfile = 0;
  PyObject *__pyx_v_hostname = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_infile,&__pyx_n_s_outfile,&__pyx_n_s_hostname,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject*)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_infile)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_outfile)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(0, 21, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_hostname);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 21, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_infile = ((PyObject*)values[0]);
    __pyx_v_outfile = ((PyObject*)values[1]);
    __pyx_v_hostname = ((PyObject*)values[2]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 21, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_infile), (&PyUnicode_Type), 1, "infile", 1))) __PYX_ERR(0, 21, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_outfile), (&PyUnicode_Type), 1, "outfile", 1))) __PYX_ERR(0, 21, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_hostname), (&PyUnicode_Type), 1, "hostname", 1))) __PYX_ERR(0, 21, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert___cinit__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), __pyx_v_infile, __pyx_v_outfile, __pyx_v_hostname);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert___cinit__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_infile, PyObject *__pyx_v_outfile, PyObject *__pyx_v_hostname) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/read/convert.pyx":22
 * cdef class Convert:
 *     def __cinit__(self, str infile, str outfile, str hostname=None):
 *         self.total = 0             # <<<<<<<<<<<<<<
 *         self.nprobes = 0
 *         self.nreplies = 0
 */
  __pyx_v_self->total = 0;

  /* "cloudtrace/read/convert.pyx":23
 *     def __cinit__(self, str infile, str outfile, str hostname=None):
 *         self.total = 0
 *         self.nprobes = 0             # <<<<<<<<<<<<<<
 *         self.nreplies = 0
 *         self.skipped = 0
 */
  __pyx_v_self->nprobes = 0;

  /* "cloudtrace/read/convert.pyx":24
 *         self.total = 0
 *         self.nprobes = 0
 *         self.nreplies = 0             # <<<<<<<<<<<<<<
 *         self.skipped = 0
 *         self.pidprobe = 0
 */
  __pyx_v_self->nreplies = 0;

  /* "cloudtrace/read/convert.pyx":25
 *         self.nprobes = 0
 *         self.nreplies = 0
 *         self.skipped = 0             # <<<<<<<<<<<<<<
 *         self.pidprobe = 0
 *         self.pidreply = 0
 */
  __pyx_v_self->skipped = 0;

  /* "cloudtrace/read/convert.pyx":26
 *         self.nreplies = 0
 *         self.skipped = 0
 *         self.pidprobe = 0             # <<<<<<<<<<<<<<
 *         self.pidreply = 0
 *         self.ttlhigh = 0
 */
  __pyx_v_self->pidprobe = 0;

  /* "cloudtrace/read/convert.pyx":27
 *         self.skipped = 0
 *         self.pidprobe = 0
 *         self.pidreply = 0             # <<<<<<<<<<<<<<
 *         self.ttlhigh = 0
 *         self.wrote = 0
 */
  __pyx_v_self->pidreply = 0;

  /* "cloudtrace/read/convert.pyx":28
 *         self.pidprobe = 0
 *         self.pidreply = 0
 *         self.ttlhigh = 0             # <<<<<<<<<<<<<<
 *         self.wrote = 0
 *         # self.times = LinkedList.__new__(LinkedList)
 */
  __pyx_v_self->ttlhigh = 0;

  /* "cloudtrace/read/convert.pyx":29
 *         self.pidreply = 0
 *         self.ttlhigh = 0
 *         self.wrote = 0             # <<<<<<<<<<<<<<
 *         # self.times = LinkedList.__new__(LinkedList)
 *         self.times = LinkedList2.__new__(LinkedList2)
 */
  __pyx_v_self->wrote = 0;

  /* "cloudtrace/read/convert.pyx":31
 *         self.wrote = 0
 *         # self.times = LinkedList.__new__(LinkedList)
 *         self.times = LinkedList2.__new__(LinkedList2)             # <<<<<<<<<<<<<<
 *         self.probes = {}
 *         self.replies = {}
 */
  __pyx_t_1 = __Pyx_tp_new(((PyObject *)__pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList2), __pyx_empty_tuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList2)))) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->times);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->times));
  __pyx_v_self->times = ((struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList2 *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":32
 *         # self.times = LinkedList.__new__(LinkedList)
 *         self.times = LinkedList2.__new__(LinkedList2)
 *         self.probes = {}             # <<<<<<<<<<<<<<
 *         self.replies = {}
 *         self.infile = infile
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->probes);
  __Pyx_DECREF(__pyx_v_self->probes);
  __pyx_v_self->probes = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":33
 *         self.times = LinkedList2.__new__(LinkedList2)
 *         self.probes = {}
 *         self.replies = {}             # <<<<<<<<<<<<<<
 *         self.infile = infile
 *         self.outfile = outfile
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->replies);
  __Pyx_DECREF(__pyx_v_self->replies);
  __pyx_v_self->replies = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":34
 *         self.probes = {}
 *         self.replies = {}
 *         self.infile = infile             # <<<<<<<<<<<<<<
 *         self.outfile = outfile
 *         # print(outfile)
 */
  __Pyx_INCREF(__pyx_v_infile);
  __Pyx_GIVEREF(__pyx_v_infile);
  __Pyx_GOTREF(__pyx_v_self->infile);
  __Pyx_DECREF(__pyx_v_self->infile);
  __pyx_v_self->infile = __pyx_v_infile;

  /* "cloudtrace/read/convert.pyx":35
 *         self.replies = {}
 *         self.infile = infile
 *         self.outfile = outfile             # <<<<<<<<<<<<<<
 *         # print(outfile)
 *         self.writefile = fopen(outfile, 'wt')
 */
  __Pyx_INCREF(__pyx_v_outfile);
  __Pyx_GIVEREF(__pyx_v_outfile);
  __Pyx_GOTREF(__pyx_v_self->outfile);
  __Pyx_DECREF(__pyx_v_self->outfile);
  __pyx_v_self->outfile = __pyx_v_outfile;

  /* "cloudtrace/read/convert.pyx":37
 *         self.outfile = outfile
 *         # print(outfile)
 *         self.writefile = fopen(outfile, 'wt')             # <<<<<<<<<<<<<<
 *         self.more = True
 *         self.hostname = hostname
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_fopen); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_outfile, __pyx_n_u_wt};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_3, __pyx_v_outfile, __pyx_n_u_wt};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_v_outfile);
    __Pyx_GIVEREF(__pyx_v_outfile);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_outfile);
    __Pyx_INCREF(__pyx_n_u_wt);
    __Pyx_GIVEREF(__pyx_n_u_wt);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_n_u_wt);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->writefile);
  __Pyx_DECREF(__pyx_v_self->writefile);
  __pyx_v_self->writefile = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":38
 *         # print(outfile)
 *         self.writefile = fopen(outfile, 'wt')
 *         self.more = True             # <<<<<<<<<<<<<<
 *         self.hostname = hostname
 * 
 */
  __pyx_v_self->more = 1;

  /* "cloudtrace/read/convert.pyx":39
 *         self.writefile = fopen(outfile, 'wt')
 *         self.more = True
 *         self.hostname = hostname             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(__pyx_v_hostname);
  __Pyx_GIVEREF(__pyx_v_hostname);
  __Pyx_GOTREF(__pyx_v_self->hostname);
  __Pyx_DECREF(__pyx_v_self->hostname);
  __pyx_v_self->hostname = __pyx_v_hostname;

  /* "cloudtrace/read/convert.pyx":21
 * 
 * cdef class Convert:
 *     def __cinit__(self, str infile, str outfile, str hostname=None):             # <<<<<<<<<<<<<<
 *         self.total = 0
 *         self.nprobes = 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pyx":41
 *         self.hostname = hostname
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static void __pyx_pw_10cloudtrace_4read_7convert_7Convert_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10cloudtrace_4read_7convert_7Convert_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10cloudtrace_4read_7convert_7Convert_2__dealloc__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10cloudtrace_4read_7convert_7Convert_2__dealloc__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/convert.pyx":44
 *         pass
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return 'Total {:,d} Probes {:,d} Replies {:,d} PIDP {:,d} PIDR {:,d} Skipped {:,d} Wrote {:,d}'.format(self.total, self.nprobes, self.nreplies, self.pidprobe, self.pidreply, self.skipped, self.wrote)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_4__repr__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_4__repr__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "cloudtrace/read/convert.pyx":45
 * 
 *     def __repr__(self):
 *         return 'Total {:,d} Probes {:,d} Replies {:,d} PIDP {:,d} PIDR {:,d} Skipped {:,d} Wrote {:,d}'.format(self.total, self.nprobes, self.nreplies, self.pidprobe, self.pidreply, self.skipped, self.wrote)             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Total_d_Probes_d_Replies_d_PIDP, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v_self->total); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_long(__pyx_v_self->nprobes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_v_self->nreplies); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_long(__pyx_v_self->pidprobe); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_long(__pyx_v_self->pidreply); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_long(__pyx_v_self->skipped); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_long(__pyx_v_self->wrote); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = NULL;
  __pyx_t_11 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_11 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[8] = {__pyx_t_10, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 7+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[8] = {__pyx_t_10, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8, __pyx_t_9};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_11, 7+__pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  } else
  #endif
  {
    __pyx_t_12 = PyTuple_New(7+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (__pyx_t_10) {
      __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_11, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_12, 3+__pyx_t_11, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_12, 4+__pyx_t_11, __pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_12, 5+__pyx_t_11, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_12, 6+__pyx_t_11, __pyx_t_9);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_t_7 = 0;
    __pyx_t_8 = 0;
    __pyx_t_9 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cloudtrace/read/convert.pyx":44
 *         pass
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return 'Total {:,d} Probes {:,d} Replies {:,d} PIDP {:,d} PIDR {:,d} Skipped {:,d} Wrote {:,d}'.format(self.total, self.nprobes, self.nreplies, self.pidprobe, self.pidreply, self.skipped, self.wrote)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pyx":47
 *         return 'Total {:,d} Probes {:,d} Replies {:,d} PIDP {:,d} PIDR {:,d} Skipped {:,d} Wrote {:,d}'.format(self.total, self.nprobes, self.nreplies, self.pidprobe, self.pidreply, self.skipped, self.wrote)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self.writefile.close()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_7close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_7convert_7Convert_6close[] = "Convert.close(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_7close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_6close(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_6close(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("close", 0);

  /* "cloudtrace/read/convert.pyx":48
 * 
 *     def close(self):
 *         self.writefile.close()             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_dst(self, bytes dst_b) except *:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->writefile, __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":47
 *         return 'Total {:,d} Probes {:,d} Replies {:,d} PIDP {:,d} PIDR {:,d} Skipped {:,d} Wrote {:,d}'.format(self.total, self.nprobes, self.nreplies, self.pidprobe, self.pidreply, self.skipped, self.wrote)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         self.writefile.close()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pyx":50
 *         self.writefile.close()
 * 
 *     cdef void write_dst(self, bytes dst_b) except *:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

static void __pyx_f_10cloudtrace_4read_7convert_7Convert_write_dst(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_dst_b) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_dst", 0);

  /* "cloudtrace/read/convert.pyx":51
 * 
 *     cdef void write_dst(self, bytes dst_b) except *:
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_old(self, timeval *currtime, int delta) except *:
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 51, __pyx_L1_error)

  /* "cloudtrace/read/convert.pyx":50
 *         self.writefile.close()
 * 
 *     cdef void write_dst(self, bytes dst_b) except *:             # <<<<<<<<<<<<<<
 *         raise NotImplementedError()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.write_dst", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/convert.pyx":53
 *         raise NotImplementedError()
 * 
 *     cdef void write_old(self, timeval *currtime, int delta) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             timeval tmptime
 */

static void __pyx_f_10cloudtrace_4read_7convert_7Convert_write_old(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, struct timeval *__pyx_v_currtime, int __pyx_v_delta) {
  struct timeval __pyx_v_tmptime;
  struct __pyx_obj_10cloudtrace_4read_10linkedlist_Node *__pyx_v_curr = 0;
  PyObject *__pyx_v_dst_b = 0;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_old", 0);

  /* "cloudtrace/read/convert.pyx":59
 *             Node curr
 *             bytes dst_b
 *         while self.times.start is not None:             # <<<<<<<<<<<<<<
 *             timersub(currtime, &self.times.start.time, &tmptime)
 *             if tmptime.tv_sec >= delta:
 */
  while (1) {
    __pyx_t_1 = (((PyObject *)__pyx_v_self->times->start) != Py_None);
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (!__pyx_t_2) break;

    /* "cloudtrace/read/convert.pyx":60
 *             bytes dst_b
 *         while self.times.start is not None:
 *             timersub(currtime, &self.times.start.time, &tmptime)             # <<<<<<<<<<<<<<
 *             if tmptime.tv_sec >= delta:
 *                 curr = self.times.pop()
 */
    timersub(__pyx_v_currtime, (&__pyx_v_self->times->start->time), (&__pyx_v_tmptime));

    /* "cloudtrace/read/convert.pyx":61
 *         while self.times.start is not None:
 *             timersub(currtime, &self.times.start.time, &tmptime)
 *             if tmptime.tv_sec >= delta:             # <<<<<<<<<<<<<<
 *                 curr = self.times.pop()
 *                 if curr is None:
 */
    __pyx_t_2 = ((__pyx_v_tmptime.tv_sec >= __pyx_v_delta) != 0);
    if (__pyx_t_2) {

      /* "cloudtrace/read/convert.pyx":62
 *             timersub(currtime, &self.times.start.time, &tmptime)
 *             if tmptime.tv_sec >= delta:
 *                 curr = self.times.pop()             # <<<<<<<<<<<<<<
 *                 if curr is None:
 *                     break
 */
      __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList2 *)__pyx_v_self->times->__pyx_vtab)->pop(__pyx_v_self->times)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 62, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_curr = ((struct __pyx_obj_10cloudtrace_4read_10linkedlist_Node *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "cloudtrace/read/convert.pyx":63
 *             if tmptime.tv_sec >= delta:
 *                 curr = self.times.pop()
 *                 if curr is None:             # <<<<<<<<<<<<<<
 *                     break
 *                 # dst_b = (<unsigned char *>curr.dst)[:curr.dstlen]
 */
      __pyx_t_2 = (((PyObject *)__pyx_v_curr) == Py_None);
      __pyx_t_1 = (__pyx_t_2 != 0);
      if (__pyx_t_1) {

        /* "cloudtrace/read/convert.pyx":64
 *                 curr = self.times.pop()
 *                 if curr is None:
 *                     break             # <<<<<<<<<<<<<<
 *                 # dst_b = (<unsigned char *>curr.dst)[:curr.dstlen]
 *                 dst_b = curr.dst
 */
        goto __pyx_L4_break;

        /* "cloudtrace/read/convert.pyx":63
 *             if tmptime.tv_sec >= delta:
 *                 curr = self.times.pop()
 *                 if curr is None:             # <<<<<<<<<<<<<<
 *                     break
 *                 # dst_b = (<unsigned char *>curr.dst)[:curr.dstlen]
 */
      }

      /* "cloudtrace/read/convert.pyx":66
 *                     break
 *                 # dst_b = (<unsigned char *>curr.dst)[:curr.dstlen]
 *                 dst_b = curr.dst             # <<<<<<<<<<<<<<
 *                 self.write_dst(dst_b)
 *                 del curr
 */
      __pyx_t_3 = __pyx_v_curr->dst;
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_dst_b, ((PyObject*)__pyx_t_3));
      __pyx_t_3 = 0;

      /* "cloudtrace/read/convert.pyx":67
 *                 # dst_b = (<unsigned char *>curr.dst)[:curr.dstlen]
 *                 dst_b = curr.dst
 *                 self.write_dst(dst_b)             # <<<<<<<<<<<<<<
 *                 del curr
 *                 # free_node(curr)
 */
      ((struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert *)__pyx_v_self->__pyx_vtab)->write_dst(__pyx_v_self, __pyx_v_dst_b); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L1_error)

      /* "cloudtrace/read/convert.pyx":68
 *                 dst_b = curr.dst
 *                 self.write_dst(dst_b)
 *                 del curr             # <<<<<<<<<<<<<<
 *                 # free_node(curr)
 *             else:
 */
      __Pyx_DECREF(((PyObject *)__pyx_v_curr));
      __pyx_v_curr = NULL;

      /* "cloudtrace/read/convert.pyx":61
 *         while self.times.start is not None:
 *             timersub(currtime, &self.times.start.time, &tmptime)
 *             if tmptime.tv_sec >= delta:             # <<<<<<<<<<<<<<
 *                 curr = self.times.pop()
 *                 if curr is None:
 */
      goto __pyx_L5;
    }

    /* "cloudtrace/read/convert.pyx":71
 *                 # free_node(curr)
 *             else:
 *                 break             # <<<<<<<<<<<<<<
 * 
 *     cdef void write_all(self) except *:
 */
    /*else*/ {
      goto __pyx_L4_break;
    }
    __pyx_L5:;
  }
  __pyx_L4_break:;

  /* "cloudtrace/read/convert.pyx":53
 *         raise NotImplementedError()
 * 
 *     cdef void write_old(self, timeval *currtime, int delta) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             timeval tmptime
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.write_old", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_curr);
  __Pyx_XDECREF(__pyx_v_dst_b);
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/convert.pyx":73
 *                 break
 * 
 *     cdef void write_all(self) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes dst_b
 */

static void __pyx_f_10cloudtrace_4read_7convert_7Convert_write_all(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_v_dst_b = 0;
  PyObject *__pyx_v_remaining = 0;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_all", 0);

  /* "cloudtrace/read/convert.pyx":76
 *         cdef:
 *             bytes dst_b
 *             list remaining = list(self.replies)             # <<<<<<<<<<<<<<
 *         for dst_b in remaining:
 *             self.write_dst(dst_b)
 */
  __pyx_t_1 = PySequence_List(__pyx_v_self->replies); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_remaining = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":77
 *             bytes dst_b
 *             list remaining = list(self.replies)
 *         for dst_b in remaining:             # <<<<<<<<<<<<<<
 *             self.write_dst(dst_b)
 * 
 */
  __pyx_t_1 = __pyx_v_remaining; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 77, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_dst_b, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "cloudtrace/read/convert.pyx":78
 *             list remaining = list(self.replies)
 *         for dst_b in remaining:
 *             self.write_dst(dst_b)             # <<<<<<<<<<<<<<
 * 
 *     cdef Reply next(self, Pcap pcap):
 */
    ((struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert *)__pyx_v_self->__pyx_vtab)->write_dst(__pyx_v_self, __pyx_v_dst_b); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 78, __pyx_L1_error)

    /* "cloudtrace/read/convert.pyx":77
 *             bytes dst_b
 *             list remaining = list(self.replies)
 *         for dst_b in remaining:             # <<<<<<<<<<<<<<
 *             self.write_dst(dst_b)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":73
 *                 break
 * 
 *     cdef void write_all(self) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes dst_b
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.write_all", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dst_b);
  __Pyx_XDECREF(__pyx_v_remaining);
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/convert.pyx":80
 *             self.write_dst(dst_b)
 * 
 *     cdef Reply next(self, Pcap pcap):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int more
 */

static struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_f_10cloudtrace_4read_7convert_7Convert_next(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *__pyx_v_pcap) {
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_p = 0;
  struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_probe = 0;
  __pyx_t_10cloudtrace_4read_7convert_probe_id __pyx_v_pid;
  PyObject *__pyx_v_pid_b = 0;
  PyObject *__pyx_v_dst_b = 0;
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_reply = 0;
  PyObject *__pyx_v_dsts = 0;
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("next", 0);

  /* "cloudtrace/read/convert.pyx":86
 *             probe_id pid
 *             bytes pid_b, dst_b
 *             Reply reply = None             # <<<<<<<<<<<<<<
 *             dict dsts
 *         self.more = pcap.next_packet()
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_reply = ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)Py_None);

  /* "cloudtrace/read/convert.pyx":88
 *             Reply reply = None
 *             dict dsts
 *         self.more = pcap.next_packet()             # <<<<<<<<<<<<<<
 *         if not self.more:
 *             # print('writing all')
 */
  __pyx_v_self->more = ((struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap *)__pyx_v_pcap->__pyx_vtab)->next_packet(__pyx_v_pcap, 0);

  /* "cloudtrace/read/convert.pyx":89
 *             dict dsts
 *         self.more = pcap.next_packet()
 *         if not self.more:             # <<<<<<<<<<<<<<
 *             # print('writing all')
 *             self.write_all()
 */
  __pyx_t_1 = ((!(__pyx_v_self->more != 0)) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/convert.pyx":91
 *         if not self.more:
 *             # print('writing all')
 *             self.write_all()             # <<<<<<<<<<<<<<
 *             # print('done writing all')
 *             pcap.close()
 */
    ((struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert *)__pyx_v_self->__pyx_vtab)->write_all(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 91, __pyx_L1_error)

    /* "cloudtrace/read/convert.pyx":93
 *             self.write_all()
 *             # print('done writing all')
 *             pcap.close()             # <<<<<<<<<<<<<<
 *             # print('closed')
 *             return None
 */
    ((struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap *)__pyx_v_pcap->__pyx_vtab)->close(__pyx_v_pcap, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 93, __pyx_L1_error)

    /* "cloudtrace/read/convert.pyx":95
 *             pcap.close()
 *             # print('closed')
 *             return None             # <<<<<<<<<<<<<<
 *         p = disassemble(pcap)
 *         self.total += 1
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "cloudtrace/read/convert.pyx":89
 *             dict dsts
 *         self.more = pcap.next_packet()
 *         if not self.more:             # <<<<<<<<<<<<<<
 *             # print('writing all')
 *             self.write_all()
 */
  }

  /* "cloudtrace/read/convert.pyx":96
 *             # print('closed')
 *             return None
 *         p = disassemble(pcap)             # <<<<<<<<<<<<<<
 *         self.total += 1
 *         if p is None:
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_10cloudtrace_4read_6packet_disassemble(__pyx_v_pcap, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_p = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":97
 *             return None
 *         p = disassemble(pcap)
 *         self.total += 1             # <<<<<<<<<<<<<<
 *         if p is None:
 *             self.skipped += 1
 */
  __pyx_v_self->total = (__pyx_v_self->total + 1);

  /* "cloudtrace/read/convert.pyx":98
 *         p = disassemble(pcap)
 *         self.total += 1
 *         if p is None:             # <<<<<<<<<<<<<<
 *             self.skipped += 1
 *             return None
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_p) == Py_None);
  __pyx_t_3 = (__pyx_t_1 != 0);
  if (__pyx_t_3) {

    /* "cloudtrace/read/convert.pyx":99
 *         self.total += 1
 *         if p is None:
 *             self.skipped += 1             # <<<<<<<<<<<<<<
 *             return None
 *         memcpy(pid.dst, p.get_dst_c(), 4)
 */
    __pyx_v_self->skipped = (__pyx_v_self->skipped + 1);

    /* "cloudtrace/read/convert.pyx":100
 *         if p is None:
 *             self.skipped += 1
 *             return None             # <<<<<<<<<<<<<<
 *         memcpy(pid.dst, p.get_dst_c(), 4)
 *         pid.ttl = p.probe_ttl()
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "cloudtrace/read/convert.pyx":98
 *         p = disassemble(pcap)
 *         self.total += 1
 *         if p is None:             # <<<<<<<<<<<<<<
 *             self.skipped += 1
 *             return None
 */
  }

  /* "cloudtrace/read/convert.pyx":101
 *             self.skipped += 1
 *             return None
 *         memcpy(pid.dst, p.get_dst_c(), 4)             # <<<<<<<<<<<<<<
 *         pid.ttl = p.probe_ttl()
 *         pid.pid = p.get_pid()
 */
  (void)(memcpy(__pyx_v_pid.dst, ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *)__pyx_v_p->__pyx_vtab)->get_dst_c(__pyx_v_p), 4));

  /* "cloudtrace/read/convert.pyx":102
 *             return None
 *         memcpy(pid.dst, p.get_dst_c(), 4)
 *         pid.ttl = p.probe_ttl()             # <<<<<<<<<<<<<<
 *         pid.pid = p.get_pid()
 *         pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]
 */
  __pyx_v_pid.ttl = ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *)__pyx_v_p->__pyx_vtab)->probe_ttl(__pyx_v_p, 0);

  /* "cloudtrace/read/convert.pyx":103
 *         memcpy(pid.dst, p.get_dst_c(), 4)
 *         pid.ttl = p.probe_ttl()
 *         pid.pid = p.get_pid()             # <<<<<<<<<<<<<<
 *         pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]
 *         dst_b = pid.dst[:4]
 */
  __pyx_v_pid.pid = ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *)__pyx_v_p->__pyx_vtab)->get_pid(__pyx_v_p, 0);

  /* "cloudtrace/read/convert.pyx":104
 *         pid.ttl = p.probe_ttl()
 *         pid.pid = p.get_pid()
 *         pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]             # <<<<<<<<<<<<<<
 *         dst_b = pid.dst[:4]
 *         if dst_b == b'\x00\x00\x00\x00':
 */
  __pyx_t_2 = __Pyx_PyBytes_FromStringAndSize(((const char*)((unsigned char *)(&__pyx_v_pid))) + 0, (sizeof(__pyx_t_10cloudtrace_4read_7convert_probe_id)) - 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_pid_b = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "cloudtrace/read/convert.pyx":105
 *         pid.pid = p.get_pid()
 *         pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]
 *         dst_b = pid.dst[:4]             # <<<<<<<<<<<<<<
 *         if dst_b == b'\x00\x00\x00\x00':
 *             return None
 */
  __pyx_t_4 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_pid.dst) + 0, 4 - 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_dst_b = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "cloudtrace/read/convert.pyx":106
 *         pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]
 *         dst_b = pid.dst[:4]
 *         if dst_b == b'\x00\x00\x00\x00':             # <<<<<<<<<<<<<<
 *             return None
 *         if p.isprobe:
 */
  __pyx_t_3 = (__Pyx_PyBytes_Equals(__pyx_v_dst_b, __pyx_kp_b_, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 106, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_3 != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/convert.pyx":107
 *         dst_b = pid.dst[:4]
 *         if dst_b == b'\x00\x00\x00\x00':
 *             return None             # <<<<<<<<<<<<<<
 *         if p.isprobe:
 *             if dst_b in self.probes:
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "cloudtrace/read/convert.pyx":106
 *         pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]
 *         dst_b = pid.dst[:4]
 *         if dst_b == b'\x00\x00\x00\x00':             # <<<<<<<<<<<<<<
 *             return None
 *         if p.isprobe:
 */
  }

  /* "cloudtrace/read/convert.pyx":108
 *         if dst_b == b'\x00\x00\x00\x00':
 *             return None
 *         if p.isprobe:             # <<<<<<<<<<<<<<
 *             if dst_b in self.probes:
 *                 dsts = self.probes[dst_b]
 */
  __pyx_t_1 = (__pyx_v_p->isprobe != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/convert.pyx":109
 *             return None
 *         if p.isprobe:
 *             if dst_b in self.probes:             # <<<<<<<<<<<<<<
 *                 dsts = self.probes[dst_b]
 *             else:
 */
    if (unlikely(__pyx_v_self->probes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 109, __pyx_L1_error)
    }
    __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_dst_b, __pyx_v_self->probes, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 109, __pyx_L1_error)
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "cloudtrace/read/convert.pyx":110
 *         if p.isprobe:
 *             if dst_b in self.probes:
 *                 dsts = self.probes[dst_b]             # <<<<<<<<<<<<<<
 *             else:
 *                 dsts = {}
 */
      if (unlikely(__pyx_v_self->probes == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 110, __pyx_L1_error)
      }
      __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_self->probes, __pyx_v_dst_b); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(PyDict_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 110, __pyx_L1_error)
      __pyx_v_dsts = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "cloudtrace/read/convert.pyx":109
 *             return None
 *         if p.isprobe:
 *             if dst_b in self.probes:             # <<<<<<<<<<<<<<
 *                 dsts = self.probes[dst_b]
 *             else:
 */
      goto __pyx_L7;
    }

    /* "cloudtrace/read/convert.pyx":112
 *                 dsts = self.probes[dst_b]
 *             else:
 *                 dsts = {}             # <<<<<<<<<<<<<<
 *                 self.probes[dst_b] = dsts
 *             dsts[pid_b] = p
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_v_dsts = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "cloudtrace/read/convert.pyx":113
 *             else:
 *                 dsts = {}
 *                 self.probes[dst_b] = dsts             # <<<<<<<<<<<<<<
 *             dsts[pid_b] = p
 *             if p.probe_ttl() == 32:
 */
      if (unlikely(__pyx_v_self->probes == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 113, __pyx_L1_error)
      }
      if (unlikely(PyDict_SetItem(__pyx_v_self->probes, __pyx_v_dst_b, __pyx_v_dsts) < 0)) __PYX_ERR(0, 113, __pyx_L1_error)
    }
    __pyx_L7:;

    /* "cloudtrace/read/convert.pyx":114
 *                 dsts = {}
 *                 self.probes[dst_b] = dsts
 *             dsts[pid_b] = p             # <<<<<<<<<<<<<<
 *             if p.probe_ttl() == 32:
 *                 self.times.add(AF_INET, pid.dst, p.time)
 */
    if (unlikely(__pyx_v_dsts == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 114, __pyx_L1_error)
    }
    if (unlikely(PyDict_SetItem(__pyx_v_dsts, __pyx_v_pid_b, ((PyObject *)__pyx_v_p)) < 0)) __PYX_ERR(0, 114, __pyx_L1_error)

    /* "cloudtrace/read/convert.pyx":115
 *                 self.probes[dst_b] = dsts
 *             dsts[pid_b] = p
 *             if p.probe_ttl() == 32:             # <<<<<<<<<<<<<<
 *                 self.times.add(AF_INET, pid.dst, p.time)
 *             self.nprobes += 1
 */
    __pyx_t_3 = ((((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet *)__pyx_v_p->__pyx_vtab)->probe_ttl(__pyx_v_p, 0) == 32) != 0);
    if (__pyx_t_3) {

      /* "cloudtrace/read/convert.pyx":116
 *             dsts[pid_b] = p
 *             if p.probe_ttl() == 32:
 *                 self.times.add(AF_INET, pid.dst, p.time)             # <<<<<<<<<<<<<<
 *             self.nprobes += 1
 *         else:
 */
      ((struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList2 *)__pyx_v_self->times->__pyx_vtab)->add(__pyx_v_self->times, AF_INET, __pyx_v_pid.dst, __pyx_v_p->time); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 116, __pyx_L1_error)

      /* "cloudtrace/read/convert.pyx":115
 *                 self.probes[dst_b] = dsts
 *             dsts[pid_b] = p
 *             if p.probe_ttl() == 32:             # <<<<<<<<<<<<<<
 *                 self.times.add(AF_INET, pid.dst, p.time)
 *             self.nprobes += 1
 */
    }

    /* "cloudtrace/read/convert.pyx":117
 *             if p.probe_ttl() == 32:
 *                 self.times.add(AF_INET, pid.dst, p.time)
 *             self.nprobes += 1             # <<<<<<<<<<<<<<
 *         else:
 *             if pid.ttl > 32 or pid.ttl == 0:
 */
    __pyx_v_self->nprobes = (__pyx_v_self->nprobes + 1);

    /* "cloudtrace/read/convert.pyx":108
 *         if dst_b == b'\x00\x00\x00\x00':
 *             return None
 *         if p.isprobe:             # <<<<<<<<<<<<<<
 *             if dst_b in self.probes:
 *                 dsts = self.probes[dst_b]
 */
    goto __pyx_L6;
  }

  /* "cloudtrace/read/convert.pyx":119
 *             self.nprobes += 1
 *         else:
 *             if pid.ttl > 32 or pid.ttl == 0:             # <<<<<<<<<<<<<<
 *                 self.ttlhigh += 1
 *             elif dst_b not in self.probes:
 */
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_pid.ttl > 32) != 0);
    if (!__pyx_t_1) {
    } else {
      __pyx_t_3 = __pyx_t_1;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_pid.ttl == 0) != 0);
    __pyx_t_3 = __pyx_t_1;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_3) {

      /* "cloudtrace/read/convert.pyx":120
 *         else:
 *             if pid.ttl > 32 or pid.ttl == 0:
 *                 self.ttlhigh += 1             # <<<<<<<<<<<<<<
 *             elif dst_b not in self.probes:
 *                 self.pidreply += 1
 */
      __pyx_v_self->ttlhigh = (__pyx_v_self->ttlhigh + 1);

      /* "cloudtrace/read/convert.pyx":119
 *             self.nprobes += 1
 *         else:
 *             if pid.ttl > 32 or pid.ttl == 0:             # <<<<<<<<<<<<<<
 *                 self.ttlhigh += 1
 *             elif dst_b not in self.probes:
 */
      goto __pyx_L9;
    }

    /* "cloudtrace/read/convert.pyx":121
 *             if pid.ttl > 32 or pid.ttl == 0:
 *                 self.ttlhigh += 1
 *             elif dst_b not in self.probes:             # <<<<<<<<<<<<<<
 *                 self.pidreply += 1
 *             else:
 */
    if (unlikely(__pyx_v_self->probes == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 121, __pyx_L1_error)
    }
    __pyx_t_3 = (__Pyx_PyDict_ContainsTF(__pyx_v_dst_b, __pyx_v_self->probes, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 121, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_3 != 0);
    if (__pyx_t_1) {

      /* "cloudtrace/read/convert.pyx":122
 *                 self.ttlhigh += 1
 *             elif dst_b not in self.probes:
 *                 self.pidreply += 1             # <<<<<<<<<<<<<<
 *             else:
 *                 dsts = self.probes[dst_b]
 */
      __pyx_v_self->pidreply = (__pyx_v_self->pidreply + 1);

      /* "cloudtrace/read/convert.pyx":121
 *             if pid.ttl > 32 or pid.ttl == 0:
 *                 self.ttlhigh += 1
 *             elif dst_b not in self.probes:             # <<<<<<<<<<<<<<
 *                 self.pidreply += 1
 *             else:
 */
      goto __pyx_L9;
    }

    /* "cloudtrace/read/convert.pyx":124
 *                 self.pidreply += 1
 *             else:
 *                 dsts = self.probes[dst_b]             # <<<<<<<<<<<<<<
 *                 if pid_b not in dsts:
 *                     self.pidreply += 1
 */
    /*else*/ {
      if (unlikely(__pyx_v_self->probes == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 124, __pyx_L1_error)
      }
      __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_self->probes, __pyx_v_dst_b); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(PyDict_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 124, __pyx_L1_error)
      __pyx_v_dsts = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "cloudtrace/read/convert.pyx":125
 *             else:
 *                 dsts = self.probes[dst_b]
 *                 if pid_b not in dsts:             # <<<<<<<<<<<<<<
 *                     self.pidreply += 1
 *                 else:
 */
      if (unlikely(__pyx_v_dsts == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 125, __pyx_L1_error)
      }
      __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_pid_b, __pyx_v_dsts, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 125, __pyx_L1_error)
      __pyx_t_3 = (__pyx_t_1 != 0);
      if (__pyx_t_3) {

        /* "cloudtrace/read/convert.pyx":126
 *                 dsts = self.probes[dst_b]
 *                 if pid_b not in dsts:
 *                     self.pidreply += 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     probe = dsts.pop(pid_b)
 */
        __pyx_v_self->pidreply = (__pyx_v_self->pidreply + 1);

        /* "cloudtrace/read/convert.pyx":125
 *             else:
 *                 dsts = self.probes[dst_b]
 *                 if pid_b not in dsts:             # <<<<<<<<<<<<<<
 *                     self.pidreply += 1
 *                 else:
 */
        goto __pyx_L12;
      }

      /* "cloudtrace/read/convert.pyx":128
 *                     self.pidreply += 1
 *                 else:
 *                     probe = dsts.pop(pid_b)             # <<<<<<<<<<<<<<
 *                     reply = <Reply>p
 *                     reply.orig_probe = probe
 */
      /*else*/ {
        if (unlikely(__pyx_v_dsts == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
          __PYX_ERR(0, 128, __pyx_L1_error)
        }
        __pyx_t_4 = __Pyx_PyDict_Pop(__pyx_v_dsts, __pyx_v_pid_b, ((PyObject *)NULL)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 128, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_10cloudtrace_4read_6packet_Packet))))) __PYX_ERR(0, 128, __pyx_L1_error)
        __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_t_4);
        __pyx_t_4 = 0;

        /* "cloudtrace/read/convert.pyx":129
 *                 else:
 *                     probe = dsts.pop(pid_b)
 *                     reply = <Reply>p             # <<<<<<<<<<<<<<
 *                     reply.orig_probe = probe
 *                     self.nreplies += 1
 */
        __pyx_t_4 = ((PyObject *)__pyx_v_p);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_DECREF_SET(__pyx_v_reply, ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "cloudtrace/read/convert.pyx":130
 *                     probe = dsts.pop(pid_b)
 *                     reply = <Reply>p
 *                     reply.orig_probe = probe             # <<<<<<<<<<<<<<
 *                     self.nreplies += 1
 *         self.write_old(p.time, 10)
 */
        __Pyx_INCREF(((PyObject *)__pyx_v_probe));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_probe));
        __Pyx_GOTREF(__pyx_v_reply->orig_probe);
        __Pyx_DECREF(((PyObject *)__pyx_v_reply->orig_probe));
        __pyx_v_reply->orig_probe = __pyx_v_probe;

        /* "cloudtrace/read/convert.pyx":131
 *                     reply = <Reply>p
 *                     reply.orig_probe = probe
 *                     self.nreplies += 1             # <<<<<<<<<<<<<<
 *         self.write_old(p.time, 10)
 *         if reply is not None:
 */
        __pyx_v_self->nreplies = (__pyx_v_self->nreplies + 1);
      }
      __pyx_L12:;
    }
    __pyx_L9:;
  }
  __pyx_L6:;

  /* "cloudtrace/read/convert.pyx":132
 *                     reply.orig_probe = probe
 *                     self.nreplies += 1
 *         self.write_old(p.time, 10)             # <<<<<<<<<<<<<<
 *         if reply is not None:
 *             return reply
 */
  ((struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert *)__pyx_v_self->__pyx_vtab)->write_old(__pyx_v_self, __pyx_v_p->time, 10); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 132, __pyx_L1_error)

  /* "cloudtrace/read/convert.pyx":133
 *                     self.nreplies += 1
 *         self.write_old(p.time, 10)
 *         if reply is not None:             # <<<<<<<<<<<<<<
 *             return reply
 *         return None
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_reply) != Py_None);
  __pyx_t_1 = (__pyx_t_3 != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/convert.pyx":134
 *         self.write_old(p.time, 10)
 *         if reply is not None:
 *             return reply             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_reply));
    __pyx_r = __pyx_v_reply;
    goto __pyx_L0;

    /* "cloudtrace/read/convert.pyx":133
 *                     self.nreplies += 1
 *         self.write_old(p.time, 10)
 *         if reply is not None:             # <<<<<<<<<<<<<<
 *             return reply
 *         return None
 */
  }

  /* "cloudtrace/read/convert.pyx":135
 *         if reply is not None:
 *             return reply
 *         return None             # <<<<<<<<<<<<<<
 * 
 * cdef int loop_check(Trace trace):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_r = ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)Py_None); __Pyx_INCREF(Py_None);
  goto __pyx_L0;

  /* "cloudtrace/read/convert.pyx":80
 *             self.write_dst(dst_b)
 * 
 *     cdef Reply next(self, Pcap pcap):             # <<<<<<<<<<<<<<
 *         cdef:
 *             int more
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_p);
  __Pyx_XDECREF((PyObject *)__pyx_v_probe);
  __Pyx_XDECREF(__pyx_v_pid_b);
  __Pyx_XDECREF(__pyx_v_dst_b);
  __Pyx_XDECREF((PyObject *)__pyx_v_reply);
  __Pyx_XDECREF(__pyx_v_dsts);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pxd":23
 *     cdef:
 *         str infile, outfile, hostname
 *         public long total, nprobes, nreplies, skipped, pidprobe, pidreply, ttlhigh, wrote             # <<<<<<<<<<<<<<
 *         # LinkedList times
 *         LinkedList2 times
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_5total_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_5total_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_5total___get__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_5total___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->total); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.total.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_5total_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_5total_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_5total_2__set__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_5total_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 23, __pyx_L1_error)
  __pyx_v_self->total = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.total.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_7nprobes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_7nprobes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_7nprobes___get__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_7nprobes___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->nprobes); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.nprobes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_7nprobes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_7nprobes_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_7nprobes_2__set__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_7nprobes_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 23, __pyx_L1_error)
  __pyx_v_self->nprobes = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.nprobes.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_8nreplies_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_8nreplies_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_8nreplies___get__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_8nreplies___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->nreplies); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.nreplies.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_8nreplies_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_8nreplies_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_8nreplies_2__set__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_8nreplies_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 23, __pyx_L1_error)
  __pyx_v_self->nreplies = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.nreplies.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_7skipped_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_7skipped_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_7skipped___get__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_7skipped___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->skipped); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.skipped.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_7skipped_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_7skipped_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_7skipped_2__set__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_7skipped_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 23, __pyx_L1_error)
  __pyx_v_self->skipped = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.skipped.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidprobe_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidprobe_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidprobe___get__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidprobe___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->pidprobe); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.pidprobe.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidprobe_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidprobe_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidprobe_2__set__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidprobe_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 23, __pyx_L1_error)
  __pyx_v_self->pidprobe = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.pidprobe.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidreply_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidreply_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidreply___get__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidreply___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->pidreply); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.pidreply.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidreply_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidreply_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidreply_2__set__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_8pidreply_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 23, __pyx_L1_error)
  __pyx_v_self->pidreply = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.pidreply.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_7ttlhigh_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_7ttlhigh_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_7ttlhigh___get__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_7ttlhigh___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->ttlhigh); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.ttlhigh.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_7ttlhigh_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_7ttlhigh_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_7ttlhigh_2__set__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_7ttlhigh_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 23, __pyx_L1_error)
  __pyx_v_self->ttlhigh = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.ttlhigh.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_5wrote_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_5wrote_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_5wrote___get__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_5wrote___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->wrote); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.wrote.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_5wrote_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_5wrote_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_5wrote_2__set__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_5wrote_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 23, __pyx_L1_error)
  __pyx_v_self->wrote = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.wrote.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pxd":27
 *         LinkedList2 times
 *         dict probes, replies
 *         public writefile             # <<<<<<<<<<<<<<
 *         Pcap pcap
 *         bint more
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_9writefile_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_9writefile_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_9writefile___get__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_9writefile___get__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->writefile);
  __pyx_r = __pyx_v_self->writefile;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_9writefile_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_9writefile_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_9writefile_2__set__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_9writefile_2__set__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->writefile);
  __Pyx_DECREF(__pyx_v_self->writefile);
  __pyx_v_self->writefile = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_9writefile_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10cloudtrace_4read_7convert_7Convert_9writefile_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_9writefile_4__del__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_4read_7convert_7Convert_9writefile_4__del__(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->writefile);
  __Pyx_DECREF(__pyx_v_self->writefile);
  __pyx_v_self->writefile = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_7convert_7Convert_8__reduce_cython__[] = "Convert.__reduce_cython__(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_8__reduce_cython__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_4read_7convert_7Convert_10__setstate_cython__[] = "Convert.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_7Convert_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_7Convert_10__setstate_cython__(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_7Convert_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_Convert *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.Convert.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pyx":137
 *         return None
 * 
 * cdef int loop_check(Trace trace):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint8_t i, j
 */

static int __pyx_f_10cloudtrace_4read_7convert_loop_check(struct __pyx_obj_10cloudtrace_4read_7combine_Trace *__pyx_v_trace) {
  uint8_t __pyx_v_i;
  uint8_t __pyx_v_j;
  __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_hop1;
  __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_hop2;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  long __pyx_t_2;
  uint8_t __pyx_t_3;
  int __pyx_t_4;
  uint16_t __pyx_t_5;
  uint16_t __pyx_t_6;
  uint8_t __pyx_t_7;
  __Pyx_RefNannySetupContext("loop_check", 0);

  /* "cloudtrace/read/convert.pyx":142
 *         hop_t *hop1
 *         hop_t *hop2
 *     for i in range(trace.hop_count - 2):             # <<<<<<<<<<<<<<
 *         hop1 = trace.hops[i]
 *         if hop1 is not NULL:
 */
  __pyx_t_1 = (__pyx_v_trace->hop_count - 2);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "cloudtrace/read/convert.pyx":143
 *         hop_t *hop2
 *     for i in range(trace.hop_count - 2):
 *         hop1 = trace.hops[i]             # <<<<<<<<<<<<<<
 *         if hop1 is not NULL:
 *             for j in range(i + 1, trace.hop_count):
 */
    __pyx_v_hop1 = (__pyx_v_trace->hops[__pyx_v_i]);

    /* "cloudtrace/read/convert.pyx":144
 *     for i in range(trace.hop_count - 2):
 *         hop1 = trace.hops[i]
 *         if hop1 is not NULL:             # <<<<<<<<<<<<<<
 *             for j in range(i + 1, trace.hop_count):
 *                 hop2 = trace.hops[j]
 */
    __pyx_t_4 = ((__pyx_v_hop1 != NULL) != 0);
    if (__pyx_t_4) {

      /* "cloudtrace/read/convert.pyx":145
 *         hop1 = trace.hops[i]
 *         if hop1 is not NULL:
 *             for j in range(i + 1, trace.hop_count):             # <<<<<<<<<<<<<<
 *                 hop2 = trace.hops[j]
 *                 if hop2 is not NULL:
 */
      __pyx_t_5 = __pyx_v_trace->hop_count;
      __pyx_t_6 = __pyx_t_5;
      for (__pyx_t_7 = (__pyx_v_i + 1); __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
        __pyx_v_j = __pyx_t_7;

        /* "cloudtrace/read/convert.pyx":146
 *         if hop1 is not NULL:
 *             for j in range(i + 1, trace.hop_count):
 *                 hop2 = trace.hops[j]             # <<<<<<<<<<<<<<
 *                 if hop2 is not NULL:
 *                     if memcmp(hop1.hop_addr, hop2.hop_addr, 16) == 0:
 */
        __pyx_v_hop2 = (__pyx_v_trace->hops[__pyx_v_j]);

        /* "cloudtrace/read/convert.pyx":147
 *             for j in range(i + 1, trace.hop_count):
 *                 hop2 = trace.hops[j]
 *                 if hop2 is not NULL:             # <<<<<<<<<<<<<<
 *                     if memcmp(hop1.hop_addr, hop2.hop_addr, 16) == 0:
 *                         if i == j - 1:
 */
        __pyx_t_4 = ((__pyx_v_hop2 != NULL) != 0);
        if (__pyx_t_4) {

          /* "cloudtrace/read/convert.pyx":148
 *                 hop2 = trace.hops[j]
 *                 if hop2 is not NULL:
 *                     if memcmp(hop1.hop_addr, hop2.hop_addr, 16) == 0:             # <<<<<<<<<<<<<<
 *                         if i == j - 1:
 *                             break
 */
          __pyx_t_4 = ((memcmp(__pyx_v_hop1->hop_addr, __pyx_v_hop2->hop_addr, 16) == 0) != 0);
          if (__pyx_t_4) {

            /* "cloudtrace/read/convert.pyx":149
 *                 if hop2 is not NULL:
 *                     if memcmp(hop1.hop_addr, hop2.hop_addr, 16) == 0:
 *                         if i == j - 1:             # <<<<<<<<<<<<<<
 *                             break
 *                         else:
 */
            __pyx_t_4 = ((__pyx_v_i == (__pyx_v_j - 1)) != 0);
            if (__pyx_t_4) {

              /* "cloudtrace/read/convert.pyx":150
 *                     if memcmp(hop1.hop_addr, hop2.hop_addr, 16) == 0:
 *                         if i == j - 1:
 *                             break             # <<<<<<<<<<<<<<
 *                         else:
 *                             return j + 1
 */
              goto __pyx_L7_break;

              /* "cloudtrace/read/convert.pyx":149
 *                 if hop2 is not NULL:
 *                     if memcmp(hop1.hop_addr, hop2.hop_addr, 16) == 0:
 *                         if i == j - 1:             # <<<<<<<<<<<<<<
 *                             break
 *                         else:
 */
            }

            /* "cloudtrace/read/convert.pyx":152
 *                             break
 *                         else:
 *                             return j + 1             # <<<<<<<<<<<<<<
 *     return -1
 * 
 */
            /*else*/ {
              __pyx_r = (__pyx_v_j + 1);
              goto __pyx_L0;
            }

            /* "cloudtrace/read/convert.pyx":148
 *                 hop2 = trace.hops[j]
 *                 if hop2 is not NULL:
 *                     if memcmp(hop1.hop_addr, hop2.hop_addr, 16) == 0:             # <<<<<<<<<<<<<<
 *                         if i == j - 1:
 *                             break
 */
          }

          /* "cloudtrace/read/convert.pyx":147
 *             for j in range(i + 1, trace.hop_count):
 *                 hop2 = trace.hops[j]
 *                 if hop2 is not NULL:             # <<<<<<<<<<<<<<
 *                     if memcmp(hop1.hop_addr, hop2.hop_addr, 16) == 0:
 *                         if i == j - 1:
 */
        }
      }
      __pyx_L7_break:;

      /* "cloudtrace/read/convert.pyx":144
 *     for i in range(trace.hop_count - 2):
 *         hop1 = trace.hops[i]
 *         if hop1 is not NULL:             # <<<<<<<<<<<<<<
 *             for j in range(i + 1, trace.hop_count):
 *                 hop2 = trace.hops[j]
 */
    }
  }

  /* "cloudtrace/read/convert.pyx":153
 *                         else:
 *                             return j + 1
 *     return -1             # <<<<<<<<<<<<<<
 * 
 * cdef str addr_tostr(char *addr):
 */
  __pyx_r = -1;
  goto __pyx_L0;

  /* "cloudtrace/read/convert.pyx":137
 *         return None
 * 
 * cdef int loop_check(Trace trace):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint8_t i, j
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pyx":155
 *     return -1
 * 
 * cdef str addr_tostr(char *addr):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char dst[46]
 */

static PyObject *__pyx_f_10cloudtrace_4read_7convert_addr_tostr(char *__pyx_v_addr) {
  char __pyx_v_dst[46];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("addr_tostr", 0);

  /* "cloudtrace/read/convert.pyx":158
 *     cdef:
 *         char dst[46]
 *     inet_ntop(AF_INET, addr, dst, INET6_ADDRSTRLEN)             # <<<<<<<<<<<<<<
 *     return (<bytes>dst).decode()
 * 
 */
  (void)(inet_ntop(AF_INET, __pyx_v_addr, __pyx_v_dst, INET6_ADDRSTRLEN));

  /* "cloudtrace/read/convert.pyx":159
 *         char dst[46]
 *     inet_ntop(AF_INET, addr, dst, INET6_ADDRSTRLEN)
 *     return (<bytes>dst).decode()             # <<<<<<<<<<<<<<
 * 
 * cdef class ConvertTrace(Convert):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_FromString(__pyx_v_dst); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
    __PYX_ERR(0, 159, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_decode_bytes(((PyObject*)__pyx_t_1), 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "cloudtrace/read/convert.pyx":155
 *     return -1
 * 
 * cdef str addr_tostr(char *addr):             # <<<<<<<<<<<<<<
 *     cdef:
 *         char dst[46]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cloudtrace.read.convert.addr_tostr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pyx":162
 * 
 * cdef class ConvertTrace(Convert):
 *     cdef dict hopdict(self, hop_t *hop):             # <<<<<<<<<<<<<<
 *         rtt = round(hop.hop_rtt.tv_sec * 1000 + (hop.hop_rtt.tv_usec / 1000), 3)
 *         return {
 */

static PyObject *__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_hopdict(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_hop) {
  PyObject *__pyx_v_rtt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("hopdict", 0);

  /* "cloudtrace/read/convert.pyx":163
 * cdef class ConvertTrace(Convert):
 *     cdef dict hopdict(self, hop_t *hop):
 *         rtt = round(hop.hop_rtt.tv_sec * 1000 + (hop.hop_rtt.tv_usec / 1000), 3)             # <<<<<<<<<<<<<<
 *         return {
 *             "addr": addr_tostr(hop.hop_addr),
 */
  __pyx_t_1 = PyFloat_FromDouble(((__pyx_v_hop->hop_rtt.tv_sec * 0x3E8) + (((double)__pyx_v_hop->hop_rtt.tv_usec) / 1000.0))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(__pyx_int_3);
  __Pyx_GIVEREF(__pyx_int_3);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_3);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_rtt = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":164
 *     cdef dict hopdict(self, hop_t *hop):
 *         rtt = round(hop.hop_rtt.tv_sec * 1000 + (hop.hop_rtt.tv_usec / 1000), 3)
 *         return {             # <<<<<<<<<<<<<<
 *             "addr": addr_tostr(hop.hop_addr),
 *             "probe_ttl": hop.hop_probe_ttl,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "cloudtrace/read/convert.pyx":165
 *         rtt = round(hop.hop_rtt.tv_sec * 1000 + (hop.hop_rtt.tv_usec / 1000), 3)
 *         return {
 *             "addr": addr_tostr(hop.hop_addr),             # <<<<<<<<<<<<<<
 *             "probe_ttl": hop.hop_probe_ttl,
 *             "probe_id": hop.hop_probe_id,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_10cloudtrace_4read_7convert_addr_tostr(__pyx_v_hop->hop_addr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_addr, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":166
 *         return {
 *             "addr": addr_tostr(hop.hop_addr),
 *             "probe_ttl": hop.hop_probe_ttl,             # <<<<<<<<<<<<<<
 *             "probe_id": hop.hop_probe_id,
 *             "probe_size": hop.hop_probe_size,
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint8_t(__pyx_v_hop->hop_probe_ttl); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_probe_ttl, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":167
 *             "addr": addr_tostr(hop.hop_addr),
 *             "probe_ttl": hop.hop_probe_ttl,
 *             "probe_id": hop.hop_probe_id,             # <<<<<<<<<<<<<<
 *             "probe_size": hop.hop_probe_size,
 *             "tx": {"sec": hop.hop_tx.tv_sec, "usec": hop.hop_tx.tv_usec},
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_v_hop->hop_probe_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_probe_id, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":168
 *             "probe_ttl": hop.hop_probe_ttl,
 *             "probe_id": hop.hop_probe_id,
 *             "probe_size": hop.hop_probe_size,             # <<<<<<<<<<<<<<
 *             "tx": {"sec": hop.hop_tx.tv_sec, "usec": hop.hop_tx.tv_usec},
 *             "rtt": rtt,
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_v_hop->hop_probe_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_probe_size, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":169
 *             "probe_id": hop.hop_probe_id,
 *             "probe_size": hop.hop_probe_size,
 *             "tx": {"sec": hop.hop_tx.tv_sec, "usec": hop.hop_tx.tv_usec},             # <<<<<<<<<<<<<<
 *             "rtt": rtt,
 *             "reply_ttl": hop.hop_reply_ttl,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_time_t(__pyx_v_hop->hop_tx.tv_sec); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_sec, __pyx_t_3) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_suseconds_t(__pyx_v_hop->hop_tx.tv_usec); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_u_usec, __pyx_t_3) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_tx, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":170
 *             "probe_size": hop.hop_probe_size,
 *             "tx": {"sec": hop.hop_tx.tv_sec, "usec": hop.hop_tx.tv_usec},
 *             "rtt": rtt,             # <<<<<<<<<<<<<<
 *             "reply_ttl": hop.hop_reply_ttl,
 *             "reply_tos": hop.hop_reply_tos,
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_rtt, __pyx_v_rtt) < 0) __PYX_ERR(0, 165, __pyx_L1_error)

  /* "cloudtrace/read/convert.pyx":171
 *             "tx": {"sec": hop.hop_tx.tv_sec, "usec": hop.hop_tx.tv_usec},
 *             "rtt": rtt,
 *             "reply_ttl": hop.hop_reply_ttl,             # <<<<<<<<<<<<<<
 *             "reply_tos": hop.hop_reply_tos,
 *             "reply_ipid": hop.hop_reply_ipid,
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint8_t(__pyx_v_hop->hop_reply_ttl); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_reply_ttl, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":172
 *             "rtt": rtt,
 *             "reply_ttl": hop.hop_reply_ttl,
 *             "reply_tos": hop.hop_reply_tos,             # <<<<<<<<<<<<<<
 *             "reply_ipid": hop.hop_reply_ipid,
 *             "reply_size": hop.hop_reply_size,
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint8_t(__pyx_v_hop->hop_reply_tos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_reply_tos, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":173
 *             "reply_ttl": hop.hop_reply_ttl,
 *             "reply_tos": hop.hop_reply_tos,
 *             "reply_ipid": hop.hop_reply_ipid,             # <<<<<<<<<<<<<<
 *             "reply_size": hop.hop_reply_size,
 *             "icmp_type": hop.hop_icmp_type,
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_v_hop->hop_reply_ipid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_reply_ipid, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":174
 *             "reply_tos": hop.hop_reply_tos,
 *             "reply_ipid": hop.hop_reply_ipid,
 *             "reply_size": hop.hop_reply_size,             # <<<<<<<<<<<<<<
 *             "icmp_type": hop.hop_icmp_type,
 *             "icmp_code": hop.hop_icmp_code,
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_v_hop->hop_reply_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_reply_size, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":175
 *             "reply_ipid": hop.hop_reply_ipid,
 *             "reply_size": hop.hop_reply_size,
 *             "icmp_type": hop.hop_icmp_type,             # <<<<<<<<<<<<<<
 *             "icmp_code": hop.hop_icmp_code,
 *             "icmp_q_ttl": hop.hop_icmp_q_ttl,
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint8_t(__pyx_v_hop->hop_icmp_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_icmp_type, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":176
 *             "reply_size": hop.hop_reply_size,
 *             "icmp_type": hop.hop_icmp_type,
 *             "icmp_code": hop.hop_icmp_code,             # <<<<<<<<<<<<<<
 *             "icmp_q_ttl": hop.hop_icmp_q_ttl,
 *             "icmp_q_ipl": hop.hop_icmp_q_ipl,
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint8_t(__pyx_v_hop->hop_icmp_code); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_icmp_code, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":177
 *             "icmp_type": hop.hop_icmp_type,
 *             "icmp_code": hop.hop_icmp_code,
 *             "icmp_q_ttl": hop.hop_icmp_q_ttl,             # <<<<<<<<<<<<<<
 *             "icmp_q_ipl": hop.hop_icmp_q_ipl,
 *             "icmp_q_tos": hop.hop_icmp_q_tos,
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint8_t(__pyx_v_hop->hop_icmp_q_ttl); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_icmp_q_ttl, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":178
 *             "icmp_code": hop.hop_icmp_code,
 *             "icmp_q_ttl": hop.hop_icmp_q_ttl,
 *             "icmp_q_ipl": hop.hop_icmp_q_ipl,             # <<<<<<<<<<<<<<
 *             "icmp_q_tos": hop.hop_icmp_q_tos,
 *             # "icmpext": [
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint16_t(__pyx_v_hop->hop_icmp_q_ipl); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_icmp_q_ipl, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":179
 *             "icmp_q_ttl": hop.hop_icmp_q_ttl,
 *             "icmp_q_ipl": hop.hop_icmp_q_ipl,
 *             "icmp_q_tos": hop.hop_icmp_q_tos,             # <<<<<<<<<<<<<<
 *             # "icmpext": [
 *             #     {"ie_cn": 1,"ie_ct": 1,"ie_dl": 8,"mpls_labels":[
 */
  __pyx_t_2 = __Pyx_PyInt_From_uint8_t(__pyx_v_hop->hop_icmp_q_tos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_icmp_q_tos, __pyx_t_2) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cloudtrace/read/convert.pyx":162
 * 
 * cdef class ConvertTrace(Convert):
 *     cdef dict hopdict(self, hop_t *hop):             # <<<<<<<<<<<<<<
 *         rtt = round(hop.hop_rtt.tv_sec * 1000 + (hop.hop_rtt.tv_usec / 1000), 3)
 *         return {
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.hopdict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_rtt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pyx":188
 *         }
 * 
 *     cdef dict tracedict(self, Trace trace):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str method, stop_reason
 */

static PyObject *__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_tracedict(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_7combine_Trace *__pyx_v_trace) {
  PyObject *__pyx_v_method = 0;
  PyObject *__pyx_v_stop_reason = 0;
  PyObject *__pyx_v_hops = 0;
  __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_sh;
  int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint16_t __pyx_t_2;
  uint16_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("tracedict", 0);

  /* "cloudtrace/read/convert.pyx":191
 *         cdef:
 *             str method, stop_reason
 *             list hops = []             # <<<<<<<<<<<<<<
 *             hop_t *sh
 *             int i
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_hops = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":194
 *             hop_t *sh
 *             int i
 *         for i in range(trace.hop_count):             # <<<<<<<<<<<<<<
 *             sh = trace.hops[i]
 *             if sh is NULL:
 */
  __pyx_t_2 = __pyx_v_trace->hop_count;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "cloudtrace/read/convert.pyx":195
 *             int i
 *         for i in range(trace.hop_count):
 *             sh = trace.hops[i]             # <<<<<<<<<<<<<<
 *             if sh is NULL:
 *                 continue
 */
    __pyx_v_sh = (__pyx_v_trace->hops[__pyx_v_i]);

    /* "cloudtrace/read/convert.pyx":196
 *         for i in range(trace.hop_count):
 *             sh = trace.hops[i]
 *             if sh is NULL:             # <<<<<<<<<<<<<<
 *                 continue
 *             hops.append(self.hopdict(sh))
 */
    __pyx_t_5 = ((__pyx_v_sh == NULL) != 0);
    if (__pyx_t_5) {

      /* "cloudtrace/read/convert.pyx":197
 *             sh = trace.hops[i]
 *             if sh is NULL:
 *                 continue             # <<<<<<<<<<<<<<
 *             hops.append(self.hopdict(sh))
 *         if trace.type == TRACE_TYPE_ICMP_ECHO_PARIS:
 */
      goto __pyx_L3_continue;

      /* "cloudtrace/read/convert.pyx":196
 *         for i in range(trace.hop_count):
 *             sh = trace.hops[i]
 *             if sh is NULL:             # <<<<<<<<<<<<<<
 *                 continue
 *             hops.append(self.hopdict(sh))
 */
    }

    /* "cloudtrace/read/convert.pyx":198
 *             if sh is NULL:
 *                 continue
 *             hops.append(self.hopdict(sh))             # <<<<<<<<<<<<<<
 *         if trace.type == TRACE_TYPE_ICMP_ECHO_PARIS:
 *             method = 'icmp-echo-paris'
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10cloudtrace_4read_7convert_ConvertTrace *)__pyx_v_self->__pyx_base.__pyx_vtab)->hopdict(__pyx_v_self, __pyx_v_sh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_hops, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_L3_continue:;
  }

  /* "cloudtrace/read/convert.pyx":199
 *                 continue
 *             hops.append(self.hopdict(sh))
 *         if trace.type == TRACE_TYPE_ICMP_ECHO_PARIS:             # <<<<<<<<<<<<<<
 *             method = 'icmp-echo-paris'
 *         else:
 */
  __pyx_t_5 = ((__pyx_v_trace->type == TRACE_TYPE_ICMP_ECHO_PARIS) != 0);
  if (__pyx_t_5) {

    /* "cloudtrace/read/convert.pyx":200
 *             hops.append(self.hopdict(sh))
 *         if trace.type == TRACE_TYPE_ICMP_ECHO_PARIS:
 *             method = 'icmp-echo-paris'             # <<<<<<<<<<<<<<
 *         else:
 *             method = 'udp-paris'
 */
    __Pyx_INCREF(__pyx_kp_u_icmp_echo_paris);
    __pyx_v_method = __pyx_kp_u_icmp_echo_paris;

    /* "cloudtrace/read/convert.pyx":199
 *                 continue
 *             hops.append(self.hopdict(sh))
 *         if trace.type == TRACE_TYPE_ICMP_ECHO_PARIS:             # <<<<<<<<<<<<<<
 *             method = 'icmp-echo-paris'
 *         else:
 */
    goto __pyx_L6;
  }

  /* "cloudtrace/read/convert.pyx":202
 *             method = 'icmp-echo-paris'
 *         else:
 *             method = 'udp-paris'             # <<<<<<<<<<<<<<
 *         if trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *             stop_reason = 'GAPLIMIT'
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_kp_u_udp_paris);
    __pyx_v_method = __pyx_kp_u_udp_paris;
  }
  __pyx_L6:;

  /* "cloudtrace/read/convert.pyx":203
 *         else:
 *             method = 'udp-paris'
 *         if trace.stop_reason == TRACE_STOP_GAPLIMIT:             # <<<<<<<<<<<<<<
 *             stop_reason = 'GAPLIMIT'
 *         elif trace.stop_reason == TRACE_STOP_COMPLETED:
 */
  __pyx_t_5 = ((__pyx_v_trace->stop_reason == TRACE_STOP_GAPLIMIT) != 0);
  if (__pyx_t_5) {

    /* "cloudtrace/read/convert.pyx":204
 *             method = 'udp-paris'
 *         if trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *             stop_reason = 'GAPLIMIT'             # <<<<<<<<<<<<<<
 *         elif trace.stop_reason == TRACE_STOP_COMPLETED:
 *             stop_reason = 'COMPLETED'
 */
    __Pyx_INCREF(__pyx_n_u_GAPLIMIT);
    __pyx_v_stop_reason = __pyx_n_u_GAPLIMIT;

    /* "cloudtrace/read/convert.pyx":203
 *         else:
 *             method = 'udp-paris'
 *         if trace.stop_reason == TRACE_STOP_GAPLIMIT:             # <<<<<<<<<<<<<<
 *             stop_reason = 'GAPLIMIT'
 *         elif trace.stop_reason == TRACE_STOP_COMPLETED:
 */
    goto __pyx_L7;
  }

  /* "cloudtrace/read/convert.pyx":205
 *         if trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *             stop_reason = 'GAPLIMIT'
 *         elif trace.stop_reason == TRACE_STOP_COMPLETED:             # <<<<<<<<<<<<<<
 *             stop_reason = 'COMPLETED'
 *             # if addr_tostr(trace.dst) != hops[-1]['addr']:
 */
  __pyx_t_5 = ((__pyx_v_trace->stop_reason == TRACE_STOP_COMPLETED) != 0);
  if (__pyx_t_5) {

    /* "cloudtrace/read/convert.pyx":206
 *             stop_reason = 'GAPLIMIT'
 *         elif trace.stop_reason == TRACE_STOP_COMPLETED:
 *             stop_reason = 'COMPLETED'             # <<<<<<<<<<<<<<
 *             # if addr_tostr(trace.dst) != hops[-1]['addr']:
 *             #     print('traceroute to {} ({} hops)'.format(addr_tostr(trace.dst), trace.hop_count))
 */
    __Pyx_INCREF(__pyx_n_u_COMPLETED);
    __pyx_v_stop_reason = __pyx_n_u_COMPLETED;

    /* "cloudtrace/read/convert.pyx":205
 *         if trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *             stop_reason = 'GAPLIMIT'
 *         elif trace.stop_reason == TRACE_STOP_COMPLETED:             # <<<<<<<<<<<<<<
 *             stop_reason = 'COMPLETED'
 *             # if addr_tostr(trace.dst) != hops[-1]['addr']:
 */
    goto __pyx_L7;
  }

  /* "cloudtrace/read/convert.pyx":217
 *             #             print('{:02d}: {:15} {} '.format(i, addr_tostr(sh.hop_addr), sh.hop_icmp_type))
 *             #     exit(1)
 *         elif trace.stop_reason == TRACE_STOP_UNREACH:             # <<<<<<<<<<<<<<
 *             stop_reason = 'UNREACH'
 *         elif trace.stop_reason == TRACE_STOP_LOOP:
 */
  __pyx_t_5 = ((__pyx_v_trace->stop_reason == TRACE_STOP_UNREACH) != 0);
  if (__pyx_t_5) {

    /* "cloudtrace/read/convert.pyx":218
 *             #     exit(1)
 *         elif trace.stop_reason == TRACE_STOP_UNREACH:
 *             stop_reason = 'UNREACH'             # <<<<<<<<<<<<<<
 *         elif trace.stop_reason == TRACE_STOP_LOOP:
 *             stop_reason = 'LOOP'
 */
    __Pyx_INCREF(__pyx_n_u_UNREACH);
    __pyx_v_stop_reason = __pyx_n_u_UNREACH;

    /* "cloudtrace/read/convert.pyx":217
 *             #             print('{:02d}: {:15} {} '.format(i, addr_tostr(sh.hop_addr), sh.hop_icmp_type))
 *             #     exit(1)
 *         elif trace.stop_reason == TRACE_STOP_UNREACH:             # <<<<<<<<<<<<<<
 *             stop_reason = 'UNREACH'
 *         elif trace.stop_reason == TRACE_STOP_LOOP:
 */
    goto __pyx_L7;
  }

  /* "cloudtrace/read/convert.pyx":219
 *         elif trace.stop_reason == TRACE_STOP_UNREACH:
 *             stop_reason = 'UNREACH'
 *         elif trace.stop_reason == TRACE_STOP_LOOP:             # <<<<<<<<<<<<<<
 *             stop_reason = 'LOOP'
 *         else:
 */
  __pyx_t_5 = ((__pyx_v_trace->stop_reason == TRACE_STOP_LOOP) != 0);
  if (__pyx_t_5) {

    /* "cloudtrace/read/convert.pyx":220
 *             stop_reason = 'UNREACH'
 *         elif trace.stop_reason == TRACE_STOP_LOOP:
 *             stop_reason = 'LOOP'             # <<<<<<<<<<<<<<
 *         else:
 *             stop_reason = 'HOPLIMIT'
 */
    __Pyx_INCREF(__pyx_n_u_LOOP);
    __pyx_v_stop_reason = __pyx_n_u_LOOP;

    /* "cloudtrace/read/convert.pyx":219
 *         elif trace.stop_reason == TRACE_STOP_UNREACH:
 *             stop_reason = 'UNREACH'
 *         elif trace.stop_reason == TRACE_STOP_LOOP:             # <<<<<<<<<<<<<<
 *             stop_reason = 'LOOP'
 *         else:
 */
    goto __pyx_L7;
  }

  /* "cloudtrace/read/convert.pyx":222
 *             stop_reason = 'LOOP'
 *         else:
 *             stop_reason = 'HOPLIMIT'             # <<<<<<<<<<<<<<
 *         return {
 *             "type": "trace",
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_n_u_HOPLIMIT);
    __pyx_v_stop_reason = __pyx_n_u_HOPLIMIT;
  }
  __pyx_L7:;

  /* "cloudtrace/read/convert.pyx":223
 *         else:
 *             stop_reason = 'HOPLIMIT'
 *         return {             # <<<<<<<<<<<<<<
 *             "type": "trace",
 *             "version": "0.1",
 */
  __Pyx_XDECREF(__pyx_r);

  /* "cloudtrace/read/convert.pyx":224
 *             stop_reason = 'HOPLIMIT'
 *         return {
 *             "type": "trace",             # <<<<<<<<<<<<<<
 *             "version": "0.1",
 *             "userid": 0,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(20); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_type, __pyx_n_u_trace) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_version, __pyx_kp_u_0_1) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_userid, __pyx_int_0) < 0) __PYX_ERR(0, 224, __pyx_L1_error)

  /* "cloudtrace/read/convert.pyx":227
 *             "version": "0.1",
 *             "userid": 0,
 *             "method": method,             # <<<<<<<<<<<<<<
 *             "src": addr_tostr(trace.src),
 *             "dst": addr_tostr(trace.dst),
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_method, __pyx_v_method) < 0) __PYX_ERR(0, 224, __pyx_L1_error)

  /* "cloudtrace/read/convert.pyx":228
 *             "userid": 0,
 *             "method": method,
 *             "src": addr_tostr(trace.src),             # <<<<<<<<<<<<<<
 *             "dst": addr_tostr(trace.dst),
 *             "icmp_sum": 47652,
 */
  __pyx_t_7 = __pyx_f_10cloudtrace_4read_7convert_addr_tostr(__pyx_v_trace->src); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_src, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":229
 *             "method": method,
 *             "src": addr_tostr(trace.src),
 *             "dst": addr_tostr(trace.dst),             # <<<<<<<<<<<<<<
 *             "icmp_sum": 47652,
 *             "stop_reason": stop_reason,
 */
  __pyx_t_7 = __pyx_f_10cloudtrace_4read_7convert_addr_tostr(__pyx_v_trace->dst); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_dst, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_icmp_sum, __pyx_int_47652) < 0) __PYX_ERR(0, 224, __pyx_L1_error)

  /* "cloudtrace/read/convert.pyx":231
 *             "dst": addr_tostr(trace.dst),
 *             "icmp_sum": 47652,
 *             "stop_reason": stop_reason,             # <<<<<<<<<<<<<<
 *             "stop_data": trace.stop_data,
 *             "start": {"sec": trace.start.tv_sec, "usec": trace.start.tv_usec, "ftime": "fake"},
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_stop_reason, __pyx_v_stop_reason) < 0) __PYX_ERR(0, 224, __pyx_L1_error)

  /* "cloudtrace/read/convert.pyx":232
 *             "icmp_sum": 47652,
 *             "stop_reason": stop_reason,
 *             "stop_data": trace.stop_data,             # <<<<<<<<<<<<<<
 *             "start": {"sec": trace.start.tv_sec, "usec": trace.start.tv_usec, "ftime": "fake"},
 *             "hop_count": trace.hop_count,
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint8_t(__pyx_v_trace->stop_data); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_stop_data, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":233
 *             "stop_reason": stop_reason,
 *             "stop_data": trace.stop_data,
 *             "start": {"sec": trace.start.tv_sec, "usec": trace.start.tv_usec, "ftime": "fake"},             # <<<<<<<<<<<<<<
 *             "hop_count": trace.hop_count,
 *             "attempts": trace.attempts,
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_time_t(__pyx_v_trace->start.tv_sec); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_sec, __pyx_t_8) < 0) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyInt_From_suseconds_t(__pyx_v_trace->start.tv_usec); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_usec, __pyx_t_8) < 0) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_u_ftime, __pyx_n_u_fake) < 0) __PYX_ERR(0, 233, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_start, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":234
 *             "stop_data": trace.stop_data,
 *             "start": {"sec": trace.start.tv_sec, "usec": trace.start.tv_usec, "ftime": "fake"},
 *             "hop_count": trace.hop_count,             # <<<<<<<<<<<<<<
 *             "attempts": trace.attempts,
 *             "hoplimit": trace.hoplimit,
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint16_t(__pyx_v_trace->hop_count); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_hop_count, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":235
 *             "start": {"sec": trace.start.tv_sec, "usec": trace.start.tv_usec, "ftime": "fake"},
 *             "hop_count": trace.hop_count,
 *             "attempts": trace.attempts,             # <<<<<<<<<<<<<<
 *             "hoplimit": trace.hoplimit,
 *             "firsthop": trace.firsthop,
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint8_t(__pyx_v_trace->attempts); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_attempts, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":236
 *             "hop_count": trace.hop_count,
 *             "attempts": trace.attempts,
 *             "hoplimit": trace.hoplimit,             # <<<<<<<<<<<<<<
 *             "firsthop": trace.firsthop,
 *             "wait": trace.wait,
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint8_t(__pyx_v_trace->hoplimit); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_hoplimit, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":237
 *             "attempts": trace.attempts,
 *             "hoplimit": trace.hoplimit,
 *             "firsthop": trace.firsthop,             # <<<<<<<<<<<<<<
 *             "wait": trace.wait,
 *             "wait_probe": trace.wait_probe,
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint8_t(__pyx_v_trace->firsthop); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_firsthop, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":238
 *             "hoplimit": trace.hoplimit,
 *             "firsthop": trace.firsthop,
 *             "wait": trace.wait,             # <<<<<<<<<<<<<<
 *             "wait_probe": trace.wait_probe,
 *             "tos": trace.tos,
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint8_t(__pyx_v_trace->wait); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_wait, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":239
 *             "firsthop": trace.firsthop,
 *             "wait": trace.wait,
 *             "wait_probe": trace.wait_probe,             # <<<<<<<<<<<<<<
 *             "tos": trace.tos,
 *             "probe_size": trace.probe_size,
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint8_t(__pyx_v_trace->wait_probe); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_wait_probe, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":240
 *             "wait": trace.wait,
 *             "wait_probe": trace.wait_probe,
 *             "tos": trace.tos,             # <<<<<<<<<<<<<<
 *             "probe_size": trace.probe_size,
 *             "probe_count": trace.probec,
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint8_t(__pyx_v_trace->tos); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_tos, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":241
 *             "wait_probe": trace.wait_probe,
 *             "tos": trace.tos,
 *             "probe_size": trace.probe_size,             # <<<<<<<<<<<<<<
 *             "probe_count": trace.probec,
 *             "hops": hops
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint16_t(__pyx_v_trace->probe_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_probe_size, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":242
 *             "tos": trace.tos,
 *             "probe_size": trace.probe_size,
 *             "probe_count": trace.probec,             # <<<<<<<<<<<<<<
 *             "hops": hops
 *         }
 */
  __pyx_t_7 = __Pyx_PyInt_From_uint16_t(__pyx_v_trace->probec); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_probe_count, __pyx_t_7) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "cloudtrace/read/convert.pyx":243
 *             "probe_size": trace.probe_size,
 *             "probe_count": trace.probec,
 *             "hops": hops             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_hops, __pyx_v_hops) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cloudtrace/read/convert.pyx":188
 *         }
 * 
 *     cdef dict tracedict(self, Trace trace):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str method, stop_reason
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.tracedict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_method);
  __Pyx_XDECREF(__pyx_v_stop_reason);
  __Pyx_XDECREF(__pyx_v_hops);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pyx":246
 *         }
 * 
 *     cdef void write_dst(self, bytes dst_b) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             timeval tmptime
 */

static void __pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_write_dst(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, PyObject *__pyx_v_dst_b) {
  struct __pyx_obj_10cloudtrace_4read_7combine_Trace *__pyx_v_trace = 0;
  int __pyx_v_loop_hop;
  PyObject *__pyx_v_output = 0;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_dst", 0);

  /* "cloudtrace/read/convert.pyx":252
 *             int loop_hop
 *             dict output
 *         if dst_b not in self.probes:             # <<<<<<<<<<<<<<
 *             # print('{}: {} not in probes. odd.'.format(self.infile, dst_b))
 *             return
 */
  if (unlikely(__pyx_v_self->__pyx_base.probes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 252, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_dst_b, __pyx_v_self->__pyx_base.probes, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 252, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "cloudtrace/read/convert.pyx":254
 *         if dst_b not in self.probes:
 *             # print('{}: {} not in probes. odd.'.format(self.infile, dst_b))
 *             return             # <<<<<<<<<<<<<<
 *         del self.probes[dst_b]
 *         if dst_b in self.replies:
 */
    goto __pyx_L0;

    /* "cloudtrace/read/convert.pyx":252
 *             int loop_hop
 *             dict output
 *         if dst_b not in self.probes:             # <<<<<<<<<<<<<<
 *             # print('{}: {} not in probes. odd.'.format(self.infile, dst_b))
 *             return
 */
  }

  /* "cloudtrace/read/convert.pyx":255
 *             # print('{}: {} not in probes. odd.'.format(self.infile, dst_b))
 *             return
 *         del self.probes[dst_b]             # <<<<<<<<<<<<<<
 *         if dst_b in self.replies:
 *             trace = self.replies.pop(dst_b)
 */
  if (unlikely(__pyx_v_self->__pyx_base.probes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 255, __pyx_L1_error)
  }
  if (unlikely(PyDict_DelItem(__pyx_v_self->__pyx_base.probes, __pyx_v_dst_b) < 0)) __PYX_ERR(0, 255, __pyx_L1_error)

  /* "cloudtrace/read/convert.pyx":256
 *             return
 *         del self.probes[dst_b]
 *         if dst_b in self.replies:             # <<<<<<<<<<<<<<
 *             trace = self.replies.pop(dst_b)
 *             loop_hop = loop_check(trace)
 */
  if (unlikely(__pyx_v_self->__pyx_base.replies == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 256, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_v_dst_b, __pyx_v_self->__pyx_base.replies, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 256, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/read/convert.pyx":257
 *         del self.probes[dst_b]
 *         if dst_b in self.replies:
 *             trace = self.replies.pop(dst_b)             # <<<<<<<<<<<<<<
 *             loop_hop = loop_check(trace)
 *             if loop_hop > 0:
 */
    if (unlikely(__pyx_v_self->__pyx_base.replies == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
      __PYX_ERR(0, 257, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_Pop(__pyx_v_self->__pyx_base.replies, __pyx_v_dst_b, ((PyObject *)NULL)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 257, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_10cloudtrace_4read_7combine_Trace))))) __PYX_ERR(0, 257, __pyx_L1_error)
    __pyx_v_trace = ((struct __pyx_obj_10cloudtrace_4read_7combine_Trace *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "cloudtrace/read/convert.pyx":258
 *         if dst_b in self.replies:
 *             trace = self.replies.pop(dst_b)
 *             loop_hop = loop_check(trace)             # <<<<<<<<<<<<<<
 *             if loop_hop > 0:
 *                 trace.hop_count = loop_hop
 */
    __pyx_v_loop_hop = __pyx_f_10cloudtrace_4read_7convert_loop_check(__pyx_v_trace);

    /* "cloudtrace/read/convert.pyx":259
 *             trace = self.replies.pop(dst_b)
 *             loop_hop = loop_check(trace)
 *             if loop_hop > 0:             # <<<<<<<<<<<<<<
 *                 trace.hop_count = loop_hop
 *                 trace.stop_reason = TRACE_STOP_LOOP
 */
    __pyx_t_1 = ((__pyx_v_loop_hop > 0) != 0);
    if (__pyx_t_1) {

      /* "cloudtrace/read/convert.pyx":260
 *             loop_hop = loop_check(trace)
 *             if loop_hop > 0:
 *                 trace.hop_count = loop_hop             # <<<<<<<<<<<<<<
 *                 trace.stop_reason = TRACE_STOP_LOOP
 *             elif trace.hop_count == 32 and trace.stop_reason == TRACE_STOP_GAPLIMIT:
 */
      __pyx_v_trace->hop_count = __pyx_v_loop_hop;

      /* "cloudtrace/read/convert.pyx":261
 *             if loop_hop > 0:
 *                 trace.hop_count = loop_hop
 *                 trace.stop_reason = TRACE_STOP_LOOP             # <<<<<<<<<<<<<<
 *             elif trace.hop_count == 32 and trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                 trace.stop_reason = TRACE_STOP_HOPLIMIT
 */
      __pyx_v_trace->stop_reason = TRACE_STOP_LOOP;

      /* "cloudtrace/read/convert.pyx":259
 *             trace = self.replies.pop(dst_b)
 *             loop_hop = loop_check(trace)
 *             if loop_hop > 0:             # <<<<<<<<<<<<<<
 *                 trace.hop_count = loop_hop
 *                 trace.stop_reason = TRACE_STOP_LOOP
 */
      goto __pyx_L5;
    }

    /* "cloudtrace/read/convert.pyx":262
 *                 trace.hop_count = loop_hop
 *                 trace.stop_reason = TRACE_STOP_LOOP
 *             elif trace.hop_count == 32 and trace.stop_reason == TRACE_STOP_GAPLIMIT:             # <<<<<<<<<<<<<<
 *                 trace.stop_reason = TRACE_STOP_HOPLIMIT
 *             output = self.tracedict(trace)
 */
    __pyx_t_2 = ((__pyx_v_trace->hop_count == 32) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_trace->stop_reason == TRACE_STOP_GAPLIMIT) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_1) {

      /* "cloudtrace/read/convert.pyx":263
 *                 trace.stop_reason = TRACE_STOP_LOOP
 *             elif trace.hop_count == 32 and trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                 trace.stop_reason = TRACE_STOP_HOPLIMIT             # <<<<<<<<<<<<<<
 *             output = self.tracedict(trace)
 *             # print(json.dumps(output))
 */
      __pyx_v_trace->stop_reason = TRACE_STOP_HOPLIMIT;

      /* "cloudtrace/read/convert.pyx":262
 *                 trace.hop_count = loop_hop
 *                 trace.stop_reason = TRACE_STOP_LOOP
 *             elif trace.hop_count == 32 and trace.stop_reason == TRACE_STOP_GAPLIMIT:             # <<<<<<<<<<<<<<
 *                 trace.stop_reason = TRACE_STOP_HOPLIMIT
 *             output = self.tracedict(trace)
 */
    }
    __pyx_L5:;

    /* "cloudtrace/read/convert.pyx":264
 *             elif trace.hop_count == 32 and trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                 trace.stop_reason = TRACE_STOP_HOPLIMIT
 *             output = self.tracedict(trace)             # <<<<<<<<<<<<<<
 *             # print(json.dumps(output))
 *             self.writefile.write('{}\n'.format(json.dumps(output)))
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_10cloudtrace_4read_7convert_ConvertTrace *)__pyx_v_self->__pyx_base.__pyx_vtab)->tracedict(__pyx_v_self, __pyx_v_trace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_output = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "cloudtrace/read/convert.pyx":266
 *             output = self.tracedict(trace)
 *             # print(json.dumps(output))
 *             self.writefile.write('{}\n'.format(json.dumps(output)))             # <<<<<<<<<<<<<<
 *             trace.hop_count = 32
 *             self.wrote += 1
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.writefile, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__4, __pyx_n_s_format); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_json); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_dumps); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_8, __pyx_v_output) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_output);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "cloudtrace/read/convert.pyx":267
 *             # print(json.dumps(output))
 *             self.writefile.write('{}\n'.format(json.dumps(output)))
 *             trace.hop_count = 32             # <<<<<<<<<<<<<<
 *             self.wrote += 1
 *             del trace
 */
    __pyx_v_trace->hop_count = 32;

    /* "cloudtrace/read/convert.pyx":268
 *             self.writefile.write('{}\n'.format(json.dumps(output)))
 *             trace.hop_count = 32
 *             self.wrote += 1             # <<<<<<<<<<<<<<
 *             del trace
 * 
 */
    __pyx_v_self->__pyx_base.wrote = (__pyx_v_self->__pyx_base.wrote + 1);

    /* "cloudtrace/read/convert.pyx":269
 *             trace.hop_count = 32
 *             self.wrote += 1
 *             del trace             # <<<<<<<<<<<<<<
 * 
 *     cdef Trace initialize_trace(self, Packet probe):
 */
    __Pyx_DECREF(((PyObject *)__pyx_v_trace));
    __pyx_v_trace = NULL;

    /* "cloudtrace/read/convert.pyx":256
 *             return
 *         del self.probes[dst_b]
 *         if dst_b in self.replies:             # <<<<<<<<<<<<<<
 *             trace = self.replies.pop(dst_b)
 *             loop_hop = loop_check(trace)
 */
  }

  /* "cloudtrace/read/convert.pyx":246
 *         }
 * 
 *     cdef void write_dst(self, bytes dst_b) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             timeval tmptime
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.write_dst", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_trace);
  __Pyx_XDECREF(__pyx_v_output);
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/read/convert.pyx":271
 *             del trace
 * 
 *     cdef Trace initialize_trace(self, Packet probe):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Trace trace = Trace()
 */

static struct __pyx_obj_10cloudtrace_4read_7combine_Trace *__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_initialize_trace(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *__pyx_v_probe) {
  struct __pyx_obj_10cloudtrace_4read_7combine_Trace *__pyx_v_trace = 0;
  struct __pyx_obj_10cloudtrace_4read_7combine_Trace *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  uint16_t __pyx_t_3;
  uint8_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("initialize_trace", 0);

  /* "cloudtrace/read/convert.pyx":273
 *     cdef Trace initialize_trace(self, Packet probe):
 *         cdef:
 *             Trace trace = Trace()             # <<<<<<<<<<<<<<
 *         memcpy(trace.src, probe.iphdr.ip_src, sizeof(trace.src))
 *         memcpy(trace.dst, probe.iphdr.ip_dst, sizeof(trace.dst))
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_10cloudtrace_4read_7combine_Trace)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_trace = ((struct __pyx_obj_10cloudtrace_4read_7combine_Trace *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":274
 *         cdef:
 *             Trace trace = Trace()
 *         memcpy(trace.src, probe.iphdr.ip_src, sizeof(trace.src))             # <<<<<<<<<<<<<<
 *         memcpy(trace.dst, probe.iphdr.ip_dst, sizeof(trace.dst))
 *         memcpy(&trace.start, &probe.time, sizeof(timeval))
 */
  if (unlikely(!__pyx_v_trace)) { __Pyx_RaiseUnboundLocalError("trace"); __PYX_ERR(0, 274, __pyx_L1_error) }
  (void)(memcpy(__pyx_v_trace->src, __pyx_v_probe->iphdr->ip_src, (sizeof(__pyx_v_trace->src))));

  /* "cloudtrace/read/convert.pyx":275
 *             Trace trace = Trace()
 *         memcpy(trace.src, probe.iphdr.ip_src, sizeof(trace.src))
 *         memcpy(trace.dst, probe.iphdr.ip_dst, sizeof(trace.dst))             # <<<<<<<<<<<<<<
 *         memcpy(&trace.start, &probe.time, sizeof(timeval))
 *         trace.probec = 32
 */
  if (unlikely(!__pyx_v_trace)) { __Pyx_RaiseUnboundLocalError("trace"); __PYX_ERR(0, 275, __pyx_L1_error) }
  (void)(memcpy(__pyx_v_trace->dst, __pyx_v_probe->iphdr->ip_dst, (sizeof(__pyx_v_trace->dst))));

  /* "cloudtrace/read/convert.pyx":276
 *         memcpy(trace.src, probe.iphdr.ip_src, sizeof(trace.src))
 *         memcpy(trace.dst, probe.iphdr.ip_dst, sizeof(trace.dst))
 *         memcpy(&trace.start, &probe.time, sizeof(timeval))             # <<<<<<<<<<<<<<
 *         trace.probec = 32
 *         if probe.ptype == ICMP_PROBE:
 */
  (void)(memcpy((&__pyx_v_trace->start), (&__pyx_v_probe->time), (sizeof(struct timeval))));

  /* "cloudtrace/read/convert.pyx":277
 *         memcpy(trace.dst, probe.iphdr.ip_dst, sizeof(trace.dst))
 *         memcpy(&trace.start, &probe.time, sizeof(timeval))
 *         trace.probec = 32             # <<<<<<<<<<<<<<
 *         if probe.ptype == ICMP_PROBE:
 *             trace.type = TRACE_TYPE_ICMP_ECHO_PARIS
 */
  __pyx_v_trace->probec = 32;

  /* "cloudtrace/read/convert.pyx":278
 *         memcpy(&trace.start, &probe.time, sizeof(timeval))
 *         trace.probec = 32
 *         if probe.ptype == ICMP_PROBE:             # <<<<<<<<<<<<<<
 *             trace.type = TRACE_TYPE_ICMP_ECHO_PARIS
 *         elif probe.ptype == UDP_PROBE:
 */
  __pyx_t_2 = ((__pyx_v_probe->ptype == __pyx_v_10cloudtrace_4read_6packet_ICMP_PROBE) != 0);
  if (__pyx_t_2) {

    /* "cloudtrace/read/convert.pyx":279
 *         trace.probec = 32
 *         if probe.ptype == ICMP_PROBE:
 *             trace.type = TRACE_TYPE_ICMP_ECHO_PARIS             # <<<<<<<<<<<<<<
 *         elif probe.ptype == UDP_PROBE:
 *             trace.type = TRACE_TYPE_UDP_PARIS
 */
    __pyx_v_trace->type = TRACE_TYPE_ICMP_ECHO_PARIS;

    /* "cloudtrace/read/convert.pyx":278
 *         memcpy(&trace.start, &probe.time, sizeof(timeval))
 *         trace.probec = 32
 *         if probe.ptype == ICMP_PROBE:             # <<<<<<<<<<<<<<
 *             trace.type = TRACE_TYPE_ICMP_ECHO_PARIS
 *         elif probe.ptype == UDP_PROBE:
 */
    goto __pyx_L3;
  }

  /* "cloudtrace/read/convert.pyx":280
 *         if probe.ptype == ICMP_PROBE:
 *             trace.type = TRACE_TYPE_ICMP_ECHO_PARIS
 *         elif probe.ptype == UDP_PROBE:             # <<<<<<<<<<<<<<
 *             trace.type = TRACE_TYPE_UDP_PARIS
 *             trace.sport = (<UDPProbe>probe).udphdr.source
 */
  __pyx_t_2 = ((__pyx_v_probe->ptype == __pyx_v_10cloudtrace_4read_6packet_UDP_PROBE) != 0);
  if (__pyx_t_2) {

    /* "cloudtrace/read/convert.pyx":281
 *             trace.type = TRACE_TYPE_ICMP_ECHO_PARIS
 *         elif probe.ptype == UDP_PROBE:
 *             trace.type = TRACE_TYPE_UDP_PARIS             # <<<<<<<<<<<<<<
 *             trace.sport = (<UDPProbe>probe).udphdr.source
 *             trace.dport = (<UDPProbe>probe).udphdr.dest
 */
    __pyx_v_trace->type = TRACE_TYPE_UDP_PARIS;

    /* "cloudtrace/read/convert.pyx":282
 *         elif probe.ptype == UDP_PROBE:
 *             trace.type = TRACE_TYPE_UDP_PARIS
 *             trace.sport = (<UDPProbe>probe).udphdr.source             # <<<<<<<<<<<<<<
 *             trace.dport = (<UDPProbe>probe).udphdr.dest
 *         trace.flags = 0
 */
    __pyx_t_3 = ((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_probe)->udphdr->source;
    __pyx_v_trace->sport = __pyx_t_3;

    /* "cloudtrace/read/convert.pyx":283
 *             trace.type = TRACE_TYPE_UDP_PARIS
 *             trace.sport = (<UDPProbe>probe).udphdr.source
 *             trace.dport = (<UDPProbe>probe).udphdr.dest             # <<<<<<<<<<<<<<
 *         trace.flags = 0
 *         trace.attempts = 1
 */
    __pyx_t_3 = ((struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe *)__pyx_v_probe)->udphdr->dest;
    __pyx_v_trace->dport = __pyx_t_3;

    /* "cloudtrace/read/convert.pyx":280
 *         if probe.ptype == ICMP_PROBE:
 *             trace.type = TRACE_TYPE_ICMP_ECHO_PARIS
 *         elif probe.ptype == UDP_PROBE:             # <<<<<<<<<<<<<<
 *             trace.type = TRACE_TYPE_UDP_PARIS
 *             trace.sport = (<UDPProbe>probe).udphdr.source
 */
  }
  __pyx_L3:;

  /* "cloudtrace/read/convert.pyx":284
 *             trace.sport = (<UDPProbe>probe).udphdr.source
 *             trace.dport = (<UDPProbe>probe).udphdr.dest
 *         trace.flags = 0             # <<<<<<<<<<<<<<
 *         trace.attempts = 1
 *         trace.hoplimit = 32
 */
  __pyx_v_trace->flags = 0;

  /* "cloudtrace/read/convert.pyx":285
 *             trace.dport = (<UDPProbe>probe).udphdr.dest
 *         trace.flags = 0
 *         trace.attempts = 1             # <<<<<<<<<<<<<<
 *         trace.hoplimit = 32
 *         trace.gaplimit = 0
 */
  __pyx_v_trace->attempts = 1;

  /* "cloudtrace/read/convert.pyx":286
 *         trace.flags = 0
 *         trace.attempts = 1
 *         trace.hoplimit = 32             # <<<<<<<<<<<<<<
 *         trace.gaplimit = 0
 *         trace.gapaction = 0
 */
  __pyx_v_trace->hoplimit = 32;

  /* "cloudtrace/read/convert.pyx":287
 *         trace.attempts = 1
 *         trace.hoplimit = 32
 *         trace.gaplimit = 0             # <<<<<<<<<<<<<<
 *         trace.gapaction = 0
 *         trace.firsthop = 1
 */
  __pyx_v_trace->gaplimit = 0;

  /* "cloudtrace/read/convert.pyx":288
 *         trace.hoplimit = 32
 *         trace.gaplimit = 0
 *         trace.gapaction = 0             # <<<<<<<<<<<<<<
 *         trace.firsthop = 1
 *         trace.tos = probe.iphdr.ip_tos
 */
  __pyx_v_trace->gapaction = 0;

  /* "cloudtrace/read/convert.pyx":289
 *         trace.gaplimit = 0
 *         trace.gapaction = 0
 *         trace.firsthop = 1             # <<<<<<<<<<<<<<
 *         trace.tos = probe.iphdr.ip_tos
 *         trace.wait = 0
 */
  __pyx_v_trace->firsthop = 1;

  /* "cloudtrace/read/convert.pyx":290
 *         trace.gapaction = 0
 *         trace.firsthop = 1
 *         trace.tos = probe.iphdr.ip_tos             # <<<<<<<<<<<<<<
 *         trace.wait = 0
 *         trace.wait_probe = 0
 */
  __pyx_t_4 = __pyx_v_probe->iphdr->ip_tos;
  __pyx_v_trace->tos = __pyx_t_4;

  /* "cloudtrace/read/convert.pyx":291
 *         trace.firsthop = 1
 *         trace.tos = probe.iphdr.ip_tos
 *         trace.wait = 0             # <<<<<<<<<<<<<<
 *         trace.wait_probe = 0
 *         trace.loops = 1
 */
  __pyx_v_trace->wait = 0;

  /* "cloudtrace/read/convert.pyx":292
 *         trace.tos = probe.iphdr.ip_tos
 *         trace.wait = 0
 *         trace.wait_probe = 0             # <<<<<<<<<<<<<<
 *         trace.loops = 1
 *         trace.loopaction = 1
 */
  __pyx_v_trace->wait_probe = 0;

  /* "cloudtrace/read/convert.pyx":293
 *         trace.wait = 0
 *         trace.wait_probe = 0
 *         trace.loops = 1             # <<<<<<<<<<<<<<
 *         trace.loopaction = 1
 *         trace.confidence = 0
 */
  __pyx_v_trace->loops = 1;

  /* "cloudtrace/read/convert.pyx":294
 *         trace.wait_probe = 0
 *         trace.loops = 1
 *         trace.loopaction = 1             # <<<<<<<<<<<<<<
 *         trace.confidence = 0
 *         trace.probe_size = probe.iphdr.ip_len
 */
  __pyx_v_trace->loopaction = 1;

  /* "cloudtrace/read/convert.pyx":295
 *         trace.loops = 1
 *         trace.loopaction = 1
 *         trace.confidence = 0             # <<<<<<<<<<<<<<
 *         trace.probe_size = probe.iphdr.ip_len
 *         trace.hop_count = 0
 */
  __pyx_v_trace->confidence = 0;

  /* "cloudtrace/read/convert.pyx":296
 *         trace.loopaction = 1
 *         trace.confidence = 0
 *         trace.probe_size = probe.iphdr.ip_len             # <<<<<<<<<<<<<<
 *         trace.hop_count = 0
 *         trace.stop_reason = TRACE_STOP_GAPLIMIT
 */
  __pyx_t_3 = __pyx_v_probe->iphdr->ip_len;
  __pyx_v_trace->probe_size = __pyx_t_3;

  /* "cloudtrace/read/convert.pyx":297
 *         trace.confidence = 0
 *         trace.probe_size = probe.iphdr.ip_len
 *         trace.hop_count = 0             # <<<<<<<<<<<<<<
 *         trace.stop_reason = TRACE_STOP_GAPLIMIT
 *         return trace
 */
  __pyx_v_trace->hop_count = 0;

  /* "cloudtrace/read/convert.pyx":298
 *         trace.probe_size = probe.iphdr.ip_len
 *         trace.hop_count = 0
 *         trace.stop_reason = TRACE_STOP_GAPLIMIT             # <<<<<<<<<<<<<<
 *         return trace
 * 
 */
  __pyx_v_trace->stop_reason = TRACE_STOP_GAPLIMIT;

  /* "cloudtrace/read/convert.pyx":299
 *         trace.hop_count = 0
 *         trace.stop_reason = TRACE_STOP_GAPLIMIT
 *         return trace             # <<<<<<<<<<<<<<
 * 
 *     cpdef void convert_trace(self, bint fix, uint8_t verbose) except *:
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_trace));
  __pyx_r = __pyx_v_trace;
  goto __pyx_L0;

  /* "cloudtrace/read/convert.pyx":271
 *             del trace
 * 
 *     cdef Trace initialize_trace(self, Packet probe):             # <<<<<<<<<<<<<<
 *         cdef:
 *             Trace trace = Trace()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.initialize_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_trace);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/read/convert.pyx":301
 *         return trace
 * 
 *     cpdef void convert_trace(self, bint fix, uint8_t verbose) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             Pcap pcap = Pcap(self.infile, fix=fix, verbose=verbose)
 */

static PyObject *__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_1convert_trace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static void __pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_convert_trace(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, int __pyx_v_fix, uint8_t __pyx_v_verbose, int __pyx_skip_dispatch) {
  struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *__pyx_v_pcap = 0;
  unsigned short __pyx_v_ttl;
  CYTHON_UNUSED unsigned short __pyx_v_ttlindex;
  __pyx_t_10cloudtrace_4read_7convert_probe_id __pyx_v_pid;
  CYTHON_UNUSED PyObject *__pyx_v_pid_b = 0;
  PyObject *__pyx_v_dst_b = 0;
  __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_v_hop;
  struct __pyx_obj_10cloudtrace_4read_7combine_Trace *__pyx_v_trace = 0;
  uint8_t __pyx_v_icmptype;
  struct __pyx_obj_10cloudtrace_4read_6packet_Reply *__pyx_v_reply = 0;
  uint32_t __pyx_v_loop;
  PyObject *__pyx_v_sidef = NULL;
  PyObject *__pyx_v_header = NULL;
  PyObject *__pyx_v_footer = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  unsigned char __pyx_t_15;
  __pyx_t_10cloudtrace_4read_7combine_hop_t *__pyx_t_16;
  uint8_t __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_trace", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_convert_trace); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_1convert_trace)) {
        __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_fix); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = __Pyx_PyInt_From_uint8_t(__pyx_v_verbose); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 301, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
        __pyx_t_7 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_7 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
          PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else
        #endif
        {
          __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 301, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
          __pyx_t_3 = 0;
          __pyx_t_4 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/read/convert.pyx":303
 *     cpdef void convert_trace(self, bint fix, uint8_t verbose) except *:
 *         cdef:
 *             Pcap pcap = Pcap(self.infile, fix=fix, verbose=verbose)             # <<<<<<<<<<<<<<
 *             unsigned short ttl, ttlindex
 *             probe_id pid
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base.infile);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base.infile);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base.infile);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_v_fix); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_fix, __pyx_t_5) < 0) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyInt_From_uint8_t(__pyx_v_verbose); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_verbose, __pyx_t_5) < 0) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10cloudtrace_4read_4pcap_Pcap), __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_pcap = ((struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cloudtrace/read/convert.pyx":311
 *             uint8_t icmptype
 *             Reply reply
 *             uint32_t loop = 0             # <<<<<<<<<<<<<<
 *         with open(self.hostname) as sidef:
 *             header = sidef.readline()
 */
  __pyx_v_loop = 0;

  /* "cloudtrace/read/convert.pyx":312
 *             Reply reply
 *             uint32_t loop = 0
 *         with open(self.hostname) as sidef:             # <<<<<<<<<<<<<<
 *             header = sidef.readline()
 *             # print(header)
 */
  /*with:*/ {
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_open, __pyx_v_self->__pyx_base.hostname); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_exit); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_enter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L3_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        /*try:*/ {
          __pyx_v_sidef = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "cloudtrace/read/convert.pyx":313
 *             uint32_t loop = 0
 *         with open(self.hostname) as sidef:
 *             header = sidef.readline()             # <<<<<<<<<<<<<<
 *             # print(header)
 *             # header = {"type":"cycle-start", "list_name":"default", "id":0, "hostname":self.hostname, "start_time":0}
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_sidef, __pyx_n_s_readline); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 313, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_v_header = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "cloudtrace/read/convert.pyx":317
 *             # header = {"type":"cycle-start", "list_name":"default", "id":0, "hostname":self.hostname, "start_time":0}
 *             # self.writefile.write(json.dumps(header) + '\n')
 *             self.writefile.write(header)             # <<<<<<<<<<<<<<
 *             while True:
 *                 loop += 1
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.writefile, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 317, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_2 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_2, __pyx_v_header) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_header);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "cloudtrace/read/convert.pyx":318
 *             # self.writefile.write(json.dumps(header) + '\n')
 *             self.writefile.write(header)
 *             while True:             # <<<<<<<<<<<<<<
 *                 loop += 1
 *                 if verbose > 0 and loop % 1000000 == 0:
 */
          while (1) {

            /* "cloudtrace/read/convert.pyx":319
 *             self.writefile.write(header)
 *             while True:
 *                 loop += 1             # <<<<<<<<<<<<<<
 *                 if verbose > 0 and loop % 1000000 == 0:
 *                     print('Loop {:,d} Wrote {:,d}'.format(loop, self.wrote))
 */
            __pyx_v_loop = (__pyx_v_loop + 1);

            /* "cloudtrace/read/convert.pyx":320
 *             while True:
 *                 loop += 1
 *                 if verbose > 0 and loop % 1000000 == 0:             # <<<<<<<<<<<<<<
 *                     print('Loop {:,d} Wrote {:,d}'.format(loop, self.wrote))
 *                 reply = self.next(pcap)
 */
            __pyx_t_14 = ((__pyx_v_verbose > 0) != 0);
            if (__pyx_t_14) {
            } else {
              __pyx_t_13 = __pyx_t_14;
              goto __pyx_L16_bool_binop_done;
            }
            __pyx_t_14 = ((__Pyx_mod_long(__pyx_v_loop, 0xF4240) == 0) != 0);
            __pyx_t_13 = __pyx_t_14;
            __pyx_L16_bool_binop_done:;
            if (__pyx_t_13) {

              /* "cloudtrace/read/convert.pyx":321
 *                 loop += 1
 *                 if verbose > 0 and loop % 1000000 == 0:
 *                     print('Loop {:,d} Wrote {:,d}'.format(loop, self.wrote))             # <<<<<<<<<<<<<<
 *                 reply = self.next(pcap)
 *                 if not self.more:
 */
              __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Loop_d_Wrote_d, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 321, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_loop); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 321, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_8 = __Pyx_PyInt_From_long(__pyx_v_self->__pyx_base.wrote); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 321, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_8);
              __pyx_t_4 = NULL;
              __pyx_t_7 = 0;
              if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
                __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
                if (likely(__pyx_t_4)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
                  __Pyx_INCREF(__pyx_t_4);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_5, function);
                  __pyx_t_7 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_t_8};
                __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L7_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
                PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_t_2, __pyx_t_8};
                __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L7_error)
                __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
              } else
              #endif
              {
                __pyx_t_3 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 321, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_3);
                if (__pyx_t_4) {
                  __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_2);
                PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_t_2);
                __Pyx_GIVEREF(__pyx_t_8);
                PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_t_8);
                __pyx_t_2 = 0;
                __pyx_t_8 = 0;
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L7_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              }
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 321, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

              /* "cloudtrace/read/convert.pyx":320
 *             while True:
 *                 loop += 1
 *                 if verbose > 0 and loop % 1000000 == 0:             # <<<<<<<<<<<<<<
 *                     print('Loop {:,d} Wrote {:,d}'.format(loop, self.wrote))
 *                 reply = self.next(pcap)
 */
            }

            /* "cloudtrace/read/convert.pyx":322
 *                 if verbose > 0 and loop % 1000000 == 0:
 *                     print('Loop {:,d} Wrote {:,d}'.format(loop, self.wrote))
 *                 reply = self.next(pcap)             # <<<<<<<<<<<<<<
 *                 if not self.more:
 *                     # print('breaking')
 */
            __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_10cloudtrace_4read_7convert_ConvertTrace *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.next(((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)__pyx_v_self), __pyx_v_pcap)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 322, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_XDECREF_SET(__pyx_v_reply, ((struct __pyx_obj_10cloudtrace_4read_6packet_Reply *)__pyx_t_5));
            __pyx_t_5 = 0;

            /* "cloudtrace/read/convert.pyx":323
 *                     print('Loop {:,d} Wrote {:,d}'.format(loop, self.wrote))
 *                 reply = self.next(pcap)
 *                 if not self.more:             # <<<<<<<<<<<<<<
 *                     # print('breaking')
 *                     break
 */
            __pyx_t_13 = ((!(__pyx_v_self->__pyx_base.more != 0)) != 0);
            if (__pyx_t_13) {

              /* "cloudtrace/read/convert.pyx":325
 *                 if not self.more:
 *                     # print('breaking')
 *                     break             # <<<<<<<<<<<<<<
 *                 if reply is None:
 *                     continue
 */
              goto __pyx_L14_break;

              /* "cloudtrace/read/convert.pyx":323
 *                     print('Loop {:,d} Wrote {:,d}'.format(loop, self.wrote))
 *                 reply = self.next(pcap)
 *                 if not self.more:             # <<<<<<<<<<<<<<
 *                     # print('breaking')
 *                     break
 */
            }

            /* "cloudtrace/read/convert.pyx":326
 *                     # print('breaking')
 *                     break
 *                 if reply is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 # print('packet')
 */
            __pyx_t_13 = (((PyObject *)__pyx_v_reply) == Py_None);
            __pyx_t_14 = (__pyx_t_13 != 0);
            if (__pyx_t_14) {

              /* "cloudtrace/read/convert.pyx":327
 *                     break
 *                 if reply is None:
 *                     continue             # <<<<<<<<<<<<<<
 *                 # print('packet')
 *                 memcpy(pid.dst, reply.get_dst_c(), 4)
 */
              goto __pyx_L13_continue;

              /* "cloudtrace/read/convert.pyx":326
 *                     # print('breaking')
 *                     break
 *                 if reply is None:             # <<<<<<<<<<<<<<
 *                     continue
 *                 # print('packet')
 */
            }

            /* "cloudtrace/read/convert.pyx":329
 *                     continue
 *                 # print('packet')
 *                 memcpy(pid.dst, reply.get_dst_c(), 4)             # <<<<<<<<<<<<<<
 *                 pid.ttl = reply.probe_ttl()
 *                 pid.pid = reply.get_pid()
 */
            (void)(memcpy(__pyx_v_pid.dst, ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply *)__pyx_v_reply->__pyx_base.__pyx_vtab)->__pyx_base.get_dst_c(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_reply)), 4));

            /* "cloudtrace/read/convert.pyx":330
 *                 # print('packet')
 *                 memcpy(pid.dst, reply.get_dst_c(), 4)
 *                 pid.ttl = reply.probe_ttl()             # <<<<<<<<<<<<<<
 *                 pid.pid = reply.get_pid()
 *                 pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]
 */
            __pyx_v_pid.ttl = ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply *)__pyx_v_reply->__pyx_base.__pyx_vtab)->__pyx_base.probe_ttl(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_reply), 0);

            /* "cloudtrace/read/convert.pyx":331
 *                 memcpy(pid.dst, reply.get_dst_c(), 4)
 *                 pid.ttl = reply.probe_ttl()
 *                 pid.pid = reply.get_pid()             # <<<<<<<<<<<<<<
 *                 pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]
 *                 dst_b = pid.dst[:4]
 */
            __pyx_v_pid.pid = ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply *)__pyx_v_reply->__pyx_base.__pyx_vtab)->__pyx_base.get_pid(((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_v_reply), 0);

            /* "cloudtrace/read/convert.pyx":332
 *                 pid.ttl = reply.probe_ttl()
 *                 pid.pid = reply.get_pid()
 *                 pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]             # <<<<<<<<<<<<<<
 *                 dst_b = pid.dst[:4]
 *                 ttl = pid.ttl
 */
            __pyx_t_5 = __Pyx_PyBytes_FromStringAndSize(((const char*)((unsigned char *)(&__pyx_v_pid))) + 0, (sizeof(__pyx_t_10cloudtrace_4read_7convert_probe_id)) - 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 332, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_1 = __pyx_t_5;
            __Pyx_INCREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF_SET(__pyx_v_pid_b, ((PyObject*)__pyx_t_1));
            __pyx_t_1 = 0;

            /* "cloudtrace/read/convert.pyx":333
 *                 pid.pid = reply.get_pid()
 *                 pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]
 *                 dst_b = pid.dst[:4]             # <<<<<<<<<<<<<<
 *                 ttl = pid.ttl
 *                 ttlindex = ttl - 1
 */
            __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_pid.dst) + 0, 4 - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 333, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_XDECREF_SET(__pyx_v_dst_b, ((PyObject*)__pyx_t_1));
            __pyx_t_1 = 0;

            /* "cloudtrace/read/convert.pyx":334
 *                 pid_b = <bytes>(<unsigned char *>&pid)[:sizeof(probe_id)]
 *                 dst_b = pid.dst[:4]
 *                 ttl = pid.ttl             # <<<<<<<<<<<<<<
 *                 ttlindex = ttl - 1
 *                 if dst_b in self.replies:
 */
            __pyx_t_15 = __pyx_v_pid.ttl;
            __pyx_v_ttl = __pyx_t_15;

            /* "cloudtrace/read/convert.pyx":335
 *                 dst_b = pid.dst[:4]
 *                 ttl = pid.ttl
 *                 ttlindex = ttl - 1             # <<<<<<<<<<<<<<
 *                 if dst_b in self.replies:
 *                     trace = self.replies[dst_b]
 */
            __pyx_v_ttlindex = (__pyx_v_ttl - 1);

            /* "cloudtrace/read/convert.pyx":336
 *                 ttl = pid.ttl
 *                 ttlindex = ttl - 1
 *                 if dst_b in self.replies:             # <<<<<<<<<<<<<<
 *                     trace = self.replies[dst_b]
 *                 else:
 */
            if (unlikely(__pyx_v_self->__pyx_base.replies == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
              __PYX_ERR(0, 336, __pyx_L7_error)
            }
            __pyx_t_14 = (__Pyx_PyDict_ContainsTF(__pyx_v_dst_b, __pyx_v_self->__pyx_base.replies, Py_EQ)); if (unlikely(__pyx_t_14 < 0)) __PYX_ERR(0, 336, __pyx_L7_error)
            __pyx_t_13 = (__pyx_t_14 != 0);
            if (__pyx_t_13) {

              /* "cloudtrace/read/convert.pyx":337
 *                 ttlindex = ttl - 1
 *                 if dst_b in self.replies:
 *                     trace = self.replies[dst_b]             # <<<<<<<<<<<<<<
 *                 else:
 *                     trace = self.initialize_trace(reply.orig_probe)
 */
              if (unlikely(__pyx_v_self->__pyx_base.replies == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 337, __pyx_L7_error)
              }
              __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->__pyx_base.replies, __pyx_v_dst_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_1);
              if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10cloudtrace_4read_7combine_Trace))))) __PYX_ERR(0, 337, __pyx_L7_error)
              __Pyx_XDECREF_SET(__pyx_v_trace, ((struct __pyx_obj_10cloudtrace_4read_7combine_Trace *)__pyx_t_1));
              __pyx_t_1 = 0;

              /* "cloudtrace/read/convert.pyx":336
 *                 ttl = pid.ttl
 *                 ttlindex = ttl - 1
 *                 if dst_b in self.replies:             # <<<<<<<<<<<<<<
 *                     trace = self.replies[dst_b]
 *                 else:
 */
              goto __pyx_L20;
            }

            /* "cloudtrace/read/convert.pyx":339
 *                     trace = self.replies[dst_b]
 *                 else:
 *                     trace = self.initialize_trace(reply.orig_probe)             # <<<<<<<<<<<<<<
 *                     self.replies[dst_b] = trace
 *                 hop = trace.new_hop(pid.ttl)
 */
            /*else*/ {
              __pyx_t_1 = ((PyObject *)__pyx_v_reply->orig_probe);
              __Pyx_INCREF(__pyx_t_1);
              __pyx_t_5 = ((PyObject *)((struct __pyx_vtabstruct_10cloudtrace_4read_7convert_ConvertTrace *)__pyx_v_self->__pyx_base.__pyx_vtab)->initialize_trace(__pyx_v_self, ((struct __pyx_obj_10cloudtrace_4read_6packet_Packet *)__pyx_t_1))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 339, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_XDECREF_SET(__pyx_v_trace, ((struct __pyx_obj_10cloudtrace_4read_7combine_Trace *)__pyx_t_5));
              __pyx_t_5 = 0;

              /* "cloudtrace/read/convert.pyx":340
 *                 else:
 *                     trace = self.initialize_trace(reply.orig_probe)
 *                     self.replies[dst_b] = trace             # <<<<<<<<<<<<<<
 *                 hop = trace.new_hop(pid.ttl)
 *                 reply.create_hop(hop)
 */
              if (unlikely(__pyx_v_self->__pyx_base.replies == Py_None)) {
                PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
                __PYX_ERR(0, 340, __pyx_L7_error)
              }
              if (unlikely(PyDict_SetItem(__pyx_v_self->__pyx_base.replies, __pyx_v_dst_b, ((PyObject *)__pyx_v_trace)) < 0)) __PYX_ERR(0, 340, __pyx_L7_error)
            }
            __pyx_L20:;

            /* "cloudtrace/read/convert.pyx":341
 *                     trace = self.initialize_trace(reply.orig_probe)
 *                     self.replies[dst_b] = trace
 *                 hop = trace.new_hop(pid.ttl)             # <<<<<<<<<<<<<<
 *                 reply.create_hop(hop)
 *                 icmptype = reply.icmphdr.type
 */
            __pyx_t_16 = ((struct __pyx_vtabstruct_10cloudtrace_4read_7combine_Trace *)__pyx_v_trace->__pyx_vtab)->new_hop(__pyx_v_trace, __pyx_v_pid.ttl); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 341, __pyx_L7_error)
            __pyx_v_hop = __pyx_t_16;

            /* "cloudtrace/read/convert.pyx":342
 *                     self.replies[dst_b] = trace
 *                 hop = trace.new_hop(pid.ttl)
 *                 reply.create_hop(hop)             # <<<<<<<<<<<<<<
 *                 icmptype = reply.icmphdr.type
 *                 if icmptype == 0 or icmptype == 3:
 */
            ((struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply *)__pyx_v_reply->__pyx_base.__pyx_vtab)->create_hop(__pyx_v_reply, __pyx_v_hop);

            /* "cloudtrace/read/convert.pyx":343
 *                 hop = trace.new_hop(pid.ttl)
 *                 reply.create_hop(hop)
 *                 icmptype = reply.icmphdr.type             # <<<<<<<<<<<<<<
 *                 if icmptype == 0 or icmptype == 3:
 *                     if ttl < trace.hop_count or trace.stop_reason == TRACE_STOP_GAPLIMIT:
 */
            __pyx_t_17 = __pyx_v_reply->icmphdr->type;
            __pyx_v_icmptype = __pyx_t_17;

            /* "cloudtrace/read/convert.pyx":344
 *                 reply.create_hop(hop)
 *                 icmptype = reply.icmphdr.type
 *                 if icmptype == 0 or icmptype == 3:             # <<<<<<<<<<<<<<
 *                     if ttl < trace.hop_count or trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                         trace.hop_count = ttl
 */
            switch (__pyx_v_icmptype) {
              case 0:
              case 3:

              /* "cloudtrace/read/convert.pyx":345
 *                 icmptype = reply.icmphdr.type
 *                 if icmptype == 0 or icmptype == 3:
 *                     if ttl < trace.hop_count or trace.stop_reason == TRACE_STOP_GAPLIMIT:             # <<<<<<<<<<<<<<
 *                         trace.hop_count = ttl
 *                         if icmptype == 0:
 */
              __pyx_t_14 = ((__pyx_v_ttl < __pyx_v_trace->hop_count) != 0);
              if (!__pyx_t_14) {
              } else {
                __pyx_t_13 = __pyx_t_14;
                goto __pyx_L22_bool_binop_done;
              }
              __pyx_t_14 = ((__pyx_v_trace->stop_reason == TRACE_STOP_GAPLIMIT) != 0);
              __pyx_t_13 = __pyx_t_14;
              __pyx_L22_bool_binop_done:;
              if (__pyx_t_13) {

                /* "cloudtrace/read/convert.pyx":346
 *                 if icmptype == 0 or icmptype == 3:
 *                     if ttl < trace.hop_count or trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                         trace.hop_count = ttl             # <<<<<<<<<<<<<<
 *                         if icmptype == 0:
 *                             trace.stop_reason = TRACE_STOP_COMPLETED
 */
                __pyx_v_trace->hop_count = __pyx_v_ttl;

                /* "cloudtrace/read/convert.pyx":347
 *                     if ttl < trace.hop_count or trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                         trace.hop_count = ttl
 *                         if icmptype == 0:             # <<<<<<<<<<<<<<
 *                             trace.stop_reason = TRACE_STOP_COMPLETED
 *                         else:
 */
                __pyx_t_13 = ((__pyx_v_icmptype == 0) != 0);
                if (__pyx_t_13) {

                  /* "cloudtrace/read/convert.pyx":348
 *                         trace.hop_count = ttl
 *                         if icmptype == 0:
 *                             trace.stop_reason = TRACE_STOP_COMPLETED             # <<<<<<<<<<<<<<
 *                         else:
 *                             trace.stop_reason = TRACE_STOP_UNREACH
 */
                  __pyx_v_trace->stop_reason = TRACE_STOP_COMPLETED;

                  /* "cloudtrace/read/convert.pyx":347
 *                     if ttl < trace.hop_count or trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                         trace.hop_count = ttl
 *                         if icmptype == 0:             # <<<<<<<<<<<<<<
 *                             trace.stop_reason = TRACE_STOP_COMPLETED
 *                         else:
 */
                  goto __pyx_L24;
                }

                /* "cloudtrace/read/convert.pyx":350
 *                             trace.stop_reason = TRACE_STOP_COMPLETED
 *                         else:
 *                             trace.stop_reason = TRACE_STOP_UNREACH             # <<<<<<<<<<<<<<
 *                 elif icmptype == 11:
 *                     if ttl > trace.hop_count and trace.stop_reason == TRACE_STOP_GAPLIMIT:
 */
                /*else*/ {
                  __pyx_v_trace->stop_reason = TRACE_STOP_UNREACH;
                }
                __pyx_L24:;

                /* "cloudtrace/read/convert.pyx":345
 *                 icmptype = reply.icmphdr.type
 *                 if icmptype == 0 or icmptype == 3:
 *                     if ttl < trace.hop_count or trace.stop_reason == TRACE_STOP_GAPLIMIT:             # <<<<<<<<<<<<<<
 *                         trace.hop_count = ttl
 *                         if icmptype == 0:
 */
              }

              /* "cloudtrace/read/convert.pyx":344
 *                 reply.create_hop(hop)
 *                 icmptype = reply.icmphdr.type
 *                 if icmptype == 0 or icmptype == 3:             # <<<<<<<<<<<<<<
 *                     if ttl < trace.hop_count or trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                         trace.hop_count = ttl
 */
              break;
              case 11:

              /* "cloudtrace/read/convert.pyx":352
 *                             trace.stop_reason = TRACE_STOP_UNREACH
 *                 elif icmptype == 11:
 *                     if ttl > trace.hop_count and trace.stop_reason == TRACE_STOP_GAPLIMIT:             # <<<<<<<<<<<<<<
 *                         trace.hop_count = ttl
 *             pcap.read_all()
 */
              __pyx_t_14 = ((__pyx_v_ttl > __pyx_v_trace->hop_count) != 0);
              if (__pyx_t_14) {
              } else {
                __pyx_t_13 = __pyx_t_14;
                goto __pyx_L26_bool_binop_done;
              }
              __pyx_t_14 = ((__pyx_v_trace->stop_reason == TRACE_STOP_GAPLIMIT) != 0);
              __pyx_t_13 = __pyx_t_14;
              __pyx_L26_bool_binop_done:;
              if (__pyx_t_13) {

                /* "cloudtrace/read/convert.pyx":353
 *                 elif icmptype == 11:
 *                     if ttl > trace.hop_count and trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                         trace.hop_count = ttl             # <<<<<<<<<<<<<<
 *             pcap.read_all()
 *             footer = sidef.readline()
 */
                __pyx_v_trace->hop_count = __pyx_v_ttl;

                /* "cloudtrace/read/convert.pyx":352
 *                             trace.stop_reason = TRACE_STOP_UNREACH
 *                 elif icmptype == 11:
 *                     if ttl > trace.hop_count and trace.stop_reason == TRACE_STOP_GAPLIMIT:             # <<<<<<<<<<<<<<
 *                         trace.hop_count = ttl
 *             pcap.read_all()
 */
              }

              /* "cloudtrace/read/convert.pyx":351
 *                         else:
 *                             trace.stop_reason = TRACE_STOP_UNREACH
 *                 elif icmptype == 11:             # <<<<<<<<<<<<<<
 *                     if ttl > trace.hop_count and trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                         trace.hop_count = ttl
 */
              break;
              default: break;
            }
            __pyx_L13_continue:;
          }
          __pyx_L14_break:;

          /* "cloudtrace/read/convert.pyx":354
 *                     if ttl > trace.hop_count and trace.stop_reason == TRACE_STOP_GAPLIMIT:
 *                         trace.hop_count = ttl
 *             pcap.read_all()             # <<<<<<<<<<<<<<
 *             footer = sidef.readline()
 *             # footer = {"type":"cycle-stop", "list_name":"default", "id":0, "hostname":self.hostname, "stop_time":0}
 */
          ((struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap *)__pyx_v_pcap->__pyx_vtab)->read_all(__pyx_v_pcap, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 354, __pyx_L7_error)

          /* "cloudtrace/read/convert.pyx":355
 *                         trace.hop_count = ttl
 *             pcap.read_all()
 *             footer = sidef.readline()             # <<<<<<<<<<<<<<
 *             # footer = {"type":"cycle-stop", "list_name":"default", "id":0, "hostname":self.hostname, "stop_time":0}
 *             # self.writefile.write(json.dumps(footer) + '\n')
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sidef, __pyx_n_s_readline); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 355, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_footer = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "cloudtrace/read/convert.pyx":358
 *             # footer = {"type":"cycle-stop", "list_name":"default", "id":0, "hostname":self.hostname, "stop_time":0}
 *             # self.writefile.write(json.dumps(footer) + '\n')
 *             self.writefile.write(footer)             # <<<<<<<<<<<<<<
 *         # print('closing')
 *         self.close()
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.writefile, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_footer) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_footer);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 358, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "cloudtrace/read/convert.pyx":312
 *             Reply reply
 *             uint32_t loop = 0
 *         with open(self.hostname) as sidef:             # <<<<<<<<<<<<<<
 *             header = sidef.readline()
 *             # print(header)
 */
        }
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L12_try_end;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.convert_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_3) < 0) __PYX_ERR(0, 312, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_8 = PyTuple_Pack(3, __pyx_t_5, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 312, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 312, __pyx_L9_except_error)
          __Pyx_GOTREF(__pyx_t_18);
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_18);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (__pyx_t_13 < 0) __PYX_ERR(0, 312, __pyx_L9_except_error)
          __pyx_t_14 = ((!(__pyx_t_13 != 0)) != 0);
          if (__pyx_t_14) {
            __Pyx_GIVEREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_XGIVEREF(__pyx_t_3);
            __Pyx_ErrRestoreWithState(__pyx_t_5, __pyx_t_1, __pyx_t_3);
            __pyx_t_5 = 0; __pyx_t_1 = 0; __pyx_t_3 = 0; 
            __PYX_ERR(0, 312, __pyx_L9_except_error)
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L8_exception_handled;
        }
        __pyx_L9_except_error:;
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        goto __pyx_L1_error;
        __pyx_L8_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        __pyx_L12_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_9) {
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_tuple__5, NULL);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 312, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        goto __pyx_L6;
      }
      __pyx_L6:;
    }
    goto __pyx_L31;
    __pyx_L3_error:;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    goto __pyx_L1_error;
    __pyx_L31:;
  }

  /* "cloudtrace/read/convert.pyx":360
 *             self.writefile.write(footer)
 *         # print('closing')
 *         self.close()             # <<<<<<<<<<<<<<
 *         # print('done')
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cloudtrace/read/convert.pyx":301
 *         return trace
 * 
 *     cpdef void convert_trace(self, bint fix, uint8_t verbose) except *:             # <<<<<<<<<<<<<<
 *         cdef:
 *             Pcap pcap = Pcap(self.infile, fix=fix, verbose=verbose)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.convert_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_pcap);
  __Pyx_XDECREF(__pyx_v_pid_b);
  __Pyx_XDECREF(__pyx_v_dst_b);
  __Pyx_XDECREF((PyObject *)__pyx_v_trace);
  __Pyx_XDECREF((PyObject *)__pyx_v_reply);
  __Pyx_XDECREF(__pyx_v_sidef);
  __Pyx_XDECREF(__pyx_v_header);
  __Pyx_XDECREF(__pyx_v_footer);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_1convert_trace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10cloudtrace_4read_7convert_12ConvertTrace_convert_trace[] = "ConvertTrace.convert_trace(self, bool fix, uint8_t verbose) -> void";
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_1convert_trace(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_fix;
  uint8_t __pyx_v_verbose;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("convert_trace (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_fix,&__pyx_n_s_verbose,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_fix)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_verbose)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("convert_trace", 1, 2, 2, 1); __PYX_ERR(0, 301, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "convert_trace") < 0)) __PYX_ERR(0, 301, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_fix = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_fix == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L3_error)
    __pyx_v_verbose = __Pyx_PyInt_As_uint8_t(values[1]); if (unlikely((__pyx_v_verbose == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("convert_trace", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 301, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.convert_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_12ConvertTrace_convert_trace(((struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *)__pyx_v_self), __pyx_v_fix, __pyx_v_verbose);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_12ConvertTrace_convert_trace(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, int __pyx_v_fix, uint8_t __pyx_v_verbose) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_trace", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_convert_trace(__pyx_v_self, __pyx_v_fix, __pyx_v_verbose, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.convert_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_4read_7convert_12ConvertTrace_2__reduce_cython__[] = "ConvertTrace.__reduce_cython__(self)";
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_12ConvertTrace_2__reduce_cython__(((struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_12ConvertTrace_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_4read_7convert_12ConvertTrace_4__setstate_cython__[] = "ConvertTrace.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_4read_7convert_12ConvertTrace_4__setstate_cython__(((struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_4read_7convert_12ConvertTrace_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.read.convert.ConvertTrace.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert __pyx_vtable_10cloudtrace_4read_7convert_Convert;

static PyObject *__pyx_tp_new_10cloudtrace_4read_7convert_Convert(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_4read_7convert_Convert *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)o);
  p->__pyx_vtab = __pyx_vtabptr_10cloudtrace_4read_7convert_Convert;
  p->infile = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->outfile = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->hostname = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->times = ((struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList2 *)Py_None); Py_INCREF(Py_None);
  p->probes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->replies = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->writefile = Py_None; Py_INCREF(Py_None);
  p->pcap = ((struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_10cloudtrace_4read_7convert_7Convert_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10cloudtrace_4read_7convert_Convert(PyObject *o) {
  struct __pyx_obj_10cloudtrace_4read_7convert_Convert *p = (struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10cloudtrace_4read_7convert_7Convert_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->infile);
  Py_CLEAR(p->outfile);
  Py_CLEAR(p->hostname);
  Py_CLEAR(p->times);
  Py_CLEAR(p->probes);
  Py_CLEAR(p->replies);
  Py_CLEAR(p->writefile);
  Py_CLEAR(p->pcap);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_10cloudtrace_4read_7convert_Convert(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10cloudtrace_4read_7convert_Convert *p = (struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)o;
  if (p->times) {
    e = (*v)(((PyObject *)p->times), a); if (e) return e;
  }
  if (p->probes) {
    e = (*v)(p->probes, a); if (e) return e;
  }
  if (p->replies) {
    e = (*v)(p->replies, a); if (e) return e;
  }
  if (p->writefile) {
    e = (*v)(p->writefile, a); if (e) return e;
  }
  if (p->pcap) {
    e = (*v)(((PyObject *)p->pcap), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10cloudtrace_4read_7convert_Convert(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10cloudtrace_4read_7convert_Convert *p = (struct __pyx_obj_10cloudtrace_4read_7convert_Convert *)o;
  tmp = ((PyObject*)p->times);
  p->times = ((struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList2 *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->probes);
  p->probes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->replies);
  p->replies = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->writefile);
  p->writefile = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->pcap);
  p->pcap = ((struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10cloudtrace_4read_7convert_7Convert_total(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_7convert_7Convert_5total_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_7convert_7Convert_total(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_5total_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_7convert_7Convert_nprobes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_7convert_7Convert_7nprobes_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_7convert_7Convert_nprobes(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_7nprobes_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_7convert_7Convert_nreplies(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_7convert_7Convert_8nreplies_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_7convert_7Convert_nreplies(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_8nreplies_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_7convert_7Convert_skipped(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_7convert_7Convert_7skipped_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_7convert_7Convert_skipped(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_7skipped_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_7convert_7Convert_pidprobe(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidprobe_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_7convert_7Convert_pidprobe(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidprobe_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_7convert_7Convert_pidreply(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidreply_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_7convert_7Convert_pidreply(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_8pidreply_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_7convert_7Convert_ttlhigh(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_7convert_7Convert_7ttlhigh_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_7convert_7Convert_ttlhigh(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_7ttlhigh_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_7convert_7Convert_wrote(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_7convert_7Convert_5wrote_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_7convert_7Convert_wrote(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_5wrote_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10cloudtrace_4read_7convert_7Convert_writefile(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10cloudtrace_4read_7convert_7Convert_9writefile_1__get__(o);
}

static int __pyx_setprop_10cloudtrace_4read_7convert_7Convert_writefile(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_9writefile_3__set__(o, v);
  }
  else {
    return __pyx_pw_10cloudtrace_4read_7convert_7Convert_9writefile_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_10cloudtrace_4read_7convert_Convert[] = {
  {"close", (PyCFunction)__pyx_pw_10cloudtrace_4read_7convert_7Convert_7close, METH_NOARGS, __pyx_doc_10cloudtrace_4read_7convert_7Convert_6close},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_7convert_7Convert_9__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_4read_7convert_7Convert_8__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_7convert_7Convert_11__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_4read_7convert_7Convert_10__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10cloudtrace_4read_7convert_Convert[] = {
  {(char *)"total", __pyx_getprop_10cloudtrace_4read_7convert_7Convert_total, __pyx_setprop_10cloudtrace_4read_7convert_7Convert_total, (char *)"total: 'long'", 0},
  {(char *)"nprobes", __pyx_getprop_10cloudtrace_4read_7convert_7Convert_nprobes, __pyx_setprop_10cloudtrace_4read_7convert_7Convert_nprobes, (char *)"nprobes: 'long'", 0},
  {(char *)"nreplies", __pyx_getprop_10cloudtrace_4read_7convert_7Convert_nreplies, __pyx_setprop_10cloudtrace_4read_7convert_7Convert_nreplies, (char *)"nreplies: 'long'", 0},
  {(char *)"skipped", __pyx_getprop_10cloudtrace_4read_7convert_7Convert_skipped, __pyx_setprop_10cloudtrace_4read_7convert_7Convert_skipped, (char *)"skipped: 'long'", 0},
  {(char *)"pidprobe", __pyx_getprop_10cloudtrace_4read_7convert_7Convert_pidprobe, __pyx_setprop_10cloudtrace_4read_7convert_7Convert_pidprobe, (char *)"pidprobe: 'long'", 0},
  {(char *)"pidreply", __pyx_getprop_10cloudtrace_4read_7convert_7Convert_pidreply, __pyx_setprop_10cloudtrace_4read_7convert_7Convert_pidreply, (char *)"pidreply: 'long'", 0},
  {(char *)"ttlhigh", __pyx_getprop_10cloudtrace_4read_7convert_7Convert_ttlhigh, __pyx_setprop_10cloudtrace_4read_7convert_7Convert_ttlhigh, (char *)"ttlhigh: 'long'", 0},
  {(char *)"wrote", __pyx_getprop_10cloudtrace_4read_7convert_7Convert_wrote, __pyx_setprop_10cloudtrace_4read_7convert_7Convert_wrote, (char *)"wrote: 'long'", 0},
  {(char *)"writefile", __pyx_getprop_10cloudtrace_4read_7convert_7Convert_writefile, __pyx_setprop_10cloudtrace_4read_7convert_7Convert_writefile, (char *)"writefile: object", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_4read_7convert_Convert = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.read.convert.Convert", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_4read_7convert_Convert), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_4read_7convert_Convert, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10cloudtrace_4read_7convert_7Convert_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10cloudtrace_4read_7convert_Convert, /*tp_traverse*/
  __pyx_tp_clear_10cloudtrace_4read_7convert_Convert, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_4read_7convert_Convert, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10cloudtrace_4read_7convert_Convert, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_4read_7convert_Convert, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_10cloudtrace_4read_7convert_ConvertTrace __pyx_vtable_10cloudtrace_4read_7convert_ConvertTrace;

static PyObject *__pyx_tp_new_10cloudtrace_4read_7convert_ConvertTrace(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *p;
  PyObject *o = __pyx_tp_new_10cloudtrace_4read_7convert_Convert(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10cloudtrace_4read_7convert_Convert*)__pyx_vtabptr_10cloudtrace_4read_7convert_ConvertTrace;
  return o;
}

static PyMethodDef __pyx_methods_10cloudtrace_4read_7convert_ConvertTrace[] = {
  {"convert_trace", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_1convert_trace, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10cloudtrace_4read_7convert_12ConvertTrace_convert_trace},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_3__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_4read_7convert_12ConvertTrace_2__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_4read_7convert_12ConvertTrace_5__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_4read_7convert_12ConvertTrace_4__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_4read_7convert_ConvertTrace = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.read.convert.ConvertTrace", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_4read_7convert_Convert, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_10cloudtrace_4read_7convert_7Convert_5__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10cloudtrace_4read_7convert_Convert, /*tp_traverse*/
  __pyx_tp_clear_10cloudtrace_4read_7convert_Convert, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_4read_7convert_ConvertTrace, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_4read_7convert_ConvertTrace, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_convert(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_convert},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "convert",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_b_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 0, 0},
  {&__pyx_kp_u_0_1, __pyx_k_0_1, sizeof(__pyx_k_0_1), 0, 1, 0, 0},
  {&__pyx_n_u_COMPLETED, __pyx_k_COMPLETED, sizeof(__pyx_k_COMPLETED), 0, 1, 0, 1},
  {&__pyx_n_s_Convert, __pyx_k_Convert, sizeof(__pyx_k_Convert), 0, 0, 1, 1},
  {&__pyx_n_s_ConvertTrace, __pyx_k_ConvertTrace, sizeof(__pyx_k_ConvertTrace), 0, 0, 1, 1},
  {&__pyx_n_u_GAPLIMIT, __pyx_k_GAPLIMIT, sizeof(__pyx_k_GAPLIMIT), 0, 1, 0, 1},
  {&__pyx_n_u_HOPLIMIT, __pyx_k_HOPLIMIT, sizeof(__pyx_k_HOPLIMIT), 0, 1, 0, 1},
  {&__pyx_n_u_LOOP, __pyx_k_LOOP, sizeof(__pyx_k_LOOP), 0, 1, 0, 1},
  {&__pyx_kp_u_Loop_d_Wrote_d, __pyx_k_Loop_d_Wrote_d, sizeof(__pyx_k_Loop_d_Wrote_d), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_TRACE_STOP_COMPLETED, __pyx_k_TRACE_STOP_COMPLETED, sizeof(__pyx_k_TRACE_STOP_COMPLETED), 0, 0, 1, 1},
  {&__pyx_n_s_TRACE_STOP_GAPLIMIT, __pyx_k_TRACE_STOP_GAPLIMIT, sizeof(__pyx_k_TRACE_STOP_GAPLIMIT), 0, 0, 1, 1},
  {&__pyx_n_s_TRACE_STOP_HOPLIMIT, __pyx_k_TRACE_STOP_HOPLIMIT, sizeof(__pyx_k_TRACE_STOP_HOPLIMIT), 0, 0, 1, 1},
  {&__pyx_n_s_TRACE_STOP_LOOP, __pyx_k_TRACE_STOP_LOOP, sizeof(__pyx_k_TRACE_STOP_LOOP), 0, 0, 1, 1},
  {&__pyx_n_s_TRACE_STOP_UNREACH, __pyx_k_TRACE_STOP_UNREACH, sizeof(__pyx_k_TRACE_STOP_UNREACH), 0, 0, 1, 1},
  {&__pyx_n_s_TRACE_TYPE_ICMP_ECHO_PARIS, __pyx_k_TRACE_TYPE_ICMP_ECHO_PARIS, sizeof(__pyx_k_TRACE_TYPE_ICMP_ECHO_PARIS), 0, 0, 1, 1},
  {&__pyx_n_s_TRACE_TYPE_UDP_PARIS, __pyx_k_TRACE_TYPE_UDP_PARIS, sizeof(__pyx_k_TRACE_TYPE_UDP_PARIS), 0, 0, 1, 1},
  {&__pyx_kp_u_Total_d_Probes_d_Replies_d_PIDP, __pyx_k_Total_d_Probes_d_Replies_d_PIDP, sizeof(__pyx_k_Total_d_Probes_d_Replies_d_PIDP), 0, 1, 0, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_u_UNREACH, __pyx_k_UNREACH, sizeof(__pyx_k_UNREACH), 0, 1, 0, 1},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_n_u_addr, __pyx_k_addr, sizeof(__pyx_k_addr), 0, 1, 0, 1},
  {&__pyx_n_u_attempts, __pyx_k_attempts, sizeof(__pyx_k_attempts), 0, 1, 0, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cloudtrace_read_pcap, __pyx_k_cloudtrace_read_pcap, sizeof(__pyx_k_cloudtrace_read_pcap), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_convert_trace, __pyx_k_convert_trace, sizeof(__pyx_k_convert_trace), 0, 0, 1, 1},
  {&__pyx_n_s_defaultdict, __pyx_k_defaultdict, sizeof(__pyx_k_defaultdict), 0, 0, 1, 1},
  {&__pyx_n_u_dst, __pyx_k_dst, sizeof(__pyx_k_dst), 0, 1, 0, 1},
  {&__pyx_n_s_dumps, __pyx_k_dumps, sizeof(__pyx_k_dumps), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_u_fake, __pyx_k_fake, sizeof(__pyx_k_fake), 0, 1, 0, 1},
  {&__pyx_n_u_firsthop, __pyx_k_firsthop, sizeof(__pyx_k_firsthop), 0, 1, 0, 1},
  {&__pyx_n_s_fix, __pyx_k_fix, sizeof(__pyx_k_fix), 0, 0, 1, 1},
  {&__pyx_n_s_fopen, __pyx_k_fopen, sizeof(__pyx_k_fopen), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_u_ftime, __pyx_k_ftime, sizeof(__pyx_k_ftime), 0, 1, 0, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_u_hop_count, __pyx_k_hop_count, sizeof(__pyx_k_hop_count), 0, 1, 0, 1},
  {&__pyx_n_u_hoplimit, __pyx_k_hoplimit, sizeof(__pyx_k_hoplimit), 0, 1, 0, 1},
  {&__pyx_n_u_hops, __pyx_k_hops, sizeof(__pyx_k_hops), 0, 1, 0, 1},
  {&__pyx_n_s_hostname, __pyx_k_hostname, sizeof(__pyx_k_hostname), 0, 0, 1, 1},
  {&__pyx_n_u_icmp_code, __pyx_k_icmp_code, sizeof(__pyx_k_icmp_code), 0, 1, 0, 1},
  {&__pyx_kp_u_icmp_echo_paris, __pyx_k_icmp_echo_paris, sizeof(__pyx_k_icmp_echo_paris), 0, 1, 0, 0},
  {&__pyx_n_u_icmp_q_ipl, __pyx_k_icmp_q_ipl, sizeof(__pyx_k_icmp_q_ipl), 0, 1, 0, 1},
  {&__pyx_n_u_icmp_q_tos, __pyx_k_icmp_q_tos, sizeof(__pyx_k_icmp_q_tos), 0, 1, 0, 1},
  {&__pyx_n_u_icmp_q_ttl, __pyx_k_icmp_q_ttl, sizeof(__pyx_k_icmp_q_ttl), 0, 1, 0, 1},
  {&__pyx_n_u_icmp_sum, __pyx_k_icmp_sum, sizeof(__pyx_k_icmp_sum), 0, 1, 0, 1},
  {&__pyx_n_u_icmp_type, __pyx_k_icmp_type, sizeof(__pyx_k_icmp_type), 0, 1, 0, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_infile, __pyx_k_infile, sizeof(__pyx_k_infile), 0, 0, 1, 1},
  {&__pyx_n_s_json, __pyx_k_json, sizeof(__pyx_k_json), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_u_method, __pyx_k_method, sizeof(__pyx_k_method), 0, 1, 0, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_outfile, __pyx_k_outfile, sizeof(__pyx_k_outfile), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_u_probe_count, __pyx_k_probe_count, sizeof(__pyx_k_probe_count), 0, 1, 0, 1},
  {&__pyx_n_u_probe_id, __pyx_k_probe_id, sizeof(__pyx_k_probe_id), 0, 1, 0, 1},
  {&__pyx_n_u_probe_size, __pyx_k_probe_size, sizeof(__pyx_k_probe_size), 0, 1, 0, 1},
  {&__pyx_n_u_probe_ttl, __pyx_k_probe_ttl, sizeof(__pyx_k_probe_ttl), 0, 1, 0, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_readline, __pyx_k_readline, sizeof(__pyx_k_readline), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_u_reply_ipid, __pyx_k_reply_ipid, sizeof(__pyx_k_reply_ipid), 0, 1, 0, 1},
  {&__pyx_n_u_reply_size, __pyx_k_reply_size, sizeof(__pyx_k_reply_size), 0, 1, 0, 1},
  {&__pyx_n_u_reply_tos, __pyx_k_reply_tos, sizeof(__pyx_k_reply_tos), 0, 1, 0, 1},
  {&__pyx_n_u_reply_ttl, __pyx_k_reply_ttl, sizeof(__pyx_k_reply_ttl), 0, 1, 0, 1},
  {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
  {&__pyx_n_u_rtt, __pyx_k_rtt, sizeof(__pyx_k_rtt), 0, 1, 0, 1},
  {&__pyx_n_u_sec, __pyx_k_sec, sizeof(__pyx_k_sec), 0, 1, 0, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_sleep, __pyx_k_sleep, sizeof(__pyx_k_sleep), 0, 0, 1, 1},
  {&__pyx_n_u_src, __pyx_k_src, sizeof(__pyx_k_src), 0, 1, 0, 1},
  {&__pyx_n_u_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 1, 0, 1},
  {&__pyx_n_u_stop_data, __pyx_k_stop_data, sizeof(__pyx_k_stop_data), 0, 1, 0, 1},
  {&__pyx_n_u_stop_reason, __pyx_k_stop_reason, sizeof(__pyx_k_stop_reason), 0, 1, 0, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_u_tos, __pyx_k_tos, sizeof(__pyx_k_tos), 0, 1, 0, 1},
  {&__pyx_n_u_trace, __pyx_k_trace, sizeof(__pyx_k_trace), 0, 1, 0, 1},
  {&__pyx_n_u_tx, __pyx_k_tx, sizeof(__pyx_k_tx), 0, 1, 0, 1},
  {&__pyx_n_u_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 1, 0, 1},
  {&__pyx_kp_u_udp_paris, __pyx_k_udp_paris, sizeof(__pyx_k_udp_paris), 0, 1, 0, 0},
  {&__pyx_n_u_usec, __pyx_k_usec, sizeof(__pyx_k_usec), 0, 1, 0, 1},
  {&__pyx_n_u_userid, __pyx_k_userid, sizeof(__pyx_k_userid), 0, 1, 0, 1},
  {&__pyx_n_s_verbose, __pyx_k_verbose, sizeof(__pyx_k_verbose), 0, 0, 1, 1},
  {&__pyx_n_u_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 1, 0, 1},
  {&__pyx_n_u_wait, __pyx_k_wait, sizeof(__pyx_k_wait), 0, 1, 0, 1},
  {&__pyx_n_u_wait_probe, __pyx_k_wait_probe, sizeof(__pyx_k_wait_probe), 0, 1, 0, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_u_wt, __pyx_k_wt, sizeof(__pyx_k_wt), 0, 1, 0, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 142, __pyx_L1_error)
  __pyx_builtin_round = __Pyx_GetBuiltinName(__pyx_n_s_round); if (!__pyx_builtin_round) __PYX_ERR(0, 163, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(0, 312, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(0, 321, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "cloudtrace/read/convert.pyx":312
 *             Reply reply
 *             uint32_t loop = 0
 *         with open(self.hostname) as sidef:             # <<<<<<<<<<<<<<
 *             header = sidef.readline()
 *             # print(header)
 */
  __pyx_tuple__5 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_47652 = PyInt_FromLong(47652L); if (unlikely(!__pyx_int_47652)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TRACE_TYPE_ICMP_ECHO_PARIS, (void *)&TRACE_TYPE_ICMP_ECHO_PARIS, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TRACE_TYPE_UDP_PARIS, (void *)&TRACE_TYPE_UDP_PARIS, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TRACE_STOP_GAPLIMIT, (void *)&TRACE_STOP_GAPLIMIT, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TRACE_STOP_COMPLETED, (void *)&TRACE_STOP_COMPLETED, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TRACE_STOP_UNREACH, (void *)&TRACE_STOP_UNREACH, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TRACE_STOP_LOOP, (void *)&TRACE_STOP_LOOP, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_TRACE_STOP_HOPLIMIT, (void *)&TRACE_STOP_HOPLIMIT, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("loop_check", (void (*)(void))__pyx_f_10cloudtrace_4read_7convert_loop_check, "int (struct __pyx_obj_10cloudtrace_4read_7combine_Trace *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("addr_tostr", (void (*)(void))__pyx_f_10cloudtrace_4read_7convert_addr_tostr, "PyObject *(char *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_10cloudtrace_4read_7convert_Convert = &__pyx_vtable_10cloudtrace_4read_7convert_Convert;
  __pyx_vtable_10cloudtrace_4read_7convert_Convert.write_dst = (void (*)(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *, PyObject *))__pyx_f_10cloudtrace_4read_7convert_7Convert_write_dst;
  __pyx_vtable_10cloudtrace_4read_7convert_Convert.next = (struct __pyx_obj_10cloudtrace_4read_6packet_Reply *(*)(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *, struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *))__pyx_f_10cloudtrace_4read_7convert_7Convert_next;
  __pyx_vtable_10cloudtrace_4read_7convert_Convert.write_old = (void (*)(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *, struct timeval *, int))__pyx_f_10cloudtrace_4read_7convert_7Convert_write_old;
  __pyx_vtable_10cloudtrace_4read_7convert_Convert.write_all = (void (*)(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *))__pyx_f_10cloudtrace_4read_7convert_7Convert_write_all;
  if (PyType_Ready(&__pyx_type_10cloudtrace_4read_7convert_Convert) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_4read_7convert_Convert.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_4read_7convert_Convert.tp_dictoffset && __pyx_type_10cloudtrace_4read_7convert_Convert.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_4read_7convert_Convert.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_4read_7convert_Convert.tp_dict, __pyx_vtabptr_10cloudtrace_4read_7convert_Convert) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Convert, (PyObject *)&__pyx_type_10cloudtrace_4read_7convert_Convert) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_4read_7convert_Convert) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_7convert_Convert = &__pyx_type_10cloudtrace_4read_7convert_Convert;
  __pyx_vtabptr_10cloudtrace_4read_7convert_ConvertTrace = &__pyx_vtable_10cloudtrace_4read_7convert_ConvertTrace;
  __pyx_vtable_10cloudtrace_4read_7convert_ConvertTrace.__pyx_base = *__pyx_vtabptr_10cloudtrace_4read_7convert_Convert;
  __pyx_vtable_10cloudtrace_4read_7convert_ConvertTrace.__pyx_base.write_dst = (void (*)(struct __pyx_obj_10cloudtrace_4read_7convert_Convert *, PyObject *))__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_write_dst;
  __pyx_vtable_10cloudtrace_4read_7convert_ConvertTrace.initialize_trace = (struct __pyx_obj_10cloudtrace_4read_7combine_Trace *(*)(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *, struct __pyx_obj_10cloudtrace_4read_6packet_Packet *))__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_initialize_trace;
  __pyx_vtable_10cloudtrace_4read_7convert_ConvertTrace.convert_trace = (void (*)(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *, int, uint8_t, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_convert_trace;
  __pyx_vtable_10cloudtrace_4read_7convert_ConvertTrace.hopdict = (PyObject *(*)(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *, __pyx_t_10cloudtrace_4read_7combine_hop_t *))__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_hopdict;
  __pyx_vtable_10cloudtrace_4read_7convert_ConvertTrace.tracedict = (PyObject *(*)(struct __pyx_obj_10cloudtrace_4read_7convert_ConvertTrace *, struct __pyx_obj_10cloudtrace_4read_7combine_Trace *))__pyx_f_10cloudtrace_4read_7convert_12ConvertTrace_tracedict;
  __pyx_type_10cloudtrace_4read_7convert_ConvertTrace.tp_base = __pyx_ptype_10cloudtrace_4read_7convert_Convert;
  if (PyType_Ready(&__pyx_type_10cloudtrace_4read_7convert_ConvertTrace) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_4read_7convert_ConvertTrace.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_4read_7convert_ConvertTrace.tp_dictoffset && __pyx_type_10cloudtrace_4read_7convert_ConvertTrace.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_4read_7convert_ConvertTrace.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_4read_7convert_ConvertTrace.tp_dict, __pyx_vtabptr_10cloudtrace_4read_7convert_ConvertTrace) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ConvertTrace, (PyObject *)&__pyx_type_10cloudtrace_4read_7convert_ConvertTrace) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_4read_7convert_ConvertTrace) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_7convert_ConvertTrace = &__pyx_type_10cloudtrace_4read_7convert_ConvertTrace;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.read.pcap"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10cloudtrace_4read_4pcap_Pcap = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.pcap", "Pcap", sizeof(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_4pcap_Pcap) __PYX_ERR(3, 22, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_4pcap_Pcap = (struct __pyx_vtabstruct_10cloudtrace_4read_4pcap_Pcap*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_4pcap_Pcap->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_4pcap_Pcap)) __PYX_ERR(3, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.read.combine"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10cloudtrace_4read_7combine_Trace = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.combine", "Trace", sizeof(struct __pyx_obj_10cloudtrace_4read_7combine_Trace), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_7combine_Trace) __PYX_ERR(4, 24, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_7combine_Trace = (struct __pyx_vtabstruct_10cloudtrace_4read_7combine_Trace*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_7combine_Trace->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_7combine_Trace)) __PYX_ERR(4, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.read.packet"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10cloudtrace_4read_6packet_Packet = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.packet", "Packet", sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_Packet), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_6packet_Packet) __PYX_ERR(5, 10, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_6packet_Packet = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Packet*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_6packet_Packet->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_6packet_Packet)) __PYX_ERR(5, 10, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_ICMPProbe = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.packet", "ICMPProbe", sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPProbe), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_6packet_ICMPProbe) __PYX_ERR(5, 26, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_6packet_ICMPProbe = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPProbe*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_6packet_ICMPProbe->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_6packet_ICMPProbe)) __PYX_ERR(5, 26, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_UDPProbe = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.packet", "UDPProbe", sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_UDPProbe), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_6packet_UDPProbe) __PYX_ERR(5, 34, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_6packet_UDPProbe = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_UDPProbe*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_6packet_UDPProbe->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_6packet_UDPProbe)) __PYX_ERR(5, 34, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_Reply = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.packet", "Reply", sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_Reply), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_6packet_Reply) __PYX_ERR(5, 40, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_6packet_Reply = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_Reply*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_6packet_Reply->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_6packet_Reply)) __PYX_ERR(5, 40, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_EchoReply = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.packet", "EchoReply", sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_EchoReply), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_6packet_EchoReply) __PYX_ERR(5, 56, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_6packet_EchoReply = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_EchoReply*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_6packet_EchoReply->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_6packet_EchoReply)) __PYX_ERR(5, 56, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_6packet_ICMPReply = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.packet", "ICMPReply", sizeof(struct __pyx_obj_10cloudtrace_4read_6packet_ICMPReply), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_6packet_ICMPReply) __PYX_ERR(5, 59, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_6packet_ICMPReply = (struct __pyx_vtabstruct_10cloudtrace_4read_6packet_ICMPReply*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_6packet_ICMPReply->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_6packet_ICMPReply)) __PYX_ERR(5, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.read.linkedlist"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.linkedlist", "LinkedList", sizeof(struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_10linkedlist_LinkedList = (struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_10linkedlist_LinkedList)) __PYX_ERR(6, 12, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_10linkedlist_Node = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.linkedlist", "Node", sizeof(struct __pyx_obj_10cloudtrace_4read_10linkedlist_Node), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_10linkedlist_Node) __PYX_ERR(6, 22, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList2 = __Pyx_ImportType(__pyx_t_1, "cloudtrace.read.linkedlist", "LinkedList2", sizeof(struct __pyx_obj_10cloudtrace_4read_10linkedlist_LinkedList2), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList2) __PYX_ERR(6, 28, __pyx_L1_error)
  __pyx_vtabptr_10cloudtrace_4read_10linkedlist_LinkedList2 = (struct __pyx_vtabstruct_10cloudtrace_4read_10linkedlist_LinkedList2*)__Pyx_GetVtable(__pyx_ptype_10cloudtrace_4read_10linkedlist_LinkedList2->tp_dict); if (unlikely(!__pyx_vtabptr_10cloudtrace_4read_10linkedlist_LinkedList2)) __PYX_ERR(6, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.read.packet"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "ICMP_PROBE", (void **)&__pyx_vp_10cloudtrace_4read_6packet_ICMP_PROBE, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "UDP_PROBE", (void **)&__pyx_vp_10cloudtrace_4read_6packet_UDP_PROBE, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "ECHO_REPLY", (void **)&__pyx_vp_10cloudtrace_4read_6packet_ECHO_REPLY, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "ICMP_REPLY", (void **)&__pyx_vp_10cloudtrace_4read_6packet_ICMP_REPLY, "uint8_t") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.read.packet"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "disassemble", (void (**)(void))&__pyx_f_10cloudtrace_4read_6packet_disassemble, "struct __pyx_obj_10cloudtrace_4read_6packet_Packet *(struct __pyx_obj_10cloudtrace_4read_4pcap_Pcap *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initconvert(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initconvert(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_convert(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_convert(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_convert(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'convert' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_convert(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("convert", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_cloudtrace__read__convert) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cloudtrace.read.convert")) {
      if (unlikely(PyDict_SetItemString(modules, "cloudtrace.read.convert", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely(__Pyx_modinit_variable_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "cloudtrace/read/convert.pyx":1
 * import json             # <<<<<<<<<<<<<<
 * from collections import defaultdict
 * from time import sleep
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_json, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_json, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":2
 * import json
 * from collections import defaultdict             # <<<<<<<<<<<<<<
 * from time import sleep
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_defaultdict);
  __Pyx_GIVEREF(__pyx_n_s_defaultdict);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_defaultdict);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_defaultdict, __pyx_t_1) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":3
 * import json
 * from collections import defaultdict
 * from time import sleep             # <<<<<<<<<<<<<<
 * 
 * from libc.string cimport memcpy, strcpy, memset, memcmp
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_sleep);
  __Pyx_GIVEREF(__pyx_n_s_sleep);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_sleep);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_time, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_sleep); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sleep, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/read/convert.pyx":8
 * from libc.stdlib cimport free, calloc
 * # from traceutils.file2.file2 import fopen
 * from cloudtrace.read.pcap import fopen             # <<<<<<<<<<<<<<
 * 
 * cdef:
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_fopen);
  __Pyx_GIVEREF(__pyx_n_s_fopen);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_fopen);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_cloudtrace_read_pcap, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_fopen); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fopen, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/read/convert.pyx":11
 * 
 * cdef:
 *     public uint8_t TRACE_TYPE_ICMP_ECHO_PARIS = 1             # <<<<<<<<<<<<<<
 *     public uint8_t TRACE_TYPE_UDP_PARIS = 2
 * 
 */
  TRACE_TYPE_ICMP_ECHO_PARIS = 1;

  /* "cloudtrace/read/convert.pyx":12
 * cdef:
 *     public uint8_t TRACE_TYPE_ICMP_ECHO_PARIS = 1
 *     public uint8_t TRACE_TYPE_UDP_PARIS = 2             # <<<<<<<<<<<<<<
 * 
 *     public uint8_t TRACE_STOP_GAPLIMIT = 1
 */
  TRACE_TYPE_UDP_PARIS = 2;

  /* "cloudtrace/read/convert.pyx":14
 *     public uint8_t TRACE_TYPE_UDP_PARIS = 2
 * 
 *     public uint8_t TRACE_STOP_GAPLIMIT = 1             # <<<<<<<<<<<<<<
 *     public uint8_t TRACE_STOP_COMPLETED = 2
 *     public uint8_t TRACE_STOP_UNREACH = 3
 */
  TRACE_STOP_GAPLIMIT = 1;

  /* "cloudtrace/read/convert.pyx":15
 * 
 *     public uint8_t TRACE_STOP_GAPLIMIT = 1
 *     public uint8_t TRACE_STOP_COMPLETED = 2             # <<<<<<<<<<<<<<
 *     public uint8_t TRACE_STOP_UNREACH = 3
 *     public uint8_t TRACE_STOP_LOOP = 4
 */
  TRACE_STOP_COMPLETED = 2;

  /* "cloudtrace/read/convert.pyx":16
 *     public uint8_t TRACE_STOP_GAPLIMIT = 1
 *     public uint8_t TRACE_STOP_COMPLETED = 2
 *     public uint8_t TRACE_STOP_UNREACH = 3             # <<<<<<<<<<<<<<
 *     public uint8_t TRACE_STOP_LOOP = 4
 *     public uint8_t TRACE_STOP_HOPLIMIT = 5
 */
  TRACE_STOP_UNREACH = 3;

  /* "cloudtrace/read/convert.pyx":17
 *     public uint8_t TRACE_STOP_COMPLETED = 2
 *     public uint8_t TRACE_STOP_UNREACH = 3
 *     public uint8_t TRACE_STOP_LOOP = 4             # <<<<<<<<<<<<<<
 *     public uint8_t TRACE_STOP_HOPLIMIT = 5
 * 
 */
  TRACE_STOP_LOOP = 4;

  /* "cloudtrace/read/convert.pyx":18
 *     public uint8_t TRACE_STOP_UNREACH = 3
 *     public uint8_t TRACE_STOP_LOOP = 4
 *     public uint8_t TRACE_STOP_HOPLIMIT = 5             # <<<<<<<<<<<<<<
 * 
 * cdef class Convert:
 */
  TRACE_STOP_HOPLIMIT = 5;

  /* "cloudtrace/read/convert.pyx":1
 * import json             # <<<<<<<<<<<<<<
 * from collections import defaultdict
 * from time import sleep
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init cloudtrace.read.convert", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cloudtrace.read.convert");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B3
    if ((1)) {
        return _PyDict_Pop(d, key, default_value);
    } else
#endif
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_pop, d, key);
    }
}

/* decode_c_bytes */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* ModInt[long] */
static CYTHON_INLINE long __Pyx_mod_long(long a, long b) {
    long r = a % b;
    r += ((r != 0) & ((r ^ b) < 0)) * b;
    return r;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint8_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint8_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint8_t) -1;
        }
    } else {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint16_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint16_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint16_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint16_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_time_t(time_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const time_t neg_one = (time_t) -1, const_zero = (time_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(time_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(time_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(time_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(time_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(time_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(time_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_suseconds_t(suseconds_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const suseconds_t neg_one = (suseconds_t) -1, const_zero = (suseconds_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(suseconds_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(suseconds_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(suseconds_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(suseconds_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(suseconds_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(suseconds_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
