/* Generated by Cython 0.29.32 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "cloudtrace.trace.probe",
        "sources": [
            "cloudtrace/trace/probe.pyx"
        ]
    },
    "module_name": "cloudtrace.trace.probe"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_32"
#define CYTHON_HEX_VERSION 0x001D20F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_HEX >= 0x07030900)
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cloudtrace__trace__probe
#define __PYX_HAVE_API__cloudtrace__trace__probe
/* Early includes */
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include "netinet/ip.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "stringsource",
  "cloudtrace/trace/probe.pyx",
};

/*--- Type declarations ---*/
struct __pyx_obj_10cloudtrace_5trace_5probe_Probe;
struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe;
struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe;
struct __pyx_t_10cloudtrace_5trace_5utils_ethhdr_s;
struct __pyx_t_10cloudtrace_5trace_5utils_ip_s;
struct __pyx_t_10cloudtrace_5trace_5utils_echo_s;
struct __pyx_t_10cloudtrace_5trace_5utils_unused_s;
union __pyx_t_10cloudtrace_5trace_5utils_icmpunion;
struct __pyx_t_10cloudtrace_5trace_5utils_icmp_s;
struct __pyx_t_10cloudtrace_5trace_5utils_udp_s;
struct __pyx_t_10cloudtrace_5trace_5utils_udppseudo_s;
struct __pyx_t_10cloudtrace_5trace_5utils_ip6_hdrctl;
union __pyx_t_10cloudtrace_5trace_5utils_ip6_un;
struct __pyx_t_10cloudtrace_5trace_5utils_icmpexthdr_s;
struct __pyx_t_10cloudtrace_5trace_5utils_icmpext_s;
struct __pyx_t_10cloudtrace_5trace_5utils_ip6_s;

/* "cloudtrace/trace/utils.pxd":26
 * cdef uint16_t checksum(void *data, int length) except -1;
 * 
 * cdef struct ethhdr_s:             # <<<<<<<<<<<<<<
 *     uint8_t h_dest[6]
 *     uint8_t h_source[6]
 */
struct __pyx_t_10cloudtrace_5trace_5utils_ethhdr_s {
  uint8_t h_dest[6];
  uint8_t h_source[6];
  uint16_t h_proto;
};

/* "cloudtrace/trace/utils.pxd":30
 *     uint8_t h_source[6]
 *     uint16_t h_proto
 * ctypedef ethhdr_s ethhdr             # <<<<<<<<<<<<<<
 * 
 * cdef struct ip_s:
 */
typedef struct __pyx_t_10cloudtrace_5trace_5utils_ethhdr_s __pyx_t_10cloudtrace_5trace_5utils_ethhdr;

/* "cloudtrace/trace/utils.pxd":32
 * ctypedef ethhdr_s ethhdr
 * 
 * cdef struct ip_s:             # <<<<<<<<<<<<<<
 *     uint8_t ip_hl
 *     uint8_t ip_tos
 */
struct __pyx_t_10cloudtrace_5trace_5utils_ip_s {
  uint8_t ip_hl;
  uint8_t ip_tos;
  uint16_t ip_len;
  uint16_t id;
  uint16_t ip_off;
  uint8_t ip_ttl;
  uint8_t ip_p;
  uint16_t ip_sum;
  uint8_t ip_src[4];
  uint8_t ip_dst[4];
};

/* "cloudtrace/trace/utils.pxd":43
 *     uint8_t ip_src[4]
 *     uint8_t ip_dst[4]
 * ctypedef ip_s ip             # <<<<<<<<<<<<<<
 * 
 * cdef void iptohost(ip *iphdr);
 */
typedef struct __pyx_t_10cloudtrace_5trace_5utils_ip_s __pyx_t_10cloudtrace_5trace_5utils_ip;

/* "cloudtrace/trace/utils.pxd":47
 * cdef void iptohost(ip *iphdr);
 * 
 * cdef struct echo_s:             # <<<<<<<<<<<<<<
 *     uint16_t id
 *     uint16_t seq
 */
struct __pyx_t_10cloudtrace_5trace_5utils_echo_s {
  uint16_t id;
  uint16_t seq;
};

/* "cloudtrace/trace/utils.pxd":50
 *     uint16_t id
 *     uint16_t seq
 * ctypedef echo_s echo             # <<<<<<<<<<<<<<
 * 
 * cdef struct unused_s:
 */
typedef struct __pyx_t_10cloudtrace_5trace_5utils_echo_s __pyx_t_10cloudtrace_5trace_5utils_echo;

/* "cloudtrace/trace/utils.pxd":52
 * ctypedef echo_s echo
 * 
 * cdef struct unused_s:             # <<<<<<<<<<<<<<
 *     uint16_t __unused
 *     uint16_t mtu
 */
struct __pyx_t_10cloudtrace_5trace_5utils_unused_s {
  uint16_t __pyx___unused;
  uint16_t mtu;
};

/* "cloudtrace/trace/utils.pxd":55
 *     uint16_t __unused
 *     uint16_t mtu
 * ctypedef unused_s unused             # <<<<<<<<<<<<<<
 * 
 * cdef union icmpunion:
 */
typedef struct __pyx_t_10cloudtrace_5trace_5utils_unused_s __pyx_t_10cloudtrace_5trace_5utils_unused;

/* "cloudtrace/trace/utils.pxd":57
 * ctypedef unused_s unused
 * 
 * cdef union icmpunion:             # <<<<<<<<<<<<<<
 *     echo_s echo
 *     unsigned int gateway
 */
union __pyx_t_10cloudtrace_5trace_5utils_icmpunion {
  struct __pyx_t_10cloudtrace_5trace_5utils_echo_s echo;
  unsigned int gateway;
  struct __pyx_t_10cloudtrace_5trace_5utils_unused_s frag;
};

/* "cloudtrace/trace/utils.pxd":62
 *     unused_s frag
 * 
 * cdef struct icmp_s:             # <<<<<<<<<<<<<<
 *     uint8_t type
 *     uint8_t code
 */
struct __pyx_t_10cloudtrace_5trace_5utils_icmp_s {
  uint8_t type;
  uint8_t code;
  uint16_t checksum;
  union __pyx_t_10cloudtrace_5trace_5utils_icmpunion un;
};

/* "cloudtrace/trace/utils.pxd":67
 *     uint16_t checksum
 *     icmpunion un
 * ctypedef icmp_s icmp             # <<<<<<<<<<<<<<
 * 
 * cdef void icmptohost(icmp *hdr);
 */
typedef struct __pyx_t_10cloudtrace_5trace_5utils_icmp_s __pyx_t_10cloudtrace_5trace_5utils_icmp;

/* "cloudtrace/trace/utils.pxd":71
 * cdef void icmptohost(icmp *hdr);
 * 
 * cdef struct udp_s:             # <<<<<<<<<<<<<<
 *     uint16_t source
 *     uint16_t dest
 */
struct __pyx_t_10cloudtrace_5trace_5utils_udp_s {
  uint16_t source;
  uint16_t dest;
  short len;
  uint16_t check;
};

/* "cloudtrace/trace/utils.pxd":76
 *     short len
 *     uint16_t check
 * ctypedef udp_s udp             # <<<<<<<<<<<<<<
 * 
 * cdef void udptohost(udp *hdr);
 */
typedef struct __pyx_t_10cloudtrace_5trace_5utils_udp_s __pyx_t_10cloudtrace_5trace_5utils_udp;

/* "cloudtrace/trace/utils.pxd":80
 * cdef void udptohost(udp *hdr);
 * 
 * cdef struct udppseudo_s:             # <<<<<<<<<<<<<<
 *     uint8_t src[4]
 *     uint8_t dst[4]
 */
struct __pyx_t_10cloudtrace_5trace_5utils_udppseudo_s {
  uint8_t src[4];
  uint8_t dst[4];
  uint8_t zeros;
  uint8_t proto;
  uint16_t length;
  uint16_t sport;
  uint16_t dport;
  uint16_t length2;
  uint16_t chksum;
  uint8_t data[0x5C0];
};

/* "cloudtrace/trace/utils.pxd":91
 *     uint16_t chksum
 *     uint8_t data[1472]
 * ctypedef udppseudo_s udppseudo             # <<<<<<<<<<<<<<
 * 
 * cdef struct ip6_hdrctl:
 */
typedef struct __pyx_t_10cloudtrace_5trace_5utils_udppseudo_s __pyx_t_10cloudtrace_5trace_5utils_udppseudo;

/* "cloudtrace/trace/utils.pxd":93
 * ctypedef udppseudo_s udppseudo
 * 
 * cdef struct ip6_hdrctl:             # <<<<<<<<<<<<<<
 *     uint32_t ip6_un1_flow
 *     uint16_t ip6_un1_plen
 */
struct __pyx_t_10cloudtrace_5trace_5utils_ip6_hdrctl {
  uint32_t ip6_un1_flow;
  uint16_t ip6_un1_plen;
  uint8_t ip6_un1_nxt;
  uint8_t ip6_un1_hlim;
};

/* "cloudtrace/trace/utils.pxd":99
 *     uint8_t ip6_un1_hlim
 * 
 * cdef union ip6_un:             # <<<<<<<<<<<<<<
 *     ip6_hdrctl ip6_un1
 *     uint8_t ip6_un2_vfc
 */
union __pyx_t_10cloudtrace_5trace_5utils_ip6_un {
  struct __pyx_t_10cloudtrace_5trace_5utils_ip6_hdrctl ip6_un1;
  uint8_t ip6_un2_vfc;
};

/* "cloudtrace/trace/utils.pxd":103
 *     uint8_t ip6_un2_vfc
 * 
 * cdef struct icmpexthdr_s:             # <<<<<<<<<<<<<<
 *     uint8_t version
 *     uint8_t reserved
 */
struct __pyx_t_10cloudtrace_5trace_5utils_icmpexthdr_s {
  uint8_t version;
  uint8_t reserved;
  uint16_t checksum;
};

/* "cloudtrace/trace/utils.pxd":107
 *     uint8_t reserved
 *     uint16_t checksum
 * ctypedef icmpexthdr_s icmpexthdr             # <<<<<<<<<<<<<<
 * 
 * cdef struct icmpext_s:
 */
typedef struct __pyx_t_10cloudtrace_5trace_5utils_icmpexthdr_s __pyx_t_10cloudtrace_5trace_5utils_icmpexthdr;

/* "cloudtrace/trace/utils.pxd":109
 * ctypedef icmpexthdr_s icmpexthdr
 * 
 * cdef struct icmpext_s:             # <<<<<<<<<<<<<<
 *     uint16_t length
 *     uint8_t classnum
 */
struct __pyx_t_10cloudtrace_5trace_5utils_icmpext_s {
  uint16_t length;
  uint8_t classnum;
  uint8_t ctype;
};

/* "cloudtrace/trace/utils.pxd":113
 *     uint8_t classnum
 *     uint8_t ctype
 * ctypedef icmpext_s icmpext             # <<<<<<<<<<<<<<
 * 
 * cdef struct ip6_s:
 */
typedef struct __pyx_t_10cloudtrace_5trace_5utils_icmpext_s __pyx_t_10cloudtrace_5trace_5utils_icmpext;

/* "cloudtrace/trace/utils.pxd":115
 * ctypedef icmpext_s icmpext
 * 
 * cdef struct ip6_s:             # <<<<<<<<<<<<<<
 *     ip6_un ip6_ctlun
 *     uint8_t ip6_src[16]
 */
struct __pyx_t_10cloudtrace_5trace_5utils_ip6_s {
  union __pyx_t_10cloudtrace_5trace_5utils_ip6_un ip6_ctlun;
  uint8_t ip6_src[16];
  uint8_t ip6_dst[16];
};
struct __pyx_t_10cloudtrace_5trace_5probe_pcapfix_global_hdr_s;
typedef struct __pyx_t_10cloudtrace_5trace_5probe_pcapfix_global_hdr_s __pyx_t_10cloudtrace_5trace_5probe_pcapfix_global_hdr_s;
struct __pyx_t_10cloudtrace_5trace_5probe_pcapfix_packet_hdr_s;
typedef struct __pyx_t_10cloudtrace_5trace_5probe_pcapfix_packet_hdr_s __pyx_t_10cloudtrace_5trace_5probe_pcapfix_packet_hdr_s;
struct __pyx_opt_args_10cloudtrace_5trace_5probe_craft;
struct __pyx_opt_args_10cloudtrace_5trace_5probe_ping;

/* "cloudtrace/trace/probe.pxd":3
 * from cloudtrace.trace.utils cimport *
 * 
 * ctypedef struct pcapfix_global_hdr_s:             # <<<<<<<<<<<<<<
 *     unsigned int magic_number
 *     unsigned short version_major
 */
struct __pyx_t_10cloudtrace_5trace_5probe_pcapfix_global_hdr_s {
  unsigned int magic_number;
  unsigned short version_major;
  unsigned short version_minor;
  int thiszone;
  unsigned int sigfigs;
  unsigned int snaplen;
  unsigned int network;
};

/* "cloudtrace/trace/probe.pxd":12
 *     unsigned int network
 * 
 * ctypedef struct pcapfix_packet_hdr_s:             # <<<<<<<<<<<<<<
 *     unsigned int ts_sec         #/[inserted by cython to avoid comment start]* timestamp seconds *[inserted by cython to avoid comment closer]/
 *     unsigned int ts_usec        #/[inserted by cython to avoid comment start]* timestamp microseconds *[inserted by cython to avoid comment closer]/
 */
struct __pyx_t_10cloudtrace_5trace_5probe_pcapfix_packet_hdr_s {
  unsigned int ts_sec;
  unsigned int ts_usec;
  unsigned int incl_len;
  unsigned int orig_len;
};

/* "cloudtrace/trace/probe.pxd":56
 * cpdef bytes fill_packet_header(uint16_t size);
 * cpdef void write_pcap_header(f);
 * cpdef void craft(str targets, uint16_t pid, file=*, int minttl=*, int maxttl=*, int cycles=*, str proto=*, uint16_t size=*, bint randomize=*) except *;             # <<<<<<<<<<<<<<
 * cpdef void ping(list targets, uint16_t pid, file=*, int numpings=*, int cycles=*, str proto=*, uint16_t size=*);
 */
struct __pyx_opt_args_10cloudtrace_5trace_5probe_craft {
  int __pyx_n;
  PyObject *file;
  int minttl;
  int maxttl;
  int cycles;
  PyObject *proto;
  uint16_t size;
  int randomize;
};

/* "cloudtrace/trace/probe.pxd":57
 * cpdef void write_pcap_header(f);
 * cpdef void craft(str targets, uint16_t pid, file=*, int minttl=*, int maxttl=*, int cycles=*, str proto=*, uint16_t size=*, bint randomize=*) except *;
 * cpdef void ping(list targets, uint16_t pid, file=*, int numpings=*, int cycles=*, str proto=*, uint16_t size=*);             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_10cloudtrace_5trace_5probe_ping {
  int __pyx_n;
  PyObject *file;
  int numpings;
  int cycles;
  PyObject *proto;
  uint16_t size;
};

/* "cloudtrace/trace/probe.pxd":20
 * cdef bint minsize(int proto) except -1;
 * 
 * cdef class Probe:             # <<<<<<<<<<<<<<
 *     cdef:
 *         void *buf
 */
struct __pyx_obj_10cloudtrace_5trace_5probe_Probe {
  PyObject_HEAD
  struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *__pyx_vtab;
  void *buf;
  __pyx_t_10cloudtrace_5trace_5utils_ethhdr *etherhdr;
  __pyx_t_10cloudtrace_5trace_5utils_ip *iphdr;
};


/* "cloudtrace/trace/probe.pxd":36
 *     cpdef void prep_trace(self, uint8_t num);
 * 
 * cdef class ICMPProbe(Probe):             # <<<<<<<<<<<<<<
 *     cdef:
 *         icmp *icmphdr
 */
struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe {
  struct __pyx_obj_10cloudtrace_5trace_5probe_Probe __pyx_base;
  __pyx_t_10cloudtrace_5trace_5utils_icmp *icmphdr;
  void *data;
};


/* "cloudtrace/trace/probe.pxd":43
 *     cpdef void set_icmp_id(self, uint16_t pid);
 * 
 * cdef class UDPProbe(Probe):             # <<<<<<<<<<<<<<
 *     cdef:
 *         udp *udphdr
 */
struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe {
  struct __pyx_obj_10cloudtrace_5trace_5probe_Probe __pyx_base;
  __pyx_t_10cloudtrace_5trace_5utils_udp *udphdr;
  void *data;
  __pyx_t_10cloudtrace_5trace_5utils_udppseudo *pseudo;
};



/* "cloudtrace/trace/probe.pyx":20
 *         return size + sizeof(icmp) + 2
 * 
 * cdef class Probe:             # <<<<<<<<<<<<<<
 *     def __cinit__(self, uint16_t size=64):
 *         self.buf = calloc(1, sizeof(ethhdr) + size)
 */

struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe {
  void (*set_dst)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, PyObject *, int __pyx_skip_dispatch);
  void (*set_src)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*to_bytes)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, int __pyx_skip_dispatch);
  void (*set_ttl)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, uint8_t, int __pyx_skip_dispatch);
  void (*prepare)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, int __pyx_skip_dispatch);
  void (*prepare_ip)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *);
  void (*copy_ether)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, void *);
  void (*create_ether)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  void (*prep_trace)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, uint8_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *__pyx_vtabptr_10cloudtrace_5trace_5probe_Probe;


/* "cloudtrace/trace/probe.pyx":66
 *         pass
 * 
 * cdef class ICMPProbe(Probe):             # <<<<<<<<<<<<<<
 *     def __cinit__(self, uint16_t size=1500):
 *         self.icmphdr = <icmp *>(self.buf + sizeof(ethhdr) + sizeof(ip))
 */

struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_ICMPProbe {
  struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe __pyx_base;
  void (*set_icmp_id)(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *, uint16_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_vtabptr_10cloudtrace_5trace_5probe_ICMPProbe;


/* "cloudtrace/trace/probe.pyx":88
 *         memcpy(self.data, &data, 2)
 * 
 * cdef class UDPProbe(Probe):             # <<<<<<<<<<<<<<
 *     def __cinit__(self, uint16_t size=1500):
 *         cdef:
 */

struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_UDPProbe {
  struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe __pyx_base;
  void (*set_sport)(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *, uint16_t, int __pyx_skip_dispatch);
  void (*set_dport)(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *, uint16_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_UDPProbe *__pyx_vtabptr_10cloudtrace_5trace_5probe_UDPProbe;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_copy_ether(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, void *__pyx_v_ether); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_create_ether(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_dst, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_set_dst(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_dst, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_set_src(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_src, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10cloudtrace_5trace_5probe_5Probe_to_bytes(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_set_ttl(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, uint8_t __pyx_v_ttl, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_prepare_ip(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_prep_trace(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, CYTHON_UNUSED uint8_t __pyx_v_num, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_set_icmp_id(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, uint16_t __pyx_v_pid, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, uint8_t __pyx_v_num, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint8_t __pyx_v_num, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_sport(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_sport, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_dport(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_dport, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_dst(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, PyObject *__pyx_v_dst, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_src(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, PyObject *__pyx_v_src, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'posix.types' */

/* Module declarations from 'posix.signal' */

/* Module declarations from 'posix.time' */

/* Module declarations from 'cloudtrace.trace.utils' */
static uint16_t (*__pyx_f_10cloudtrace_5trace_5utils_checksum)(void *, int); /*proto*/

/* Module declarations from 'cloudtrace.trace.probe' */
static PyTypeObject *__pyx_ptype_10cloudtrace_5trace_5probe_Probe = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_5trace_5probe_ICMPProbe = 0;
static PyTypeObject *__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe = 0;
static int __pyx_f_10cloudtrace_5trace_5probe_minsize(int); /*proto*/
static uint8_t __pyx_f_10cloudtrace_5trace_5probe_proto_to_num(PyObject *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_f_10cloudtrace_5trace_5probe_create_probe(PyObject *, uint8_t, uint16_t, uint16_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_10cloudtrace_5trace_5probe_fill_packet_header(uint16_t, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_write_pcap_header(PyObject *, int __pyx_skip_dispatch); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_craft(PyObject *, uint16_t, int __pyx_skip_dispatch, struct __pyx_opt_args_10cloudtrace_5trace_5probe_craft *__pyx_optional_args); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_ping(PyObject *, uint16_t, int __pyx_skip_dispatch, struct __pyx_opt_args_10cloudtrace_5trace_5probe_ping *__pyx_optional_args); /*proto*/
#define __Pyx_MODULE_NAME "cloudtrace.trace.probe"
extern int __pyx_module_is_main_cloudtrace__trace__probe;
int __pyx_module_is_main_cloudtrace__trace__probe = 0;

/* Implementation of 'cloudtrace.trace.probe' */
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_open;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_print;
static const char __pyx_k_f[] = "f";
static const char __pyx_k_IP[] = "IP";
static const char __pyx_k__8[] = ".";
static const char __pyx_k__9[] = "{}.{}";
static const char __pyx_k_rb[] = "rb";
static const char __pyx_k_wb[] = "wb";
static const char __pyx_k__11[] = "{}";
static const char __pyx_k_dst[] = "dst";
static const char __pyx_k_num[] = "num";
static const char __pyx_k_pid[] = "pid";
static const char __pyx_k_sIP[] = "sIP";
static const char __pyx_k_src[] = "src";
static const char __pyx_k_ttl[] = "ttl";
static const char __pyx_k_udp[] = "udp";
static const char __pyx_k_ICMP[] = "ICMP";
static const char __pyx_k_conf[] = "conf";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_icmp[] = "icmp";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_ping[] = "ping";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_Ether[] = "Ether";
static const char __pyx_k_Probe[] = "Probe";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_craft[] = "craft";
static const char __pyx_k_dport[] = "dport";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_file2[] = "file2";
static const char __pyx_k_fopen[] = "fopen";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_proto[] = "proto";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_route[] = "route";
static const char __pyx_k_sICMP[] = "sICMP";
static const char __pyx_k_sport[] = "sport";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_No_TCP[] = "No TCP!";
static const char __pyx_k_cycles[] = "cycles";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_maxttl[] = "maxttl";
static const char __pyx_k_minttl[] = "minttl";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_0_0_0_0[] = "0.0.0.0";
static const char __pyx_k_ipproto[] = "ipproto";
static const char __pyx_k_prepare[] = "prepare";
static const char __pyx_k_randint[] = "randint";
static const char __pyx_k_set_dst[] = "set_dst";
static const char __pyx_k_set_src[] = "set_src";
static const char __pyx_k_set_ttl[] = "set_ttl";
static const char __pyx_k_targets[] = "targets";
static const char __pyx_k_UDPProbe[] = "UDPProbe";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_numpings[] = "numpings";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_to_bytes[] = "to_bytes";
static const char __pyx_k_ICMPProbe[] = "ICMPProbe";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_randomize[] = "randomize";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_set_dport[] = "set_dport";
static const char __pyx_k_set_sport[] = "set_sport";
static const char __pyx_k_getmacbyip[] = "getmacbyip";
static const char __pyx_k_prep_trace[] = "prep_trace";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_rpartition[] = "rpartition";
static const char __pyx_k_scapy_arch[] = "scapy.arch";
static const char __pyx_k_set_icmp_id[] = "set_icmp_id";
static const char __pyx_k_create_ether[] = "create_ether";
static const char __pyx_k_create_probe[] = "create_probe";
static const char __pyx_k_proto_to_num[] = "proto_to_num";
static const char __pyx_k_scapy_config[] = "scapy.config";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_Probe_prepare[] = "Probe.prepare";
static const char __pyx_k_Probe_set_dst[] = "Probe.set_dst";
static const char __pyx_k_Probe_set_src[] = "Probe.set_src";
static const char __pyx_k_Probe_set_ttl[] = "Probe.set_ttl";
static const char __pyx_k_get_if_hwaddr[] = "get_if_hwaddr";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_192_172_226_78[] = "192.172.226.78";
static const char __pyx_k_Probe_to_bytes[] = "Probe.to_bytes";
static const char __pyx_k_scapy_layers_l2[] = "scapy.layers.l2";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_Probe_prep_trace[] = "Probe.prep_trace";
static const char __pyx_k_Too_fast_to_time[] = "Too fast to time.";
static const char __pyx_k_UDPProbe_prepare[] = "UDPProbe.prepare";
static const char __pyx_k_UDPProbe_set_dst[] = "UDPProbe.set_dst";
static const char __pyx_k_UDPProbe_set_src[] = "UDPProbe.set_src";
static const char __pyx_k_ICMPProbe_prepare[] = "ICMPProbe.prepare";
static const char __pyx_k_scapy_layers_inet[] = "scapy.layers.inet";
static const char __pyx_k_write_pcap_header[] = "write_pcap_header";
static const char __pyx_k_Probe_create_ether[] = "Probe.create_ether";
static const char __pyx_k_UDPProbe_set_dport[] = "UDPProbe.set_dport";
static const char __pyx_k_UDPProbe_set_sport[] = "UDPProbe.set_sport";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_fill_packet_header[] = "fill_packet_header";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_UDPProbe_prep_trace[] = "UDPProbe.prep_trace";
static const char __pyx_k_ICMPProbe_prep_trace[] = "ICMPProbe.prep_trace";
static const char __pyx_k_ICMPProbe_set_icmp_id[] = "ICMPProbe.set_icmp_id";
static const char __pyx_k_Probe___reduce_cython[] = "Probe.__reduce_cython__";
static const char __pyx_k_Time_2f_seconds_2f_pps[] = "Time: {:,.2f} seconds. {:.2f} pps";
static const char __pyx_k_cloudtrace_trace_probe[] = "cloudtrace.trace.probe";
static const char __pyx_k_Probe___setstate_cython[] = "Probe.__setstate_cython__";
static const char __pyx_k_UDPProbe___reduce_cython[] = "UDPProbe.__reduce_cython__";
static const char __pyx_k_ICMPProbe___reduce_cython[] = "ICMPProbe.__reduce_cython__";
static const char __pyx_k_UDPProbe___setstate_cython[] = "UDPProbe.__setstate_cython__";
static const char __pyx_k_cloudtrace_trace_probe_pyx[] = "cloudtrace/trace/probe.pyx";
static const char __pyx_k_ICMPProbe___setstate_cython[] = "ICMPProbe.__setstate_cython__";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static PyObject *__pyx_kp_u_0_0_0_0;
static PyObject *__pyx_kp_u_192_172_226_78;
static PyObject *__pyx_n_s_Ether;
static PyObject *__pyx_n_s_ICMP;
static PyObject *__pyx_n_s_ICMPProbe;
static PyObject *__pyx_n_s_ICMPProbe___reduce_cython;
static PyObject *__pyx_n_s_ICMPProbe___setstate_cython;
static PyObject *__pyx_n_s_ICMPProbe_prep_trace;
static PyObject *__pyx_n_s_ICMPProbe_prepare;
static PyObject *__pyx_n_s_ICMPProbe_set_icmp_id;
static PyObject *__pyx_n_s_IP;
static PyObject *__pyx_kp_u_No_TCP;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_Probe;
static PyObject *__pyx_n_s_Probe___reduce_cython;
static PyObject *__pyx_n_s_Probe___setstate_cython;
static PyObject *__pyx_n_s_Probe_create_ether;
static PyObject *__pyx_n_s_Probe_prep_trace;
static PyObject *__pyx_n_s_Probe_prepare;
static PyObject *__pyx_n_s_Probe_set_dst;
static PyObject *__pyx_n_s_Probe_set_src;
static PyObject *__pyx_n_s_Probe_set_ttl;
static PyObject *__pyx_n_s_Probe_to_bytes;
static PyObject *__pyx_kp_u_Time_2f_seconds_2f_pps;
static PyObject *__pyx_kp_u_Too_fast_to_time;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_UDPProbe;
static PyObject *__pyx_n_s_UDPProbe___reduce_cython;
static PyObject *__pyx_n_s_UDPProbe___setstate_cython;
static PyObject *__pyx_n_s_UDPProbe_prep_trace;
static PyObject *__pyx_n_s_UDPProbe_prepare;
static PyObject *__pyx_n_s_UDPProbe_set_dport;
static PyObject *__pyx_n_s_UDPProbe_set_dst;
static PyObject *__pyx_n_s_UDPProbe_set_sport;
static PyObject *__pyx_n_s_UDPProbe_set_src;
static PyObject *__pyx_kp_u__11;
static PyObject *__pyx_kp_u__8;
static PyObject *__pyx_kp_u__9;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cloudtrace_trace_probe;
static PyObject *__pyx_kp_s_cloudtrace_trace_probe_pyx;
static PyObject *__pyx_n_s_conf;
static PyObject *__pyx_n_s_craft;
static PyObject *__pyx_n_s_create_ether;
static PyObject *__pyx_n_s_create_probe;
static PyObject *__pyx_n_s_cycles;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_dport;
static PyObject *__pyx_n_s_dst;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_file2;
static PyObject *__pyx_n_s_fill_packet_header;
static PyObject *__pyx_n_s_fopen;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_get_if_hwaddr;
static PyObject *__pyx_n_s_getmacbyip;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_u_icmp;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_ipproto;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_maxttl;
static PyObject *__pyx_n_s_minttl;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_num;
static PyObject *__pyx_n_s_numpings;
static PyObject *__pyx_n_s_open;
static PyObject *__pyx_n_s_pid;
static PyObject *__pyx_n_s_ping;
static PyObject *__pyx_n_s_prep_trace;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_proto;
static PyObject *__pyx_n_s_proto_to_num;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_randint;
static PyObject *__pyx_n_s_random;
static PyObject *__pyx_n_s_randomize;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_u_rb;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_route;
static PyObject *__pyx_n_s_rpartition;
static PyObject *__pyx_n_s_sICMP;
static PyObject *__pyx_n_s_sIP;
static PyObject *__pyx_n_s_scapy_arch;
static PyObject *__pyx_n_s_scapy_config;
static PyObject *__pyx_n_s_scapy_layers_inet;
static PyObject *__pyx_n_s_scapy_layers_l2;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_set_dport;
static PyObject *__pyx_n_s_set_dst;
static PyObject *__pyx_n_s_set_icmp_id;
static PyObject *__pyx_n_s_set_sport;
static PyObject *__pyx_n_s_set_src;
static PyObject *__pyx_n_s_set_ttl;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_sport;
static PyObject *__pyx_n_s_src;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_targets;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_to_bytes;
static PyObject *__pyx_n_s_ttl;
static PyObject *__pyx_n_u_udp;
static PyObject *__pyx_n_u_wb;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_write_pcap_header;
static int __pyx_pf_10cloudtrace_5trace_5probe_5Probe___cinit__(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, uint16_t __pyx_v_size); /* proto */
static void __pyx_pf_10cloudtrace_5trace_5probe_5Probe_2__dealloc__(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_4create_ether(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_dst); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_6set_dst(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_dst); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_8set_src(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_src); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_10to_bytes(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_12set_ttl(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, uint8_t __pyx_v_ttl); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_14prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_16prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, uint8_t __pyx_v_num); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe___cinit__(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, CYTHON_UNUSED uint16_t __pyx_v_size); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_2set_icmp_id(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, uint16_t __pyx_v_pid); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_4prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_6prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, uint8_t __pyx_v_num); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe___cinit__(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_size); /* proto */
static void __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_2__dealloc__(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_4prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_6prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint8_t __pyx_v_num); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_8set_sport(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_sport); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_10set_dport(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_dport); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_12set_dst(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, PyObject *__pyx_v_dst); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_14set_src(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, PyObject *__pyx_v_src); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_proto_to_num(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_proto); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_2create_probe(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dst, uint8_t __pyx_v_ipproto, uint16_t __pyx_v_size, uint16_t __pyx_v_pid); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_4fill_packet_header(CYTHON_UNUSED PyObject *__pyx_self, uint16_t __pyx_v_size); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_6write_pcap_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8craft(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_targets, uint16_t __pyx_v_pid, PyObject *__pyx_v_file, int __pyx_v_minttl, int __pyx_v_maxttl, int __pyx_v_cycles, PyObject *__pyx_v_proto, uint16_t __pyx_v_size, int __pyx_v_randomize); /* proto */
static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_10ping(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_targets, uint16_t __pyx_v_pid, PyObject *__pyx_v_file, int __pyx_v_numpings, int __pyx_v_cycles, PyObject *__pyx_v_proto, uint16_t __pyx_v_size); /* proto */
static PyObject *__pyx_tp_new_10cloudtrace_5trace_5probe_Probe(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10cloudtrace_5trace_5probe_ICMPProbe(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10cloudtrace_5trace_5probe_UDPProbe(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_255;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__41;
static PyObject *__pyx_codeobj__43;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__49;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__55;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__59;
static PyObject *__pyx_codeobj__61;
static PyObject *__pyx_codeobj__63;
static PyObject *__pyx_codeobj__65;
static PyObject *__pyx_codeobj__67;
static PyObject *__pyx_codeobj__69;
/* Late includes */

/* "cloudtrace/trace/probe.pyx":12
 * from file2 import fopen
 * 
 * cdef bint minsize(int proto) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint16_t size = sizeof(ip)
 */

static int __pyx_f_10cloudtrace_5trace_5probe_minsize(int __pyx_v_proto) {
  uint16_t __pyx_v_size;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("minsize", 0);

  /* "cloudtrace/trace/probe.pyx":14
 * cdef bint minsize(int proto) except -1:
 *     cdef:
 *         uint16_t size = sizeof(ip)             # <<<<<<<<<<<<<<
 *     if proto == IPPROTO_UDP:
 *         return size + sizeof(udp) + 4
 */
  __pyx_v_size = (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ip));

  /* "cloudtrace/trace/probe.pyx":15
 *     cdef:
 *         uint16_t size = sizeof(ip)
 *     if proto == IPPROTO_UDP:             # <<<<<<<<<<<<<<
 *         return size + sizeof(udp) + 4
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_proto == IPPROTO_UDP) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/trace/probe.pyx":16
 *         uint16_t size = sizeof(ip)
 *     if proto == IPPROTO_UDP:
 *         return size + sizeof(udp) + 4             # <<<<<<<<<<<<<<
 *     else:
 *         return size + sizeof(icmp) + 2
 */
    __pyx_r = ((__pyx_v_size + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_udp))) + 4);
    goto __pyx_L0;

    /* "cloudtrace/trace/probe.pyx":15
 *     cdef:
 *         uint16_t size = sizeof(ip)
 *     if proto == IPPROTO_UDP:             # <<<<<<<<<<<<<<
 *         return size + sizeof(udp) + 4
 *     else:
 */
  }

  /* "cloudtrace/trace/probe.pyx":18
 *         return size + sizeof(udp) + 4
 *     else:
 *         return size + sizeof(icmp) + 2             # <<<<<<<<<<<<<<
 * 
 * cdef class Probe:
 */
  /*else*/ {
    __pyx_r = ((__pyx_v_size + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_icmp))) + 2);
    goto __pyx_L0;
  }

  /* "cloudtrace/trace/probe.pyx":12
 * from file2 import fopen
 * 
 * cdef bint minsize(int proto) except -1:             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint16_t size = sizeof(ip)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":21
 * 
 * cdef class Probe:
 *     def __cinit__(self, uint16_t size=64):             # <<<<<<<<<<<<<<
 *         self.buf = calloc(1, sizeof(ethhdr) + size)
 *         self.etherhdr = <ethhdr *>self.buf
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_5trace_5probe_5Probe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_5trace_5probe_5Probe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint16_t __pyx_v_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 21, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_uint16_t(values[0]); if (unlikely((__pyx_v_size == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 21, __pyx_L3_error)
    } else {
      __pyx_v_size = ((uint16_t)64);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 21, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe___cinit__(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self), __pyx_v_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_5trace_5probe_5Probe___cinit__(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, uint16_t __pyx_v_size) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/trace/probe.pyx":22
 * cdef class Probe:
 *     def __cinit__(self, uint16_t size=64):
 *         self.buf = calloc(1, sizeof(ethhdr) + size)             # <<<<<<<<<<<<<<
 *         self.etherhdr = <ethhdr *>self.buf
 *         self.iphdr = <ip *>(self.buf + sizeof(ethhdr))
 */
  __pyx_v_self->buf = calloc(1, ((sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr)) + __pyx_v_size));

  /* "cloudtrace/trace/probe.pyx":23
 *     def __cinit__(self, uint16_t size=64):
 *         self.buf = calloc(1, sizeof(ethhdr) + size)
 *         self.etherhdr = <ethhdr *>self.buf             # <<<<<<<<<<<<<<
 *         self.iphdr = <ip *>(self.buf + sizeof(ethhdr))
 *         self.iphdr.ip_hl = 0x45
 */
  __pyx_v_self->etherhdr = ((__pyx_t_10cloudtrace_5trace_5utils_ethhdr *)__pyx_v_self->buf);

  /* "cloudtrace/trace/probe.pyx":24
 *         self.buf = calloc(1, sizeof(ethhdr) + size)
 *         self.etherhdr = <ethhdr *>self.buf
 *         self.iphdr = <ip *>(self.buf + sizeof(ethhdr))             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_hl = 0x45
 *         self.iphdr.ip_off = htons(0x4000)
 */
  __pyx_v_self->iphdr = ((__pyx_t_10cloudtrace_5trace_5utils_ip *)(__pyx_v_self->buf + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr))));

  /* "cloudtrace/trace/probe.pyx":25
 *         self.etherhdr = <ethhdr *>self.buf
 *         self.iphdr = <ip *>(self.buf + sizeof(ethhdr))
 *         self.iphdr.ip_hl = 0x45             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_off = htons(0x4000)
 *         self.iphdr.ip_len = htons(size)
 */
  __pyx_v_self->iphdr->ip_hl = 0x45;

  /* "cloudtrace/trace/probe.pyx":26
 *         self.iphdr = <ip *>(self.buf + sizeof(ethhdr))
 *         self.iphdr.ip_hl = 0x45
 *         self.iphdr.ip_off = htons(0x4000)             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_len = htons(size)
 *         self.iphdr.ip_ttl = 64
 */
  __pyx_v_self->iphdr->ip_off = htons(0x4000);

  /* "cloudtrace/trace/probe.pyx":27
 *         self.iphdr.ip_hl = 0x45
 *         self.iphdr.ip_off = htons(0x4000)
 *         self.iphdr.ip_len = htons(size)             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_ttl = 64
 * 
 */
  __pyx_v_self->iphdr->ip_len = htons(__pyx_v_size);

  /* "cloudtrace/trace/probe.pyx":28
 *         self.iphdr.ip_off = htons(0x4000)
 *         self.iphdr.ip_len = htons(size)
 *         self.iphdr.ip_ttl = 64             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->iphdr->ip_ttl = 64;

  /* "cloudtrace/trace/probe.pyx":21
 * 
 * cdef class Probe:
 *     def __cinit__(self, uint16_t size=64):             # <<<<<<<<<<<<<<
 *         self.buf = calloc(1, sizeof(ethhdr) + size)
 *         self.etherhdr = <ethhdr *>self.buf
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":30
 *         self.iphdr.ip_ttl = 64
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.buf is not NULL:
 *             free(self.buf)
 */

/* Python wrapper */
static void __pyx_pw_10cloudtrace_5trace_5probe_5Probe_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10cloudtrace_5trace_5probe_5Probe_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10cloudtrace_5trace_5probe_5Probe_2__dealloc__(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10cloudtrace_5trace_5probe_5Probe_2__dealloc__(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "cloudtrace/trace/probe.pyx":31
 * 
 *     def __dealloc__(self):
 *         if self.buf is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.buf)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->buf != NULL) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/trace/probe.pyx":32
 *     def __dealloc__(self):
 *         if self.buf is not NULL:
 *             free(self.buf)             # <<<<<<<<<<<<<<
 * 
 *     cdef void copy_ether(self, void *ether):
 */
    free(__pyx_v_self->buf);

    /* "cloudtrace/trace/probe.pyx":31
 * 
 *     def __dealloc__(self):
 *         if self.buf is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.buf)
 * 
 */
  }

  /* "cloudtrace/trace/probe.pyx":30
 *         self.iphdr.ip_ttl = 64
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.buf is not NULL:
 *             free(self.buf)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/trace/probe.pyx":34
 *             free(self.buf)
 * 
 *     cdef void copy_ether(self, void *ether):             # <<<<<<<<<<<<<<
 *         memcpy(self.etherhdr, ether, sizeof(ethhdr))
 * 
 */

static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_copy_ether(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, void *__pyx_v_ether) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_ether", 0);

  /* "cloudtrace/trace/probe.pyx":35
 * 
 *     cdef void copy_ether(self, void *ether):
 *         memcpy(self.etherhdr, ether, sizeof(ethhdr))             # <<<<<<<<<<<<<<
 * 
 *     cpdef void create_ether(self, src, dst):
 */
  (void)(memcpy(__pyx_v_self->etherhdr, __pyx_v_ether, (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr))));

  /* "cloudtrace/trace/probe.pyx":34
 *             free(self.buf)
 * 
 *     cdef void copy_ether(self, void *ether):             # <<<<<<<<<<<<<<
 *         memcpy(self.etherhdr, ether, sizeof(ethhdr))
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/trace/probe.pyx":37
 *         memcpy(self.etherhdr, ether, sizeof(ethhdr))
 * 
 *     cpdef void create_ether(self, src, dst):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes ether
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_5create_ether(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_create_ether(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_dst, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_ether = 0;
  PyObject *__pyx_v_e = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  unsigned char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_ether", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_create_ether); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 37, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_5create_ether)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_src, __pyx_v_dst};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 37, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_src, __pyx_v_dst};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 37, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 37, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_src);
          __Pyx_GIVEREF(__pyx_v_src);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_src);
          __Pyx_INCREF(__pyx_v_dst);
          __Pyx_GIVEREF(__pyx_v_dst);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_dst);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 37, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":40
 *         cdef:
 *             bytes ether
 *         e = Ether(src=src, dst=dst) / sIP(dst='192.172.226.78') / sICMP()             # <<<<<<<<<<<<<<
 *         ether = bytes(e)[:sizeof(ethhdr)]
 *         self.copy_ether(<unsigned char *>ether)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Ether); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_src, __pyx_v_src) < 0) __PYX_ERR(1, 40, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_dst, __pyx_v_dst) < 0) __PYX_ERR(1, 40, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_sIP); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_dst, __pyx_kp_u_192_172_226_78) < 0) __PYX_ERR(1, 40, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_sICMP); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_6 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_e = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "cloudtrace/trace/probe.pyx":41
 *             bytes ether
 *         e = Ether(src=src, dst=dst) / sIP(dst='192.172.226.78') / sICMP()
 *         ether = bytes(e)[:sizeof(ethhdr)]             # <<<<<<<<<<<<<<
 *         self.copy_ether(<unsigned char *>ether)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyBytes_Type)), __pyx_v_e); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = PySequence_GetSlice(__pyx_t_3, 0, (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr))); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ether = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "cloudtrace/trace/probe.pyx":42
 *         e = Ether(src=src, dst=dst) / sIP(dst='192.172.226.78') / sICMP()
 *         ether = bytes(e)[:sizeof(ethhdr)]
 *         self.copy_ether(<unsigned char *>ether)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_dst(self, bytes dst) except *:
 */
  __pyx_t_7 = __Pyx_PyBytes_AsWritableUString(__pyx_v_ether); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(1, 42, __pyx_L1_error)
  ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self->__pyx_vtab)->copy_ether(__pyx_v_self, ((unsigned char *)__pyx_t_7));

  /* "cloudtrace/trace/probe.pyx":37
 *         memcpy(self.etherhdr, ether, sizeof(ethhdr))
 * 
 *     cpdef void create_ether(self, src, dst):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes ether
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.Probe.create_ether", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ether);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_5create_ether(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_5Probe_4create_ether[] = "Probe.create_ether(self, src, dst) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5Probe_5create_ether = {"create_ether", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_5create_ether, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_4create_ether};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_5create_ether(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_src = 0;
  PyObject *__pyx_v_dst = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_ether (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_src,&__pyx_n_s_dst,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_src)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dst)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_ether", 1, 2, 2, 1); __PYX_ERR(1, 37, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_ether") < 0)) __PYX_ERR(1, 37, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_src = values[0];
    __pyx_v_dst = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_ether", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 37, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.create_ether", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe_4create_ether(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self), __pyx_v_src, __pyx_v_dst);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_4create_ether(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_src, PyObject *__pyx_v_dst) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_ether", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_5Probe_create_ether(__pyx_v_self, __pyx_v_src, __pyx_v_dst, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.create_ether", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":44
 *         self.copy_ether(<unsigned char *>ether)
 * 
 *     cpdef void set_dst(self, bytes dst) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_7set_dst(PyObject *__pyx_v_self, PyObject *__pyx_v_dst); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_set_dst(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_dst, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_dst", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_dst); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 44, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_7set_dst)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_dst) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_dst);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 44, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":45
 * 
 *     cpdef void set_dst(self, bytes dst) except *:
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_src(self, bytes src) except *:
 */
  if (unlikely(__pyx_v_dst == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 45, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_dst); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(1, 45, __pyx_L1_error)
  (void)(inet_pton(AF_INET, __pyx_t_5, __pyx_v_self->iphdr->ip_dst));

  /* "cloudtrace/trace/probe.pyx":44
 *         self.copy_ether(<unsigned char *>ether)
 * 
 *     cpdef void set_dst(self, bytes dst) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.set_dst", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_7set_dst(PyObject *__pyx_v_self, PyObject *__pyx_v_dst); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_5Probe_6set_dst[] = "Probe.set_dst(self, bytes dst) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5Probe_7set_dst = {"set_dst", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_7set_dst, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_6set_dst};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_7set_dst(PyObject *__pyx_v_self, PyObject *__pyx_v_dst) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_dst (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dst), (&PyBytes_Type), 1, "dst", 1))) __PYX_ERR(1, 44, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe_6set_dst(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self), ((PyObject*)__pyx_v_dst));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_6set_dst(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_dst) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_dst", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_10cloudtrace_5trace_5probe_5Probe_set_dst(__pyx_v_self, __pyx_v_dst, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 44, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.set_dst", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":47
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 * 
 *     cpdef void set_src(self, bytes src) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_9set_src(PyObject *__pyx_v_self, PyObject *__pyx_v_src); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_set_src(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_src, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_src", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_src); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_9set_src)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_src) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_src);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 47, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":48
 * 
 *     cpdef void set_src(self, bytes src) except *:
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bytes to_bytes(self):
 */
  if (unlikely(__pyx_v_src == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 48, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_src); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(1, 48, __pyx_L1_error)
  (void)(inet_pton(AF_INET, __pyx_t_5, __pyx_v_self->iphdr->ip_src));

  /* "cloudtrace/trace/probe.pyx":47
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 * 
 *     cpdef void set_src(self, bytes src) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.set_src", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_9set_src(PyObject *__pyx_v_self, PyObject *__pyx_v_src); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_5Probe_8set_src[] = "Probe.set_src(self, bytes src) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5Probe_9set_src = {"set_src", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_9set_src, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_8set_src};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_9set_src(PyObject *__pyx_v_self, PyObject *__pyx_v_src) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_src (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_src), (&PyBytes_Type), 1, "src", 1))) __PYX_ERR(1, 47, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe_8set_src(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self), ((PyObject*)__pyx_v_src));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_8set_src(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, PyObject *__pyx_v_src) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_src", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_10cloudtrace_5trace_5probe_5Probe_set_src(__pyx_v_self, __pyx_v_src, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 47, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.set_src", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":50
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 * 
 *     cpdef bytes to_bytes(self):             # <<<<<<<<<<<<<<
 *         return <bytes>((<char *>self.buf)[:ntohs(self.iphdr.ip_len) + sizeof(ethhdr)])
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_11to_bytes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_10cloudtrace_5trace_5probe_5Probe_to_bytes(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_bytes", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_to_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_11to_bytes)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 50, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(1, 50, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":51
 * 
 *     cpdef bytes to_bytes(self):
 *         return <bytes>((<char *>self.buf)[:ntohs(self.iphdr.ip_len) + sizeof(ethhdr)])             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_ttl(self, uint8_t ttl) except *:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((char *)__pyx_v_self->buf) + 0, (ntohs(__pyx_v_self->iphdr->ip_len) + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr))) - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cloudtrace/trace/probe.pyx":50
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 * 
 *     cpdef bytes to_bytes(self):             # <<<<<<<<<<<<<<
 *         return <bytes>((<char *>self.buf)[:ntohs(self.iphdr.ip_len) + sizeof(ethhdr)])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.to_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_11to_bytes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_5Probe_10to_bytes[] = "Probe.to_bytes(self) -> bytes";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5Probe_11to_bytes = {"to_bytes", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_11to_bytes, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_10to_bytes};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_11to_bytes(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_bytes (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe_10to_bytes(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_10to_bytes(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_bytes", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10cloudtrace_5trace_5probe_5Probe_to_bytes(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.to_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":53
 *         return <bytes>((<char *>self.buf)[:ntohs(self.iphdr.ip_len) + sizeof(ethhdr)])
 * 
 *     cpdef void set_ttl(self, uint8_t ttl) except *:             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_ttl = ttl
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_13set_ttl(PyObject *__pyx_v_self, PyObject *__pyx_arg_ttl); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_set_ttl(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, uint8_t __pyx_v_ttl, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_ttl", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_ttl); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 53, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_13set_ttl)) {
        __pyx_t_3 = __Pyx_PyInt_From_uint8_t(__pyx_v_ttl); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 53, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 53, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":54
 * 
 *     cpdef void set_ttl(self, uint8_t ttl) except *:
 *         self.iphdr.ip_ttl = ttl             # <<<<<<<<<<<<<<
 * 
 *     cpdef void prepare(self):
 */
  __pyx_v_self->iphdr->ip_ttl = __pyx_v_ttl;

  /* "cloudtrace/trace/probe.pyx":53
 *         return <bytes>((<char *>self.buf)[:ntohs(self.iphdr.ip_len) + sizeof(ethhdr)])
 * 
 *     cpdef void set_ttl(self, uint8_t ttl) except *:             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_ttl = ttl
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.set_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_13set_ttl(PyObject *__pyx_v_self, PyObject *__pyx_arg_ttl); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_5Probe_12set_ttl[] = "Probe.set_ttl(self, uint8_t ttl) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5Probe_13set_ttl = {"set_ttl", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_13set_ttl, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_12set_ttl};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_13set_ttl(PyObject *__pyx_v_self, PyObject *__pyx_arg_ttl) {
  uint8_t __pyx_v_ttl;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_ttl (wrapper)", 0);
  assert(__pyx_arg_ttl); {
    __pyx_v_ttl = __Pyx_PyInt_As_uint8_t(__pyx_arg_ttl); if (unlikely((__pyx_v_ttl == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 53, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.set_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe_12set_ttl(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self), ((uint8_t)__pyx_v_ttl));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_12set_ttl(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, uint8_t __pyx_v_ttl) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_ttl", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_10cloudtrace_5trace_5probe_5Probe_set_ttl(__pyx_v_self, __pyx_v_ttl, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 53, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.set_ttl", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":56
 *         self.iphdr.ip_ttl = ttl
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_15prepare(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prepare", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_prepare); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 56, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_15prepare)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 56, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":57
 * 
 *     cpdef void prepare(self):
 *         self.prepare_ip()             # <<<<<<<<<<<<<<
 * 
 *     cdef void prepare_ip(self):
 */
  ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self->__pyx_vtab)->prepare_ip(__pyx_v_self);

  /* "cloudtrace/trace/probe.pyx":56
 *         self.iphdr.ip_ttl = ttl
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.Probe.prepare", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_15prepare(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_5Probe_14prepare[] = "Probe.prepare(self) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5Probe_15prepare = {"prepare", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_15prepare, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_14prepare};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_15prepare(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("prepare (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe_14prepare(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_14prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prepare", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_5Probe_prepare(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.prepare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":59
 *         self.prepare_ip()
 * 
 *     cdef void prepare_ip(self):             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_sum = 0
 *         self.iphdr.ip_sum = htons(checksum(self.iphdr, sizeof(ip)))
 */

static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_prepare_ip(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  uint16_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prepare_ip", 0);

  /* "cloudtrace/trace/probe.pyx":60
 * 
 *     cdef void prepare_ip(self):
 *         self.iphdr.ip_sum = 0             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_sum = htons(checksum(self.iphdr, sizeof(ip)))
 * 
 */
  __pyx_v_self->iphdr->ip_sum = 0;

  /* "cloudtrace/trace/probe.pyx":61
 *     cdef void prepare_ip(self):
 *         self.iphdr.ip_sum = 0
 *         self.iphdr.ip_sum = htons(checksum(self.iphdr, sizeof(ip)))             # <<<<<<<<<<<<<<
 * 
 *     cpdef void prep_trace(self, uint8_t num):
 */
  __pyx_t_1 = __pyx_f_10cloudtrace_5trace_5utils_checksum(__pyx_v_self->iphdr, (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ip))); if (unlikely(__pyx_t_1 == ((uint16_t)-1))) __PYX_ERR(1, 61, __pyx_L1_error)
  __pyx_v_self->iphdr->ip_sum = htons(__pyx_t_1);

  /* "cloudtrace/trace/probe.pyx":59
 *         self.prepare_ip()
 * 
 *     cdef void prepare_ip(self):             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_sum = 0
 *         self.iphdr.ip_sum = htons(checksum(self.iphdr, sizeof(ip)))
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.Probe.prepare_ip", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/trace/probe.pyx":63
 *         self.iphdr.ip_sum = htons(checksum(self.iphdr, sizeof(ip)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_17prep_trace(PyObject *__pyx_v_self, PyObject *__pyx_arg_num); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_5Probe_prep_trace(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, CYTHON_UNUSED uint8_t __pyx_v_num, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prep_trace", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_prep_trace); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 63, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_17prep_trace)) {
        __pyx_t_3 = __Pyx_PyInt_From_uint8_t(__pyx_v_num); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 63, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 63, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":64
 * 
 *     cpdef void prep_trace(self, uint8_t num):
 *         pass             # <<<<<<<<<<<<<<
 * 
 * cdef class ICMPProbe(Probe):
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.Probe.prep_trace", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/trace/probe.pyx":63
 *         self.iphdr.ip_sum = htons(checksum(self.iphdr, sizeof(ip)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_17prep_trace(PyObject *__pyx_v_self, PyObject *__pyx_arg_num); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_5Probe_16prep_trace[] = "Probe.prep_trace(self, uint8_t num) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5Probe_17prep_trace = {"prep_trace", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_17prep_trace, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_16prep_trace};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_17prep_trace(PyObject *__pyx_v_self, PyObject *__pyx_arg_num) {
  uint8_t __pyx_v_num;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("prep_trace (wrapper)", 0);
  assert(__pyx_arg_num); {
    __pyx_v_num = __Pyx_PyInt_As_uint8_t(__pyx_arg_num); if (unlikely((__pyx_v_num == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 63, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.prep_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe_16prep_trace(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self), ((uint8_t)__pyx_v_num));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_16prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, uint8_t __pyx_v_num) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prep_trace", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_5Probe_prep_trace(__pyx_v_self, __pyx_v_num, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.prep_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_5Probe_18__reduce_cython__[] = "Probe.__reduce_cython__(self)";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5Probe_19__reduce_cython__ = {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_19__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_18__reduce_cython__};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe_18__reduce_cython__(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_5Probe_20__setstate_cython__[] = "Probe.__setstate_cython__(self, __pyx_state)";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5Probe_21__setstate_cython__ = {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_21__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_20__setstate_cython__};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5Probe_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_5Probe_20__setstate_cython__(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_5Probe_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.Probe.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":67
 * 
 * cdef class ICMPProbe(Probe):
 *     def __cinit__(self, uint16_t size=1500):             # <<<<<<<<<<<<<<
 *         self.icmphdr = <icmp *>(self.buf + sizeof(ethhdr) + sizeof(ip))
 *         self.iphdr.ip_p = IPPROTO_ICMP
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED uint16_t __pyx_v_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 67, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_uint16_t(values[0]); if (unlikely((__pyx_v_size == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 67, __pyx_L3_error)
    } else {
      __pyx_v_size = ((uint16_t)0x5DC);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 67, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.ICMPProbe.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe___cinit__(((struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *)__pyx_v_self), __pyx_v_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe___cinit__(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, CYTHON_UNUSED uint16_t __pyx_v_size) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/trace/probe.pyx":68
 * cdef class ICMPProbe(Probe):
 *     def __cinit__(self, uint16_t size=1500):
 *         self.icmphdr = <icmp *>(self.buf + sizeof(ethhdr) + sizeof(ip))             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_p = IPPROTO_ICMP
 *         self.icmphdr.type = 8
 */
  __pyx_v_self->icmphdr = ((__pyx_t_10cloudtrace_5trace_5utils_icmp *)((__pyx_v_self->__pyx_base.buf + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr))) + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ip))));

  /* "cloudtrace/trace/probe.pyx":69
 *     def __cinit__(self, uint16_t size=1500):
 *         self.icmphdr = <icmp *>(self.buf + sizeof(ethhdr) + sizeof(ip))
 *         self.iphdr.ip_p = IPPROTO_ICMP             # <<<<<<<<<<<<<<
 *         self.icmphdr.type = 8
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(icmp)
 */
  __pyx_v_self->__pyx_base.iphdr->ip_p = IPPROTO_ICMP;

  /* "cloudtrace/trace/probe.pyx":70
 *         self.icmphdr = <icmp *>(self.buf + sizeof(ethhdr) + sizeof(ip))
 *         self.iphdr.ip_p = IPPROTO_ICMP
 *         self.icmphdr.type = 8             # <<<<<<<<<<<<<<
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(icmp)
 * 
 */
  __pyx_v_self->icmphdr->type = 8;

  /* "cloudtrace/trace/probe.pyx":71
 *         self.iphdr.ip_p = IPPROTO_ICMP
 *         self.icmphdr.type = 8
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(icmp)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_icmp_id(self, uint16_t pid):
 */
  __pyx_v_self->data = (((__pyx_v_self->__pyx_base.buf + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr))) + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ip))) + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_icmp)));

  /* "cloudtrace/trace/probe.pyx":67
 * 
 * cdef class ICMPProbe(Probe):
 *     def __cinit__(self, uint16_t size=1500):             # <<<<<<<<<<<<<<
 *         self.icmphdr = <icmp *>(self.buf + sizeof(ethhdr) + sizeof(ip))
 *         self.iphdr.ip_p = IPPROTO_ICMP
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":73
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(icmp)
 * 
 *     cpdef void set_icmp_id(self, uint16_t pid):             # <<<<<<<<<<<<<<
 *         self.icmphdr.un.echo.id = htons(pid)
 * 
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_3set_icmp_id(PyObject *__pyx_v_self, PyObject *__pyx_arg_pid); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_set_icmp_id(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, uint16_t __pyx_v_pid, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_icmp_id", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_icmp_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 73, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_3set_icmp_id)) {
        __pyx_t_3 = __Pyx_PyInt_From_uint16_t(__pyx_v_pid); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 73, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 73, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":74
 * 
 *     cpdef void set_icmp_id(self, uint16_t pid):
 *         self.icmphdr.un.echo.id = htons(pid)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void prepare(self):
 */
  __pyx_v_self->icmphdr->un.echo.id = htons(__pyx_v_pid);

  /* "cloudtrace/trace/probe.pyx":73
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(icmp)
 * 
 *     cpdef void set_icmp_id(self, uint16_t pid):             # <<<<<<<<<<<<<<
 *         self.icmphdr.un.echo.id = htons(pid)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.ICMPProbe.set_icmp_id", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_3set_icmp_id(PyObject *__pyx_v_self, PyObject *__pyx_arg_pid); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_2set_icmp_id[] = "ICMPProbe.set_icmp_id(self, uint16_t pid) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_3set_icmp_id = {"set_icmp_id", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_3set_icmp_id, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_2set_icmp_id};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_3set_icmp_id(PyObject *__pyx_v_self, PyObject *__pyx_arg_pid) {
  uint16_t __pyx_v_pid;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_icmp_id (wrapper)", 0);
  assert(__pyx_arg_pid); {
    __pyx_v_pid = __Pyx_PyInt_As_uint16_t(__pyx_arg_pid); if (unlikely((__pyx_v_pid == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 73, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.ICMPProbe.set_icmp_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_2set_icmp_id(((struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *)__pyx_v_self), ((uint16_t)__pyx_v_pid));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_2set_icmp_id(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, uint16_t __pyx_v_pid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_icmp_id", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_set_icmp_id(__pyx_v_self, __pyx_v_pid, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.ICMPProbe.set_icmp_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":76
 *         self.icmphdr.un.echo.id = htons(pid)
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 *         self.icmphdr.checksum = 0
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_5prepare(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prepare", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_prepare); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_5prepare)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 76, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":77
 * 
 *     cpdef void prepare(self):
 *         self.prepare_ip()             # <<<<<<<<<<<<<<
 *         self.icmphdr.checksum = 0
 *         self.icmphdr.checksum = htons(checksum(self.icmphdr, ntohs(self.iphdr.ip_len) - sizeof(ip)))
 */
  ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_ICMPProbe *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.prepare_ip(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self));

  /* "cloudtrace/trace/probe.pyx":78
 *     cpdef void prepare(self):
 *         self.prepare_ip()
 *         self.icmphdr.checksum = 0             # <<<<<<<<<<<<<<
 *         self.icmphdr.checksum = htons(checksum(self.icmphdr, ntohs(self.iphdr.ip_len) - sizeof(ip)))
 * 
 */
  __pyx_v_self->icmphdr->checksum = 0;

  /* "cloudtrace/trace/probe.pyx":79
 *         self.prepare_ip()
 *         self.icmphdr.checksum = 0
 *         self.icmphdr.checksum = htons(checksum(self.icmphdr, ntohs(self.iphdr.ip_len) - sizeof(ip)))             # <<<<<<<<<<<<<<
 * 
 *     cpdef void prep_trace(self, uint8_t num):
 */
  __pyx_t_5 = __pyx_f_10cloudtrace_5trace_5utils_checksum(__pyx_v_self->icmphdr, (ntohs(__pyx_v_self->__pyx_base.iphdr->ip_len) - (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ip)))); if (unlikely(__pyx_t_5 == ((uint16_t)-1))) __PYX_ERR(1, 79, __pyx_L1_error)
  __pyx_v_self->icmphdr->checksum = htons(__pyx_t_5);

  /* "cloudtrace/trace/probe.pyx":76
 *         self.icmphdr.un.echo.id = htons(pid)
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 *         self.icmphdr.checksum = 0
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.ICMPProbe.prepare", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_5prepare(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_4prepare[] = "ICMPProbe.prepare(self) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_5prepare = {"prepare", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_5prepare, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_4prepare};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_5prepare(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("prepare (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_4prepare(((struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_4prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prepare", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_prepare(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.ICMPProbe.prepare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":81
 *         self.icmphdr.checksum = htons(checksum(self.icmphdr, ntohs(self.iphdr.ip_len) - sizeof(ip)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint16_t data = 0x0608 - num
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_7prep_trace(PyObject *__pyx_v_self, PyObject *__pyx_arg_num); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, uint8_t __pyx_v_num, int __pyx_skip_dispatch) {
  uint16_t __pyx_v_data;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prep_trace", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_prep_trace); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 81, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_7prep_trace)) {
        __pyx_t_3 = __Pyx_PyInt_From_uint8_t(__pyx_v_num); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 81, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 81, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":83
 *     cpdef void prep_trace(self, uint8_t num):
 *         cdef:
 *             uint16_t data = 0x0608 - num             # <<<<<<<<<<<<<<
 *         self.icmphdr.un.echo.seq = htons(num)
 *         data = htons(data)
 */
  __pyx_v_data = (0x0608 - __pyx_v_num);

  /* "cloudtrace/trace/probe.pyx":84
 *         cdef:
 *             uint16_t data = 0x0608 - num
 *         self.icmphdr.un.echo.seq = htons(num)             # <<<<<<<<<<<<<<
 *         data = htons(data)
 *         memcpy(self.data, &data, 2)
 */
  __pyx_v_self->icmphdr->un.echo.seq = htons(__pyx_v_num);

  /* "cloudtrace/trace/probe.pyx":85
 *             uint16_t data = 0x0608 - num
 *         self.icmphdr.un.echo.seq = htons(num)
 *         data = htons(data)             # <<<<<<<<<<<<<<
 *         memcpy(self.data, &data, 2)
 * 
 */
  __pyx_v_data = htons(__pyx_v_data);

  /* "cloudtrace/trace/probe.pyx":86
 *         self.icmphdr.un.echo.seq = htons(num)
 *         data = htons(data)
 *         memcpy(self.data, &data, 2)             # <<<<<<<<<<<<<<
 * 
 * cdef class UDPProbe(Probe):
 */
  (void)(memcpy(__pyx_v_self->data, (&__pyx_v_data), 2));

  /* "cloudtrace/trace/probe.pyx":81
 *         self.icmphdr.checksum = htons(checksum(self.icmphdr, ntohs(self.iphdr.ip_len) - sizeof(ip)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint16_t data = 0x0608 - num
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.ICMPProbe.prep_trace", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_7prep_trace(PyObject *__pyx_v_self, PyObject *__pyx_arg_num); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_6prep_trace[] = "ICMPProbe.prep_trace(self, uint8_t num) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_7prep_trace = {"prep_trace", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_7prep_trace, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_6prep_trace};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_7prep_trace(PyObject *__pyx_v_self, PyObject *__pyx_arg_num) {
  uint8_t __pyx_v_num;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("prep_trace (wrapper)", 0);
  assert(__pyx_arg_num); {
    __pyx_v_num = __Pyx_PyInt_As_uint8_t(__pyx_arg_num); if (unlikely((__pyx_v_num == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 81, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.ICMPProbe.prep_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_6prep_trace(((struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *)__pyx_v_self), ((uint8_t)__pyx_v_num));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_6prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, uint8_t __pyx_v_num) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prep_trace", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_prep_trace(__pyx_v_self, __pyx_v_num, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.ICMPProbe.prep_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_8__reduce_cython__[] = "ICMPProbe.__reduce_cython__(self)";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_9__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_8__reduce_cython__};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_8__reduce_cython__(((struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.ICMPProbe.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_10__setstate_cython__[] = "ICMPProbe.__setstate_cython__(self, __pyx_state)";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_11__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_10__setstate_cython__};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_10__setstate_cython__(((struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_9ICMPProbe_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.ICMPProbe.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":89
 * 
 * cdef class UDPProbe(Probe):
 *     def __cinit__(self, uint16_t size=1500):             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint16_t length = htons(size - sizeof(ip))
 */

/* Python wrapper */
static int __pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  uint16_t __pyx_v_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 89, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_uint16_t(values[0]); if (unlikely((__pyx_v_size == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 89, __pyx_L3_error)
    } else {
      __pyx_v_size = ((uint16_t)0x5DC);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 89, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe___cinit__(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self), __pyx_v_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe___cinit__(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_size) {
  uint16_t __pyx_v_length;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  uint8_t __pyx_t_1;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "cloudtrace/trace/probe.pyx":91
 *     def __cinit__(self, uint16_t size=1500):
 *         cdef:
 *             uint16_t length = htons(size - sizeof(ip))             # <<<<<<<<<<<<<<
 *         self.udphdr = <udp *>(self.buf + sizeof(ethhdr) + sizeof(ip))
 *         self.iphdr.ip_p = IPPROTO_UDP
 */
  __pyx_v_length = htons((__pyx_v_size - (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ip))));

  /* "cloudtrace/trace/probe.pyx":92
 *         cdef:
 *             uint16_t length = htons(size - sizeof(ip))
 *         self.udphdr = <udp *>(self.buf + sizeof(ethhdr) + sizeof(ip))             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_p = IPPROTO_UDP
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(udp) + 2
 */
  __pyx_v_self->udphdr = ((__pyx_t_10cloudtrace_5trace_5utils_udp *)((__pyx_v_self->__pyx_base.buf + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr))) + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ip))));

  /* "cloudtrace/trace/probe.pyx":93
 *             uint16_t length = htons(size - sizeof(ip))
 *         self.udphdr = <udp *>(self.buf + sizeof(ethhdr) + sizeof(ip))
 *         self.iphdr.ip_p = IPPROTO_UDP             # <<<<<<<<<<<<<<
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(udp) + 2
 *         self.pseudo = <udppseudo *>calloc(1, sizeof(udppseudo))
 */
  __pyx_v_self->__pyx_base.iphdr->ip_p = IPPROTO_UDP;

  /* "cloudtrace/trace/probe.pyx":94
 *         self.udphdr = <udp *>(self.buf + sizeof(ethhdr) + sizeof(ip))
 *         self.iphdr.ip_p = IPPROTO_UDP
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(udp) + 2             # <<<<<<<<<<<<<<
 *         self.pseudo = <udppseudo *>calloc(1, sizeof(udppseudo))
 *         self.pseudo.proto = self.iphdr.ip_p
 */
  __pyx_v_self->data = ((((__pyx_v_self->__pyx_base.buf + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr))) + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ip))) + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_udp))) + 2);

  /* "cloudtrace/trace/probe.pyx":95
 *         self.iphdr.ip_p = IPPROTO_UDP
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(udp) + 2
 *         self.pseudo = <udppseudo *>calloc(1, sizeof(udppseudo))             # <<<<<<<<<<<<<<
 *         self.pseudo.proto = self.iphdr.ip_p
 *         self.udphdr.len = length
 */
  __pyx_v_self->pseudo = ((__pyx_t_10cloudtrace_5trace_5utils_udppseudo *)calloc(1, (sizeof(__pyx_t_10cloudtrace_5trace_5utils_udppseudo))));

  /* "cloudtrace/trace/probe.pyx":96
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(udp) + 2
 *         self.pseudo = <udppseudo *>calloc(1, sizeof(udppseudo))
 *         self.pseudo.proto = self.iphdr.ip_p             # <<<<<<<<<<<<<<
 *         self.udphdr.len = length
 *         self.pseudo.length = length
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.iphdr->ip_p;
  __pyx_v_self->pseudo->proto = __pyx_t_1;

  /* "cloudtrace/trace/probe.pyx":97
 *         self.pseudo = <udppseudo *>calloc(1, sizeof(udppseudo))
 *         self.pseudo.proto = self.iphdr.ip_p
 *         self.udphdr.len = length             # <<<<<<<<<<<<<<
 *         self.pseudo.length = length
 *         self.pseudo.length2 = length
 */
  __pyx_v_self->udphdr->len = __pyx_v_length;

  /* "cloudtrace/trace/probe.pyx":98
 *         self.pseudo.proto = self.iphdr.ip_p
 *         self.udphdr.len = length
 *         self.pseudo.length = length             # <<<<<<<<<<<<<<
 *         self.pseudo.length2 = length
 * 
 */
  __pyx_v_self->pseudo->length = __pyx_v_length;

  /* "cloudtrace/trace/probe.pyx":99
 *         self.udphdr.len = length
 *         self.pseudo.length = length
 *         self.pseudo.length2 = length             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->pseudo->length2 = __pyx_v_length;

  /* "cloudtrace/trace/probe.pyx":89
 * 
 * cdef class UDPProbe(Probe):
 *     def __cinit__(self, uint16_t size=1500):             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint16_t length = htons(size - sizeof(ip))
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":101
 *         self.pseudo.length2 = length
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.pseudo is not NULL:
 *             free(self.pseudo)
 */

/* Python wrapper */
static void __pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_2__dealloc__(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_2__dealloc__(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "cloudtrace/trace/probe.pyx":102
 * 
 *     def __dealloc__(self):
 *         if self.pseudo is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.pseudo)
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->pseudo != NULL) != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/trace/probe.pyx":103
 *     def __dealloc__(self):
 *         if self.pseudo is not NULL:
 *             free(self.pseudo)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void prepare(self):
 */
    free(__pyx_v_self->pseudo);

    /* "cloudtrace/trace/probe.pyx":102
 * 
 *     def __dealloc__(self):
 *         if self.pseudo is not NULL:             # <<<<<<<<<<<<<<
 *             free(self.pseudo)
 * 
 */
  }

  /* "cloudtrace/trace/probe.pyx":101
 *         self.pseudo.length2 = length
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self.pseudo is not NULL:
 *             free(self.pseudo)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "cloudtrace/trace/probe.pyx":105
 *             free(self.pseudo)
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 *         self.udphdr.check = 0
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_5prepare(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  uint16_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prepare", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_prepare); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 105, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_5prepare)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 105, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":106
 * 
 *     cpdef void prepare(self):
 *         self.prepare_ip()             # <<<<<<<<<<<<<<
 *         self.udphdr.check = 0
 *         self.udphdr.check = htons(checksum(self.pseudo, sizeof(udppseudo)))
 */
  ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.prepare_ip(((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_self));

  /* "cloudtrace/trace/probe.pyx":107
 *     cpdef void prepare(self):
 *         self.prepare_ip()
 *         self.udphdr.check = 0             # <<<<<<<<<<<<<<
 *         self.udphdr.check = htons(checksum(self.pseudo, sizeof(udppseudo)))
 * 
 */
  __pyx_v_self->udphdr->check = 0;

  /* "cloudtrace/trace/probe.pyx":108
 *         self.prepare_ip()
 *         self.udphdr.check = 0
 *         self.udphdr.check = htons(checksum(self.pseudo, sizeof(udppseudo)))             # <<<<<<<<<<<<<<
 * 
 *     cpdef void prep_trace(self, uint8_t num):
 */
  __pyx_t_5 = __pyx_f_10cloudtrace_5trace_5utils_checksum(__pyx_v_self->pseudo, (sizeof(__pyx_t_10cloudtrace_5trace_5utils_udppseudo))); if (unlikely(__pyx_t_5 == ((uint16_t)-1))) __PYX_ERR(1, 108, __pyx_L1_error)
  __pyx_v_self->udphdr->check = htons(__pyx_t_5);

  /* "cloudtrace/trace/probe.pyx":105
 *             free(self.pseudo)
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 *         self.udphdr.check = 0
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.UDPProbe.prepare", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_5prepare(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_4prepare[] = "UDPProbe.prepare(self) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_5prepare = {"prepare", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_5prepare, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_4prepare};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_5prepare(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("prepare (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_4prepare(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_4prepare(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prepare", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_prepare(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.prepare", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":110
 *         self.udphdr.check = htons(checksum(self.pseudo, sizeof(udppseudo)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         cdef:
 *             unsigned short data, chksum
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_7prep_trace(PyObject *__pyx_v_self, PyObject *__pyx_arg_num); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint8_t __pyx_v_num, int __pyx_skip_dispatch) {
  unsigned short __pyx_v_data;
  unsigned short __pyx_v_chksum;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  uint16_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prep_trace", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_prep_trace); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_7prep_trace)) {
        __pyx_t_3 = __Pyx_PyInt_From_uint8_t(__pyx_v_num); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 110, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 110, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":114
 *             unsigned short data, chksum
 *             unsigned char[2] test
 *         chksum = checksum(self.pseudo, 20)             # <<<<<<<<<<<<<<
 *         data = htons(chksum - num)
 *         memcpy(self.data, &data, 2)
 */
  __pyx_t_6 = __pyx_f_10cloudtrace_5trace_5utils_checksum(__pyx_v_self->pseudo, 20); if (unlikely(__pyx_t_6 == ((uint16_t)-1))) __PYX_ERR(1, 114, __pyx_L1_error)
  __pyx_v_chksum = __pyx_t_6;

  /* "cloudtrace/trace/probe.pyx":115
 *             unsigned char[2] test
 *         chksum = checksum(self.pseudo, 20)
 *         data = htons(chksum - num)             # <<<<<<<<<<<<<<
 *         memcpy(self.data, &data, 2)
 *         memcpy(self.pseudo.data, &data, 2)
 */
  __pyx_v_data = htons((__pyx_v_chksum - __pyx_v_num));

  /* "cloudtrace/trace/probe.pyx":116
 *         chksum = checksum(self.pseudo, 20)
 *         data = htons(chksum - num)
 *         memcpy(self.data, &data, 2)             # <<<<<<<<<<<<<<
 *         memcpy(self.pseudo.data, &data, 2)
 * 
 */
  (void)(memcpy(__pyx_v_self->data, (&__pyx_v_data), 2));

  /* "cloudtrace/trace/probe.pyx":117
 *         data = htons(chksum - num)
 *         memcpy(self.data, &data, 2)
 *         memcpy(self.pseudo.data, &data, 2)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_sport(self, uint16_t sport):
 */
  (void)(memcpy(__pyx_v_self->pseudo->data, (&__pyx_v_data), 2));

  /* "cloudtrace/trace/probe.pyx":110
 *         self.udphdr.check = htons(checksum(self.pseudo, sizeof(udppseudo)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         cdef:
 *             unsigned short data, chksum
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.UDPProbe.prep_trace", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_7prep_trace(PyObject *__pyx_v_self, PyObject *__pyx_arg_num); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_6prep_trace[] = "UDPProbe.prep_trace(self, uint8_t num) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_7prep_trace = {"prep_trace", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_7prep_trace, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_6prep_trace};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_7prep_trace(PyObject *__pyx_v_self, PyObject *__pyx_arg_num) {
  uint8_t __pyx_v_num;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("prep_trace (wrapper)", 0);
  assert(__pyx_arg_num); {
    __pyx_v_num = __Pyx_PyInt_As_uint8_t(__pyx_arg_num); if (unlikely((__pyx_v_num == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 110, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.prep_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_6prep_trace(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self), ((uint8_t)__pyx_v_num));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_6prep_trace(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint8_t __pyx_v_num) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("prep_trace", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_prep_trace(__pyx_v_self, __pyx_v_num, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.prep_trace", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":119
 *         memcpy(self.pseudo.data, &data, 2)
 * 
 *     cpdef void set_sport(self, uint16_t sport):             # <<<<<<<<<<<<<<
 *         self.udphdr.source = htons(sport)
 *         self.pseudo.sport = htons(sport)
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_9set_sport(PyObject *__pyx_v_self, PyObject *__pyx_arg_sport); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_sport(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_sport, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_sport", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_sport); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_9set_sport)) {
        __pyx_t_3 = __Pyx_PyInt_From_uint16_t(__pyx_v_sport); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 119, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":120
 * 
 *     cpdef void set_sport(self, uint16_t sport):
 *         self.udphdr.source = htons(sport)             # <<<<<<<<<<<<<<
 *         self.pseudo.sport = htons(sport)
 * 
 */
  __pyx_v_self->udphdr->source = htons(__pyx_v_sport);

  /* "cloudtrace/trace/probe.pyx":121
 *     cpdef void set_sport(self, uint16_t sport):
 *         self.udphdr.source = htons(sport)
 *         self.pseudo.sport = htons(sport)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_dport(self, uint16_t dport):
 */
  __pyx_v_self->pseudo->sport = htons(__pyx_v_sport);

  /* "cloudtrace/trace/probe.pyx":119
 *         memcpy(self.pseudo.data, &data, 2)
 * 
 *     cpdef void set_sport(self, uint16_t sport):             # <<<<<<<<<<<<<<
 *         self.udphdr.source = htons(sport)
 *         self.pseudo.sport = htons(sport)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.UDPProbe.set_sport", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_9set_sport(PyObject *__pyx_v_self, PyObject *__pyx_arg_sport); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_8set_sport[] = "UDPProbe.set_sport(self, uint16_t sport) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_9set_sport = {"set_sport", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_9set_sport, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_8set_sport};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_9set_sport(PyObject *__pyx_v_self, PyObject *__pyx_arg_sport) {
  uint16_t __pyx_v_sport;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_sport (wrapper)", 0);
  assert(__pyx_arg_sport); {
    __pyx_v_sport = __Pyx_PyInt_As_uint16_t(__pyx_arg_sport); if (unlikely((__pyx_v_sport == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 119, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.set_sport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_8set_sport(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self), ((uint16_t)__pyx_v_sport));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_8set_sport(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_sport) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_sport", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_sport(__pyx_v_self, __pyx_v_sport, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.set_sport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":123
 *         self.pseudo.sport = htons(sport)
 * 
 *     cpdef void set_dport(self, uint16_t dport):             # <<<<<<<<<<<<<<
 *         self.udphdr.dest = htons(dport)
 *         self.pseudo.dport = htons(dport)
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_11set_dport(PyObject *__pyx_v_self, PyObject *__pyx_arg_dport); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_dport(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_dport, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_dport", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_dport); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_11set_dport)) {
        __pyx_t_3 = __Pyx_PyInt_From_uint16_t(__pyx_v_dport); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 123, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 123, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":124
 * 
 *     cpdef void set_dport(self, uint16_t dport):
 *         self.udphdr.dest = htons(dport)             # <<<<<<<<<<<<<<
 *         self.pseudo.dport = htons(dport)
 * 
 */
  __pyx_v_self->udphdr->dest = htons(__pyx_v_dport);

  /* "cloudtrace/trace/probe.pyx":125
 *     cpdef void set_dport(self, uint16_t dport):
 *         self.udphdr.dest = htons(dport)
 *         self.pseudo.dport = htons(dport)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_dst(self, bytes dst) except *:
 */
  __pyx_v_self->pseudo->dport = htons(__pyx_v_dport);

  /* "cloudtrace/trace/probe.pyx":123
 *         self.pseudo.sport = htons(sport)
 * 
 *     cpdef void set_dport(self, uint16_t dport):             # <<<<<<<<<<<<<<
 *         self.udphdr.dest = htons(dport)
 *         self.pseudo.dport = htons(dport)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.UDPProbe.set_dport", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_11set_dport(PyObject *__pyx_v_self, PyObject *__pyx_arg_dport); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_10set_dport[] = "UDPProbe.set_dport(self, uint16_t dport) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_11set_dport = {"set_dport", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_11set_dport, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_10set_dport};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_11set_dport(PyObject *__pyx_v_self, PyObject *__pyx_arg_dport) {
  uint16_t __pyx_v_dport;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_dport (wrapper)", 0);
  assert(__pyx_arg_dport); {
    __pyx_v_dport = __Pyx_PyInt_As_uint16_t(__pyx_arg_dport); if (unlikely((__pyx_v_dport == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 123, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.set_dport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_10set_dport(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self), ((uint16_t)__pyx_v_dport));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_10set_dport(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, uint16_t __pyx_v_dport) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_dport", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_dport(__pyx_v_self, __pyx_v_dport, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.set_dport", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":127
 *         self.pseudo.dport = htons(dport)
 * 
 *     cpdef void set_dst(self, bytes dst) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_13set_dst(PyObject *__pyx_v_self, PyObject *__pyx_v_dst); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_dst(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, PyObject *__pyx_v_dst, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_dst", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_dst); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_13set_dst)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_dst) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_dst);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 127, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":128
 * 
 *     cpdef void set_dst(self, bytes dst) except *:
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)             # <<<<<<<<<<<<<<
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)
 * 
 */
  if (unlikely(__pyx_v_dst == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 128, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_dst); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(1, 128, __pyx_L1_error)
  (void)(inet_pton(AF_INET, __pyx_t_5, __pyx_v_self->__pyx_base.iphdr->ip_dst));

  /* "cloudtrace/trace/probe.pyx":129
 *     cpdef void set_dst(self, bytes dst) except *:
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_src(self, bytes src) except *:
 */
  (void)(memcpy(__pyx_v_self->pseudo->dst, __pyx_v_self->__pyx_base.iphdr->ip_dst, 4));

  /* "cloudtrace/trace/probe.pyx":127
 *         self.pseudo.dport = htons(dport)
 * 
 *     cpdef void set_dst(self, bytes dst) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.set_dst", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_13set_dst(PyObject *__pyx_v_self, PyObject *__pyx_v_dst); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_12set_dst[] = "UDPProbe.set_dst(self, bytes dst) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_13set_dst = {"set_dst", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_13set_dst, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_12set_dst};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_13set_dst(PyObject *__pyx_v_self, PyObject *__pyx_v_dst) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_dst (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dst), (&PyBytes_Type), 1, "dst", 1))) __PYX_ERR(1, 127, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_12set_dst(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self), ((PyObject*)__pyx_v_dst));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_12set_dst(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, PyObject *__pyx_v_dst) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_dst", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_dst(__pyx_v_self, __pyx_v_dst, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 127, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.set_dst", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":131
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)
 * 
 *     cpdef void set_src(self, bytes src) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_15set_src(PyObject *__pyx_v_self, PyObject *__pyx_v_src); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_src(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, PyObject *__pyx_v_src, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  char *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_src", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_src); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_15set_src)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_src) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_src);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 131, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "cloudtrace/trace/probe.pyx":132
 * 
 *     cpdef void set_src(self, bytes src) except *:
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)             # <<<<<<<<<<<<<<
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)
 * 
 */
  if (unlikely(__pyx_v_src == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 132, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_src); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(1, 132, __pyx_L1_error)
  (void)(inet_pton(AF_INET, __pyx_t_5, __pyx_v_self->__pyx_base.iphdr->ip_src));

  /* "cloudtrace/trace/probe.pyx":133
 *     cpdef void set_src(self, bytes src) except *:
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)             # <<<<<<<<<<<<<<
 * 
 * cpdef uint8_t proto_to_num(str proto):
 */
  (void)(memcpy(__pyx_v_self->pseudo->src, __pyx_v_self->__pyx_base.iphdr->ip_src, 4));

  /* "cloudtrace/trace/probe.pyx":131
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)
 * 
 *     cpdef void set_src(self, bytes src) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.set_src", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_15set_src(PyObject *__pyx_v_self, PyObject *__pyx_v_src); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_14set_src[] = "UDPProbe.set_src(self, bytes src) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_15set_src = {"set_src", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_15set_src, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_14set_src};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_15set_src(PyObject *__pyx_v_self, PyObject *__pyx_v_src) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_src (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_src), (&PyBytes_Type), 1, "src", 1))) __PYX_ERR(1, 131, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_14set_src(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self), ((PyObject*)__pyx_v_src));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_14set_src(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, PyObject *__pyx_v_src) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_src", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_src(__pyx_v_self, __pyx_v_src, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 131, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.set_src", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_16__reduce_cython__[] = "UDPProbe.__reduce_cython__(self)";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_17__reduce_cython__ = {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_17__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_16__reduce_cython__};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_16__reduce_cython__(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_16__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_18__setstate_cython__[] = "UDPProbe.__setstate_cython__(self, __pyx_state)";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_19__setstate_cython__ = {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_19__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_18__setstate_cython__};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_18__setstate_cython__(((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8UDPProbe_18__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.UDPProbe.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":135
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)
 * 
 * cpdef uint8_t proto_to_num(str proto):             # <<<<<<<<<<<<<<
 *     if proto == 'icmp':
 *         return IPPROTO_ICMP
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_1proto_to_num(PyObject *__pyx_self, PyObject *__pyx_v_proto); /*proto*/
static uint8_t __pyx_f_10cloudtrace_5trace_5probe_proto_to_num(PyObject *__pyx_v_proto, CYTHON_UNUSED int __pyx_skip_dispatch) {
  uint8_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("proto_to_num", 0);

  /* "cloudtrace/trace/probe.pyx":136
 * 
 * cpdef uint8_t proto_to_num(str proto):
 *     if proto == 'icmp':             # <<<<<<<<<<<<<<
 *         return IPPROTO_ICMP
 *     elif proto == 'udp':
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_proto, __pyx_n_u_icmp, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 136, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "cloudtrace/trace/probe.pyx":137
 * cpdef uint8_t proto_to_num(str proto):
 *     if proto == 'icmp':
 *         return IPPROTO_ICMP             # <<<<<<<<<<<<<<
 *     elif proto == 'udp':
 *         return IPPROTO_UDP
 */
    __pyx_r = IPPROTO_ICMP;
    goto __pyx_L0;

    /* "cloudtrace/trace/probe.pyx":136
 * 
 * cpdef uint8_t proto_to_num(str proto):
 *     if proto == 'icmp':             # <<<<<<<<<<<<<<
 *         return IPPROTO_ICMP
 *     elif proto == 'udp':
 */
  }

  /* "cloudtrace/trace/probe.pyx":138
 *     if proto == 'icmp':
 *         return IPPROTO_ICMP
 *     elif proto == 'udp':             # <<<<<<<<<<<<<<
 *         return IPPROTO_UDP
 *     else:
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_proto, __pyx_n_u_udp, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 138, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (likely(__pyx_t_1)) {

    /* "cloudtrace/trace/probe.pyx":139
 *         return IPPROTO_ICMP
 *     elif proto == 'udp':
 *         return IPPROTO_UDP             # <<<<<<<<<<<<<<
 *     else:
 *         raise NotImplementedError()
 */
    __pyx_r = IPPROTO_UDP;
    goto __pyx_L0;

    /* "cloudtrace/trace/probe.pyx":138
 *     if proto == 'icmp':
 *         return IPPROTO_ICMP
 *     elif proto == 'udp':             # <<<<<<<<<<<<<<
 *         return IPPROTO_UDP
 *     else:
 */
  }

  /* "cloudtrace/trace/probe.pyx":141
 *         return IPPROTO_UDP
 *     else:
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 * cpdef Probe create_probe(str dst, uint8_t ipproto, uint16_t size, uint16_t pid):
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 141, __pyx_L1_error)
  }

  /* "cloudtrace/trace/probe.pyx":135
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)
 * 
 * cpdef uint8_t proto_to_num(str proto):             # <<<<<<<<<<<<<<
 *     if proto == 'icmp':
 *         return IPPROTO_ICMP
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.proto_to_num", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_1proto_to_num(PyObject *__pyx_self, PyObject *__pyx_v_proto); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_proto_to_num[] = "proto_to_num(unicode proto) -> uint8_t";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_1proto_to_num = {"proto_to_num", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_1proto_to_num, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_proto_to_num};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_1proto_to_num(PyObject *__pyx_self, PyObject *__pyx_v_proto) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("proto_to_num (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_proto), (&PyUnicode_Type), 1, "proto", 1))) __PYX_ERR(1, 135, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_proto_to_num(__pyx_self, ((PyObject*)__pyx_v_proto));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_proto_to_num(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_proto) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("proto_to_num", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint8_t(__pyx_f_10cloudtrace_5trace_5probe_proto_to_num(__pyx_v_proto, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.proto_to_num", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":143
 *         raise NotImplementedError()
 * 
 * cpdef Probe create_probe(str dst, uint8_t ipproto, uint16_t size, uint16_t pid):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str iface, src, nexthop, macdst, srcmac
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_3create_probe(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_f_10cloudtrace_5trace_5probe_create_probe(PyObject *__pyx_v_dst, uint8_t __pyx_v_ipproto, uint16_t __pyx_v_size, uint16_t __pyx_v_pid, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_iface = 0;
  PyObject *__pyx_v_src = 0;
  PyObject *__pyx_v_nexthop = 0;
  PyObject *__pyx_v_macdst = 0;
  PyObject *__pyx_v_srcmac = 0;
  struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_probe = 0;
  struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *__pyx_v_iprobe = 0;
  struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *__pyx_v_uprobe = 0;
  struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_probe", 0);

  /* "cloudtrace/trace/probe.pyx":149
 *         ICMPProbe iprobe
 *         UDPProbe uprobe
 *     iface, src, nexthop = conf.route.route(dst)             # <<<<<<<<<<<<<<
 *     macdst = getmacbyip(nexthop)
 *     srcmac = get_if_hwaddr(iface)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_conf); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_route); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_route); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_dst) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_dst);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 149, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 3) < 0) __PYX_ERR(1, 149, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(1, 149, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(1, 149, __pyx_L1_error)
  if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(1, 149, __pyx_L1_error)
  if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 149, __pyx_L1_error)
  __pyx_v_iface = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_v_src = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_v_nexthop = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "cloudtrace/trace/probe.pyx":150
 *         UDPProbe uprobe
 *     iface, src, nexthop = conf.route.route(dst)
 *     macdst = getmacbyip(nexthop)             # <<<<<<<<<<<<<<
 *     srcmac = get_if_hwaddr(iface)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_getmacbyip); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_nexthop) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_nexthop);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 150, __pyx_L1_error)
  __pyx_v_macdst = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/trace/probe.pyx":151
 *     iface, src, nexthop = conf.route.route(dst)
 *     macdst = getmacbyip(nexthop)
 *     srcmac = get_if_hwaddr(iface)             # <<<<<<<<<<<<<<
 * 
 *     if ipproto == IPPROTO_ICMP:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_get_if_hwaddr); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_iface) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_iface);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 151, __pyx_L1_error)
  __pyx_v_srcmac = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cloudtrace/trace/probe.pyx":153
 *     srcmac = get_if_hwaddr(iface)
 * 
 *     if ipproto == IPPROTO_ICMP:             # <<<<<<<<<<<<<<
 *         iprobe = ICMPProbe(size)
 *         iprobe.set_icmp_id(pid)
 */
  __pyx_t_7 = ((__pyx_v_ipproto == IPPROTO_ICMP) != 0);
  if (__pyx_t_7) {

    /* "cloudtrace/trace/probe.pyx":154
 * 
 *     if ipproto == IPPROTO_ICMP:
 *         iprobe = ICMPProbe(size)             # <<<<<<<<<<<<<<
 *         iprobe.set_icmp_id(pid)
 *         probe = iprobe
 */
    __pyx_t_1 = __Pyx_PyInt_From_uint16_t(__pyx_v_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_ICMPProbe), __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_iprobe = ((struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "cloudtrace/trace/probe.pyx":155
 *     if ipproto == IPPROTO_ICMP:
 *         iprobe = ICMPProbe(size)
 *         iprobe.set_icmp_id(pid)             # <<<<<<<<<<<<<<
 *         probe = iprobe
 *     elif ipproto == IPPROTO_UDP:
 */
    ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_ICMPProbe *)__pyx_v_iprobe->__pyx_base.__pyx_vtab)->set_icmp_id(__pyx_v_iprobe, __pyx_v_pid, 0);

    /* "cloudtrace/trace/probe.pyx":156
 *         iprobe = ICMPProbe(size)
 *         iprobe.set_icmp_id(pid)
 *         probe = iprobe             # <<<<<<<<<<<<<<
 *     elif ipproto == IPPROTO_UDP:
 *         uprobe = UDPProbe(size)
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_iprobe));
    __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_iprobe);

    /* "cloudtrace/trace/probe.pyx":153
 *     srcmac = get_if_hwaddr(iface)
 * 
 *     if ipproto == IPPROTO_ICMP:             # <<<<<<<<<<<<<<
 *         iprobe = ICMPProbe(size)
 *         iprobe.set_icmp_id(pid)
 */
    goto __pyx_L5;
  }

  /* "cloudtrace/trace/probe.pyx":157
 *         iprobe.set_icmp_id(pid)
 *         probe = iprobe
 *     elif ipproto == IPPROTO_UDP:             # <<<<<<<<<<<<<<
 *         uprobe = UDPProbe(size)
 *         uprobe.set_sport(pid)
 */
  __pyx_t_7 = ((__pyx_v_ipproto == IPPROTO_UDP) != 0);
  if (likely(__pyx_t_7)) {

    /* "cloudtrace/trace/probe.pyx":158
 *         probe = iprobe
 *     elif ipproto == IPPROTO_UDP:
 *         uprobe = UDPProbe(size)             # <<<<<<<<<<<<<<
 *         uprobe.set_sport(pid)
 *         uprobe.set_dport(30000)
 */
    __pyx_t_4 = __Pyx_PyInt_From_uint16_t(__pyx_v_size); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe), __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_uprobe = ((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "cloudtrace/trace/probe.pyx":159
 *     elif ipproto == IPPROTO_UDP:
 *         uprobe = UDPProbe(size)
 *         uprobe.set_sport(pid)             # <<<<<<<<<<<<<<
 *         uprobe.set_dport(30000)
 *         probe = uprobe
 */
    ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_uprobe->__pyx_base.__pyx_vtab)->set_sport(__pyx_v_uprobe, __pyx_v_pid, 0);

    /* "cloudtrace/trace/probe.pyx":160
 *         uprobe = UDPProbe(size)
 *         uprobe.set_sport(pid)
 *         uprobe.set_dport(30000)             # <<<<<<<<<<<<<<
 *         probe = uprobe
 *     else:
 */
    ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_UDPProbe *)__pyx_v_uprobe->__pyx_base.__pyx_vtab)->set_dport(__pyx_v_uprobe, 0x7530, 0);

    /* "cloudtrace/trace/probe.pyx":161
 *         uprobe.set_sport(pid)
 *         uprobe.set_dport(30000)
 *         probe = uprobe             # <<<<<<<<<<<<<<
 *     else:
 *         raise Exception('No TCP!')
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_uprobe));
    __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_v_uprobe);

    /* "cloudtrace/trace/probe.pyx":157
 *         iprobe.set_icmp_id(pid)
 *         probe = iprobe
 *     elif ipproto == IPPROTO_UDP:             # <<<<<<<<<<<<<<
 *         uprobe = UDPProbe(size)
 *         uprobe.set_sport(pid)
 */
    goto __pyx_L5;
  }

  /* "cloudtrace/trace/probe.pyx":163
 *         probe = uprobe
 *     else:
 *         raise Exception('No TCP!')             # <<<<<<<<<<<<<<
 *     probe.create_ether(srcmac, macdst)
 *     probe.set_src(src.encode())
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 163, __pyx_L1_error)
  }
  __pyx_L5:;

  /* "cloudtrace/trace/probe.pyx":164
 *     else:
 *         raise Exception('No TCP!')
 *     probe.create_ether(srcmac, macdst)             # <<<<<<<<<<<<<<
 *     probe.set_src(src.encode())
 *     return probe
 */
  ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->create_ether(__pyx_v_probe, __pyx_v_srcmac, __pyx_v_macdst, 0);

  /* "cloudtrace/trace/probe.pyx":165
 *         raise Exception('No TCP!')
 *     probe.create_ether(srcmac, macdst)
 *     probe.set_src(src.encode())             # <<<<<<<<<<<<<<
 *     return probe
 * 
 */
  if (unlikely(__pyx_v_src == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(1, 165, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_src, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->set_src(__pyx_v_probe, ((PyObject*)__pyx_t_1), 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/trace/probe.pyx":166
 *     probe.create_ether(srcmac, macdst)
 *     probe.set_src(src.encode())
 *     return probe             # <<<<<<<<<<<<<<
 * 
 * cpdef bytes fill_packet_header(uint16_t size):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_probe));
  __pyx_r = __pyx_v_probe;
  goto __pyx_L0;

  /* "cloudtrace/trace/probe.pyx":143
 *         raise NotImplementedError()
 * 
 * cpdef Probe create_probe(str dst, uint8_t ipproto, uint16_t size, uint16_t pid):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str iface, src, nexthop, macdst, srcmac
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cloudtrace.trace.probe.create_probe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_iface);
  __Pyx_XDECREF(__pyx_v_src);
  __Pyx_XDECREF(__pyx_v_nexthop);
  __Pyx_XDECREF(__pyx_v_macdst);
  __Pyx_XDECREF(__pyx_v_srcmac);
  __Pyx_XDECREF((PyObject *)__pyx_v_probe);
  __Pyx_XDECREF((PyObject *)__pyx_v_iprobe);
  __Pyx_XDECREF((PyObject *)__pyx_v_uprobe);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_3create_probe(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_2create_probe[] = "create_probe(unicode dst, uint8_t ipproto, uint16_t size, uint16_t pid) -> Probe";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_3create_probe = {"create_probe", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10cloudtrace_5trace_5probe_3create_probe, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10cloudtrace_5trace_5probe_2create_probe};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_3create_probe(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_dst = 0;
  uint8_t __pyx_v_ipproto;
  uint16_t __pyx_v_size;
  uint16_t __pyx_v_pid;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_probe (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_dst,&__pyx_n_s_ipproto,&__pyx_n_s_size,&__pyx_n_s_pid,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dst)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ipproto)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_probe", 1, 4, 4, 1); __PYX_ERR(1, 143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_probe", 1, 4, 4, 2); __PYX_ERR(1, 143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("create_probe", 1, 4, 4, 3); __PYX_ERR(1, 143, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "create_probe") < 0)) __PYX_ERR(1, 143, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_dst = ((PyObject*)values[0]);
    __pyx_v_ipproto = __Pyx_PyInt_As_uint8_t(values[1]); if (unlikely((__pyx_v_ipproto == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 143, __pyx_L3_error)
    __pyx_v_size = __Pyx_PyInt_As_uint16_t(values[2]); if (unlikely((__pyx_v_size == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 143, __pyx_L3_error)
    __pyx_v_pid = __Pyx_PyInt_As_uint16_t(values[3]); if (unlikely((__pyx_v_pid == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 143, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_probe", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 143, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.create_probe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dst), (&PyUnicode_Type), 1, "dst", 1))) __PYX_ERR(1, 143, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_2create_probe(__pyx_self, __pyx_v_dst, __pyx_v_ipproto, __pyx_v_size, __pyx_v_pid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_2create_probe(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_dst, uint8_t __pyx_v_ipproto, uint16_t __pyx_v_size, uint16_t __pyx_v_pid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_probe", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10cloudtrace_5trace_5probe_create_probe(__pyx_v_dst, __pyx_v_ipproto, __pyx_v_size, __pyx_v_pid, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.create_probe", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":168
 *     return probe
 * 
 * cpdef bytes fill_packet_header(uint16_t size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t caplen = size + sizeof(ethhdr)
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5fill_packet_header(PyObject *__pyx_self, PyObject *__pyx_arg_size); /*proto*/
static PyObject *__pyx_f_10cloudtrace_5trace_5probe_fill_packet_header(uint16_t __pyx_v_size, CYTHON_UNUSED int __pyx_skip_dispatch) {
  uint32_t __pyx_v_caplen;
  __pyx_t_10cloudtrace_5trace_5probe_pcapfix_packet_hdr_s __pyx_v_ph;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_packet_header", 0);

  /* "cloudtrace/trace/probe.pyx":170
 * cpdef bytes fill_packet_header(uint16_t size):
 *     cdef:
 *         uint32_t caplen = size + sizeof(ethhdr)             # <<<<<<<<<<<<<<
 *         pcapfix_packet_hdr_s ph
 *     ph.incl_len = caplen
 */
  __pyx_v_caplen = (__pyx_v_size + (sizeof(__pyx_t_10cloudtrace_5trace_5utils_ethhdr)));

  /* "cloudtrace/trace/probe.pyx":172
 *         uint32_t caplen = size + sizeof(ethhdr)
 *         pcapfix_packet_hdr_s ph
 *     ph.incl_len = caplen             # <<<<<<<<<<<<<<
 *     ph.orig_len = caplen
 *     return <bytes>(<unsigned char *>&ph)[:sizeof(pcapfix_packet_hdr_s)]
 */
  __pyx_v_ph.incl_len = __pyx_v_caplen;

  /* "cloudtrace/trace/probe.pyx":173
 *         pcapfix_packet_hdr_s ph
 *     ph.incl_len = caplen
 *     ph.orig_len = caplen             # <<<<<<<<<<<<<<
 *     return <bytes>(<unsigned char *>&ph)[:sizeof(pcapfix_packet_hdr_s)]
 * 
 */
  __pyx_v_ph.orig_len = __pyx_v_caplen;

  /* "cloudtrace/trace/probe.pyx":174
 *     ph.incl_len = caplen
 *     ph.orig_len = caplen
 *     return <bytes>(<unsigned char *>&ph)[:sizeof(pcapfix_packet_hdr_s)]             # <<<<<<<<<<<<<<
 * 
 * cpdef void write_pcap_header(f):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(((const char*)((unsigned char *)(&__pyx_v_ph))) + 0, (sizeof(__pyx_t_10cloudtrace_5trace_5probe_pcapfix_packet_hdr_s)) - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject*)__pyx_t_1));
  __pyx_r = ((PyObject*)__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cloudtrace/trace/probe.pyx":168
 *     return probe
 * 
 * cpdef bytes fill_packet_header(uint16_t size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t caplen = size + sizeof(ethhdr)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.fill_packet_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5fill_packet_header(PyObject *__pyx_self, PyObject *__pyx_arg_size); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_4fill_packet_header[] = "fill_packet_header(uint16_t size) -> bytes";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_5fill_packet_header = {"fill_packet_header", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5fill_packet_header, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_4fill_packet_header};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_5fill_packet_header(PyObject *__pyx_self, PyObject *__pyx_arg_size) {
  uint16_t __pyx_v_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("fill_packet_header (wrapper)", 0);
  assert(__pyx_arg_size); {
    __pyx_v_size = __Pyx_PyInt_As_uint16_t(__pyx_arg_size); if (unlikely((__pyx_v_size == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 168, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.fill_packet_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_4fill_packet_header(__pyx_self, ((uint16_t)__pyx_v_size));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_4fill_packet_header(CYTHON_UNUSED PyObject *__pyx_self, uint16_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("fill_packet_header", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10cloudtrace_5trace_5probe_fill_packet_header(__pyx_v_size, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.fill_packet_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":176
 *     return <bytes>(<unsigned char *>&ph)[:sizeof(pcapfix_packet_hdr_s)]
 * 
 * cpdef void write_pcap_header(f):             # <<<<<<<<<<<<<<
 *     cdef:
 *         pcapfix_global_hdr_s *gh = <pcapfix_global_hdr_s *>calloc(1, sizeof(pcapfix_global_hdr_s))
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_7write_pcap_header(PyObject *__pyx_self, PyObject *__pyx_v_f); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_write_pcap_header(PyObject *__pyx_v_f, CYTHON_UNUSED int __pyx_skip_dispatch) {
  __pyx_t_10cloudtrace_5trace_5probe_pcapfix_global_hdr_s *__pyx_v_gh;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_pcap_header", 0);

  /* "cloudtrace/trace/probe.pyx":178
 * cpdef void write_pcap_header(f):
 *     cdef:
 *         pcapfix_global_hdr_s *gh = <pcapfix_global_hdr_s *>calloc(1, sizeof(pcapfix_global_hdr_s))             # <<<<<<<<<<<<<<
 *     gh.magic_number = 0xa1b2c3d4
 *     gh.version_major = 2
 */
  __pyx_v_gh = ((__pyx_t_10cloudtrace_5trace_5probe_pcapfix_global_hdr_s *)calloc(1, (sizeof(__pyx_t_10cloudtrace_5trace_5probe_pcapfix_global_hdr_s))));

  /* "cloudtrace/trace/probe.pyx":179
 *     cdef:
 *         pcapfix_global_hdr_s *gh = <pcapfix_global_hdr_s *>calloc(1, sizeof(pcapfix_global_hdr_s))
 *     gh.magic_number = 0xa1b2c3d4             # <<<<<<<<<<<<<<
 *     gh.version_major = 2
 *     gh.version_minor = 4
 */
  __pyx_v_gh->magic_number = 0xa1b2c3d4;

  /* "cloudtrace/trace/probe.pyx":180
 *         pcapfix_global_hdr_s *gh = <pcapfix_global_hdr_s *>calloc(1, sizeof(pcapfix_global_hdr_s))
 *     gh.magic_number = 0xa1b2c3d4
 *     gh.version_major = 2             # <<<<<<<<<<<<<<
 *     gh.version_minor = 4
 *     gh.thiszone = 0
 */
  __pyx_v_gh->version_major = 2;

  /* "cloudtrace/trace/probe.pyx":181
 *     gh.magic_number = 0xa1b2c3d4
 *     gh.version_major = 2
 *     gh.version_minor = 4             # <<<<<<<<<<<<<<
 *     gh.thiszone = 0
 *     gh.sigfigs = 0
 */
  __pyx_v_gh->version_minor = 4;

  /* "cloudtrace/trace/probe.pyx":182
 *     gh.version_major = 2
 *     gh.version_minor = 4
 *     gh.thiszone = 0             # <<<<<<<<<<<<<<
 *     gh.sigfigs = 0
 *     gh.snaplen = 0xffff
 */
  __pyx_v_gh->thiszone = 0;

  /* "cloudtrace/trace/probe.pyx":183
 *     gh.version_minor = 4
 *     gh.thiszone = 0
 *     gh.sigfigs = 0             # <<<<<<<<<<<<<<
 *     gh.snaplen = 0xffff
 *     gh.network = 1
 */
  __pyx_v_gh->sigfigs = 0;

  /* "cloudtrace/trace/probe.pyx":184
 *     gh.thiszone = 0
 *     gh.sigfigs = 0
 *     gh.snaplen = 0xffff             # <<<<<<<<<<<<<<
 *     gh.network = 1
 *     f.write(<bytes>((<unsigned char *>gh)[:sizeof(pcapfix_global_hdr_s)]))
 */
  __pyx_v_gh->snaplen = 0xffff;

  /* "cloudtrace/trace/probe.pyx":185
 *     gh.sigfigs = 0
 *     gh.snaplen = 0xffff
 *     gh.network = 1             # <<<<<<<<<<<<<<
 *     f.write(<bytes>((<unsigned char *>gh)[:sizeof(pcapfix_global_hdr_s)]))
 *     free(gh)
 */
  __pyx_v_gh->network = 1;

  /* "cloudtrace/trace/probe.pyx":186
 *     gh.snaplen = 0xffff
 *     gh.network = 1
 *     f.write(<bytes>((<unsigned char *>gh)[:sizeof(pcapfix_global_hdr_s)]))             # <<<<<<<<<<<<<<
 *     free(gh)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBytes_FromStringAndSize(((const char*)((unsigned char *)__pyx_v_gh)) + 0, (sizeof(__pyx_t_10cloudtrace_5trace_5probe_pcapfix_global_hdr_s)) - 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/trace/probe.pyx":187
 *     gh.network = 1
 *     f.write(<bytes>((<unsigned char *>gh)[:sizeof(pcapfix_global_hdr_s)]))
 *     free(gh)             # <<<<<<<<<<<<<<
 * 
 * cpdef void craft(str targets, uint16_t pid, file=None, int minttl=1, int maxttl=32, int cycles=1, str proto='icmp', uint16_t size=64, bint randomize=False) except *:
 */
  free(__pyx_v_gh);

  /* "cloudtrace/trace/probe.pyx":176
 *     return <bytes>(<unsigned char *>&ph)[:sizeof(pcapfix_packet_hdr_s)]
 * 
 * cpdef void write_pcap_header(f):             # <<<<<<<<<<<<<<
 *     cdef:
 *         pcapfix_global_hdr_s *gh = <pcapfix_global_hdr_s *>calloc(1, sizeof(pcapfix_global_hdr_s))
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.write_pcap_header", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_7write_pcap_header(PyObject *__pyx_self, PyObject *__pyx_v_f); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_6write_pcap_header[] = "write_pcap_header(f) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_7write_pcap_header = {"write_pcap_header", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_7write_pcap_header, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_6write_pcap_header};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_7write_pcap_header(PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write_pcap_header (wrapper)", 0);
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_6write_pcap_header(__pyx_self, ((PyObject *)__pyx_v_f));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_6write_pcap_header(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_f) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("write_pcap_header", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_10cloudtrace_5trace_5probe_write_pcap_header(__pyx_v_f, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cloudtrace.trace.probe.write_pcap_header", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":189
 *     free(gh)
 * 
 * cpdef void craft(str targets, uint16_t pid, file=None, int minttl=1, int maxttl=32, int cycles=1, str proto='icmp', uint16_t size=64, bint randomize=False) except *:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes dst
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9craft(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_craft(PyObject *__pyx_v_targets, uint16_t __pyx_v_pid, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_10cloudtrace_5trace_5probe_craft *__pyx_optional_args) {
  PyObject *__pyx_v_file = ((PyObject *)Py_None);
  int __pyx_v_minttl = ((int)1);
  int __pyx_v_maxttl = ((int)32);
  int __pyx_v_cycles = ((int)1);
  PyObject *__pyx_v_proto = ((PyObject*)__pyx_n_u_icmp);
  uint16_t __pyx_v_size = ((uint16_t)64);
  int __pyx_v_randomize = ((int)0);
  PyObject *__pyx_v_dst = 0;
  uint8_t __pyx_v_ttl;
  struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_probe = 0;
  PyObject *__pyx_v_pkt = 0;
  PyObject *__pyx_v_caplen_line = 0;
  uint16_t __pyx_v_ipproto;
  uint32_t __pyx_v_num_targets;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_start = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_g = NULL;
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_v_addr = NULL;
  PyObject *__pyx_v_end = NULL;
  PyObject *__pyx_v_duration = NULL;
  PyObject *__pyx_v_pps = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  Py_ssize_t __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *(*__pyx_t_19)(PyObject *);
  long __pyx_t_20;
  long __pyx_t_21;
  uint8_t __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_t_24;
  char const *__pyx_t_25;
  PyObject *__pyx_t_26 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("craft", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_file = __pyx_optional_args->file;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_minttl = __pyx_optional_args->minttl;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_maxttl = __pyx_optional_args->maxttl;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_cycles = __pyx_optional_args->cycles;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_proto = __pyx_optional_args->proto;
              if (__pyx_optional_args->__pyx_n > 5) {
                __pyx_v_size = __pyx_optional_args->size;
                if (__pyx_optional_args->__pyx_n > 6) {
                  __pyx_v_randomize = __pyx_optional_args->randomize;
                }
              }
            }
          }
        }
      }
    }
  }

  /* "cloudtrace/trace/probe.pyx":196
 *         bytes pkt, caplen_line
 *         uint16_t ipproto
 *         uint32_t num_targets = 0             # <<<<<<<<<<<<<<
 * 
 *     ipproto = proto_to_num(proto)
 */
  __pyx_v_num_targets = 0;

  /* "cloudtrace/trace/probe.pyx":198
 *         uint32_t num_targets = 0
 * 
 *     ipproto = proto_to_num(proto)             # <<<<<<<<<<<<<<
 *     if size < minsize(ipproto):
 *         size = minsize(ipproto)
 */
  __pyx_v_ipproto = __pyx_f_10cloudtrace_5trace_5probe_proto_to_num(__pyx_v_proto, 0);

  /* "cloudtrace/trace/probe.pyx":199
 * 
 *     ipproto = proto_to_num(proto)
 *     if size < minsize(ipproto):             # <<<<<<<<<<<<<<
 *         size = minsize(ipproto)
 *     elif size > 1500:
 */
  __pyx_t_1 = __pyx_f_10cloudtrace_5trace_5probe_minsize(__pyx_v_ipproto); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 199, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_v_size < __pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "cloudtrace/trace/probe.pyx":200
 *     ipproto = proto_to_num(proto)
 *     if size < minsize(ipproto):
 *         size = minsize(ipproto)             # <<<<<<<<<<<<<<
 *     elif size > 1500:
 *         size = 1500
 */
    __pyx_t_2 = __pyx_f_10cloudtrace_5trace_5probe_minsize(__pyx_v_ipproto); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 200, __pyx_L1_error)
    __pyx_v_size = __pyx_t_2;

    /* "cloudtrace/trace/probe.pyx":199
 * 
 *     ipproto = proto_to_num(proto)
 *     if size < minsize(ipproto):             # <<<<<<<<<<<<<<
 *         size = minsize(ipproto)
 *     elif size > 1500:
 */
    goto __pyx_L3;
  }

  /* "cloudtrace/trace/probe.pyx":201
 *     if size < minsize(ipproto):
 *         size = minsize(ipproto)
 *     elif size > 1500:             # <<<<<<<<<<<<<<
 *         size = 1500
 * 
 */
  __pyx_t_2 = ((__pyx_v_size > 0x5DC) != 0);
  if (__pyx_t_2) {

    /* "cloudtrace/trace/probe.pyx":202
 *         size = minsize(ipproto)
 *     elif size > 1500:
 *         size = 1500             # <<<<<<<<<<<<<<
 * 
 *     probe = create_probe('0.0.0.0', ipproto, size, pid)
 */
    __pyx_v_size = 0x5DC;

    /* "cloudtrace/trace/probe.pyx":201
 *     if size < minsize(ipproto):
 *         size = minsize(ipproto)
 *     elif size > 1500:             # <<<<<<<<<<<<<<
 *         size = 1500
 * 
 */
  }
  __pyx_L3:;

  /* "cloudtrace/trace/probe.pyx":204
 *         size = 1500
 * 
 *     probe = create_probe('0.0.0.0', ipproto, size, pid)             # <<<<<<<<<<<<<<
 *     caplen_line = fill_packet_header(size)
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_10cloudtrace_5trace_5probe_create_probe(__pyx_kp_u_0_0_0_0, __pyx_v_ipproto, __pyx_v_size, __pyx_v_pid, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "cloudtrace/trace/probe.pyx":205
 * 
 *     probe = create_probe('0.0.0.0', ipproto, size, pid)
 *     caplen_line = fill_packet_header(size)             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(file, str):
 */
  __pyx_t_3 = __pyx_f_10cloudtrace_5trace_5probe_fill_packet_header(__pyx_v_size, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_caplen_line = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "cloudtrace/trace/probe.pyx":207
 *     caplen_line = fill_packet_header(size)
 * 
 *     if isinstance(file, str):             # <<<<<<<<<<<<<<
 *         f = open(file, 'wb')
 *     else:
 */
  __pyx_t_2 = PyUnicode_Check(__pyx_v_file); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/trace/probe.pyx":208
 * 
 *     if isinstance(file, str):
 *         f = open(file, 'wb')             # <<<<<<<<<<<<<<
 *     else:
 *         f = file
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_file);
    __Pyx_GIVEREF(__pyx_v_file);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_file);
    __Pyx_INCREF(__pyx_n_u_wb);
    __Pyx_GIVEREF(__pyx_n_u_wb);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_n_u_wb);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_f = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "cloudtrace/trace/probe.pyx":207
 *     caplen_line = fill_packet_header(size)
 * 
 *     if isinstance(file, str):             # <<<<<<<<<<<<<<
 *         f = open(file, 'wb')
 *     else:
 */
    goto __pyx_L4;
  }

  /* "cloudtrace/trace/probe.pyx":210
 *         f = open(file, 'wb')
 *     else:
 *         f = file             # <<<<<<<<<<<<<<
 *     try:
 *         write_pcap_header(f)
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_file);
    __pyx_v_f = __pyx_v_file;
  }
  __pyx_L4:;

  /* "cloudtrace/trace/probe.pyx":211
 *     else:
 *         f = file
 *     try:             # <<<<<<<<<<<<<<
 *         write_pcap_header(f)
 *         start = time.time()
 */
  /*try:*/ {

    /* "cloudtrace/trace/probe.pyx":212
 *         f = file
 *     try:
 *         write_pcap_header(f)             # <<<<<<<<<<<<<<
 *         start = time.time()
 *         for _ in range(cycles):
 */
    __pyx_f_10cloudtrace_5trace_5probe_write_pcap_header(__pyx_v_f, 0);

    /* "cloudtrace/trace/probe.pyx":213
 *     try:
 *         write_pcap_header(f)
 *         start = time.time()             # <<<<<<<<<<<<<<
 *         for _ in range(cycles):
 *             with fopen(targets, 'rb') as g:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 213, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 213, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 213, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_start = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "cloudtrace/trace/probe.pyx":214
 *         write_pcap_header(f)
 *         start = time.time()
 *         for _ in range(cycles):             # <<<<<<<<<<<<<<
 *             with fopen(targets, 'rb') as g:
 *                 for line in g:
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 214, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 214, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_4 = __pyx_t_5; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 214, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 214, __pyx_L6_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(1, 214, __pyx_L6_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 214, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(1, 214, __pyx_L6_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 214, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_7(__pyx_t_4);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(1, 214, __pyx_L6_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "cloudtrace/trace/probe.pyx":215
 *         start = time.time()
 *         for _ in range(cycles):
 *             with fopen(targets, 'rb') as g:             # <<<<<<<<<<<<<<
 *                 for line in g:
 *                     if randomize:
 */
      /*with:*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_fopen); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 215, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_8 = NULL;
        __pyx_t_9 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_9 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_targets, __pyx_n_u_rb};
          __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 215, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_targets, __pyx_n_u_rb};
          __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 215, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_5);
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 215, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_10);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_INCREF(__pyx_v_targets);
          __Pyx_GIVEREF(__pyx_v_targets);
          PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_v_targets);
          __Pyx_INCREF(__pyx_n_u_rb);
          __Pyx_GIVEREF(__pyx_n_u_rb);
          PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_n_u_rb);
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 215, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_exit); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 215, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_10 = __Pyx_PyObject_LookupSpecial(__pyx_t_5, __pyx_n_s_enter); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 215, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 215, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __pyx_t_3;
        __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            /*try:*/ {
              __Pyx_XDECREF_SET(__pyx_v_g, __pyx_t_10);
              __pyx_t_10 = 0;

              /* "cloudtrace/trace/probe.pyx":216
 *         for _ in range(cycles):
 *             with fopen(targets, 'rb') as g:
 *                 for line in g:             # <<<<<<<<<<<<<<
 *                     if randomize:
 *                         addr, _, _ = line.decode().rpartition('.')
 */
              if (likely(PyList_CheckExact(__pyx_v_g)) || PyTuple_CheckExact(__pyx_v_g)) {
                __pyx_t_10 = __pyx_v_g; __Pyx_INCREF(__pyx_t_10); __pyx_t_15 = 0;
                __pyx_t_16 = NULL;
              } else {
                __pyx_t_15 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_v_g); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 216, __pyx_L16_error)
                __Pyx_GOTREF(__pyx_t_10);
                __pyx_t_16 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 216, __pyx_L16_error)
              }
              for (;;) {
                if (likely(!__pyx_t_16)) {
                  if (likely(PyList_CheckExact(__pyx_t_10))) {
                    if (__pyx_t_15 >= PyList_GET_SIZE(__pyx_t_10)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_5 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_15); __Pyx_INCREF(__pyx_t_5); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(1, 216, __pyx_L16_error)
                    #else
                    __pyx_t_5 = PySequence_ITEM(__pyx_t_10, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 216, __pyx_L16_error)
                    __Pyx_GOTREF(__pyx_t_5);
                    #endif
                  } else {
                    if (__pyx_t_15 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_15); __Pyx_INCREF(__pyx_t_5); __pyx_t_15++; if (unlikely(0 < 0)) __PYX_ERR(1, 216, __pyx_L16_error)
                    #else
                    __pyx_t_5 = PySequence_ITEM(__pyx_t_10, __pyx_t_15); __pyx_t_15++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 216, __pyx_L16_error)
                    __Pyx_GOTREF(__pyx_t_5);
                    #endif
                  }
                } else {
                  __pyx_t_5 = __pyx_t_16(__pyx_t_10);
                  if (unlikely(!__pyx_t_5)) {
                    PyObject* exc_type = PyErr_Occurred();
                    if (exc_type) {
                      if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                      else __PYX_ERR(1, 216, __pyx_L16_error)
                    }
                    break;
                  }
                  __Pyx_GOTREF(__pyx_t_5);
                }
                __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_5);
                __pyx_t_5 = 0;

                /* "cloudtrace/trace/probe.pyx":217
 *             with fopen(targets, 'rb') as g:
 *                 for line in g:
 *                     if randomize:             # <<<<<<<<<<<<<<
 *                         addr, _, _ = line.decode().rpartition('.')
 *                         addr = '{}.{}'.format(addr, random.randint(0, 255))
 */
                __pyx_t_1 = (__pyx_v_randomize != 0);
                if (__pyx_t_1) {

                  /* "cloudtrace/trace/probe.pyx":218
 *                 for line in g:
 *                     if randomize:
 *                         addr, _, _ = line.decode().rpartition('.')             # <<<<<<<<<<<<<<
 *                         addr = '{}.{}'.format(addr, random.randint(0, 255))
 *                         dst = '{}'.format(addr).encode()
 */
                  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_decode); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 218, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __pyx_t_17 = NULL;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
                    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_8);
                    if (likely(__pyx_t_17)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                      __Pyx_INCREF(__pyx_t_17);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_8, function);
                    }
                  }
                  __pyx_t_3 = (__pyx_t_17) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_17) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
                  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 218, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_rpartition); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 218, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __pyx_t_3 = NULL;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
                    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_8);
                    if (likely(__pyx_t_3)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
                      __Pyx_INCREF(__pyx_t_3);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_8, function);
                    }
                  }
                  __pyx_t_5 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_3, __pyx_kp_u__8) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_u__8);
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 218, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
                    PyObject* sequence = __pyx_t_5;
                    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
                    if (unlikely(size != 3)) {
                      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                      __PYX_ERR(1, 218, __pyx_L16_error)
                    }
                    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                    if (likely(PyTuple_CheckExact(sequence))) {
                      __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
                      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
                      __pyx_t_17 = PyTuple_GET_ITEM(sequence, 2); 
                    } else {
                      __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
                      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
                      __pyx_t_17 = PyList_GET_ITEM(sequence, 2); 
                    }
                    __Pyx_INCREF(__pyx_t_8);
                    __Pyx_INCREF(__pyx_t_3);
                    __Pyx_INCREF(__pyx_t_17);
                    #else
                    __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 218, __pyx_L16_error)
                    __Pyx_GOTREF(__pyx_t_8);
                    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 218, __pyx_L16_error)
                    __Pyx_GOTREF(__pyx_t_3);
                    __pyx_t_17 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 218, __pyx_L16_error)
                    __Pyx_GOTREF(__pyx_t_17);
                    #endif
                    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                  } else {
                    Py_ssize_t index = -1;
                    __pyx_t_18 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 218, __pyx_L16_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                    __pyx_t_19 = Py_TYPE(__pyx_t_18)->tp_iternext;
                    index = 0; __pyx_t_8 = __pyx_t_19(__pyx_t_18); if (unlikely(!__pyx_t_8)) goto __pyx_L27_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_8);
                    index = 1; __pyx_t_3 = __pyx_t_19(__pyx_t_18); if (unlikely(!__pyx_t_3)) goto __pyx_L27_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_3);
                    index = 2; __pyx_t_17 = __pyx_t_19(__pyx_t_18); if (unlikely(!__pyx_t_17)) goto __pyx_L27_unpacking_failed;
                    __Pyx_GOTREF(__pyx_t_17);
                    if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_18), 3) < 0) __PYX_ERR(1, 218, __pyx_L16_error)
                    __pyx_t_19 = NULL;
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    goto __pyx_L28_unpacking_done;
                    __pyx_L27_unpacking_failed:;
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                    __pyx_t_19 = NULL;
                    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
                    __PYX_ERR(1, 218, __pyx_L16_error)
                    __pyx_L28_unpacking_done:;
                  }
                  __Pyx_XDECREF_SET(__pyx_v_addr, __pyx_t_8);
                  __pyx_t_8 = 0;
                  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_3);
                  __pyx_t_3 = 0;
                  __Pyx_DECREF_SET(__pyx_v__, __pyx_t_17);
                  __pyx_t_17 = 0;

                  /* "cloudtrace/trace/probe.pyx":219
 *                     if randomize:
 *                         addr, _, _ = line.decode().rpartition('.')
 *                         addr = '{}.{}'.format(addr, random.randint(0, 255))             # <<<<<<<<<<<<<<
 *                         dst = '{}'.format(addr).encode()
 *                         # targets.append('{}'.format(addr).encode())
 */
                  __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__9, __pyx_n_s_format); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 219, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_random); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 219, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_randint); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 219, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_8);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 219, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
                  __pyx_t_8 = NULL;
                  __pyx_t_9 = 0;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
                    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_17);
                    if (likely(__pyx_t_8)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                      __Pyx_INCREF(__pyx_t_8);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_17, function);
                      __pyx_t_9 = 1;
                    }
                  }
                  #if CYTHON_FAST_PYCALL
                  if (PyFunction_Check(__pyx_t_17)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_addr, __pyx_t_3};
                    __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 219, __pyx_L16_error)
                    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                    __Pyx_GOTREF(__pyx_t_5);
                    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  } else
                  #endif
                  #if CYTHON_FAST_PYCCALL
                  if (__Pyx_PyFastCFunction_Check(__pyx_t_17)) {
                    PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_v_addr, __pyx_t_3};
                    __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 219, __pyx_L16_error)
                    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
                    __Pyx_GOTREF(__pyx_t_5);
                    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  } else
                  #endif
                  {
                    __pyx_t_18 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 219, __pyx_L16_error)
                    __Pyx_GOTREF(__pyx_t_18);
                    if (__pyx_t_8) {
                      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_8); __pyx_t_8 = NULL;
                    }
                    __Pyx_INCREF(__pyx_v_addr);
                    __Pyx_GIVEREF(__pyx_v_addr);
                    PyTuple_SET_ITEM(__pyx_t_18, 0+__pyx_t_9, __pyx_v_addr);
                    __Pyx_GIVEREF(__pyx_t_3);
                    PyTuple_SET_ITEM(__pyx_t_18, 1+__pyx_t_9, __pyx_t_3);
                    __pyx_t_3 = 0;
                    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_18, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 219, __pyx_L16_error)
                    __Pyx_GOTREF(__pyx_t_5);
                    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  }
                  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  __Pyx_DECREF_SET(__pyx_v_addr, __pyx_t_5);
                  __pyx_t_5 = 0;

                  /* "cloudtrace/trace/probe.pyx":220
 *                         addr, _, _ = line.decode().rpartition('.')
 *                         addr = '{}.{}'.format(addr, random.randint(0, 255))
 *                         dst = '{}'.format(addr).encode()             # <<<<<<<<<<<<<<
 *                         # targets.append('{}'.format(addr).encode())
 *                     else:
 */
                  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__11, __pyx_n_s_format); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 220, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __pyx_t_3 = NULL;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
                    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_18);
                    if (likely(__pyx_t_3)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                      __Pyx_INCREF(__pyx_t_3);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_18, function);
                    }
                  }
                  __pyx_t_17 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_3, __pyx_v_addr) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_v_addr);
                  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                  if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 220, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_17);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_t_17, __pyx_n_s_encode); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 220, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
                  __pyx_t_17 = NULL;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
                    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_18);
                    if (likely(__pyx_t_17)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                      __Pyx_INCREF(__pyx_t_17);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_18, function);
                    }
                  }
                  __pyx_t_5 = (__pyx_t_17) ? __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_17) : __Pyx_PyObject_CallNoArg(__pyx_t_18);
                  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                  if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 220, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(1, 220, __pyx_L16_error)
                  __Pyx_XDECREF_SET(__pyx_v_dst, ((PyObject*)__pyx_t_5));
                  __pyx_t_5 = 0;

                  /* "cloudtrace/trace/probe.pyx":217
 *             with fopen(targets, 'rb') as g:
 *                 for line in g:
 *                     if randomize:             # <<<<<<<<<<<<<<
 *                         addr, _, _ = line.decode().rpartition('.')
 *                         addr = '{}.{}'.format(addr, random.randint(0, 255))
 */
                  goto __pyx_L26;
                }

                /* "cloudtrace/trace/probe.pyx":223
 *                         # targets.append('{}'.format(addr).encode())
 *                     else:
 *                         dst = line.strip()             # <<<<<<<<<<<<<<
 *                         # targets.append(line.strip().encode())
 *                     probe.set_dst(dst)
 */
                /*else*/ {
                  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_line, __pyx_n_s_strip); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 223, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __pyx_t_17 = NULL;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
                    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_18);
                    if (likely(__pyx_t_17)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                      __Pyx_INCREF(__pyx_t_17);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_18, function);
                    }
                  }
                  __pyx_t_5 = (__pyx_t_17) ? __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_17) : __Pyx_PyObject_CallNoArg(__pyx_t_18);
                  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                  if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 223, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(1, 223, __pyx_L16_error)
                  __Pyx_XDECREF_SET(__pyx_v_dst, ((PyObject*)__pyx_t_5));
                  __pyx_t_5 = 0;
                }
                __pyx_L26:;

                /* "cloudtrace/trace/probe.pyx":225
 *                         dst = line.strip()
 *                         # targets.append(line.strip().encode())
 *                     probe.set_dst(dst)             # <<<<<<<<<<<<<<
 *                     for ttl in range(minttl, maxttl+1, 1):
 *                         probe.set_ttl(ttl)
 */
                ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->set_dst(__pyx_v_probe, __pyx_v_dst, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 225, __pyx_L16_error)

                /* "cloudtrace/trace/probe.pyx":226
 *                         # targets.append(line.strip().encode())
 *                     probe.set_dst(dst)
 *                     for ttl in range(minttl, maxttl+1, 1):             # <<<<<<<<<<<<<<
 *                         probe.set_ttl(ttl)
 *                         probe.prep_trace(ttl)
 */
                __pyx_t_20 = (__pyx_v_maxttl + 1);
                __pyx_t_21 = __pyx_t_20;
                for (__pyx_t_22 = __pyx_v_minttl; __pyx_t_22 < __pyx_t_21; __pyx_t_22+=1) {
                  __pyx_v_ttl = __pyx_t_22;

                  /* "cloudtrace/trace/probe.pyx":227
 *                     probe.set_dst(dst)
 *                     for ttl in range(minttl, maxttl+1, 1):
 *                         probe.set_ttl(ttl)             # <<<<<<<<<<<<<<
 *                         probe.prep_trace(ttl)
 *                         probe.prepare()
 */
                  ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->set_ttl(__pyx_v_probe, __pyx_v_ttl, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 227, __pyx_L16_error)

                  /* "cloudtrace/trace/probe.pyx":228
 *                     for ttl in range(minttl, maxttl+1, 1):
 *                         probe.set_ttl(ttl)
 *                         probe.prep_trace(ttl)             # <<<<<<<<<<<<<<
 *                         probe.prepare()
 *                         pkt = probe.to_bytes()
 */
                  ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->prep_trace(__pyx_v_probe, __pyx_v_ttl, 0);

                  /* "cloudtrace/trace/probe.pyx":229
 *                         probe.set_ttl(ttl)
 *                         probe.prep_trace(ttl)
 *                         probe.prepare()             # <<<<<<<<<<<<<<
 *                         pkt = probe.to_bytes()
 *                         f.write(caplen_line)
 */
                  ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->prepare(__pyx_v_probe, 0);

                  /* "cloudtrace/trace/probe.pyx":230
 *                         probe.prep_trace(ttl)
 *                         probe.prepare()
 *                         pkt = probe.to_bytes()             # <<<<<<<<<<<<<<
 *                         f.write(caplen_line)
 *                         f.write(pkt)
 */
                  __pyx_t_5 = ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->to_bytes(__pyx_v_probe, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 230, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_XDECREF_SET(__pyx_v_pkt, ((PyObject*)__pyx_t_5));
                  __pyx_t_5 = 0;

                  /* "cloudtrace/trace/probe.pyx":231
 *                         probe.prepare()
 *                         pkt = probe.to_bytes()
 *                         f.write(caplen_line)             # <<<<<<<<<<<<<<
 *                         f.write(pkt)
 *                     num_targets += 1
 */
                  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 231, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __pyx_t_17 = NULL;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
                    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_18);
                    if (likely(__pyx_t_17)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                      __Pyx_INCREF(__pyx_t_17);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_18, function);
                    }
                  }
                  __pyx_t_5 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_17, __pyx_v_caplen_line) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_v_caplen_line);
                  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                  if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 231, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

                  /* "cloudtrace/trace/probe.pyx":232
 *                         pkt = probe.to_bytes()
 *                         f.write(caplen_line)
 *                         f.write(pkt)             # <<<<<<<<<<<<<<
 *                     num_targets += 1
 *     finally:
 */
                  __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 232, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  __pyx_t_17 = NULL;
                  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
                    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_18);
                    if (likely(__pyx_t_17)) {
                      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
                      __Pyx_INCREF(__pyx_t_17);
                      __Pyx_INCREF(function);
                      __Pyx_DECREF_SET(__pyx_t_18, function);
                    }
                  }
                  __pyx_t_5 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_18, __pyx_t_17, __pyx_v_pkt) : __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_v_pkt);
                  __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
                  if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 232, __pyx_L16_error)
                  __Pyx_GOTREF(__pyx_t_5);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
                }

                /* "cloudtrace/trace/probe.pyx":233
 *                         f.write(caplen_line)
 *                         f.write(pkt)
 *                     num_targets += 1             # <<<<<<<<<<<<<<
 *     finally:
 *         f.close()
 */
                __pyx_v_num_targets = (__pyx_v_num_targets + 1);

                /* "cloudtrace/trace/probe.pyx":216
 *         for _ in range(cycles):
 *             with fopen(targets, 'rb') as g:
 *                 for line in g:             # <<<<<<<<<<<<<<
 *                     if randomize:
 *                         addr, _, _ = line.decode().rpartition('.')
 */
              }
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

              /* "cloudtrace/trace/probe.pyx":215
 *         start = time.time()
 *         for _ in range(cycles):
 *             with fopen(targets, 'rb') as g:             # <<<<<<<<<<<<<<
 *                 for line in g:
 *                     if randomize:
 */
            }
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            goto __pyx_L23_try_end;
            __pyx_L16_error:;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("cloudtrace.trace.probe.craft", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_10, &__pyx_t_5, &__pyx_t_18) < 0) __PYX_ERR(1, 215, __pyx_L18_except_error)
              __Pyx_GOTREF(__pyx_t_10);
              __Pyx_GOTREF(__pyx_t_5);
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_17 = PyTuple_Pack(3, __pyx_t_10, __pyx_t_5, __pyx_t_18); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 215, __pyx_L18_except_error)
              __Pyx_GOTREF(__pyx_t_17);
              __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_17, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
              if (unlikely(!__pyx_t_23)) __PYX_ERR(1, 215, __pyx_L18_except_error)
              __Pyx_GOTREF(__pyx_t_23);
              __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_23);
              __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
              if (__pyx_t_1 < 0) __PYX_ERR(1, 215, __pyx_L18_except_error)
              __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
              if (__pyx_t_2) {
                __Pyx_GIVEREF(__pyx_t_10);
                __Pyx_GIVEREF(__pyx_t_5);
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_ErrRestoreWithState(__pyx_t_10, __pyx_t_5, __pyx_t_18);
                __pyx_t_10 = 0; __pyx_t_5 = 0; __pyx_t_18 = 0; 
                __PYX_ERR(1, 215, __pyx_L18_except_error)
              }
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
              goto __pyx_L17_exception_handled;
            }
            __pyx_L18_except_error:;
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            goto __pyx_L6_error;
            __pyx_L17_exception_handled:;
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            __pyx_L23_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_11) {
              __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__12, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 215, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            }
            goto __pyx_L15;
          }
          __pyx_L15:;
        }
        goto __pyx_L34;
        __pyx_L10_error:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L6_error;
        __pyx_L34:;
      }

      /* "cloudtrace/trace/probe.pyx":214
 *         write_pcap_header(f)
 *         start = time.time()
 *         for _ in range(cycles):             # <<<<<<<<<<<<<<
 *             with fopen(targets, 'rb') as g:
 *                 for line in g:
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "cloudtrace/trace/probe.pyx":235
 *                     num_targets += 1
 *     finally:
 *         f.close()             # <<<<<<<<<<<<<<
 *     end = time.time()
 *     duration = end - start
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_close); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_18);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_18);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_18, function);
        }
      }
      __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_18);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_11 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_23 = 0; __pyx_t_26 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_23, &__pyx_t_26);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_14, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_14, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_23);
      __Pyx_XGOTREF(__pyx_t_26);
      __pyx_t_9 = __pyx_lineno; __pyx_t_24 = __pyx_clineno; __pyx_t_25 = __pyx_filename;
      {
        __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_close); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 235, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_18))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_18);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_18);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_18, function);
          }
        }
        __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_18, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_18);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 235, __pyx_L36_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_23);
        __Pyx_XGIVEREF(__pyx_t_26);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_23, __pyx_t_26);
      }
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ErrRestore(__pyx_t_11, __pyx_t_14, __pyx_t_13);
      __pyx_t_11 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_23 = 0; __pyx_t_26 = 0;
      __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_24; __pyx_filename = __pyx_t_25;
      goto __pyx_L1_error;
      __pyx_L36_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_23);
        __Pyx_XGIVEREF(__pyx_t_26);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_23, __pyx_t_26);
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_12 = 0; __pyx_t_23 = 0; __pyx_t_26 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "cloudtrace/trace/probe.pyx":236
 *     finally:
 *         f.close()
 *     end = time.time()             # <<<<<<<<<<<<<<
 *     duration = end - start
 *     if duration > 0:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_18, __pyx_n_s_time); if (unlikely(!__pyx_t_18)) __PYX_ERR(1, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  __pyx_t_18 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_18)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_18);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_4 = (__pyx_t_18) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_18) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_end = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "cloudtrace/trace/probe.pyx":237
 *         f.close()
 *     end = time.time()
 *     duration = end - start             # <<<<<<<<<<<<<<
 *     if duration > 0:
 *         pps = num_targets * 32 / duration
 */
  __pyx_t_4 = PyNumber_Subtract(__pyx_v_end, __pyx_v_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_duration = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "cloudtrace/trace/probe.pyx":238
 *     end = time.time()
 *     duration = end - start
 *     if duration > 0:             # <<<<<<<<<<<<<<
 *         pps = num_targets * 32 / duration
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_duration, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 238, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 238, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_2) {

    /* "cloudtrace/trace/probe.pyx":239
 *     duration = end - start
 *     if duration > 0:
 *         pps = num_targets * 32 / duration             # <<<<<<<<<<<<<<
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))
 *     else:
 */
    __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_num_targets * 32)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_4, __pyx_v_duration); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_pps = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "cloudtrace/trace/probe.pyx":240
 *     if duration > 0:
 *         pps = num_targets * 32 / duration
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))             # <<<<<<<<<<<<<<
 *     else:
 *         print('Too fast to time.')
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Time_2f_seconds_2f_pps, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_18 = NULL;
    __pyx_t_24 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_18)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_18);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_24 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_18, __pyx_v_duration, __pyx_v_pps};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 240, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_18, __pyx_v_duration, __pyx_v_pps};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_24, 2+__pyx_t_24); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 240, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_24); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_18) {
        __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_18); __pyx_t_18 = NULL;
      }
      __Pyx_INCREF(__pyx_v_duration);
      __Pyx_GIVEREF(__pyx_v_duration);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_24, __pyx_v_duration);
      __Pyx_INCREF(__pyx_v_pps);
      __Pyx_GIVEREF(__pyx_v_pps);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_24, __pyx_v_pps);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_10, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 240, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "cloudtrace/trace/probe.pyx":238
 *     end = time.time()
 *     duration = end - start
 *     if duration > 0:             # <<<<<<<<<<<<<<
 *         pps = num_targets * 32 / duration
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))
 */
    goto __pyx_L37;
  }

  /* "cloudtrace/trace/probe.pyx":242
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))
 *     else:
 *         print('Too fast to time.')             # <<<<<<<<<<<<<<
 * 
 * cpdef void ping(list targets, uint16_t pid, file=None, int numpings=1, int cycles=1, str proto='icmp', uint16_t size=64):
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L37:;

  /* "cloudtrace/trace/probe.pyx":189
 *     free(gh)
 * 
 * cpdef void craft(str targets, uint16_t pid, file=None, int minttl=1, int maxttl=32, int cycles=1, str proto='icmp', uint16_t size=64, bint randomize=False) except *:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes dst
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("cloudtrace.trace.probe.craft", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dst);
  __Pyx_XDECREF((PyObject *)__pyx_v_probe);
  __Pyx_XDECREF(__pyx_v_pkt);
  __Pyx_XDECREF(__pyx_v_caplen_line);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_g);
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_addr);
  __Pyx_XDECREF(__pyx_v_end);
  __Pyx_XDECREF(__pyx_v_duration);
  __Pyx_XDECREF(__pyx_v_pps);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9craft(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_8craft[] = "craft(unicode targets, uint16_t pid, file=None, int minttl=1, int maxttl=32, int cycles=1, unicode proto=u'icmp', uint16_t size=64, bool randomize=False) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_9craft = {"craft", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10cloudtrace_5trace_5probe_9craft, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10cloudtrace_5trace_5probe_8craft};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_9craft(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_targets = 0;
  uint16_t __pyx_v_pid;
  PyObject *__pyx_v_file = 0;
  int __pyx_v_minttl;
  int __pyx_v_maxttl;
  int __pyx_v_cycles;
  PyObject *__pyx_v_proto = 0;
  uint16_t __pyx_v_size;
  int __pyx_v_randomize;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("craft (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_targets,&__pyx_n_s_pid,&__pyx_n_s_file,&__pyx_n_s_minttl,&__pyx_n_s_maxttl,&__pyx_n_s_cycles,&__pyx_n_s_proto,&__pyx_n_s_size,&__pyx_n_s_randomize,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[2] = ((PyObject *)Py_None);
    values[6] = ((PyObject*)__pyx_n_u_icmp);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_targets)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("craft", 0, 2, 9, 1); __PYX_ERR(1, 189, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_minttl);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxttl);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cycles);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_proto);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_randomize);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "craft") < 0)) __PYX_ERR(1, 189, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_targets = ((PyObject*)values[0]);
    __pyx_v_pid = __Pyx_PyInt_As_uint16_t(values[1]); if (unlikely((__pyx_v_pid == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L3_error)
    __pyx_v_file = values[2];
    if (values[3]) {
      __pyx_v_minttl = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_minttl == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L3_error)
    } else {
      __pyx_v_minttl = ((int)1);
    }
    if (values[4]) {
      __pyx_v_maxttl = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_maxttl == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L3_error)
    } else {
      __pyx_v_maxttl = ((int)32);
    }
    if (values[5]) {
      __pyx_v_cycles = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_cycles == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L3_error)
    } else {
      __pyx_v_cycles = ((int)1);
    }
    __pyx_v_proto = ((PyObject*)values[6]);
    if (values[7]) {
      __pyx_v_size = __Pyx_PyInt_As_uint16_t(values[7]); if (unlikely((__pyx_v_size == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L3_error)
    } else {
      __pyx_v_size = ((uint16_t)64);
    }
    if (values[8]) {
      __pyx_v_randomize = __Pyx_PyObject_IsTrue(values[8]); if (unlikely((__pyx_v_randomize == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L3_error)
    } else {
      __pyx_v_randomize = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("craft", 0, 2, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 189, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.craft", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_targets), (&PyUnicode_Type), 1, "targets", 1))) __PYX_ERR(1, 189, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_proto), (&PyUnicode_Type), 1, "proto", 1))) __PYX_ERR(1, 189, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_8craft(__pyx_self, __pyx_v_targets, __pyx_v_pid, __pyx_v_file, __pyx_v_minttl, __pyx_v_maxttl, __pyx_v_cycles, __pyx_v_proto, __pyx_v_size, __pyx_v_randomize);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_8craft(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_targets, uint16_t __pyx_v_pid, PyObject *__pyx_v_file, int __pyx_v_minttl, int __pyx_v_maxttl, int __pyx_v_cycles, PyObject *__pyx_v_proto, uint16_t __pyx_v_size, int __pyx_v_randomize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct __pyx_opt_args_10cloudtrace_5trace_5probe_craft __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("craft", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1.__pyx_n = 7;
  __pyx_t_1.file = __pyx_v_file;
  __pyx_t_1.minttl = __pyx_v_minttl;
  __pyx_t_1.maxttl = __pyx_v_maxttl;
  __pyx_t_1.cycles = __pyx_v_cycles;
  __pyx_t_1.proto = __pyx_v_proto;
  __pyx_t_1.size = __pyx_v_size;
  __pyx_t_1.randomize = __pyx_v_randomize;
  __pyx_f_10cloudtrace_5trace_5probe_craft(__pyx_v_targets, __pyx_v_pid, 0, &__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 189, __pyx_L1_error)
  __pyx_t_2 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cloudtrace.trace.probe.craft", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cloudtrace/trace/probe.pyx":244
 *         print('Too fast to time.')
 * 
 * cpdef void ping(list targets, uint16_t pid, file=None, int numpings=1, int cycles=1, str proto='icmp', uint16_t size=64):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes dst
 */

static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_11ping(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static void __pyx_f_10cloudtrace_5trace_5probe_ping(PyObject *__pyx_v_targets, uint16_t __pyx_v_pid, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_10cloudtrace_5trace_5probe_ping *__pyx_optional_args) {
  PyObject *__pyx_v_file = ((PyObject *)Py_None);
  int __pyx_v_numpings = ((int)1);
  int __pyx_v_cycles = ((int)1);
  PyObject *__pyx_v_proto = ((PyObject*)__pyx_n_u_icmp);
  uint16_t __pyx_v_size = ((uint16_t)64);
  PyObject *__pyx_v_dst = 0;
  uint8_t __pyx_v_seq;
  struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *__pyx_v_probe = 0;
  PyObject *__pyx_v_pkt = 0;
  PyObject *__pyx_v_caplen_line = 0;
  uint16_t __pyx_v_ipproto;
  PyObject *__pyx_v_f = NULL;
  PyObject *__pyx_v_start = NULL;
  CYTHON_UNUSED int __pyx_v__;
  PyObject *__pyx_v_end = NULL;
  PyObject *__pyx_v_duration = NULL;
  PyObject *__pyx_v_pps = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  uint8_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ping", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_file = __pyx_optional_args->file;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_numpings = __pyx_optional_args->numpings;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_cycles = __pyx_optional_args->cycles;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_proto = __pyx_optional_args->proto;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_size = __pyx_optional_args->size;
            }
          }
        }
      }
    }
  }

  /* "cloudtrace/trace/probe.pyx":252
 *         uint16_t ipproto
 * 
 *     ipproto = proto_to_num(proto)             # <<<<<<<<<<<<<<
 *     ipproto = proto_to_num(proto)
 *     if size < minsize(ipproto):
 */
  __pyx_v_ipproto = __pyx_f_10cloudtrace_5trace_5probe_proto_to_num(__pyx_v_proto, 0);

  /* "cloudtrace/trace/probe.pyx":253
 * 
 *     ipproto = proto_to_num(proto)
 *     ipproto = proto_to_num(proto)             # <<<<<<<<<<<<<<
 *     if size < minsize(ipproto):
 *         size = minsize(ipproto)
 */
  __pyx_v_ipproto = __pyx_f_10cloudtrace_5trace_5probe_proto_to_num(__pyx_v_proto, 0);

  /* "cloudtrace/trace/probe.pyx":254
 *     ipproto = proto_to_num(proto)
 *     ipproto = proto_to_num(proto)
 *     if size < minsize(ipproto):             # <<<<<<<<<<<<<<
 *         size = minsize(ipproto)
 *     elif size > 1500:
 */
  __pyx_t_1 = __pyx_f_10cloudtrace_5trace_5probe_minsize(__pyx_v_ipproto); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(1, 254, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_v_size < __pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "cloudtrace/trace/probe.pyx":255
 *     ipproto = proto_to_num(proto)
 *     if size < minsize(ipproto):
 *         size = minsize(ipproto)             # <<<<<<<<<<<<<<
 *     elif size > 1500:
 *         size = 1500
 */
    __pyx_t_2 = __pyx_f_10cloudtrace_5trace_5probe_minsize(__pyx_v_ipproto); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(1, 255, __pyx_L1_error)
    __pyx_v_size = __pyx_t_2;

    /* "cloudtrace/trace/probe.pyx":254
 *     ipproto = proto_to_num(proto)
 *     ipproto = proto_to_num(proto)
 *     if size < minsize(ipproto):             # <<<<<<<<<<<<<<
 *         size = minsize(ipproto)
 *     elif size > 1500:
 */
    goto __pyx_L3;
  }

  /* "cloudtrace/trace/probe.pyx":256
 *     if size < minsize(ipproto):
 *         size = minsize(ipproto)
 *     elif size > 1500:             # <<<<<<<<<<<<<<
 *         size = 1500
 * 
 */
  __pyx_t_2 = ((__pyx_v_size > 0x5DC) != 0);
  if (__pyx_t_2) {

    /* "cloudtrace/trace/probe.pyx":257
 *         size = minsize(ipproto)
 *     elif size > 1500:
 *         size = 1500             # <<<<<<<<<<<<<<
 * 
 *     probe = create_probe(targets[0].decode(), ipproto, size, pid)
 */
    __pyx_v_size = 0x5DC;

    /* "cloudtrace/trace/probe.pyx":256
 *     if size < minsize(ipproto):
 *         size = minsize(ipproto)
 *     elif size > 1500:             # <<<<<<<<<<<<<<
 *         size = 1500
 * 
 */
  }
  __pyx_L3:;

  /* "cloudtrace/trace/probe.pyx":259
 *         size = 1500
 * 
 *     probe = create_probe(targets[0].decode(), ipproto, size, pid)             # <<<<<<<<<<<<<<
 *     caplen_line = fill_packet_header(size)
 * 
 */
  if (unlikely(__pyx_v_targets == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 259, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_GetItemInt_List(__pyx_v_targets, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_decode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "unicode", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(1, 259, __pyx_L1_error)
  __pyx_t_5 = ((PyObject *)__pyx_f_10cloudtrace_5trace_5probe_create_probe(((PyObject*)__pyx_t_3), __pyx_v_ipproto, __pyx_v_size, __pyx_v_pid, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_probe = ((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cloudtrace/trace/probe.pyx":260
 * 
 *     probe = create_probe(targets[0].decode(), ipproto, size, pid)
 *     caplen_line = fill_packet_header(size)             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(file, str):
 */
  __pyx_t_5 = __pyx_f_10cloudtrace_5trace_5probe_fill_packet_header(__pyx_v_size, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_caplen_line = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "cloudtrace/trace/probe.pyx":262
 *     caplen_line = fill_packet_header(size)
 * 
 *     if isinstance(file, str):             # <<<<<<<<<<<<<<
 *         f = open(file, 'wb')
 *     else:
 */
  __pyx_t_2 = PyUnicode_Check(__pyx_v_file); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "cloudtrace/trace/probe.pyx":263
 * 
 *     if isinstance(file, str):
 *         f = open(file, 'wb')             # <<<<<<<<<<<<<<
 *     else:
 *         f = file
 */
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_file);
    __Pyx_GIVEREF(__pyx_v_file);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_file);
    __Pyx_INCREF(__pyx_n_u_wb);
    __Pyx_GIVEREF(__pyx_n_u_wb);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_n_u_wb);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_open, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_f = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cloudtrace/trace/probe.pyx":262
 *     caplen_line = fill_packet_header(size)
 * 
 *     if isinstance(file, str):             # <<<<<<<<<<<<<<
 *         f = open(file, 'wb')
 *     else:
 */
    goto __pyx_L4;
  }

  /* "cloudtrace/trace/probe.pyx":265
 *         f = open(file, 'wb')
 *     else:
 *         f = file             # <<<<<<<<<<<<<<
 *     try:
 *         write_pcap_header(f)
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_file);
    __pyx_v_f = __pyx_v_file;
  }
  __pyx_L4:;

  /* "cloudtrace/trace/probe.pyx":266
 *     else:
 *         f = file
 *     try:             # <<<<<<<<<<<<<<
 *         write_pcap_header(f)
 *         start = time.time()
 */
  /*try:*/ {

    /* "cloudtrace/trace/probe.pyx":267
 *         f = file
 *     try:
 *         write_pcap_header(f)             # <<<<<<<<<<<<<<
 *         start = time.time()
 *         for _ in range(cycles):
 */
    __pyx_f_10cloudtrace_5trace_5probe_write_pcap_header(__pyx_v_f, 0);

    /* "cloudtrace/trace/probe.pyx":268
 *     try:
 *         write_pcap_header(f)
 *         start = time.time()             # <<<<<<<<<<<<<<
 *         for _ in range(cycles):
 *             for dst in targets:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 268, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 268, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 268, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_start = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cloudtrace/trace/probe.pyx":269
 *         write_pcap_header(f)
 *         start = time.time()
 *         for _ in range(cycles):             # <<<<<<<<<<<<<<
 *             for dst in targets:
 *                 probe.set_dst(dst)
 */
    __pyx_t_6 = __pyx_v_cycles;
    __pyx_t_7 = __pyx_t_6;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_v__ = __pyx_t_8;

      /* "cloudtrace/trace/probe.pyx":270
 *         start = time.time()
 *         for _ in range(cycles):
 *             for dst in targets:             # <<<<<<<<<<<<<<
 *                 probe.set_dst(dst)
 *                 for seq in range(numpings):
 */
      if (unlikely(__pyx_v_targets == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(1, 270, __pyx_L6_error)
      }
      __pyx_t_3 = __pyx_v_targets; __Pyx_INCREF(__pyx_t_3); __pyx_t_9 = 0;
      for (;;) {
        if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(1, 270, __pyx_L6_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 270, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
        if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 270, __pyx_L6_error)
        __Pyx_XDECREF_SET(__pyx_v_dst, ((PyObject*)__pyx_t_4));
        __pyx_t_4 = 0;

        /* "cloudtrace/trace/probe.pyx":271
 *         for _ in range(cycles):
 *             for dst in targets:
 *                 probe.set_dst(dst)             # <<<<<<<<<<<<<<
 *                 for seq in range(numpings):
 *                     probe.prep_trace(seq)
 */
        ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->set_dst(__pyx_v_probe, __pyx_v_dst, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 271, __pyx_L6_error)

        /* "cloudtrace/trace/probe.pyx":272
 *             for dst in targets:
 *                 probe.set_dst(dst)
 *                 for seq in range(numpings):             # <<<<<<<<<<<<<<
 *                     probe.prep_trace(seq)
 *                     probe.prepare()
 */
        __pyx_t_10 = __pyx_v_numpings;
        __pyx_t_11 = __pyx_t_10;
        for (__pyx_t_12 = 0; __pyx_t_12 < __pyx_t_11; __pyx_t_12+=1) {
          __pyx_v_seq = __pyx_t_12;

          /* "cloudtrace/trace/probe.pyx":273
 *                 probe.set_dst(dst)
 *                 for seq in range(numpings):
 *                     probe.prep_trace(seq)             # <<<<<<<<<<<<<<
 *                     probe.prepare()
 *                     pkt = probe.to_bytes()
 */
          ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->prep_trace(__pyx_v_probe, __pyx_v_seq, 0);

          /* "cloudtrace/trace/probe.pyx":274
 *                 for seq in range(numpings):
 *                     probe.prep_trace(seq)
 *                     probe.prepare()             # <<<<<<<<<<<<<<
 *                     pkt = probe.to_bytes()
 *                     f.write(caplen_line)
 */
          ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->prepare(__pyx_v_probe, 0);

          /* "cloudtrace/trace/probe.pyx":275
 *                     probe.prep_trace(seq)
 *                     probe.prepare()
 *                     pkt = probe.to_bytes()             # <<<<<<<<<<<<<<
 *                     f.write(caplen_line)
 *                     f.write(pkt)
 */
          __pyx_t_4 = ((struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe *)__pyx_v_probe->__pyx_vtab)->to_bytes(__pyx_v_probe, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 275, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_XDECREF_SET(__pyx_v_pkt, ((PyObject*)__pyx_t_4));
          __pyx_t_4 = 0;

          /* "cloudtrace/trace/probe.pyx":276
 *                     probe.prepare()
 *                     pkt = probe.to_bytes()
 *                     f.write(caplen_line)             # <<<<<<<<<<<<<<
 *                     f.write(pkt)
 *                     # print(pkt)
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 276, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          __pyx_t_4 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_13, __pyx_v_caplen_line) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_caplen_line);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 276, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "cloudtrace/trace/probe.pyx":277
 *                     pkt = probe.to_bytes()
 *                     f.write(caplen_line)
 *                     f.write(pkt)             # <<<<<<<<<<<<<<
 *                     # print(pkt)
 *     finally:
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 277, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          __pyx_t_4 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_13, __pyx_v_pkt) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_pkt);
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 277, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }

        /* "cloudtrace/trace/probe.pyx":270
 *         start = time.time()
 *         for _ in range(cycles):
 *             for dst in targets:             # <<<<<<<<<<<<<<
 *                 probe.set_dst(dst)
 *                 for seq in range(numpings):
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  }

  /* "cloudtrace/trace/probe.pyx":280
 *                     # print(pkt)
 *     finally:
 *         f.close()             # <<<<<<<<<<<<<<
 *     end = time.time()
 *     duration = end - start
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_close); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 280, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __pyx_t_6 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
      {
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_f, __pyx_n_s_close); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 280, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 280, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
      }
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_14;
      goto __pyx_L1_error;
      __pyx_L15_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
      }
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "cloudtrace/trace/probe.pyx":281
 *     finally:
 *         f.close()
 *     end = time.time()             # <<<<<<<<<<<<<<
 *     duration = end - start
 *     if duration > 0:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_end = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "cloudtrace/trace/probe.pyx":282
 *         f.close()
 *     end = time.time()
 *     duration = end - start             # <<<<<<<<<<<<<<
 *     if duration > 0:
 *         pps = len(targets) * 32 / duration
 */
  __pyx_t_3 = PyNumber_Subtract(__pyx_v_end, __pyx_v_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_duration = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "cloudtrace/trace/probe.pyx":283
 *     end = time.time()
 *     duration = end - start
 *     if duration > 0:             # <<<<<<<<<<<<<<
 *         pps = len(targets) * 32 / duration
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_duration, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 283, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 283, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "cloudtrace/trace/probe.pyx":284
 *     duration = end - start
 *     if duration > 0:
 *         pps = len(targets) * 32 / duration             # <<<<<<<<<<<<<<
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))
 *     else:
 */
    if (unlikely(__pyx_v_targets == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 284, __pyx_L1_error)
    }
    __pyx_t_9 = PyList_GET_SIZE(__pyx_v_targets); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(1, 284, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t((__pyx_t_9 * 32)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_v_duration); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_pps = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "cloudtrace/trace/probe.pyx":285
 *     if duration > 0:
 *         pps = len(targets) * 32 / duration
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))             # <<<<<<<<<<<<<<
 *     else:
 *         print('Too fast to time.')
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Time_2f_seconds_2f_pps, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_duration, __pyx_v_pps};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 285, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_duration, __pyx_v_pps};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 285, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_duration);
      __Pyx_GIVEREF(__pyx_v_duration);
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_7, __pyx_v_duration);
      __Pyx_INCREF(__pyx_v_pps);
      __Pyx_GIVEREF(__pyx_v_pps);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_7, __pyx_v_pps);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_13, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_print, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "cloudtrace/trace/probe.pyx":283
 *     end = time.time()
 *     duration = end - start
 *     if duration > 0:             # <<<<<<<<<<<<<<
 *         pps = len(targets) * 32 / duration
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))
 */
    goto __pyx_L16;
  }

  /* "cloudtrace/trace/probe.pyx":287
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))
 *     else:
 *         print('Too fast to time.')             # <<<<<<<<<<<<<<
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_print, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L16:;

  /* "cloudtrace/trace/probe.pyx":244
 *         print('Too fast to time.')
 * 
 * cpdef void ping(list targets, uint16_t pid, file=None, int numpings=1, int cycles=1, str proto='icmp', uint16_t size=64):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes dst
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_WriteUnraisable("cloudtrace.trace.probe.ping", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dst);
  __Pyx_XDECREF((PyObject *)__pyx_v_probe);
  __Pyx_XDECREF(__pyx_v_pkt);
  __Pyx_XDECREF(__pyx_v_caplen_line);
  __Pyx_XDECREF(__pyx_v_f);
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_end);
  __Pyx_XDECREF(__pyx_v_duration);
  __Pyx_XDECREF(__pyx_v_pps);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_11ping(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_10cloudtrace_5trace_5probe_10ping[] = "ping(list targets, uint16_t pid, file=None, int numpings=1, int cycles=1, unicode proto=u'icmp', uint16_t size=64) -> void";
static PyMethodDef __pyx_mdef_10cloudtrace_5trace_5probe_11ping = {"ping", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_10cloudtrace_5trace_5probe_11ping, METH_VARARGS|METH_KEYWORDS, __pyx_doc_10cloudtrace_5trace_5probe_10ping};
static PyObject *__pyx_pw_10cloudtrace_5trace_5probe_11ping(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_targets = 0;
  uint16_t __pyx_v_pid;
  PyObject *__pyx_v_file = 0;
  int __pyx_v_numpings;
  int __pyx_v_cycles;
  PyObject *__pyx_v_proto = 0;
  uint16_t __pyx_v_size;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ping (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_targets,&__pyx_n_s_pid,&__pyx_n_s_file,&__pyx_n_s_numpings,&__pyx_n_s_cycles,&__pyx_n_s_proto,&__pyx_n_s_size,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[2] = ((PyObject *)Py_None);
    values[5] = ((PyObject*)__pyx_n_u_icmp);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_targets)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pid)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ping", 0, 2, 7, 1); __PYX_ERR(1, 244, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_numpings);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_cycles);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_proto);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ping") < 0)) __PYX_ERR(1, 244, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_targets = ((PyObject*)values[0]);
    __pyx_v_pid = __Pyx_PyInt_As_uint16_t(values[1]); if (unlikely((__pyx_v_pid == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 244, __pyx_L3_error)
    __pyx_v_file = values[2];
    if (values[3]) {
      __pyx_v_numpings = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_numpings == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 244, __pyx_L3_error)
    } else {
      __pyx_v_numpings = ((int)1);
    }
    if (values[4]) {
      __pyx_v_cycles = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_cycles == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 244, __pyx_L3_error)
    } else {
      __pyx_v_cycles = ((int)1);
    }
    __pyx_v_proto = ((PyObject*)values[5]);
    if (values[6]) {
      __pyx_v_size = __Pyx_PyInt_As_uint16_t(values[6]); if (unlikely((__pyx_v_size == ((uint16_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 244, __pyx_L3_error)
    } else {
      __pyx_v_size = ((uint16_t)64);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ping", 0, 2, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 244, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cloudtrace.trace.probe.ping", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_targets), (&PyList_Type), 1, "targets", 1))) __PYX_ERR(1, 244, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_proto), (&PyUnicode_Type), 1, "proto", 1))) __PYX_ERR(1, 244, __pyx_L1_error)
  __pyx_r = __pyx_pf_10cloudtrace_5trace_5probe_10ping(__pyx_self, __pyx_v_targets, __pyx_v_pid, __pyx_v_file, __pyx_v_numpings, __pyx_v_cycles, __pyx_v_proto, __pyx_v_size);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10cloudtrace_5trace_5probe_10ping(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_targets, uint16_t __pyx_v_pid, PyObject *__pyx_v_file, int __pyx_v_numpings, int __pyx_v_cycles, PyObject *__pyx_v_proto, uint16_t __pyx_v_size) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  struct __pyx_opt_args_10cloudtrace_5trace_5probe_ping __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ping", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1.__pyx_n = 5;
  __pyx_t_1.file = __pyx_v_file;
  __pyx_t_1.numpings = __pyx_v_numpings;
  __pyx_t_1.cycles = __pyx_v_cycles;
  __pyx_t_1.proto = __pyx_v_proto;
  __pyx_t_1.size = __pyx_v_size;
  __pyx_f_10cloudtrace_5trace_5probe_ping(__pyx_v_targets, __pyx_v_pid, 0, &__pyx_t_1); 
  __pyx_t_2 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cloudtrace.trace.probe.ping", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe __pyx_vtable_10cloudtrace_5trace_5probe_Probe;

static PyObject *__pyx_tp_new_10cloudtrace_5trace_5probe_Probe(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *)o);
  p->__pyx_vtab = __pyx_vtabptr_10cloudtrace_5trace_5probe_Probe;
  if (unlikely(__pyx_pw_10cloudtrace_5trace_5probe_5Probe_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10cloudtrace_5trace_5probe_Probe(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10cloudtrace_5trace_5probe_5Probe_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_10cloudtrace_5trace_5probe_Probe[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_19__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_18__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_5Probe_21__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_5Probe_20__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_5trace_5probe_Probe = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.trace.probe.Probe", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_5trace_5probe_Probe, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_5trace_5probe_Probe, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_5trace_5probe_Probe, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_ICMPProbe __pyx_vtable_10cloudtrace_5trace_5probe_ICMPProbe;

static PyObject *__pyx_tp_new_10cloudtrace_5trace_5probe_ICMPProbe(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *p;
  PyObject *o = __pyx_tp_new_10cloudtrace_5trace_5probe_Probe(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe*)__pyx_vtabptr_10cloudtrace_5trace_5probe_ICMPProbe;
  if (unlikely(__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static PyMethodDef __pyx_methods_10cloudtrace_5trace_5probe_ICMPProbe[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_9__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_8__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_9ICMPProbe_11__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_9ICMPProbe_10__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_5trace_5probe_ICMPProbe = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.trace.probe.ICMPProbe", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_5trace_5probe_Probe, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_5trace_5probe_ICMPProbe, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_5trace_5probe_ICMPProbe, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_UDPProbe __pyx_vtable_10cloudtrace_5trace_5probe_UDPProbe;

static PyObject *__pyx_tp_new_10cloudtrace_5trace_5probe_UDPProbe(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *p;
  PyObject *o = __pyx_tp_new_10cloudtrace_5trace_5probe_Probe(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10cloudtrace_5trace_5probe_Probe*)__pyx_vtabptr_10cloudtrace_5trace_5probe_UDPProbe;
  if (unlikely(__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10cloudtrace_5trace_5probe_UDPProbe(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __pyx_tp_dealloc_10cloudtrace_5trace_5probe_Probe(o);
}

static PyMethodDef __pyx_methods_10cloudtrace_5trace_5probe_UDPProbe[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_17__reduce_cython__, METH_NOARGS, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_16__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_10cloudtrace_5trace_5probe_8UDPProbe_19__setstate_cython__, METH_O, __pyx_doc_10cloudtrace_5trace_5probe_8UDPProbe_18__setstate_cython__},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_10cloudtrace_5trace_5probe_UDPProbe = {
  PyVarObject_HEAD_INIT(0, 0)
  "cloudtrace.trace.probe.UDPProbe", /*tp_name*/
  sizeof(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10cloudtrace_5trace_5probe_UDPProbe, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10cloudtrace_5trace_5probe_UDPProbe, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10cloudtrace_5trace_5probe_UDPProbe, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_probe(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_probe},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "probe",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_0_0_0_0, __pyx_k_0_0_0_0, sizeof(__pyx_k_0_0_0_0), 0, 1, 0, 0},
  {&__pyx_kp_u_192_172_226_78, __pyx_k_192_172_226_78, sizeof(__pyx_k_192_172_226_78), 0, 1, 0, 0},
  {&__pyx_n_s_Ether, __pyx_k_Ether, sizeof(__pyx_k_Ether), 0, 0, 1, 1},
  {&__pyx_n_s_ICMP, __pyx_k_ICMP, sizeof(__pyx_k_ICMP), 0, 0, 1, 1},
  {&__pyx_n_s_ICMPProbe, __pyx_k_ICMPProbe, sizeof(__pyx_k_ICMPProbe), 0, 0, 1, 1},
  {&__pyx_n_s_ICMPProbe___reduce_cython, __pyx_k_ICMPProbe___reduce_cython, sizeof(__pyx_k_ICMPProbe___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_ICMPProbe___setstate_cython, __pyx_k_ICMPProbe___setstate_cython, sizeof(__pyx_k_ICMPProbe___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_ICMPProbe_prep_trace, __pyx_k_ICMPProbe_prep_trace, sizeof(__pyx_k_ICMPProbe_prep_trace), 0, 0, 1, 1},
  {&__pyx_n_s_ICMPProbe_prepare, __pyx_k_ICMPProbe_prepare, sizeof(__pyx_k_ICMPProbe_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_ICMPProbe_set_icmp_id, __pyx_k_ICMPProbe_set_icmp_id, sizeof(__pyx_k_ICMPProbe_set_icmp_id), 0, 0, 1, 1},
  {&__pyx_n_s_IP, __pyx_k_IP, sizeof(__pyx_k_IP), 0, 0, 1, 1},
  {&__pyx_kp_u_No_TCP, __pyx_k_No_TCP, sizeof(__pyx_k_No_TCP), 0, 1, 0, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_Probe, __pyx_k_Probe, sizeof(__pyx_k_Probe), 0, 0, 1, 1},
  {&__pyx_n_s_Probe___reduce_cython, __pyx_k_Probe___reduce_cython, sizeof(__pyx_k_Probe___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Probe___setstate_cython, __pyx_k_Probe___setstate_cython, sizeof(__pyx_k_Probe___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_Probe_create_ether, __pyx_k_Probe_create_ether, sizeof(__pyx_k_Probe_create_ether), 0, 0, 1, 1},
  {&__pyx_n_s_Probe_prep_trace, __pyx_k_Probe_prep_trace, sizeof(__pyx_k_Probe_prep_trace), 0, 0, 1, 1},
  {&__pyx_n_s_Probe_prepare, __pyx_k_Probe_prepare, sizeof(__pyx_k_Probe_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_Probe_set_dst, __pyx_k_Probe_set_dst, sizeof(__pyx_k_Probe_set_dst), 0, 0, 1, 1},
  {&__pyx_n_s_Probe_set_src, __pyx_k_Probe_set_src, sizeof(__pyx_k_Probe_set_src), 0, 0, 1, 1},
  {&__pyx_n_s_Probe_set_ttl, __pyx_k_Probe_set_ttl, sizeof(__pyx_k_Probe_set_ttl), 0, 0, 1, 1},
  {&__pyx_n_s_Probe_to_bytes, __pyx_k_Probe_to_bytes, sizeof(__pyx_k_Probe_to_bytes), 0, 0, 1, 1},
  {&__pyx_kp_u_Time_2f_seconds_2f_pps, __pyx_k_Time_2f_seconds_2f_pps, sizeof(__pyx_k_Time_2f_seconds_2f_pps), 0, 1, 0, 0},
  {&__pyx_kp_u_Too_fast_to_time, __pyx_k_Too_fast_to_time, sizeof(__pyx_k_Too_fast_to_time), 0, 1, 0, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe, __pyx_k_UDPProbe, sizeof(__pyx_k_UDPProbe), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe___reduce_cython, __pyx_k_UDPProbe___reduce_cython, sizeof(__pyx_k_UDPProbe___reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe___setstate_cython, __pyx_k_UDPProbe___setstate_cython, sizeof(__pyx_k_UDPProbe___setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe_prep_trace, __pyx_k_UDPProbe_prep_trace, sizeof(__pyx_k_UDPProbe_prep_trace), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe_prepare, __pyx_k_UDPProbe_prepare, sizeof(__pyx_k_UDPProbe_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe_set_dport, __pyx_k_UDPProbe_set_dport, sizeof(__pyx_k_UDPProbe_set_dport), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe_set_dst, __pyx_k_UDPProbe_set_dst, sizeof(__pyx_k_UDPProbe_set_dst), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe_set_sport, __pyx_k_UDPProbe_set_sport, sizeof(__pyx_k_UDPProbe_set_sport), 0, 0, 1, 1},
  {&__pyx_n_s_UDPProbe_set_src, __pyx_k_UDPProbe_set_src, sizeof(__pyx_k_UDPProbe_set_src), 0, 0, 1, 1},
  {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
  {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
  {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cloudtrace_trace_probe, __pyx_k_cloudtrace_trace_probe, sizeof(__pyx_k_cloudtrace_trace_probe), 0, 0, 1, 1},
  {&__pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_k_cloudtrace_trace_probe_pyx, sizeof(__pyx_k_cloudtrace_trace_probe_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_conf, __pyx_k_conf, sizeof(__pyx_k_conf), 0, 0, 1, 1},
  {&__pyx_n_s_craft, __pyx_k_craft, sizeof(__pyx_k_craft), 0, 0, 1, 1},
  {&__pyx_n_s_create_ether, __pyx_k_create_ether, sizeof(__pyx_k_create_ether), 0, 0, 1, 1},
  {&__pyx_n_s_create_probe, __pyx_k_create_probe, sizeof(__pyx_k_create_probe), 0, 0, 1, 1},
  {&__pyx_n_s_cycles, __pyx_k_cycles, sizeof(__pyx_k_cycles), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_dport, __pyx_k_dport, sizeof(__pyx_k_dport), 0, 0, 1, 1},
  {&__pyx_n_s_dst, __pyx_k_dst, sizeof(__pyx_k_dst), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_file2, __pyx_k_file2, sizeof(__pyx_k_file2), 0, 0, 1, 1},
  {&__pyx_n_s_fill_packet_header, __pyx_k_fill_packet_header, sizeof(__pyx_k_fill_packet_header), 0, 0, 1, 1},
  {&__pyx_n_s_fopen, __pyx_k_fopen, sizeof(__pyx_k_fopen), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_get_if_hwaddr, __pyx_k_get_if_hwaddr, sizeof(__pyx_k_get_if_hwaddr), 0, 0, 1, 1},
  {&__pyx_n_s_getmacbyip, __pyx_k_getmacbyip, sizeof(__pyx_k_getmacbyip), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_u_icmp, __pyx_k_icmp, sizeof(__pyx_k_icmp), 0, 1, 0, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_ipproto, __pyx_k_ipproto, sizeof(__pyx_k_ipproto), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_maxttl, __pyx_k_maxttl, sizeof(__pyx_k_maxttl), 0, 0, 1, 1},
  {&__pyx_n_s_minttl, __pyx_k_minttl, sizeof(__pyx_k_minttl), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_num, __pyx_k_num, sizeof(__pyx_k_num), 0, 0, 1, 1},
  {&__pyx_n_s_numpings, __pyx_k_numpings, sizeof(__pyx_k_numpings), 0, 0, 1, 1},
  {&__pyx_n_s_open, __pyx_k_open, sizeof(__pyx_k_open), 0, 0, 1, 1},
  {&__pyx_n_s_pid, __pyx_k_pid, sizeof(__pyx_k_pid), 0, 0, 1, 1},
  {&__pyx_n_s_ping, __pyx_k_ping, sizeof(__pyx_k_ping), 0, 0, 1, 1},
  {&__pyx_n_s_prep_trace, __pyx_k_prep_trace, sizeof(__pyx_k_prep_trace), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_proto, __pyx_k_proto, sizeof(__pyx_k_proto), 0, 0, 1, 1},
  {&__pyx_n_s_proto_to_num, __pyx_k_proto_to_num, sizeof(__pyx_k_proto_to_num), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_randint, __pyx_k_randint, sizeof(__pyx_k_randint), 0, 0, 1, 1},
  {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
  {&__pyx_n_s_randomize, __pyx_k_randomize, sizeof(__pyx_k_randomize), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_u_rb, __pyx_k_rb, sizeof(__pyx_k_rb), 0, 1, 0, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_route, __pyx_k_route, sizeof(__pyx_k_route), 0, 0, 1, 1},
  {&__pyx_n_s_rpartition, __pyx_k_rpartition, sizeof(__pyx_k_rpartition), 0, 0, 1, 1},
  {&__pyx_n_s_sICMP, __pyx_k_sICMP, sizeof(__pyx_k_sICMP), 0, 0, 1, 1},
  {&__pyx_n_s_sIP, __pyx_k_sIP, sizeof(__pyx_k_sIP), 0, 0, 1, 1},
  {&__pyx_n_s_scapy_arch, __pyx_k_scapy_arch, sizeof(__pyx_k_scapy_arch), 0, 0, 1, 1},
  {&__pyx_n_s_scapy_config, __pyx_k_scapy_config, sizeof(__pyx_k_scapy_config), 0, 0, 1, 1},
  {&__pyx_n_s_scapy_layers_inet, __pyx_k_scapy_layers_inet, sizeof(__pyx_k_scapy_layers_inet), 0, 0, 1, 1},
  {&__pyx_n_s_scapy_layers_l2, __pyx_k_scapy_layers_l2, sizeof(__pyx_k_scapy_layers_l2), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_set_dport, __pyx_k_set_dport, sizeof(__pyx_k_set_dport), 0, 0, 1, 1},
  {&__pyx_n_s_set_dst, __pyx_k_set_dst, sizeof(__pyx_k_set_dst), 0, 0, 1, 1},
  {&__pyx_n_s_set_icmp_id, __pyx_k_set_icmp_id, sizeof(__pyx_k_set_icmp_id), 0, 0, 1, 1},
  {&__pyx_n_s_set_sport, __pyx_k_set_sport, sizeof(__pyx_k_set_sport), 0, 0, 1, 1},
  {&__pyx_n_s_set_src, __pyx_k_set_src, sizeof(__pyx_k_set_src), 0, 0, 1, 1},
  {&__pyx_n_s_set_ttl, __pyx_k_set_ttl, sizeof(__pyx_k_set_ttl), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_sport, __pyx_k_sport, sizeof(__pyx_k_sport), 0, 0, 1, 1},
  {&__pyx_n_s_src, __pyx_k_src, sizeof(__pyx_k_src), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_targets, __pyx_k_targets, sizeof(__pyx_k_targets), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_to_bytes, __pyx_k_to_bytes, sizeof(__pyx_k_to_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_ttl, __pyx_k_ttl, sizeof(__pyx_k_ttl), 0, 0, 1, 1},
  {&__pyx_n_u_udp, __pyx_k_udp, sizeof(__pyx_k_udp), 0, 1, 0, 1},
  {&__pyx_n_u_wb, __pyx_k_wb, sizeof(__pyx_k_wb), 0, 1, 0, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_write_pcap_header, __pyx_k_write_pcap_header, sizeof(__pyx_k_write_pcap_header), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(1, 141, __pyx_L1_error)
  __pyx_builtin_open = __Pyx_GetBuiltinName(__pyx_n_s_open); if (!__pyx_builtin_open) __PYX_ERR(1, 208, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(1, 214, __pyx_L1_error)
  __pyx_builtin_print = __Pyx_GetBuiltinName(__pyx_n_s_print); if (!__pyx_builtin_print) __PYX_ERR(1, 240, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "cloudtrace/trace/probe.pyx":163
 *         probe = uprobe
 *     else:
 *         raise Exception('No TCP!')             # <<<<<<<<<<<<<<
 *     probe.create_ether(srcmac, macdst)
 *     probe.set_src(src.encode())
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_No_TCP); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "cloudtrace/trace/probe.pyx":219
 *                     if randomize:
 *                         addr, _, _ = line.decode().rpartition('.')
 *                         addr = '{}.{}'.format(addr, random.randint(0, 255))             # <<<<<<<<<<<<<<
 *                         dst = '{}'.format(addr).encode()
 *                         # targets.append('{}'.format(addr).encode())
 */
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_255); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "cloudtrace/trace/probe.pyx":215
 *         start = time.time()
 *         for _ in range(cycles):
 *             with fopen(targets, 'rb') as g:             # <<<<<<<<<<<<<<
 *                 for line in g:
 *                     if randomize:
 */
  __pyx_tuple__12 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(1, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "cloudtrace/trace/probe.pyx":242
 *         print('Time: {:,.2f} seconds. {:.2f} pps'.format(duration, pps))
 *     else:
 *         print('Too fast to time.')             # <<<<<<<<<<<<<<
 * 
 * cpdef void ping(list targets, uint16_t pid, file=None, int numpings=1, int cycles=1, str proto='icmp', uint16_t size=64):
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_u_Too_fast_to_time); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "cloudtrace/trace/probe.pyx":37
 *         memcpy(self.etherhdr, ether, sizeof(ethhdr))
 * 
 *     cpdef void create_ether(self, src, dst):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes ether
 */
  __pyx_tuple__14 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_src, __pyx_n_s_dst); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(3, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_create_ether, 37, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(1, 37, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":44
 *         self.copy_ether(<unsigned char *>ether)
 * 
 *     cpdef void set_dst(self, bytes dst) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 * 
 */
  __pyx_tuple__16 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_dst); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_set_dst, 44, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(1, 44, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":47
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 * 
 *     cpdef void set_src(self, bytes src) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 * 
 */
  __pyx_tuple__18 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_src); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_set_src, 47, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(1, 47, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":50
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 * 
 *     cpdef bytes to_bytes(self):             # <<<<<<<<<<<<<<
 *         return <bytes>((<char *>self.buf)[:ntohs(self.iphdr.ip_len) + sizeof(ethhdr)])
 * 
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_to_bytes, 50, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(1, 50, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":53
 *         return <bytes>((<char *>self.buf)[:ntohs(self.iphdr.ip_len) + sizeof(ethhdr)])
 * 
 *     cpdef void set_ttl(self, uint8_t ttl) except *:             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_ttl = ttl
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_ttl, __pyx_n_s_ttl); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_set_ttl, 53, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(1, 53, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":56
 *         self.iphdr.ip_ttl = ttl
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 * 
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_prepare, 56, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(1, 56, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":63
 *         self.iphdr.ip_sum = htons(checksum(self.iphdr, sizeof(ip)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_num, __pyx_n_s_num); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_prep_trace, 63, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(1, 63, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__30 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":73
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(icmp)
 * 
 *     cpdef void set_icmp_id(self, uint16_t pid):             # <<<<<<<<<<<<<<
 *         self.icmphdr.un.echo.id = htons(pid)
 * 
 */
  __pyx_tuple__32 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_pid, __pyx_n_s_pid); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_set_icmp_id, 73, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(1, 73, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":76
 *         self.icmphdr.un.echo.id = htons(pid)
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 *         self.icmphdr.checksum = 0
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_prepare, 76, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(1, 76, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":81
 *         self.icmphdr.checksum = htons(checksum(self.icmphdr, ntohs(self.iphdr.ip_len) - sizeof(ip)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint16_t data = 0x0608 - num
 */
  __pyx_tuple__36 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_num, __pyx_n_s_num); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_prep_trace, 81, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(1, 81, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__40 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":105
 *             free(self.pseudo)
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 *         self.udphdr.check = 0
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__42, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_prepare, 105, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(1, 105, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":110
 *         self.udphdr.check = htons(checksum(self.pseudo, sizeof(udppseudo)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         cdef:
 *             unsigned short data, chksum
 */
  __pyx_tuple__44 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_num, __pyx_n_s_num); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(1, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__44, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_prep_trace, 110, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(1, 110, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":119
 *         memcpy(self.pseudo.data, &data, 2)
 * 
 *     cpdef void set_sport(self, uint16_t sport):             # <<<<<<<<<<<<<<
 *         self.udphdr.source = htons(sport)
 *         self.pseudo.sport = htons(sport)
 */
  __pyx_tuple__46 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_sport, __pyx_n_s_sport); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(1, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_set_sport, 119, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(1, 119, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":123
 *         self.pseudo.sport = htons(sport)
 * 
 *     cpdef void set_dport(self, uint16_t dport):             # <<<<<<<<<<<<<<
 *         self.udphdr.dest = htons(dport)
 *         self.pseudo.dport = htons(dport)
 */
  __pyx_tuple__48 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_dport, __pyx_n_s_dport); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(2, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_set_dport, 123, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(1, 123, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":127
 *         self.pseudo.dport = htons(dport)
 * 
 *     cpdef void set_dst(self, bytes dst) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)
 */
  __pyx_tuple__50 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_dst); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_set_dst, 127, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(1, 127, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":131
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)
 * 
 *     cpdef void set_src(self, bytes src) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)
 */
  __pyx_tuple__52 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_src); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_set_src, 131, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(1, 131, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__56 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__56, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 3, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":135
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)
 * 
 * cpdef uint8_t proto_to_num(str proto):             # <<<<<<<<<<<<<<
 *     if proto == 'icmp':
 *         return IPPROTO_ICMP
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_n_s_proto); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_proto_to_num, 135, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(1, 135, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":143
 *         raise NotImplementedError()
 * 
 * cpdef Probe create_probe(str dst, uint8_t ipproto, uint16_t size, uint16_t pid):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str iface, src, nexthop, macdst, srcmac
 */
  __pyx_tuple__60 = PyTuple_Pack(4, __pyx_n_s_dst, __pyx_n_s_ipproto, __pyx_n_s_size, __pyx_n_s_pid); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(4, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__60, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_create_probe, 143, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(1, 143, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":168
 *     return probe
 * 
 * cpdef bytes fill_packet_header(uint16_t size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t caplen = size + sizeof(ethhdr)
 */
  __pyx_tuple__62 = PyTuple_Pack(2, __pyx_n_s_size, __pyx_n_s_size); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__62, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_fill_packet_header, 168, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(1, 168, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":176
 *     return <bytes>(<unsigned char *>&ph)[:sizeof(pcapfix_packet_hdr_s)]
 * 
 * cpdef void write_pcap_header(f):             # <<<<<<<<<<<<<<
 *     cdef:
 *         pcapfix_global_hdr_s *gh = <pcapfix_global_hdr_s *>calloc(1, sizeof(pcapfix_global_hdr_s))
 */
  __pyx_tuple__64 = PyTuple_Pack(1, __pyx_n_s_f); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__64, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_write_pcap_header, 176, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(1, 176, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":189
 *     free(gh)
 * 
 * cpdef void craft(str targets, uint16_t pid, file=None, int minttl=1, int maxttl=32, int cycles=1, str proto='icmp', uint16_t size=64, bint randomize=False) except *:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes dst
 */
  __pyx_tuple__66 = PyTuple_Pack(9, __pyx_n_s_targets, __pyx_n_s_pid, __pyx_n_s_file, __pyx_n_s_minttl, __pyx_n_s_maxttl, __pyx_n_s_cycles, __pyx_n_s_proto, __pyx_n_s_size, __pyx_n_s_randomize); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(9, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_craft, 189, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(1, 189, __pyx_L1_error)

  /* "cloudtrace/trace/probe.pyx":244
 *         print('Too fast to time.')
 * 
 * cpdef void ping(list targets, uint16_t pid, file=None, int numpings=1, int cycles=1, str proto='icmp', uint16_t size=64):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes dst
 */
  __pyx_tuple__68 = PyTuple_Pack(7, __pyx_n_s_targets, __pyx_n_s_pid, __pyx_n_s_file, __pyx_n_s_numpings, __pyx_n_s_cycles, __pyx_n_s_proto, __pyx_n_s_size); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(1, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(7, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cloudtrace_trace_probe_pyx, __pyx_n_s_ping, 244, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(1, 244, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_int_255 = PyInt_FromLong(255); if (unlikely(!__pyx_int_255)) __PYX_ERR(1, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("minsize", (void (*)(void))__pyx_f_10cloudtrace_5trace_5probe_minsize, "int (int)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("proto_to_num", (void (*)(void))__pyx_f_10cloudtrace_5trace_5probe_proto_to_num, "uint8_t (PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("create_probe", (void (*)(void))__pyx_f_10cloudtrace_5trace_5probe_create_probe, "struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *(PyObject *, uint8_t, uint16_t, uint16_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("fill_packet_header", (void (*)(void))__pyx_f_10cloudtrace_5trace_5probe_fill_packet_header, "PyObject *(uint16_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("write_pcap_header", (void (*)(void))__pyx_f_10cloudtrace_5trace_5probe_write_pcap_header, "void (PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("craft", (void (*)(void))__pyx_f_10cloudtrace_5trace_5probe_craft, "void (PyObject *, uint16_t, int __pyx_skip_dispatch, struct __pyx_opt_args_10cloudtrace_5trace_5probe_craft *__pyx_optional_args)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ping", (void (*)(void))__pyx_f_10cloudtrace_5trace_5probe_ping, "void (PyObject *, uint16_t, int __pyx_skip_dispatch, struct __pyx_opt_args_10cloudtrace_5trace_5probe_ping *__pyx_optional_args)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_10cloudtrace_5trace_5probe_Probe = &__pyx_vtable_10cloudtrace_5trace_5probe_Probe;
  __pyx_vtable_10cloudtrace_5trace_5probe_Probe.set_dst = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_5Probe_set_dst;
  __pyx_vtable_10cloudtrace_5trace_5probe_Probe.set_src = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_5Probe_set_src;
  __pyx_vtable_10cloudtrace_5trace_5probe_Probe.to_bytes = (PyObject *(*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_5Probe_to_bytes;
  __pyx_vtable_10cloudtrace_5trace_5probe_Probe.set_ttl = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, uint8_t, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_5Probe_set_ttl;
  __pyx_vtable_10cloudtrace_5trace_5probe_Probe.prepare = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_5Probe_prepare;
  __pyx_vtable_10cloudtrace_5trace_5probe_Probe.prepare_ip = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *))__pyx_f_10cloudtrace_5trace_5probe_5Probe_prepare_ip;
  __pyx_vtable_10cloudtrace_5trace_5probe_Probe.copy_ether = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, void *))__pyx_f_10cloudtrace_5trace_5probe_5Probe_copy_ether;
  __pyx_vtable_10cloudtrace_5trace_5probe_Probe.create_ether = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_5Probe_create_ether;
  __pyx_vtable_10cloudtrace_5trace_5probe_Probe.prep_trace = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, uint8_t, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_5Probe_prep_trace;
  if (PyType_Ready(&__pyx_type_10cloudtrace_5trace_5probe_Probe) < 0) __PYX_ERR(1, 20, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_5trace_5probe_Probe.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_5trace_5probe_Probe.tp_dictoffset && __pyx_type_10cloudtrace_5trace_5probe_Probe.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_5trace_5probe_Probe.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_5trace_5probe_Probe.tp_dict, __pyx_vtabptr_10cloudtrace_5trace_5probe_Probe) < 0) __PYX_ERR(1, 20, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Probe, (PyObject *)&__pyx_type_10cloudtrace_5trace_5probe_Probe) < 0) __PYX_ERR(1, 20, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_5trace_5probe_Probe) < 0) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_5trace_5probe_Probe = &__pyx_type_10cloudtrace_5trace_5probe_Probe;
  __pyx_vtabptr_10cloudtrace_5trace_5probe_ICMPProbe = &__pyx_vtable_10cloudtrace_5trace_5probe_ICMPProbe;
  __pyx_vtable_10cloudtrace_5trace_5probe_ICMPProbe.__pyx_base = *__pyx_vtabptr_10cloudtrace_5trace_5probe_Probe;
  __pyx_vtable_10cloudtrace_5trace_5probe_ICMPProbe.__pyx_base.prepare = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_prepare;
  __pyx_vtable_10cloudtrace_5trace_5probe_ICMPProbe.__pyx_base.prep_trace = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, uint8_t, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_prep_trace;
  __pyx_vtable_10cloudtrace_5trace_5probe_ICMPProbe.set_icmp_id = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_ICMPProbe *, uint16_t, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_9ICMPProbe_set_icmp_id;
  __pyx_type_10cloudtrace_5trace_5probe_ICMPProbe.tp_base = __pyx_ptype_10cloudtrace_5trace_5probe_Probe;
  if (PyType_Ready(&__pyx_type_10cloudtrace_5trace_5probe_ICMPProbe) < 0) __PYX_ERR(1, 66, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_5trace_5probe_ICMPProbe.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_5trace_5probe_ICMPProbe.tp_dictoffset && __pyx_type_10cloudtrace_5trace_5probe_ICMPProbe.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_5trace_5probe_ICMPProbe.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_5trace_5probe_ICMPProbe.tp_dict, __pyx_vtabptr_10cloudtrace_5trace_5probe_ICMPProbe) < 0) __PYX_ERR(1, 66, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ICMPProbe, (PyObject *)&__pyx_type_10cloudtrace_5trace_5probe_ICMPProbe) < 0) __PYX_ERR(1, 66, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_5trace_5probe_ICMPProbe) < 0) __PYX_ERR(1, 66, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_5trace_5probe_ICMPProbe = &__pyx_type_10cloudtrace_5trace_5probe_ICMPProbe;
  __pyx_vtabptr_10cloudtrace_5trace_5probe_UDPProbe = &__pyx_vtable_10cloudtrace_5trace_5probe_UDPProbe;
  __pyx_vtable_10cloudtrace_5trace_5probe_UDPProbe.__pyx_base = *__pyx_vtabptr_10cloudtrace_5trace_5probe_Probe;
  __pyx_vtable_10cloudtrace_5trace_5probe_UDPProbe.__pyx_base.set_dst = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_dst;
  __pyx_vtable_10cloudtrace_5trace_5probe_UDPProbe.__pyx_base.set_src = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, PyObject *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_src;
  __pyx_vtable_10cloudtrace_5trace_5probe_UDPProbe.__pyx_base.prepare = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_prepare;
  __pyx_vtable_10cloudtrace_5trace_5probe_UDPProbe.__pyx_base.prep_trace = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_Probe *, uint8_t, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_prep_trace;
  __pyx_vtable_10cloudtrace_5trace_5probe_UDPProbe.set_sport = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *, uint16_t, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_sport;
  __pyx_vtable_10cloudtrace_5trace_5probe_UDPProbe.set_dport = (void (*)(struct __pyx_obj_10cloudtrace_5trace_5probe_UDPProbe *, uint16_t, int __pyx_skip_dispatch))__pyx_f_10cloudtrace_5trace_5probe_8UDPProbe_set_dport;
  __pyx_type_10cloudtrace_5trace_5probe_UDPProbe.tp_base = __pyx_ptype_10cloudtrace_5trace_5probe_Probe;
  if (PyType_Ready(&__pyx_type_10cloudtrace_5trace_5probe_UDPProbe) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_10cloudtrace_5trace_5probe_UDPProbe.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10cloudtrace_5trace_5probe_UDPProbe.tp_dictoffset && __pyx_type_10cloudtrace_5trace_5probe_UDPProbe.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10cloudtrace_5trace_5probe_UDPProbe.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_10cloudtrace_5trace_5probe_UDPProbe.tp_dict, __pyx_vtabptr_10cloudtrace_5trace_5probe_UDPProbe) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_UDPProbe, (PyObject *)&__pyx_type_10cloudtrace_5trace_5probe_UDPProbe) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_10cloudtrace_5trace_5probe_UDPProbe) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  __pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe = &__pyx_type_10cloudtrace_5trace_5probe_UDPProbe;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("cloudtrace.trace.utils"); if (!__pyx_t_1) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "checksum", (void (**)(void))&__pyx_f_10cloudtrace_5trace_5utils_checksum, "uint16_t (void *, int)") < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initprobe(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initprobe(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_probe(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_probe(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_probe(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'probe' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_probe(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(1, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("probe", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(1, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(1, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_cloudtrace__trace__probe) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(1, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cloudtrace.trace.probe")) {
      if (unlikely(PyDict_SetItemString(modules, "cloudtrace.trace.probe", __pyx_m) < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(1, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  #endif

  /* "cloudtrace/trace/probe.pyx":1
 * import random             # <<<<<<<<<<<<<<
 * import struct
 * import time
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_random, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_random, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/trace/probe.pyx":2
 * import random
 * import struct             # <<<<<<<<<<<<<<
 * import time
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_struct, __pyx_t_1) < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/trace/probe.pyx":3
 * import random
 * import struct
 * import time             # <<<<<<<<<<<<<<
 * 
 * from scapy.arch import get_if_hwaddr
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_time, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_1) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/trace/probe.pyx":5
 * import time
 * 
 * from scapy.arch import get_if_hwaddr             # <<<<<<<<<<<<<<
 * from scapy.config import conf
 * from scapy.layers.inet import IP as sIP, ICMP as sICMP
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_get_if_hwaddr);
  __Pyx_GIVEREF(__pyx_n_s_get_if_hwaddr);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_get_if_hwaddr);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_scapy_arch, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_get_if_hwaddr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_if_hwaddr, __pyx_t_1) < 0) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":6
 * 
 * from scapy.arch import get_if_hwaddr
 * from scapy.config import conf             # <<<<<<<<<<<<<<
 * from scapy.layers.inet import IP as sIP, ICMP as sICMP
 * from scapy.layers.l2 import getmacbyip, Ether
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_conf);
  __Pyx_GIVEREF(__pyx_n_s_conf);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_conf);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_scapy_config, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_conf); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_conf, __pyx_t_2) < 0) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/trace/probe.pyx":7
 * from scapy.arch import get_if_hwaddr
 * from scapy.config import conf
 * from scapy.layers.inet import IP as sIP, ICMP as sICMP             # <<<<<<<<<<<<<<
 * from scapy.layers.l2 import getmacbyip, Ether
 * 
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_IP);
  __Pyx_GIVEREF(__pyx_n_s_IP);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_IP);
  __Pyx_INCREF(__pyx_n_s_ICMP);
  __Pyx_GIVEREF(__pyx_n_s_ICMP);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_ICMP);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_scapy_layers_inet, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_IP); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sIP, __pyx_t_1) < 0) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ICMP); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sICMP, __pyx_t_1) < 0) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":8
 * from scapy.config import conf
 * from scapy.layers.inet import IP as sIP, ICMP as sICMP
 * from scapy.layers.l2 import getmacbyip, Ether             # <<<<<<<<<<<<<<
 * 
 * from file2 import fopen
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_getmacbyip);
  __Pyx_GIVEREF(__pyx_n_s_getmacbyip);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_getmacbyip);
  __Pyx_INCREF(__pyx_n_s_Ether);
  __Pyx_GIVEREF(__pyx_n_s_Ether);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_Ether);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_scapy_layers_l2, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_getmacbyip); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_getmacbyip, __pyx_t_2) < 0) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Ether); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Ether, __pyx_t_2) < 0) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cloudtrace/trace/probe.pyx":10
 * from scapy.layers.l2 import getmacbyip, Ether
 * 
 * from file2 import fopen             # <<<<<<<<<<<<<<
 * 
 * cdef bint minsize(int proto) except -1:
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_fopen);
  __Pyx_GIVEREF(__pyx_n_s_fopen);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_fopen);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_file2, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_fopen); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fopen, __pyx_t_1) < 0) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":37
 *         memcpy(self.etherhdr, ether, sizeof(ethhdr))
 * 
 *     cpdef void create_ether(self, src, dst):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bytes ether
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5Probe_5create_ether, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Probe_create_ether, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__15)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_Probe->tp_dict, __pyx_n_s_create_ether, __pyx_t_2) < 0) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_Probe);

  /* "cloudtrace/trace/probe.pyx":44
 *         self.copy_ether(<unsigned char *>ether)
 * 
 *     cpdef void set_dst(self, bytes dst) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5Probe_7set_dst, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Probe_set_dst, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__17)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_Probe->tp_dict, __pyx_n_s_set_dst, __pyx_t_2) < 0) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_Probe);

  /* "cloudtrace/trace/probe.pyx":47
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 * 
 *     cpdef void set_src(self, bytes src) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5Probe_9set_src, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Probe_set_src, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__19)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_Probe->tp_dict, __pyx_n_s_set_src, __pyx_t_2) < 0) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_Probe);

  /* "cloudtrace/trace/probe.pyx":50
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 * 
 *     cpdef bytes to_bytes(self):             # <<<<<<<<<<<<<<
 *         return <bytes>((<char *>self.buf)[:ntohs(self.iphdr.ip_len) + sizeof(ethhdr)])
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5Probe_11to_bytes, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Probe_to_bytes, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__21)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_Probe->tp_dict, __pyx_n_s_to_bytes, __pyx_t_2) < 0) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_Probe);

  /* "cloudtrace/trace/probe.pyx":53
 *         return <bytes>((<char *>self.buf)[:ntohs(self.iphdr.ip_len) + sizeof(ethhdr)])
 * 
 *     cpdef void set_ttl(self, uint8_t ttl) except *:             # <<<<<<<<<<<<<<
 *         self.iphdr.ip_ttl = ttl
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5Probe_13set_ttl, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Probe_set_ttl, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_Probe->tp_dict, __pyx_n_s_set_ttl, __pyx_t_2) < 0) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_Probe);

  /* "cloudtrace/trace/probe.pyx":56
 *         self.iphdr.ip_ttl = ttl
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5Probe_15prepare, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Probe_prepare, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_Probe->tp_dict, __pyx_n_s_prepare, __pyx_t_2) < 0) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_Probe);

  /* "cloudtrace/trace/probe.pyx":63
 *         self.iphdr.ip_sum = htons(checksum(self.iphdr, sizeof(ip)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5Probe_17prep_trace, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Probe_prep_trace, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_Probe->tp_dict, __pyx_n_s_prep_trace, __pyx_t_2) < 0) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_Probe);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5Probe_19__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Probe___reduce_cython, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5Probe_21__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Probe___setstate_cython, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":73
 *         self.data = self.buf + sizeof(ethhdr) + sizeof(ip) + sizeof(icmp)
 * 
 *     cpdef void set_icmp_id(self, uint16_t pid):             # <<<<<<<<<<<<<<
 *         self.icmphdr.un.echo.id = htons(pid)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_3set_icmp_id, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ICMPProbe_set_icmp_id, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_ICMPProbe->tp_dict, __pyx_n_s_set_icmp_id, __pyx_t_2) < 0) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_ICMPProbe);

  /* "cloudtrace/trace/probe.pyx":76
 *         self.icmphdr.un.echo.id = htons(pid)
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 *         self.icmphdr.checksum = 0
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_5prepare, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ICMPProbe_prepare, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_ICMPProbe->tp_dict, __pyx_n_s_prepare, __pyx_t_2) < 0) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_ICMPProbe);

  /* "cloudtrace/trace/probe.pyx":81
 *         self.icmphdr.checksum = htons(checksum(self.icmphdr, ntohs(self.iphdr.ip_len) - sizeof(ip)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         cdef:
 *             uint16_t data = 0x0608 - num
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_7prep_trace, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ICMPProbe_prep_trace, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_ICMPProbe->tp_dict, __pyx_n_s_prep_trace, __pyx_t_2) < 0) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_ICMPProbe);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ICMPProbe___reduce_cython, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_9ICMPProbe_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ICMPProbe___setstate_cython, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":105
 *             free(self.pseudo)
 * 
 *     cpdef void prepare(self):             # <<<<<<<<<<<<<<
 *         self.prepare_ip()
 *         self.udphdr.check = 0
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_5prepare, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPProbe_prepare, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe->tp_dict, __pyx_n_s_prepare, __pyx_t_2) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe);

  /* "cloudtrace/trace/probe.pyx":110
 *         self.udphdr.check = htons(checksum(self.pseudo, sizeof(udppseudo)))
 * 
 *     cpdef void prep_trace(self, uint8_t num):             # <<<<<<<<<<<<<<
 *         cdef:
 *             unsigned short data, chksum
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_7prep_trace, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPProbe_prep_trace, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe->tp_dict, __pyx_n_s_prep_trace, __pyx_t_2) < 0) __PYX_ERR(1, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe);

  /* "cloudtrace/trace/probe.pyx":119
 *         memcpy(self.pseudo.data, &data, 2)
 * 
 *     cpdef void set_sport(self, uint16_t sport):             # <<<<<<<<<<<<<<
 *         self.udphdr.source = htons(sport)
 *         self.pseudo.sport = htons(sport)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_9set_sport, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPProbe_set_sport, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__47)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe->tp_dict, __pyx_n_s_set_sport, __pyx_t_2) < 0) __PYX_ERR(1, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe);

  /* "cloudtrace/trace/probe.pyx":123
 *         self.pseudo.sport = htons(sport)
 * 
 *     cpdef void set_dport(self, uint16_t dport):             # <<<<<<<<<<<<<<
 *         self.udphdr.dest = htons(dport)
 *         self.pseudo.dport = htons(dport)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_11set_dport, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPProbe_set_dport, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe->tp_dict, __pyx_n_s_set_dport, __pyx_t_2) < 0) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe);

  /* "cloudtrace/trace/probe.pyx":127
 *         self.pseudo.dport = htons(dport)
 * 
 *     cpdef void set_dst(self, bytes dst) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, dst, self.iphdr.ip_dst)
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_13set_dst, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPProbe_set_dst, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe->tp_dict, __pyx_n_s_set_dst, __pyx_t_2) < 0) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe);

  /* "cloudtrace/trace/probe.pyx":131
 *         memcpy(self.pseudo.dst, self.iphdr.ip_dst, 4)
 * 
 *     cpdef void set_src(self, bytes src) except *:             # <<<<<<<<<<<<<<
 *         inet_pton(AF_INET, src, self.iphdr.ip_src)
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_15set_src, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPProbe_set_src, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe->tp_dict, __pyx_n_s_set_src, __pyx_t_2) < 0) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10cloudtrace_5trace_5probe_UDPProbe);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_17__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPProbe___reduce_cython, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_8UDPProbe_19__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_UDPProbe___setstate_cython, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__57)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":135
 *         memcpy(self.pseudo.src, self.iphdr.ip_src, 4)
 * 
 * cpdef uint8_t proto_to_num(str proto):             # <<<<<<<<<<<<<<
 *     if proto == 'icmp':
 *         return IPPROTO_ICMP
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_1proto_to_num, 0, __pyx_n_s_proto_to_num, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__59)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_proto_to_num, __pyx_t_2) < 0) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":143
 *         raise NotImplementedError()
 * 
 * cpdef Probe create_probe(str dst, uint8_t ipproto, uint16_t size, uint16_t pid):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str iface, src, nexthop, macdst, srcmac
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_3create_probe, 0, __pyx_n_s_create_probe, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_create_probe, __pyx_t_2) < 0) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":168
 *     return probe
 * 
 * cpdef bytes fill_packet_header(uint16_t size):             # <<<<<<<<<<<<<<
 *     cdef:
 *         uint32_t caplen = size + sizeof(ethhdr)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_5fill_packet_header, 0, __pyx_n_s_fill_packet_header, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fill_packet_header, __pyx_t_2) < 0) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":176
 *     return <bytes>(<unsigned char *>&ph)[:sizeof(pcapfix_packet_hdr_s)]
 * 
 * cpdef void write_pcap_header(f):             # <<<<<<<<<<<<<<
 *     cdef:
 *         pcapfix_global_hdr_s *gh = <pcapfix_global_hdr_s *>calloc(1, sizeof(pcapfix_global_hdr_s))
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_7write_pcap_header, 0, __pyx_n_s_write_pcap_header, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__65)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_write_pcap_header, __pyx_t_2) < 0) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":189
 *     free(gh)
 * 
 * cpdef void craft(str targets, uint16_t pid, file=None, int minttl=1, int maxttl=32, int cycles=1, str proto='icmp', uint16_t size=64, bint randomize=False) except *:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes dst
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_9craft, 0, __pyx_n_s_craft, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_craft, __pyx_t_2) < 0) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":244
 *         print('Too fast to time.')
 * 
 * cpdef void ping(list targets, uint16_t pid, file=None, int numpings=1, int cycles=1, str proto='icmp', uint16_t size=64):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bytes dst
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10cloudtrace_5trace_5probe_11ping, 0, __pyx_n_s_ping, NULL, __pyx_n_s_cloudtrace_trace_probe, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ping, __pyx_t_2) < 0) __PYX_ERR(1, 244, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "cloudtrace/trace/probe.pyx":1
 * import random             # <<<<<<<<<<<<<<
 * import struct
 * import time
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init cloudtrace.trace.probe", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cloudtrace.trace.probe");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE uint16_t __Pyx_PyInt_As_uint16_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint16_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint16_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint16_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint16_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint16_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 2 * PyLong_SHIFT) {
                            return (uint16_t) (((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 3 * PyLong_SHIFT) {
                            return (uint16_t) (((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) >= 4 * PyLong_SHIFT) {
                            return (uint16_t) (((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint16_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint16_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint16_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint16_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint16_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint16_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint16_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint16_t) ((((((uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint16_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint16_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint16_t) ((((((((uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint16_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint16_t) (((uint16_t)-1)*(((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint16_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint16_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint16_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint16_t) ((((((((((uint16_t)digits[3]) << PyLong_SHIFT) | (uint16_t)digits[2]) << PyLong_SHIFT) | (uint16_t)digits[1]) << PyLong_SHIFT) | (uint16_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint16_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint16_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint16_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint16_t) -1;
        }
    } else {
        uint16_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint16_t) -1;
        val = __Pyx_PyInt_As_uint16_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint16_t");
    return (uint16_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint16_t");
    return (uint16_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint8_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint8_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint8_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint8_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint8_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint8_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint8_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint8_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint8_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint8_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint8_t) -1;
        }
    } else {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint16_t(uint16_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint16_t neg_one = (uint16_t) -1, const_zero = (uint16_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint16_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint16_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint16_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint16_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint16_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
